%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% atp.oz ; This file is part of the MathWeb system
%
%   major updates: 09.08.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% concurrent atp service (still specialized for DORIS)
%%

functor

import
   OS(getEnv system unlink tmpnam) Property
   % Connection Pickle
   Browser(browse: Browse)
   
   XApplication at 'x-ozlib://mathweb/share/XApplication.ozf'
   XTime        at 'x-ozlib://mathweb/share/XTime.ozf'
   File         at 'x-ozlib://mathweb/share/File.ozf'
   Buffer(collector: Collector fifo:FIFO)
   at 'x-ozlib://mathweb/share/Buffer.ozf'
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   
define
   VS2A = VirtualString.toAtom
   VS2S = VirtualString.toString
   VS2BS = VirtualString.toByteString
   
   AppName     = 'atp'
   ServiceName = 'ATP'
   App         = {New XApplication.manager init}
   
   PServices = provers(bliksem: 'BLIKSEM'
		       chorus:  'CHORUS'
		       fdpll:   'FDPLL'
		       kimba:   'KIMBA'
		       mace:    'MACE'
		       otter:   'OTTER'
		       protein: 'PROTEIN'
		       satchmo: 'SATCHMO'
		       spass:   'SPASS'
		      )
   
   SyntaxMap = map(bliksem: dfg
		   spass:   dfg
		   fdpll:   dfg
		   otter:   otter
		   mace:    otter
		   protein: protein
		   satchmo: otter
		   kimba:   kimba
		   chorus:  chorus)
   ALL  = {VS2BS "all"}
   %%
   %% time statistics
   %%
   local
      VS2S = VirtualString.toString
      
      proc {PPrint VS N Mode}
         S = {VS2S VS}
         L = {Length S}
         R = N - L
      in
         case Mode
         of left then
            {App.printInfo VS}
            {For 1 R 1 proc {$ _} {App.printInfo " "} end}
         [] right then
            {For 1 R 1 proc {$ _} {App.printInfo " "} end}
            {App.printInfo VS}
         end
      end

      MinInit = 100000
      MaxInit =      0
      SumInit =      0
      
      MinZip = fun {$ X Y} {Min X Y} end
      MaxZip = fun {$ X Y} {Max X Y} end
      SumZip = fun {$ X Y} X + Y end
      
      TimeFeatures = [total totalLocal both totalNet
                      real user sys
                      saveVS sendrest send
                      waiting result returnrest return]

      TimeFeatures0 = {Map TimeFeatures fun {$ X} X#0 end}
      
      DefaultTimes = {Record.adjoinList defaults TimeFeatures0}
   in
      %%
      %% time statistics class
      %%
      
      class TimeStat
         
         feat
            minTimes: {NewDictionary}
            maxTimes: {NewDictionary} 
            sumTimes: {NewDictionary} 
            count:    {NewDictionary}

         meth init(ProverNames)
            Count = self.count
         in
            {ForAll ProverNames
             proc {$ PName}
                {ForAll [self.minTimes#MinInit
                         self.maxTimes#MaxInit
                         self.sumTimes#SumInit]
                 proc {$ T#V}
                    R = {Record.make msecs TimeFeatures} in
                    {ForAll TimeFeatures proc {$ F} R.F = V end}
                    {Dictionary.put T PName R}
                 end}
                {Dictionary.put Count PName 0}
             end}
         end
         
         meth putTimes(PName PTimes0)
            PTimes = {Adjoin DefaultTimes PTimes0} 
            MinTimes = self.minTimes
            MaxTimes = self.maxTimes
            SumTimes = self.sumTimes
            Count    = self.count
            Min = {Dictionary.get MinTimes PName}
            Max = {Dictionary.get MaxTimes PName}
            Sum = {Dictionary.get SumTimes PName}
         in
            case PTimes
            of unit then skip
            else
               {Dictionary.put MinTimes PName {Record.zip Min PTimes MinZip}}
               {Dictionary.put MaxTimes PName {Record.zip Max PTimes MaxZip}}
               {Dictionary.put SumTimes PName {Record.zip Sum PTimes SumZip}}
               {Dictionary.put Count    PName {Dictionary.get Count PName}+1}
            end
         end

         meth statistics(ProverNames)
            MinTimes = self.minTimes
            MaxTimes = self.maxTimes
            SumTimes = self.sumTimes
            Count    = self.count
         in
            {App.showInfo "\n*** time statistics\n"}
            {ForAll ProverNames
             proc {$ PName}
                N = {Dictionary.get Count PName} in
                if N > 0 then
                   {App.showInfo PName#" ("#N#" problems)"}
                   Min = {Dictionary.get MinTimes PName}
                   Max = {Dictionary.get MaxTimes PName}
                   Sum = {Dictionary.get SumTimes PName}
                   Avg = {Record.map Sum fun {$ X} X div N end}
                in
                   {ForAll TimeFeatures
                    proc {$ F}
                       {PPrint F 10 left}
                       {App.printInfo ': '}
                       {PPrint Min.F 10 right}
                       {PPrint Max.F 10 right}
                       {PPrint Avg.F 10 right}
                       {App.showInfo ''}
                    end}
                end
             end}      
         end
                  
      end
   end
   
   ProverNames = {Arity PServices}

   Stat = {New TimeStat init(ProverNames)}
   
   proc {StatisticsLoop}
      try
         {Thread.suspend {Thread.this}}
      catch Err then
         case Err of system(kernel(terminate...)...) then raise Err end
         else {Delay 1000} {Stat statistics(ProverNames)} end
      end
      {StatisticsLoop}
   end
   {App newThread(StatisticsLoop detach:true info:'ATP.statistics')}
   
   %%
   %% ATP Service
   %%
   
   local   
      fun {GetTime}
         {Property.get 'time.total'}
      end
      
      fun {Enumerate Xs Sep}
         fun {DoEnumerate Ys}
            case Ys of nil then ""
            [] Y|Yr then Sep#Y#{DoEnumerate Yr} end
         end
      in
         case Xs of nil then ""
         [] X|Xr then X#{DoEnumerate Xr} end
      end
      fun {List2VS L}
         "["#{Enumerate L ","}#"]"
      end
      fun {Features Xs Feat}
         {List2VS {Map Xs fun {$ X} X.Feat end}}
      end
      fun {ToBS X}
	 if {IsInt X} then X
	 else {VS2BS X} end
      end

      fun {FeaturesXML Xs Feat}
	 {Map Xs fun {$ X} {ToBS X.Feat} end}
      end

      fun {FeaturesOz Xs Feat}
	 {Map Xs fun {$ X} X.Feat end}
      end

      fun {FindMin Xs Default}
         case Xs of nil then Default
         [] X|Xr then {FindMin Xr {Min X Default}} end
      end
      fun {Find Xs Time}
         X|Xr = Xs in
         if X.time == Time
         then X
         else {Find Xr Time}
         end
      end
      fun {FindFirst Xs}
         T1|Tr = {Map Xs fun {$ P} P.time end}
         Tmin  = {FindMin Tr T1}
      in
         {Find Xs Tmin}
      end

      class ProverProxy
         
         feat
            name
            host
            object
            
         meth init(PObject) = M
            self.object = PObject
            {PObject get(name:?self.name
                         host:?self.host)}
         end

         meth get(name:?UniqueName <= _
                  host:?ProverHost <= _)
            UniqueName = self.name
            ProverHost = self.host
         end
         
         meth prove(ProblemBS
                    syntax:    Syntax <= dfg
                    timeOut:   TimeOut
                    replyTo:   ReplyProc
                    replyWith: ReplyWith) = M

            ?Host ?Times ?Result ?Models
            proc {Reply}
               Answer = {Record.mapInd ReplyWith
                         fun {$ Feature Field}
                            case Feature 
			    of result then Result
			    [] host   then Host
                            [] times  then Times
                            [] models then Models
			    else Field
                            end
                         end}
            in
               {ReplyProc Answer}
            end
            
            START = {GetTime}
            TOTAL TotalLocal T
% 	    ProverResult = {self.object prove(ProblemBS $
%                                               syntax:Syntax
%                                               times:?T)}
	    ProverResult = {self.object prove(ProblemBS $
                                              syntax:Syntax
                                              replyWith: reply(state: unit
                                                               name:  unit
                                                               host:  unit
                                                               time:  unit
                                                               ouput: unit
%                                                               models: unit
                                                              ))}
         in
            try
               {WaitOr ProverResult {Alarm TimeOut*1000}}
               TOTAL = {GetTime} - START
               if {IsDet ProverResult} then
                  State
               in
%                  {Browse proverResult(ProverResult)}
                  T = {CondSelect ProverResult time unit}
%                  {Browse time(T)}
                  TotalLocal = T.return2 - T.send1
                  Times = {Adjoin msecs(total:      TOTAL
                                        totalLocal: TotalLocal
                                        both:       T.send+T.return
                                        totalNet:   TOTAL   - TotalLocal
                                        sendrest:   T.send  - T.saveVS
                                        returnrest: T.return-(T.waiting+T.result)
                                       )
                           T}
%                  {Browse times(Times)}
                  Host  = {CondSelect ProverResult host unit}
                  Models= {Value.condSelect ProverResult models "[]"}
%                  if ProverResult.bool %Failed?

                  %  DORIS: valid --> no
                  %       invalid --> yes
                  case {CondSelect ProverResult state unit} 
                  of unit then          Result = error %invalid
                  [] undetermined then  Result = error
                  [] error then         Result = error
                  [] timeout then       Result = timeout
                  [] satisfiable then   Result = invalid
                  [] proof then         Result = valid
                  else
                     Result = error
                  end
                  {Reply}
%        old version:                  
%                   Models= {Value.condSelect X models "[]"}
%                   if X.bool %Failed?
%                   then 
%                      try
%                         case {ByteString.toString X.string} %Proof
%                         of "syntax error\n"   then Result = error
%                         [] "unknown result\n" then Result = error
%                         [] "satisfied\n"      then Result = satisfied
%                         [] "notsatisfied\n"   then Result = notsatisfied
%                         else                       Result = invalid
%                         end
%                      catch _ then
%                         if {IsDet Result}==false then
%                            Result = error
%                         end
%                      end
%                   else
%                      Result = valid
%                   end
%                   {Reply}
               else % the ProverResult is not determined, so we have a timeout
                  Times = msecs(total:      TOTAL
                                totalLocal: TOTAL
                                both:       TOTAL
                                totalNet:   0
                                real:       TOTAL
                                user:       TOTAL
                                sys:        0
                                saveVS:     0
                                sendrest:   0
                                send:       0
                                waiting:    TOTAL
                                result:     0
                                returnrest: 0
                                return:     TOTAL)
                  Host   = self.host
                  Models = "[]"
                  Result = timeout
                  {Reply}
                  {self.object kill}
               end
            catch Err then
               {App.showInfo "[ProverProxy]: Error catched:"}
               {App printExc(Err)}
               try
                  {App.showInfo "[ProverProxy]: trying to kill object..."}
                  {self.object kill}
                  {App.showInfo "[ProverProxy]: object killed"}
               catch SubErr then
                  {App.showInfo "[ProverProxy]: ERROR while killing object:"}
                  {App printExc(SubErr)}
               end
               {App.showInfo "[ProverProxy]: raising Error now..."}
               raise Err end
            end
         end

	 
         meth leave = M
            try {self.object leave}
            catch _ then skip end
            {Thread.terminate {Thread.this}}
         end

         meth ping($)
            unit
         end
      end
      
      class Manager
         feat
            problems
            provers
            idle
            name
            
         meth init(Provers name:PName<=unknown) = M
            self.name     = PName
            self.problems = {New FIFO init}
            self.provers  = Provers
            self.idle     = {New FIFO init}
            {ForAll Provers proc {$ P} {self.idle put(P)} end}
         end
         meth Idle(Prover) = M
            case {self.problems get($)} of false then
               {self.idle put(Prover)}
            elseof Problem then
               Manager, Apply(Prover Problem)
            end
         end
         meth prove(...) = Problem
            if {HasFeature Problem started}==false orelse
               {IsFree Problem.started}
            then
               case {self.idle get($)} of false then
                  {self.problems put(Problem)}
               elseof Prover then
                  Manager, Apply(Prover Problem)
               end
            end
         end
         meth Apply(Prover Problem)
            ProblemStarted = {Value.condSelect Problem started _}
            ProblemSolved  = {Value.condSelect Problem solved _}
         in
            if {IsFree ProblemSolved} then
               try
                  fun {`Record.subtractList` R L}
                     {Record.filterInd R fun {$ F _} {Member F L}==false end}
                  end
		  ProverHost ProverName
                  ProverSignature = self.name#ProverHost#ProverName
               in
                  ProblemStarted = ProverSignature
                  {Prover get(host:?ProverHost name:?ProverName)}  %% these values are never used!
                  {Prover {`Record.subtractList` Problem [started solved]}}
                  thread
                     {Wait {Prover ping($)}}
                     try ProblemSolved = ProverSignature
                     catch failure(...) then skip end
                     {self.serverProc Idle(Prover)}
                  end
               catch failure(...)
               then Manager, Idle(Prover) end
            end
         end 
         meth terminate = M
            {ForAll self.provers
             proc {$ P}
                try {P leave}
                catch _ then skip end
             end}
            {Thread.terminate {Thread.this}}
         end
      end
      
      proc {RequireAll PServices ?PObjects}
	 Broker = ServMan.broker
	 Sync = {NewCell unit}
      in
         %%{App.printInfo "[ATP service]: Requesting "}
	 %%{Record.forAll PServices
	 %% proc {$ PServiceName} 
	 %%    {App.printInfo "`"#PServiceName#"',"}
	 %% end}
	 %%{App.printInfo "..."}
	 PObjects = {Broker requestAll(PServices $ client:'ATP')}
	 %%{App.showInfo ' done.'}
	 {Record.forAllInd PObjects
	  proc {$ PName PList}
	     OldSync NewSync in
	     {Exchange Sync OldSync NewSync}
	     thread
		{ForAll PList Wait}
		%%{App.printInfo "[ATP service]: "#
                %% "Entering '"#PServices.PName#"' services..."}
                local
		   ClientInfo = 'ATP'(name:    'ATP'
				      display: {OS.getEnv 'DISPLAY'})
                in
		   {ForAll PList
                    proc {$ Prover} {Prover enter(client:ClientInfo)} end}
                end
		%%{App.showInfo ' done.'}
		OldSync = NewSync
	     end
	  end}
	 {Wait {Access Sync}}
      end

      %%
      %% Send notification by email if there are different results 
      %%
      proc {MailTo Addr Subj Body}
         Tmp = {OS.tmpnam} in
         {File.saveVS Body Tmp}
         %%{OS.system "mail -s \""#Subj#"\" "#Addr#" <"#Tmp _}
         {OS.system "mail "#Addr#" <"#Tmp _}
         {OS.unlink Tmp}
      end
      
      proc {ProverError ID Results Names Times Hosts Models}
	 {MailTo "afranke@ags.uni-sb.de" "prover error"
          "id     : "#ID#"\n"#
          "results: "#Results#"\n"#
          "provers: "#Names#"\n"#
          "times  : "#Times#"\n"#
          "hosts  : "#Hosts#"\n"#
          "models : "#Models#"\n"}
      end

      %%
      %% select an appropriate syntax %hack% and what a hack (JZ)
      %%
       fun {SelectSyntax PName}
	  {CondSelect SyntaxMap PName dfg} 
       end

       fun {SelectFile PName Files}
	  if {HasFeature Files PName} then
	     case Files.PName
             of '#'(Problem Syntax) then
                Problem#Syntax
             [] Problem then 
                S = {SelectSyntax PName}
             in
                Problem#S
             end
          else
	     S = {SelectSyntax PName} in
	     if {HasFeature Files S} then
		Files.S#S
	     else
		false
	     end
	  end
       end
       
       class SecondLevelCollector
	  attr
	     stack: nil
	     counter: 0
	     limit: 1
	     result: nil
             mode: unit
             
	  meth init(Mode)
	     stack <- nil
	     counter <- 0
             mode <- Mode                
	  end
	  
	  meth fireAt(Limit ?Result)
	     {self init(@mode)}
	     limit <- Limit
	     result <- ?Result
	  end

          meth accumulate(?Result)
             result <- ?Result
          end
          
	  meth push(X)
%             {Browse push(X)}
	     stack <- X|@stack
	     counter <- @counter+1
%             {Browse @counter#'results arrived'}
	     if @counter == @limit then
                if @mode == xml then 
                   @result = {List.toTuple '#' {Reverse @stack}}
                end
             end
	  end
          
          meth pushInc(X)
%             {Browse pushInc(X)}
	     Rest
          in
             @result = X|Rest
             result <- Rest
	     counter <- @counter+1
	  end
       end
       
   in
      
      %%
      %% ATP service class
      %%
      
      class AtpService
	 
         attr
            provers:   unit
	    collector: unit
	    collector2: unit
	    
         meth init(...) = M
            if @provers == unit then
               provers <- {NewDictionary}
            else
               raise atp(init) end
            end
         end

         meth prove(id:      ID              % some integer as an ID
		    problem: Files           % record unit(bliksem: ProblemVS1 spass: ProblemVS2 ...)
                    provers: Provers <= all  %
                    result:  ?Result <= _    % is eventually bound to the first result.
                    mode:    Mode    <= all
                    timeOut: TimeOut <= 300) = M

            SelectedProvers = if Provers == all then
                                 {Arity @provers}
			      else
				 {Filter {Arity @provers}
				  fun {$ PName}
				     {Member PName Provers}
				  end}
                              end
	    OneOnly
	    RLabel    = case Mode
                        of all   then all
                        [] inc   then inc
                        [] any   then oneOf
                        [] first then oneOf
			end
	    ExpectRec = {MakeRecord RLabel SelectedProvers}
         in
	    {ForAll SelectedProvers
	     proc {$ PName} ExpectRec.PName = unit end}
            
	    {@collector expect(ID ExpectRec)}
%            {Browse selected(SelectedProvers)}
	    {ForAll SelectedProvers
             proc {$ PNameVS}
                PName = {VS2A PNameVS}
             in
%                {Browse {SelectFile PName Files}}
		case {SelectFile PName Files} % take a problem for this prover
                of false then                 % there is no appropriate input syntax for this prover
		   {self answer(id:     ID
				result: error
				mode:   Mode
                                name:   PName
                                host:   unknown
                                times:  unknown
                                models: "[]")}
		[] ProblemBS#Syntax then   
		   Query  = prove(ProblemBS
                                  syntax:    Syntax
                                  timeOut:   TimeOut
                                  replyTo:   self.serverProc
				  replyWith: answer(id:     ID
						    result: unit
						    mode:   Mode
						    name:   PName
                                                    host:   unit
						    times:  unit
                                                    models: unit)
                                 )
                in
                   case Mode
		   of all   then {@provers.PName Query}  
		   [] inc   then {@provers.PName Query}  
		   [] any   then {@provers.PName
				  {AdjoinAt Query started OneOnly}}
		   [] first then {@provers.PName
				  {AdjoinAt Query solved  OneOnly}}
		   end
		end
	     end}
	 end

%      ProblemRec == (id:      ID
%                     problem: unit(bliksem: "..." spass: "...")
%                     provers: bliksem#spass#....
%                     mode:    byte string: "all" "any" or "first"
%                     timeOut: integer 
	 meth proveProblem(ProblemRec ?Result)
%	    {Browse ProblemRec}
	    ID =      {CondSelect ProblemRec id unit}
	    Problem = {CondSelect ProblemRec problem unit}
	    Provers = case {CondSelect ProblemRec provers ALL}
		      of !ALL then all
                      [] all then all
                      [] any then any
                      [] first then first
		      [] T then
			 Ts = if {IsList T} then T
                              elseif {IsRecord T} then {Record.toList T}
                              end
                      in
                         {Map Ts
                          fun{$ PNameBS}
                             {VS2A PNameBS}
                          end}
		      end
	    Mode =    {VS2A {CondSelect ProblemRec mode all}}
	    
	    TimeOut = {CondSelect ProblemRec timeout 300}
	 in	    
    	    if (Problem == unit) then
	       raise fault(44 "You must at least give a problem description")
	       end
	    end
%            {Browse waiting_for#{Length {Arity Problem}}#results}
	    case Mode
            of inc then
               {@collector2 accumulate(?Result)}
            [] all then 
               {@collector2 fireAt(1   %{Length {Arity Problem}}  %% wait until all results are available
                                   ?Result)}
            else
               {@collector2 fireAt({Length {Arity Problem}}  %% wait until all results are available
                                   ?Result)}
            end
	    AtpService, prove(id:      ID
			      problem: Problem
			      provers: Provers
			      mode:    Mode
			      timeOut: TimeOut)
	 end
      
	   
%      ProblemRec == unit(mbaseID: ID to Mbase object
% 			  provers: List of provers to be called on the problem
% 			  timeOut: time out for the provers in miliseconds)
	 %% for future use...
% 	 meth proveMbaseProblem(ProblemRec)
	    
% 	    MbaseID = {CondSelect ProblemRec mbaseID unit}
% 	    Provers = {CondSelect ProblemRec provers DefaultProvers}
% 	    TimeOut = {CondSelect ProblemRec timeOut 300}
% 	    if (MbaseID == unit) then
% 	       raise fault(44 "You must give an MBaseID to the problem description")
% 	       end
% 	    end	   
% 	    Broker = ServMan.broker
% 	    MBase = {Broker.request('MBASE' $)}
% 	    if MBase == false then
% 	       raise fault(0 "MBase not found!")
% 	       end
% 	    end
% 	    OMDocProblem = {MBase getAssertion(MbaseID format: xml)}
% 	 end
         
         meth answer(id:     ID
                     result: Result
		     mode:   Mode
		     name:   PName
                     host:   Host
                     times:  Times
                     models: Models <= "[]"
                    )
	    {Stat putTimes(PName Times)}
	    Time = {Value.condSelect Times total 0}%alternatively: user...
	    {Wait Time}
	    Answer = answer(result: Result
			    name:   PName
			    host:   "'"#Host#"'"
			    time:   Time
			    models: Models)
	 in
	    {App.showInfo {XTime.localTimeHHMMSS}
	     #"  No. "#ID#" is "#Result
	     #" ( "#PName#" @ "#Host#" , UserCPU: "#Time#" ms )"}
	    if Mode == first andthen Result == error then
	       Current =  {@collector current($)}
	    in
	       case Current 
	       of nil then skip
	       [] [_#(_#Rec)] then 
		  Number = {Length {Arity {Record.filter {Record.subtract Rec PName}
					   fun{$ X}
					      unit#Y = X
					   in
					      {IsDet Y}
					   end}}}
	       in
		  if Number == 0 then
		     {@collector add(ID PName Answer)}
		  else skip end
	       else skip end
	    else
               {@collector add(ID PName Answer)}
	    end
	 end
         
	 meth returnXml(X 'proc':Return) = M
%	    {Browse M}
            ID#R       = X
            Mode       = {Label R}
            AnswersRec = {Record.map R fun {$ _#A} A end}
            Answers    = {Record.toList AnswersRec}  
         in
            case Answers
	    of nil then
	       {Return unit(id: ID result: {VS2BS error})}
	    [] [P] then
	       {Return unit(id: ID
			    result:  {VS2BS P.result}
			    provers: [{VS2BS P.name}]
			    times:   '#'P.time
			    hosts:    [{VS2BS P.host}]
			   )}
	    else
	       ProversOk ProversErr Results Names Times Hosts % Models
	    in
	       {List.partition Answers
		fun {$ P}
		   {IsDet P.result} andthen
		   {Member P.result [valid invalid satisfied notsatisfied]}
		end
		?ProversOk ?ProversErr}
               
	       case Mode
	       of first then
                  case ProversOk of nil then
                     if {All ProversErr fun {$ P} P.result==timeout end}
                     then
                        Results = timeout
                     else
                        Results = error
                     end
                     Names = {FeaturesXML ProversErr name}
		     Times = {FeaturesXML ProversErr time}
                     Hosts = {FeaturesXML ProversErr host}
%                     Models= "[]"
                  [] FirstProverOk|_ then
                     R = FirstProverOk.result in 
                     if {All ProversOk fun {$ P} P.result==R end} then
                        Results = R % (valid|invalid|satisfied|notsatisfied)
                        local
                           First = {FindFirst ProversOk}
                        in
                           Names = First.name
                           Times = First.time
                           Hosts = First.host
%                           Models= First.models
                        end
                     else
                        Results = {FeaturesXML ProversOk result}
                        Names   = {FeaturesXML ProversOk name}
                        Times   = {FeaturesXML ProversOk time}
                        Hosts   = {FeaturesXML ProversOk host}
%                        Models  = {FeaturesXML ProversOk models}
                     end
                  end
               [] all then
                  if ProversOk == nil then
                     if {All ProversErr fun {$ P} P.result==timeout end}
                     then
                        Results = timeout
                     elseif {All ProversErr fun {$ P} P.result==error end}
                     then
                        Results = error
                     else
                        Results = {FeaturesXML ProversErr result} 
                     end
                     Names = {FeaturesXML ProversErr name}
                     Times = {FeaturesXML ProversErr time}
                     Hosts = {FeaturesXML ProversErr host}
%                     Models= "[]"
                  else
                     if {All ProversOk fun {$ P} P.result==valid end} then
                        Results = valid
                     elseif {All ProversOk fun {$ P} P.result==invalid end}
                     then
                        Results = invalid
                     else
                        Results = {FeaturesXML ProversOk result}
                     end
                     Names = {FeaturesXML ProversOk name}
                     Times = {FeaturesXML ProversOk time}
                     Hosts = {FeaturesXML ProversOk host}
%                     Models= {Features ProversOk models}
                  end
               end
               %%
               %% send email in case of different results
               %%
               case {VirtualString.toString Results}
               of &[|_ then {ProverError ID Results Names Times Hosts nil}
               else skip end
               
	       {Return unit(id: ID
			    result: Results
			    provers: Names
			    times:   {List.toTuple '#' Times}
			    hosts:   Hosts
			   )}
            end
	 end
         
         meth returnOz(X 'proc':Return) = M
%	    {Browse M}
            ID#R       = X
            Mode       = {Label R}
            AnswersRec = {Record.map R fun {$ _#A} A end}
            Answers    = {Record.toList AnswersRec}  
         in
%            {Browse answers(Answers)}
%            {Browse mode(Mode)}
            case Answers
	    of nil then
	       {Return unit(id: ID result: error)}
	    [] [P] then
	       {Return unit(id: ID
			    result:  P.result
			    provers: [P.name]
			    times:   P.time
			    hosts:   [{VS2S P.host}]
			   )}
	    else
	       ProversOk ProversErr Results Names Times Hosts % Models
	    in
	       {List.partition Answers
		fun {$ P}
		   {IsDet P.result} andthen
		   {Member P.result [valid invalid satisfied notsatisfied]}
		end
		?ProversOk ?ProversErr}
               
	       case Mode
	       of first then
                  case ProversOk of nil then
                     if {All ProversErr fun {$ P} P.result==timeout end}
                     then
                        Results = timeout
                     else
                        Results = error
                     end
                     Names = {FeaturesOz ProversErr name}
		     Times = {FeaturesOz ProversErr time}
                     Hosts = {FeaturesOz ProversErr host}
%                     Models= "[]"
                  [] FirstProverOk|_ then
                     R = FirstProverOk.result in 
                     if {All ProversOk fun {$ P} P.result==R end} then
                        Results = R % (valid|invalid|satisfied|notsatisfied)
                        local
                           First = {FindFirst ProversOk}
                        in
                           Names = First.name
                           Times = First.time
                           Hosts = First.host
%                           Models= First.models
                        end
                     else
                        Results = {FeaturesOz ProversOk result}
                        Names   = {FeaturesOz ProversOk name}
                        Times   = {FeaturesOz ProversOk time}
                        Hosts   = {FeaturesOz ProversOk host}
%                        Models  = {FeaturesOz ProversOk models}
                     end
                  end
               elseif {Member Mode [all inc]} then
                  if ProversOk == nil then
                     if {All ProversErr fun {$ P} P.result==timeout end}
                     then
                        Results = timeout
                     elseif {All ProversErr fun {$ P} P.result==error end}
                     then
                        Results = error
                     else
                        Results = {FeaturesOz ProversErr result} 
                     end
                     Names = {FeaturesOz ProversErr name}
                     Times = {FeaturesOz ProversErr time}
                     Hosts = {FeaturesOz ProversErr host}
%                     Models= "[]"
                  else
                     if {All ProversOk fun {$ P} P.result==valid end} then
                        Results = valid
                     elseif {All ProversOk fun {$ P} P.result==invalid end}
                     then
                        Results = invalid
                     else
                        Results = {FeaturesOz ProversOk result}
                     end
                     Names = {FeaturesOz ProversOk name}
                     Times = {FeaturesOz ProversOk time}
                     Hosts = {FeaturesOz ProversOk host}
%                     Models= {Features ProversOk models}
                  end
               end
               %%
               %% send email in case of different results
               %%
               case {VirtualString.toString Results}
               of &[|_ then {ProverError ID Results Names Times Hosts nil}
               else skip end

               {Return unit(id: ID
                            result: Results
                            provers: Names %{Nth I Names}
                            times:   Times %{Nth I Times}
                            hosts:   Hosts %{Nth I Hosts}
                           )}
            end
         end
         meth return(X 'proc':Return) = M
            ID#R       = X
            Mode       = {Label R}
            AnswersRec = {Record.map R fun {$ _#A} A end}
            Answers    = {Record.toList AnswersRec}  
         in
            case Answers
            of nil then
               {Return "answer("#ID#",error,[],[],[],[])"}
            [] [P] then
               {Return "answer("#ID#","#P.result#","#P.name#","#P.time#","
                #P.host#","#P.models#")"}
            else
               ProversOk ProversErr Results Names Times Hosts Models
            in
               {List.partition Answers
                fun {$ P}
                   {IsDet P.result} andthen
                   {Member P.result [valid invalid satisfied notsatisfied]}
                end
                ?ProversOk ?ProversErr}
               
               case Mode
               of first then
                  case ProversOk of nil then
                     if {All ProversErr fun {$ P} P.result==timeout end}
                     then
                        Results = timeout
                     else
                        Results = error
                     end
                     Names = {Features ProversErr name}
                     Times = {Features ProversErr time}
                     Hosts = {Features ProversErr host}
                     Models= "[]"
                  [] FirstProverOk|_ then
                     R = FirstProverOk.result in 
                     if {All ProversOk fun {$ P} P.result==R end} then
                        Results = R % (valid|invalid|satisfied|notsatisfied)
                        local
                           First = {FindFirst ProversOk}
                        in
                           Names = First.name
                           Times = First.time
                           Hosts = First.host
                           Models= First.models
                        end
                     else
                        Results = {Features ProversOk result}
                        Names   = {Features ProversOk name}
                        Times   = {Features ProversOk time}
                        Hosts   = {Features ProversOk host}
                        Models  = {Features ProversOk models}
                     end
                  end
               [] all then
                  if ProversOk == nil then
                     if {All ProversErr fun {$ P} P.result==timeout end}
                     then
                        Results = timeout
                     elseif {All ProversErr fun {$ P} P.result==error end}
                     then
                        Results = error
                     else
                        Results = {Features ProversErr result} 
                     end
                     Names = {Features ProversErr name}
                     Times = {Features ProversErr time}
                     Hosts = {Features ProversErr host}
                     Models= "[]"
                  else
                     if {All ProversOk fun {$ P} P.result==valid end} then
                        Results = valid
                     elseif {All ProversOk fun {$ P} P.result==invalid end}
                     then
                        Results = invalid
                     else
                        Results = {Features ProversOk result}
                     end
                     Names = {Features ProversOk name}
                     Times = {Features ProversOk time}
                     Hosts = {Features ProversOk host}
                     Models= {Features ProversOk models}
                  end
               end
               %%
               %% send email in case of different results
               %%
               case {VirtualString.toString Results}
               of &[|_ then {ProverError ID Results Names Times Hosts Models}
               else skip end
               
               {Return "answer("#ID#","#Results#","#Names#","#Times#","#Hosts#","#Models#")"}
            end
         end
         
         meth returnRestAndLeave('proc':Return done:Done<=_)
            try
               {ForAll {@collector current($)}
                proc {$ ID#(_#R)}
                   R1 = {Record.filter R fun {$ _#Y} {IsDet Y} end} in
                   if {IsAtom R1}==false then
                      {self return(ID#R1 'proc':Return)} 
                   end
                end}
               {self leave}
            finally
               Done=unit
            end
         end
         
         meth enter(return:    Return   <= unit              %% should be a unary procedure
                    mode:      Mode     <= xml
                    services:  Services <= {Arity PServices}
                    done:      ?Done    <= _         ...)
            ServicesRec = {Record.make services {Filter Services IsAtom}}
            MyPServices = {Record.filterInd PServices
                           fun {$ PName _} {HasFeature ServicesRec PName} end}
            provers <- {Record.mapInd {RequireAll MyPServices}
			fun {$ PName PObjects}
			   fun {MkProxy PObject}
			      {App newServer(ProverProxy init(PObject) $
					     info: MyPServices.PName#".proxy")}
			   end
			   ProverProxys = {Map PObjects MkProxy}
                        in
			   if ProverProxys == nil then
			      nil
			   else
			      {App newServer(Manager init(ProverProxys name:PName)
					     $
					     info: MyPServices.PName#".manager")}
			   end
			end}
	    provers <- {Record.filter @provers fun{$ P} P\=nil end}
	    collector2 <- {New SecondLevelCollector init(Mode)}

	    proc {DoReturn Answers}
	       if Return == unit then
		  case Mode
                  of xml then
                     {self returnXml(Answers 'proc': proc{$ X}
                                                        {@collector2 push(X)}
                                                     end)}
                  else
                     {self returnOz(Answers 'proc': proc{$ X}
                                                       {@collector2 pushInc(X)}
                                                    end)}
                  end
               else
		  {self return(Answers 'proc': Return)}
	       end
	    end
         in
	    collector <- {New Collector init(DoReturn)}
	    
	    Done = unit
	    %%{App newServer(Collector init(DoReturn) $
	    %%               detach: false
	    %%               info:   'result-collector')}
	 end
         
	 meth availableProvers($) = M
	    {Map {Arity @provers}
	     fun {$ P}
		case P
                of bliksem then "'Bliksem 1.03'"
		[] spass   then "'Spass 0.97'"
		[] fdpll   then "'FDPLL 0.9.8'"
		[] otter   then "'Otter 3.0.5'"
		[] mace    then "'MACE'"
		[] satchmo then "'SATCHMO'"
		[] kimba   then "'KIMBA'"
		[] chorus  then "'CHORUS'"
                else "'"#P#"'"
                end
             end}
         end
         
         meth leave(...) = M
            {Record.forAll @provers proc {$ PManager} {PManager terminate} end}
            provers   <- unit
            collector <- unit
            %%{@collector terminate}
            {Thread.terminate {Thread.this}}
         end
         
      end

   end

   /*
   class LimitedCounter
      prop
         locking
      attr
         val:0
         max:1
         signal:unit
      meth init(Max)
         val <- {Min 0 Max}
         max <- Max
         LimitedCounter, signalMax
      end
      meth signalMax
         if @val==@max then
            signal <- _
         end 
      end
      meth get($)
         lock @val end
      end
      meth inc
         lock
            val<-@val+1
            LimitedCounter, signalMax
         end
      end
      meth dec
         lock
            val<-@val-1
            @signal=unit
         end
      end
      meth signal($)
         lock @signal end
      end
   end
   
   class Limit
      prop
         locking
      attr
         counter: unit
      meth init(Max)
         counter <- {New LimitedCounter init(Max)}
      end
      meth inc
         lock
            {Wait {@counter signal($)}}
            {@counter inc}
         end
      end
      meth dec
         {@counter dec}
      end
   end
   */
   
   ServMan = {New ServiceManager init(ServiceName AtpService
                                      appName: AppName
                                      appMan:  App)}
   
   {ServMan offerService}   
   {App.exit 0}
end







