%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% bliksem.oz ; This file is part of the MathWeb system
%
%   major updates: 09.08.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% BLIKSEM service
%%%

functor
   
import
   OS(getEnv /*tmpnam*/)
%   Browser(browse: Browse)
   
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'bliksem'
   ServiceName = 'BLIKSEM'
   App         = {New XApplication.manager init}

   ProofBegin = "found a proof!"
   Saturation = "found a saturation!"       %% neg-satisfiable or satisfiable?
   Unknown    = "Did not understand the following:"
   
   BLIKSEM = {OS.getEnv "BLIKSEMHOME"}
   DFG2BLK = {OS.getEnv "DFG2BLK"}
   
   class BliksemService from ShellProver
      feat
         name: bliksem
         
      meth submitProblem(ProblemBS % a ByteString
                         syntax: Syntax <= bliksem) = M
         BliksemIn  = "bliksem.in"  %{OS.tmpnam}
         BliksemOut = "bliksem.out" %{OS.tmpnam}
         CallBliksem = BLIKSEM#" <"#BliksemIn#" >"#BliksemOut#" 2>&1"
      in
         case Syntax
         of dfg then
            if DFG2BLK == false then
               {App.showError 'DFG2BLK not found.'}
               raise dfg2blk(not_found) end
            end
            ProblemDFG = "problem.dfg" %{OS.tmpnam}
            CallDfg2Blk = DFG2BLK#" <"#ProblemDFG#" >"#BliksemIn
         in
            {self saveVS(ProblemBS ProblemDFG)}
            {self exec(CallDfg2Blk)}
            {self start(CallBliksem)}
            %%tmpfiles <- [ProblemDFG BliksemIn]
         [] bliksem then
            {self saveVS(ProblemBS BliksemIn)}
            {self start(CallBliksem)}
            %%tmpfiles <- [BliksemIn]
         end
         %%outfile <- BliksemOut
      end
      
      meth analyzeResult($)
         {self save}
	 F 
         fun {ProofFound}
            case {F getS($)}
            of false       then undetermined
            [] !Saturation then satisfiable      %% is that right?
            [] !ProofBegin then proof
            [] !Unknown    then error
            [] &s|&y|&n|&t|&a|&x|& |&e|&r|&r|&o|&r|_ then error
            else {ProofFound}
            end
         end
      in
	 try
            ProofOk Result            
         in
            F = {self openRead("bliksem.out" /*@outfile*/ $)}       
	    ProofOk = {ProofFound}
            Result = atp_result(state: ProofOk
                                name:  self.name
                                host:  self.host
                                output: {self readAll(F $)}
                               )
	    {F close}
            Result
%            s(string:{ByteString.make Proof} bool:Failed? host:self.host)
	 catch _ then Output in
	    try
               Output = {self readAll(F $)}
               {F close}
            catch _ then skip end
	    atp_result(state: error
                       name:  self.name
                       host:  self.host
                       output: if {IsDet Output} then Output else nil end
                      )
	 end
      end
   end
   ServMan = {New ServiceManager init(ServiceName BliksemService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end

% 	       case {F getS($)}
% 	       of false       then unknown
% 	       [] !Saturation then false
% 	       [] !ProofBegin then true
% 	       [] &s|&y|&n|&t|&a|&x|& |&e|&r|&r|&o|&r|_ then error
% 	       else {ProofFound}
% 	       end
% 	    end
% 	    ProofOk = {ProofFound}
% 	    ?Failed = {Member ProofOk [false error unknown]}
% 	 in
% 	    case ProofOk
% 	    of true    then Proof = {self readAll(F $)}
% 	    [] false   then Proof = "-------------------------------\n"
%             [] error   then Proof = "syntax error\n"
% 	    [] unknown then Proof = "unknown result\n"
% 	    end      
