%%%
%%% FDPLL agent
%%%

functor
   
import
   OS(getEnv)
   Browser(browse: Browse)
   
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'fdpll'
   ServiceName = 'FDPLL'
   App         = {New XApplication.manager init}

   FDPLL         = {OS.getEnv "FDPLLHOME"}
   FDPLLSTRATEGY = {OS.getEnv "FDPLLSTRATEGY"}
   TIME_LIMIT    = 300
   DFG2TME       = {OS.getEnv "DFG2TME"}
   TME2X         = {OS.getEnv "TME2X"}

   FileBase = "problem"
   
   class FdpllService from ShellProver
      feat
         name: fdpll
% example call for solaris:
% cat p1.dfg | /project/omega/atp/bin/solaris/dfg2tme > p1.tme
% /project/omega/atp/bin/solaris/tme2X tmeeqax p1
%  /project/omega4/atp/fdpll/bin/fdpll -q -a -H /project/omega4/atp/fdpll/lib/fdpll/bottom_up.fdpll -t 40 p1-eqax | tee problem.prot

      meth submitProblem(ProblemBS syntax:Syntax<=dfg)
         
         case Syntax of dfg then
            
            try
               {self saveVS(ProblemBS FileBase)}
               
               %% convert to tme syntax              :  *     --> *.tme
               %% add equality axioms (if neccessary):  *.tme --> *-eqax.tme
               
               {self exec("cat "#FileBase#" | "#DFG2TME#" >"#FileBase#".tme"
                          #" && "#TME2X#" tmeeqax "#FileBase)}
               
               %% implicit eqtrafo:    *-eqax.tme --> *-eqax-eqt.tme
               %% fdpll prover                    --> *.prot
               
               {self start(FDPLL#" -q -a -H "#FDPLLSTRATEGY#" -t "#TIME_LIMIT
                           #" "#FileBase#"-eqax | tee "#FileBase#".prot")}
            catch E then
               {Browse E}
            end
         else
            raise unknownSyntax(Syntax) end
         end
      end

      meth analyzeResult($)
         
         %% *.prot should contain one line beginning like this:
         %%     Result...................: REFUTATION FOUND
         %%     Result...................: SATISFIABLE with model...

         try
            {self exec("grep 'Result...................: ' "#FileBase#".prot"#
                       " | awk '{printf $2}' >"#FileBase#".result")}
            FL = {self loadS(FileBase#".result" $)}
            State = case FL
                    of "REFUTATION"  then proof
                    [] "SATISFIABLE" then satisfiable
                    else undetermined
                    end
            
            %% cleanup
            {self exec("rm -f "
                       #FileBase#".tme "
                       #FileBase#"-eqax.tme "
                       #FileBase#"-eqax-eqt.tme "
                       #FileBase#".prot"
                       #FileBase#".result"
                      )}
         in
            atp_result(state: State
                       host:  self.host
                       name:  self.name
                       output: "dummy output")
         catch _ then
            atp_result(state: error
                       host:  self.host
                       name:  self.name
                       output: "an error occured")
         end
      end       
   end
   
   ServMan = {New ServiceManager init(ServiceName FdpllService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end

%       meth analyzeResult($)
       
%          %% *.prot should contain one line beginning like this:
%          %%     Result...................: REFUTATION FOUND
%          %%     Result...................: SATISFIABLE with model ...
         
%          try
%             {self exec("grep 'Result...................: ' "#FileBase#".prot"#
%                        " | awk '{printf $2}' >"#FileBase#".result")}

%             ProofOk = case {self loadS(FileBase#".result" $)}
%                       of "REFUTATION"  then true
%                       [] "SATISFIABLE" then false
%                       else unknown
%                       end
            
%             Proof = case ProofOk
%                     of true    then "proof found\n"
%                     [] false   then "-------------------------------\n"
%                     [] unknown then "unknown result\n"
%                     end
            
%             %% cleanup
            
%             {self exec("rm -f "
%                        #FileBase#".tme "
%                        #FileBase#"-eqax.tme "
%                        #FileBase#"-eqax-eqt.tme "
%                        #FileBase#".prot"
%                        #FileBase#".result"
%                       )}
%          in
%             s(string: {ByteString.make Proof}
%               bool  : ProofOk\=true
%               host  : self.host)
%          catch _ then
%             s(string: "-------------------------------\n"
%               bool  : false
%               host  : self.host)
%          end
%       end       
