%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% geoff.oz ; This file is part of the MathWeb system
%
%   major updates: 24.08.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% Geoff's SystemOnTPTP online service wrapped in a MathWeb service shell
%%%

functor
   
import
   OS Open
   Browser(browse: Browse)
   System
   
   HttpClient
   at 'x-ozlib://mathweb/httpd/HttpClient.ozf'
   XURL
   at 'x-ozlib://mathweb/share/XURL.ozf'

   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   SimpleServiceClass('class': SimpleService)
   at 'x-ozlib://mathweb/sys/SimpleService.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   BufferClass(simple: Buffer)
   at 'x-ozlib://mathweb/share/Buffer.ozf'
%    ProversList(provers: Provers)
%    at 'x-ozlib://mathweb/atp/SystemsOnTPTP.ozf'
   
define
   VS2S = VirtualString.toString
   VS2A = VirtualString.toAtom
   
   AppName     = 'system-on-TPTP-service'
   ServiceName = 'GEOFF'
   App         = {New XApplication.manager init}

   ArgSpec = record(kqmlagent(single char:[&a] type:bool default: false))
   Args    
   try
      Args = {App.getCmdArgs ArgSpec}
   catch error(ap(usage S)...) then
      {App.showError 'otter: '#S}
      {App.exit 1}
   end
   
   GeoffURL = case {OS.getEnv "SYSTEM-ON-TPTP-URL"}
              of false then
                 "http://www.cs.jcu.edu.au/cgi-bin/tptp/SystemOnTPTPFormReply"
              [] T then T end
   
   MyURL = "http://www.ags.uni-sb.de/~jzimmer/cgi-bin/env.cgi"
   
   DefaultTimeLimit = 100
   ListSystemsQueryURL = GeoffURL#'?'#{HttpClient.recordToQuery 
                                       flags('QuietFlag': "-q2"
                                             'SubmitButton': "ListSystems"
                                             'NoHTML': 1)}

   Provers
   
   try
      Res = {XURL.loadVS ListSystemsQueryURL}
   in
      case Res
      of nil then
         Provers = nil
      [] R then
         StrList = {String.tokens R &\n $}
      in
      Provers = {Map StrList VS2A}
      else
         Provers = nil
      end
   catch _ then
      {App.showInfo '*****************************************************************'}
      {App.showInfo 'GEOFF: SystemOnTPTP service not reachable :-('}
      {App.showInfo '*****************************************************************'}
      {App.exit 1}
   end

   FixedFlags = fixed('UPLOADProblem': nil
                      'ReportPassword': nil
                      'CPUPassword': nil
                      'ReportFlag': "-q2"     
                      'QuietFlag': "-q2"
                      'AutoMode': "-cE"
                      'AutoModeSystemsLimit': 3
                      'Completeness': "Completeness"
                      'NoHTML': 1
                     )
   FixedQuery = {HttpClient.recordToQuery FixedFlags}
   
   OtherFlags = flags('ProblemSource': "TPTP"
                      'TPTPProblem': "PLA001-1"
                      'FORMULAEProblem': nil
                      'AutoModeTimeLimit': 300
                      'SubmitButton': "RunSelectedSystems"
                     )
   
% Waldmeister---601   system information being retrieved
% Waldmeister---601's non-default parameters being retrieved
% Waldmeister---601's SYN054-1.p being prepared by tptp2X
% Waldmeister---601   being executed on SYN054-1
% RESULT: SYN054-1 - Waldmeister---601 says Unknown - CPU = 0 WC = 0 
   
   fun {ReadSpaces S} 
      case S of nil then nil
      [] C|Sr then
         if {Char.isSpace C}
         then {ReadSpaces Sr}
         else S
         end
         end
   end
   proc {ReadTillSpace S ?Str ?Rest}
      case S of nil then Str=nil Rest=nil
      [] C|Sr then
         if {Char.isSpace C}
         then Str=nil Rest=S 
         else Str=C|{ReadTillSpace Sr $ ?Rest}
         end
      end
   end
   
% RESULT: SYN054-1 - Bliksem---1.12 says Theorem - CPU = 0 WC = 0  Terms = 152 Clauses = 470 Generated = 9 Kept = 9
   fun{ReadResult Output}
      ProbName Rest1 Rest2 Rest3 Prover State Answer
   in
      {String.token Output & ProbName Rest1}
      {String.token Rest1 & _ Rest2}
      {String.token Rest2 & Prover Rest3}
      case Rest3
      of &s|&a|&y|&s|& |&T|&h|&e|&o|&r|&e|&m|O then
         State = proof
         Answer = O
      [] &s|&a|&y|&s|& |&U|&n|&k|&n|&o|&w|&n|O then
         State = undetermined     
         Answer = O
      else
         State = undetermined     %% ToDo, what other states can be there?
         Answer = Rest2
      end
      atp_result(state: State
                 name: {VS2A Prover}
                 host: unknown
                 output: Answer)
   end
   
   fun{ReadResults Text}
      S = {Text getS($)}
   in
      case S 
      of false  then nil
      [] nil    then nil
      [] &E|&R|&R|&O|&R|&:|& |Output then
         atp_result(state: error
                    output: Output)
      [] &R|&E|&S|&U|&L|&T|&:|& |Rest then
         {ReadResult Rest}|{ReadResults Text}
      else
         {ReadResults Text}
      end
   end
   
   
%%%===========================================
   class SystemOnTPTPService from SimpleService
      feat
         name: geoff
      attr
         provers
         
%          Problem == prob(flags: f('ProblemSource': "TPTP"
%                                   'TPTPProblem': "SYN054-1"
%                                   'FORMULAEProblem': nil
%                                   'AutoModeTimeLimit': 300
%                                   'SubmitButton': "RunSelectedSystems")
%                          provers: ['Bliksem---1.12'
%                                    'SPASS---1.03'
%                                    'Otter---3.2'
%                                    'Waldmeister---601'])
         
      meth prove(Problem    ?Result
                 syntax:    Syntax    <= tptp
                 replyWith: ReplyWith <= unit(state: _)
                 timeout:   TimeOut   <= unit)
         
         
         NewFlags
         NewFlags = case {CondSelect Problem flags unit}
                    of unit then
                       OtherFlags
                    [] Fs then 
                       {Record.mapInd OtherFlags
                        fun{$ Feat Val}
                           case {CondSelect Fs Feat unit}
                           of unit then Val
                           [] V then V end
                        end}
                    end
         NewQuery = {HttpClient.recordToQuery NewFlags}
         
         NewProvers = case {CondSelect Problem provers unit}
                      of unit then
                         Provers
                      [] Ps then
                         PAtoms = {Map Ps VS2A}
                      in
                         {Filter Provers
                          fun{$ P}
                             PName = {Label P}
                          in
                             {Member PName PAtoms}
                          end}
                      end
         
         ProversList = {FoldR
                        NewProvers
                        fun{$ Prover Rest}
                           Name = {Label Prover}
                        in
                           {Append [('System___'#Name)#Name
                                    ('TimeLimit___'#Name)#{CondSelect Prover timelimit DefaultTimeLimit}
%    the default value are perfect for these:
%                                    ('Transform___'#Name)#Prover.transform
%                                    ('Format___'#Name)#Prover.format
%                                    ('Command___'#Name)#Prover.command
                                   ]
                            Rest}
                        end
                        nil}
         
         ProversQuery = {HttpClient.pairListToQuery ProversList}
         
         Query = '?'#FixedQuery#"&"#NewQuery#"&"#ProversQuery
         {App.showInfo Query}
         Time1 = {OS.time}
         Res = {XURL.loadVS GeoffURL#Query}
         
%   Result = {HttpClient.post request(url:  MyURL
         {Wait Res}
         Time2 = {OS.time}
         Text = {New Buffer init}
      in
         {Text put(Res)}

         Result = {ReadResults Text}
         {App.showInfo "time: "#(Time2 - Time1)}
      end
      
   end % class
%%%===========================================
   
   if {IsFree Provers} orelse Provers == nil then
      {App.showInfo '\n*****************************************************************'}
      {App.showInfo 'GEOFF: No provers available for SystemOnTPTP :-('}
      {App.showInfo '*****************************************************************'}
      {App.exit 1}
   else
      ServMan = {New ServiceManager init(ServiceName
                                         SystemOnTPTPService
                                         appName: AppName
                                         appMan:  App
                                         argSpec: ArgSpec
                                        )}
   in
      {ServMan offerService}
      {App.exit 0}
   end
end
   
%   {Browse Result}
%    Result = {HttpClient.post request(url:  GeoffURL
%                                      headers: h('content-type': 'multipart/form-data; boundary='#Boundary)
%                                      content: Content)}
%   {System.showInfo Result.content}
