%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% otter.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% The OTTER service
%%%

functor
   
import
   OS(getEnv)

   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   AtpAgent('class': ATPAgent)
   at 'x-ozlib://mathweb/kqml/AtpAgent.ozf'
   
define
%   VS2S = VirtualString.toString

   AppName     = 'otter'
   ServiceName = 'OTTER'
   App         = {New XApplication.manager init}

   ProofBegin1 = "Proof object:"
   ProofEnd1   = "End of proof object."
   ProofBegin2 = ";; BEGINNING OF PROOF OBJECT"
   ProofEnd2   = ";; END OF PROOF OBJECT"
   
   OTTER     = {OS.getEnv "OTTERHOME"}
   DFG2TPTP  = {OS.getEnv "DFG2TPTP"}
   TPTP2X    = {OS.getEnv "TPTP2X"}
   DFG2OTTER = {OS.getEnv "DFG2OTTER"} 
   
   %  Special CMD-Args because of Kqml agent mode
   ArgSpec = record(kqmlagent(single char:[&a] type:bool default: false))
   Args    
   try
      Args = {App.getCmdArgs ArgSpec}
   catch error(ap(usage S)...) then
      {App.showError 'otter: '#S}
      {App.exit 1}
   end
   
   class OtterService from ShellProver
      feat
         name: otter
         
      meth submitProblem(ProblemBS % a ByteString
			 syntax: Syntax <= otter)
	 OtterIn  = "otter.in"  %{OS.tmpnam}
	 OtterOut = "otter.out" %{OS.tmpnam}
         CallOtter = OTTER#" <"#OtterIn#" >"#OtterOut#" 2>&1"
      in
         case Syntax
         of dfg then
            if DFG2OTTER == false then
               {App.showError 'DFG2OTTER not found.'}
               raise dfg2otter(not_found) end
            end
            ProblemDFG = "problem.dfg"  %{OS.tmpnam}
            ProblemTPTP= "problem-tptp" %{OS.tmpnam}
%%            CallDfg2Otter = DFG2OTTER#" "#ProblemDFG#" "#OtterIn
            CallDfg2Tptp  = DFG2TPTP#" "#ProblemDFG#" "#ProblemTPTP
            CallTptp2Otter= TPTP2X#" -fotter -d"#@tmpdir#"/ "#ProblemTPTP  %% -> problem+rm_eq_stfp.in 
            OtterInVS
            %%CallMove      = "mv "#ProblemTPTP#"+rm_eq_stfp.in "#OtterIn
         in
            {self saveVS(ProblemBS ProblemDFG)}
            %%            {self exec(CallDfg2Otter)}
            {self exec(CallDfg2Tptp)}
            {self exec(CallTptp2Otter)}
            %%{self exec(CallMove)}
            {self loadS(ProblemTPTP#"+rm_eq_stfp.in" ?OtterInVS)}
            {self saveVS("set(build_proof_object).\n"#OtterInVS OtterIn)}
            {self start(CallOtter)}
            %%tmpfiles <- [ProblemDFG OtterIn ProblemTPTP#"+rm_eq_stfp.in"]
         [] otter then
            {self saveVS("set(build_proof_object).\n"#ProblemBS OtterIn)}
            {self start(CallOtter)}
            %%tmpfiles <- [OtterIn]
         end 
         %%outfile  <- OtterOut
      end
      
      meth analyzeResult($)
         F
         fun {ProofFound}
            case {F getS($)}
            of false        then undetermined
            [] !ProofBegin1 then proof
            [] !ProofBegin2 then proof
            else {ProofFound}
            end
         end
         fun {Take}
            case {F getS($)}
            of false then nil
            [] !ProofEnd1 then ProofEnd1#"\n"
            [] !ProofEnd2 then ProofEnd2#"\n"
            [] &S|&e|&a|&r|&c|&h|& |&s|&t|&o|&p|&p|&e|&d|_ then '\n'#{Take}    %% comes from 2>&1 in the sh
            [] _|&S|&e|&a|&r|&c|&h|& |&s|&t|&o|&p|&p|&e|&d|_ then '\n'#{Take}  %% sometimes there is a ^G ... :-(
            [] Line  then Line#'\n'#{Take}
            end
         end
      in
         try
            State Result
         in
            F = {self openRead("otter.out" $)}
            State = {ProofFound}
            Result = atp_result(state: State
                                name:  self.name
                                host:  self.host
                                output: (if State == proof then "Proof object:\n" else nil end)#{Take}
                               )
            {F close}
            Result
	 catch _ then Output in
	    try
               Output = {Take}
               {F close}
            catch _ then skip end
	    atp_result(state: error
                       name:  self.name
                       host:  self.host
                       output: if {IsDet Output} then Output else nil end
                      )
	 end
      end
   end
   
   OtterAgent = {ByNeed fun{$}
			   class $
			      from OtterService ATPAgent
				 
			      meth init(url:    BrokerURL
					ticket: BrokerTicket
					cmdArgs: CmdArgs)
				 OtterService, init(url:     BrokerURL
						    ticket:  BrokerTicket
						    cmdArgs: CmdArgs)

				 ATPAgent, init(app:  App
						name: ServiceName
						port: unit
					       )
			      end
			   end
			end}
   ServiceClass = if Args.kqmlagent then
		     OtterAgent
		  else
		     OtterService
		  end
   
   ServMan = {New ServiceManager init(ServiceName
				      ServiceClass
				      appName: AppName
				      appMan:  App
				      argSpec: ArgSpec
				     )}
   {ServMan offerService}
   {App.exit 0}
end

%       meth analyzeResult($)
%          Proof
%          local
%             F = {self openRead("otter.out" $)}
%             fun {Skip}
%                case {F getS($)}
%                of false        then true
%                [] !ProofBegin1 then false
%                [] !ProofBegin2 then false
%                else {Skip}
%                end
%             end
%             fun {Take}
%                case {F getS($)}
%                of false      then nil
%                [] !ProofEnd1 then ProofEnd1#"\n"
%                [] !ProofEnd2 then ProofEnd2#"\n"
%                [] Line       then Line#"\n"#{Take}
%                end
%             end
%          in
%             Failed? = {Skip}
%             if Failed?
%             then Proof = "-------------------------------\n"
%             else Proof = "Proof object:\n"#{Take}
%             end
%             {F close}
%          end
%       in
%          s(string:{ByteString.make Proof} bool:Failed? host:self.host)
%       end
%    end
