%%%
%%% PVS agent
%%%

functor
   
import
   System(printInfo) OS(getEnv) 

   %% If we need the browser for debugging purposes:
   %%   Browser(browse)
   
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'

   CasService('class':     CASService)
   at 'x-ozlib://mathweb/sys/CasService.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'

define
   AppName     = 'pvs'
   ServiceName = 'PVS'
   App	 = {New XApplication.manager init}
   
   PVSHOME     = {OS.getEnv 'PVSHOME'}
   if PVSHOME==false then
      raise undefinedVariable('PVSHOME') end
   end
   PVS  = PVSHOME#"/pvs"
   
   Cmd     = PVS
   Args    = ["-raw"]
   Prelude0 = "(load \""#PVSHOME#"/pvs-loader.lsp\")"
   Prelude1 = "(in-package pvs)"
   
   VS2BS = VirtualString.toByteString
   
   fun {DropSpaces S} {List.dropWhile S Char.isSpace} end
   fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end
   %%fun {NoNewlineOrBackslash C} C\=&\n andthen C\=&\\ end
   fun {IsAlphaOrSpace C} {Char.isAlpha C} orelse {Char.isSpace C} end
   fun {MkError S} error({VS2BS S}) end
   
   class PvsService from CASService
		       
      meth startCAS
	 CASService, startCAS(cmd:Cmd args:Args
			      suffix: "\n"
			      prompts: ["PVS: " 
					"\n(Yes or No) "
					":end-pvs-eval\n"#50
					"USER: "
				       ]
			      err:    "error"
			      prelude: Prelude0
			      printInfo: System.printInfo)
	 {self eval("t" _)}
	 {self eval("t" _)}
	 {self eval("t" _)}
	 {self eval(Prelude1 _)}
      end
      /*
      meth preProcess(ExprVS $)
	 ExprVS
      end
      */
   
      %%/*
      meth postProcess(S $)
	 
	 Errchkstr
	 Errchkres
  
	 fun {Pvserrchk Strs}
	    {List.isPrefix "rre-svp-dne:" Strs}
	 end

	 fun{IsT X}
	    if x == true then true else false end
	 end

      in
	 {String.tokens {Reverse S} &\n ?Errchkstr}
	 {List.map Errchkstr Pvserrchk ?Errchkres}
	 if {List.all Errchkres IsT}
	 then
	    %{System.printInfo S}
	    %{System.printInfo "Error string recognised! :end-pvs-err\n"}
	    {MkError S}
	 else
	    %%{VS2BS {Clear S}}
	    S
	 end
      end
      %%*/
      
      meth eval(Expr ?Returned)
	 CASService, eval(Expr ?Returned)
	 {Wait Returned}
      end
      
      meth write(Expr)
	 {@cas write(Expr)}
      end
      
      meth prove(Expr  $
		 tactic:  Tactic  <= "grind"
		 library: Library <= "top_analysis"
		 trueorfalse: Trueorfalse <= false)
	 
	 %% Wrap conjecture in appropriate PVS command.
	 %% The tactic, library and trueorfalse variables
	 %% allow for other variants to be defined on top of this.
	 InputVS =
	 "(pvs::prove-formula-decl \"test: conjecture "#Expr#"\""
	 #"\n"#"'|"#Library#"| '(\"\" ("#Tactic#")))"
	 InputNVS =
	 "(pvs::prove-formula-decl \"test: conjecture NOT("#Expr#")\""
	 #"\n"#"'|"#Library#"| '(\"\" ("#Tactic#")))"
	 QED
	 QEN
	 Result
	 
	 proc {DoThread}
	    %%Input the conjecture with wrapping as applied above.
	    %%	 {Browser.browse InputVS}
	    {self eval(InputVS _)}
	    
	    {self eval("t" ?QED)}
	    
	    {self eval("t" _)}
	    
	    %% If PVS has not succeeded in proving the conjecture, it will
	    %% ask: Would you like the partial proof to be saved?
	    %% (***Old proof will be overwritten.***)
	    %% (Yes or No) so we must check for this question and
	    %% send "no" if it is asked. The only weay currently to
	    %% check for this question if to see if the proof attempt
	    %% failed: i.e. no Q.E.D. so that is what we look for.
	    

		  case QED 
     		  of error(Str) then	 
		     {App.showInfo Str}
		  else
		     if {List.isPrefix ".D.E.Q"
			 {String.token {DropSpaces {Reverse QED}} &\n _}}
		     then
			{self eval("t" _)}
			{self eval("t" _)}
			%%If we found a Q.E.D. then we return TRUE.
			Result = "True"
			
			%% If we didn't find a Q.E.D. then we seer if "trueorfalse" has
			%% been set to true. If "trueorfalse" is set to true, we try to
			%% prove the negation of the conjecture. If not, we return
			%% "Unknown.".
		     else
			{Delay 500}
			{self eval("no" _)}
			{self eval("t" _)}
			{self eval("t" _)}
			if Trueorfalse
			then
			   {self eval(InputNVS _)}
			   {self eval("t" ?QEN)}
			   {self eval("t" _)}
			   case QEN 
			   of error(Str) then	 
			      {App.showInfo Str}
			   else
			      if {List.isPrefix ".D.E.Q"
				  {String.token {DropSpaces {Reverse QEN}} &\n _}}
			      then
				 {self eval("t" _)}
				 {self eval("t" _)}
				 %%If we found a Q.E.D. then we return FALSE.
				 Result = "False"
			      else
				 {Delay 500}
				 {self eval("no" _)}
				 {self eval("t" _)}
				 {self eval("t" _)}
				 Result = "Unknown."
			      end
			   end	
			else
			   Result = "Unknown."
			end
		     end
		     %% Either after saying "don't save partial proof" or after a
		     %% successful proof attempt, PVS reports that it's ready and
		     %% needs to be told to proceed before we get the top-level
		     %% prompt back, so in both cases above we evaluate "t" twice.
		     %% All these extra tokens are needed because we're inserting at
		     %% the level below the standar emacs interface which sends such
		     %% tokens to indicate it is ready.
		  end
	 end
      in
	 {App newThread(DoThread info:'pvs_prove')}         
	 
	 %% Return Result
	 Result
      end
      
      meth provetf(Expr $)
	 {self prove(Expr $ trueorfalse:true)}
      end
      
      meth provects(Expr $)
	 {self prove(Expr $ tactic:"cts")}
      end
      
      meth leave
	 {WaitOr CASService, eval(":exit" $) {Alarm 10000}}
	 CASService, leave
      end
      
   end
   
   ServMan = {New ServiceManager init(ServiceName PvsService
				      appName: AppName
				      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end
