%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% tps-agent.oz ; This file is part of the MathWeb system
%
%   major updates: 04.08.2001
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@mathweb.org
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% This file contains TPS server
%%%

functor

import 
   Error(exceptionToMessage messageToVirtualString)
   OS(getEnv tmpnam unlink time)
   Resolve(localize)
   Property
%   System
   Browser(browse:Browse)
   
   
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XOpen(server:Stream pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   File
   at 'x-ozlib://mathweb/share/File.ozf'
   Process(sigF)
   at 'x-ozlib://mathweb/share/Process.ozf'

   LispSocket(manager: Lisp2OzSocket)
   at 'x-ozlib://mathweb/sys/LispSocket.ozf'
   
   XService(threadManager: ThreadManager
	    clientManager: ClientManager
	    faultManager:  FaultManager           
	    manager:       ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'
   
   OM2TPS(om2Tps: OM2Tps)
   at 'x-ozlib://mathweb/omutils/OM2Tps.ozf'

   %%----------------------------------
   %% KQML agent functors
   %%----------------------------------
   KqmlAgent('class': KQMLAgent)
   at 'x-ozlib://mathweb/kqml/KqmlAgent.ozf'
   KqmlLispSocket('class': KqmlLisp2OzSocket)
   at 'x-ozlib://mathweb/kqml/KqmlLispSocket.ozf'
   
define
   VS2S = VirtualString.toString
%   VS2A = VirtualString.toAtom
   
   {Property.put 'print.depth' 100}
   {Property.put 'print.width' 100}

   AppName    = 'tps'
   ServerName = 'TPS'

   App        = {New XApplication.manager init}
   Host       = {OS.getEnv 'HOST'}

   %  Special CMD-Args because of Kqml agent mode
   ArgSpec = record(kqmlagent(single char:[&a] type:bool default: false))

   TPS           = {OS.getEnv "TPSPROGRAM"}
   TPSEMACS      = {OS.getEnv "LOUIEMACS"}
   TPSEMACS_INIT = {OS.getEnv "LOUIEMACS_INIT"} % 'ilisp.emacs'
   LOUIXTERM     = {OS.getEnv "LOUIXTERM"}
   TPSPROGRAM    = {OS.getEnv "TPSPROGRAM"}
   TPSDUMP       = {OS.getEnv "TPSDUMP"}
   ACLVERSION    = case {OS.getEnv "ACLVERSION"}
		   of false then false
		   [] S     then {StringToAtom S}
		   end
   ALLEGRO       = {OS.getEnv "ALLEGRO"}
   GEOMETRY      = case {OS.getEnv "EMACSGEOMETRY"}  %  for Volker :-)
		   of false then
		      "138x53+0+0"
		   [] Geometry then
		      Geometry
		   end
   local
      proc {ShowVar Name Val}
	 if Val==false
	 then {App.showInfo "$"#Name#" is undefined."}
	 else {App.showInfo "$"#Name#" = "#Val}
	 end
      end
   in
      {ShowVar "TPSEMACS"      TPSEMACS}
      {ShowVar "TPSEMACS_INIT" TPSEMACS_INIT}
      {ShowVar "LOUIXTERM"     LOUIXTERM}
      {ShowVar "TPSPROGRAM"    TPSPROGRAM}
      {ShowVar "TPSDUMP"       TPSDUMP}
      case ACLVERSION of false then skip else
         {ShowVar "ACLVERSION" ACLVERSION}
         {ShowVar "ALLEGRO"    ALLEGRO}
      end
   end
   
   EOF = 128
   
   fun {ErrToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end
   
   fun {Concat Ls}
      case Ls of nil then nil
      [] L1|Lr then {Append L1 {Concat Lr}} end
   end
   
   local
      proc {DoSplit S M L ?S1 ?S2}
         case S of nil then S1=nil S2=nil
         [] C|Sr then
            if {List.take S L}==M
            then S1=nil S2={List.drop S L}
            else S1=C|{DoSplit Sr M L $ ?S2}
            end
         end
      end
   in
      proc {SplitString Ss M ?As ?Us}
         {DoSplit Ss M {Length M} ?As ?Us}
      end
   end
   
   fun {MkWorkFile WorkFile TpsProblem TpsPrt TpsPrf}
      "restoreproof \""#TpsProblem#"\"\n"#
      WorkFile#
      "printproof \""#TpsPrt#"\"\n"#
      "allscopeflag t\n"#
      "printtypes t\n"#
      "printtypes-all t\n"#
      "infix-notation nil\n"#
      "saveproof \""#TpsPrf#"\"\n"#
      "exit"
   end

   %%
   %% emacs mode
   %%
   local
      proc {LocalShowInfo Line}
         {App.showInfo "[EMACS]: "#Line}
      end
   in
      class TpsEmacs
         from PipeGeneric 
            
         feat error: {NewCell nil}
            
         meth start(display   : Display
		    clientName: ClientName
		    host      : Host
		    inoutPort : InOut
		    servPort  : Serv
		    cmdsVS    : _
		    showInfo  : ShowInfo <= LocalShowInfo
		    onExit    : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end)
	    
	    IlispTmp = {OS.tmpnam}
	    DumpArgs   = if ACLVERSION\=false andthen ACLVERSION>='5'
			 then ALLEGRO#" -I "#TPSDUMP
			 else TPSDUMP
			 end	
	    IlispTmpVS = {File.loadVS TPSEMACS_INIT}#
	    '(setq tps-program "'#DumpArgs#' -qq -- -service '#Host#' '#InOut#' '#Serv#'")'#
	    '(delete-file "'#IlispTmp#'")'
	 in
	    {File.saveVS IlispTmpVS IlispTmp}
	    
	    PipeGeneric, init(cmd:  "env"
			      args: ["DISPLAY="#Display
				     TPSEMACS %"-display" Display
				     "-debug-init" "-name" "Tps Service@"#Host#" (Emacs)"
				     "-geometry" GEOMETRY
				     "-fn" "8x13"
				     "-q" "-l" IlispTmp "-f" "tps"]
			      showInfo:ShowInfo onExit:ExitProc)
	 end
         
         meth close
            PipeGeneric, close(kill:true)
         end      
      end
      
      TpsEmacsAgent = {ByNeed
		       fun {$}
			  class $
			     from TpsEmacs KQMLAgent
			     meth init(...) = Init
				TpsEmacs, Init
				KQMLAgent, init(app: App
                                                  name: 'TpsAgent')
			     end
			     meth test
				{App.showInfo 'Hi, I can speak KQML!'}
			     end
			  end % class ..Agent
		       end}
      
   end
   
   %%
   %% xterm mode
   %%
   local
      proc {LocalShowInfo Line}
         {App.showInfo "[XTERM]: "#Line}
      end
   in
      class TpsXterm
         from PipeGeneric

         meth start(display : Display
		    clientName: ClientName
		    host    : Host
		    inoutPort : InOut
		    servPort  : Serv
		    cmdsVS    : CmdsVS
		    showInfo: ShowInfo <= LocalShowInfo
		    onExit  : ExitProc <=
		    proc {$} {ShowInfo "Process terminated."} end)
	    
	    InitTmp  = {OS.tmpnam}#'.lisp'
	    
	    %% create temporary init file
	    {File.saveVS '(delete-file "'#InitTmp#'")\n'#CmdsVS InitTmp}
	    %% xterm -geometry 80x48+4+16 '#+963+651' -fn vtsingle -fb vtsymbold -n Tps -T Tps -sb -e /project/omega/tps/mathwebtps -- -service blah 2 3
%	    (serv~enter "TPS" :mode "xterm")
	 in
	    PipeGeneric, init(cmd:  "env"
			      args: ["DISPLAY="#Display
				     LOUIXTERM %"-display" Display
				     "-geometry" "80x25"
				     "-title" "Omega Service@"#Host#" (xterm)"
				     "-fn" "vtsingle"
				     "-fb" "vtsymbold"
				     "-sb"
				     "-e"
				     TPSPROGRAM
				     "--" "-service" Host InOut Serv]
			      showInfo:ShowInfo onExit:ExitProc)
	 end
	 
	 meth close
            PipeGeneric, close(kill:true)
         end
      end
      
      TpsXtermAgent = {ByNeed
			 fun {$}
			    class $
			       from TpsXterm KQMLAgent
			       meth init(...) = Init
				  TpsXterm, Init
				  KQMLAgent, init(app: App
                                                  name: 'TpsAgent')
			       end
			    end % class ..Agent
			 end}
   end


   %%
   %% quiet mode
   %%
   class TpsQuiet
      from PipeGeneric
	    
      meth start(display  : Display
		 inoutPort: InOut
		 servPort : Serv
		 cmdsVS   : _
		 showInfo : ShowInfo <= /*LocalShowInfo*/ proc {$ _} skip end
		 onExit   : ExitProc <= proc {$} {ShowInfo "Process terminated."} end
		 ...)
% changed jzimmer 9.3.2001 we don't need two dumps anymore...
	 DumpArgs = if ACLVERSION\=false andthen ACLVERSION>='5'
		    then [ALLEGRO "-I" TPSDUMP]
		    else [TPSDUMP]
		    end
	 EnvArgs = {Concat
		    [["DISPLAY="#Display]
		     DumpArgs
		     ["--" "-service" Host InOut Serv]]}
      in
% 	 env /project/omega/lisp/solaris/acl5.0.1/lisp -I /project/omega/tps/tps3-ultra.dxl -- -service "leibniz" 23423 23344
	 PipeGeneric, init(cmd:  "env"
			   args: EnvArgs
			   showInfo:ShowInfo onExit:ExitProc)
      end
      
      meth close
	 PipeGeneric, close(kill:true)
      end
   end
   
   TpsQuietAgent = {ByNeed
		    fun {$}
		       class $
			  from TpsXterm KQMLAgent
			  meth init(...) = Init
			     TpsQuiet, Init
			     KQMLAgent, init(app: App
					     name: 'TpsAgent')
			  end
		       end % class ..Agent
		    end}
   
   %%
   %% main
   %%
   class TpsService from
                       ThreadManager
                       ClientManager
                       FaultManager
                       ShellProver                        
      feat
         serverName
         brokerURL
         brokerTicket
	 agent

         name: tps
	 
	 inout % input/output stream
         serv  % mathweb stream

         TpsPipe 
         ServObj   % the Object providing the Oz-Services 

	 serverMode % emacs or xterm
	 
      meth init(url:    URL      <= false
		ticket: Ticket   <= false
		cmdArgs: CmdArgs <= unit
	       )
	 self.agent        = {HasFeature CmdArgs kqmlagent} andthen (CmdArgs.kqmlagent == true)
	 self.serverName   = if self.agent then
				{VS2S ServerName#'-agent'}
			     else ServerName end
	 self.brokerURL    = URL
	 self.brokerTicket = Ticket
	 ThreadManager, init
	 if self.agent then
	    {App.showInfo 'I am a KQML agent!'}
	 end

      end
      
      meth StartTpsPipe(ServerArgs ClientInfo exit:?ExitSignal)
    
	 %% Oz <-> Lisp stream connections
	 ClientName = {CondSelect ClientInfo name unit}
	 InOut Serv
	 self.inout = {New Stream init(port: ?InOut readStopC: EOF)}
	 self.serv  = {New Stream init(port: ?Serv  readStopC: EOF)}
	 
	 %% stuff for xterm mode...
	 CmdsVS =
	 '(auto::socket~reset)\n'#
	 '(auto::socket~define :inout)\n'#
	 '(auto::socket~connect "'#Host#'" '#InOut#' :inout)\n'#
	 '(auto::socket~define :service)\n'#
	 '(auto::socket~connect "'#Host#'" '#Serv #' :service)\n'
	 
	 %% auxiliuary procedures
         proc {SendClient Msg}
	    try {Send self.clientPort Msg}
            catch _ then skip end
         end
         proc {SendError VS}
            {SendClient error(VS)}
         end
         proc {DoExit}
            {SendClient exit}
            ExitSignal=unit
         end

         self.serverMode = {CondSelect ServerArgs mode quiet} 

	 %% pipe class
	 PipeClass = case self.serverMode
		     of emacs then
			if self.agent then TpsEmacsAgent else TpsEmacs end
		     [] xterm then
			if self.agent then TpsXtermAgent else TpsXterm end
		     [] quiet then
			if self.agent then TpsQuietAgent else TpsQuiet end
		     [] Mode then
			{SendError "Mode `"#Mode#"' unknown."}
			{DoExit}
			class
			   meth start(...) skip end
			   meth close(...) skip end
			end
                     end

	 %% init method
	 Init = local
		   DefaultInit = start(display : self.clientDisplay
				       clientName: ClientName
				       host    : Host
				       inoutPort: InOut
				       servPort: Serv
				       cmdsVS    : CmdsVS
				       onExit  : DoExit) 
		in
		   if {HasFeature ServerArgs outputLabel} then
		      OutLabel = ServerArgs.outputLabel
		      proc {ShowInfo Line} {SendClient OutLabel(Line)} end
		   in
		      {Adjoin x(showInfo:ShowInfo) DefaultInit}
                   else
                      DefaultInit
                   end
                end
      in
         self.TpsPipe = {New PipeClass Init}        
      end
      
      /** This method enters an tps service */
      /** NOTE: ClientInfo = ClientName(name: user: host: display: port:) */
      /**       ServerArgs = unit(mode:) */
      meth enter(client: ClientInfo <= unit
		 args:   ServerArgs <= unit)
	 
	 ClientManager, init(ClientInfo)
	 ClientManager, showEnter(self.appManager.showInfo)
	 ClientManager, startThreads

	 ExitSignal
      in
	 TpsService, StartTpsPipe(ServerArgs ClientInfo exit:?ExitSignal)
	 
         {Delay 2500} % important !!!
	 
	 if {IsFree ExitSignal} then
	    {ForAll [inout serv] proc {$ Sock} {self.Sock accept} end}
%            TpsService, StartReader
	    TpsService, StartServObj
            TpsService, InstallWatcher    
	 end
      end
  
      meth StartServObj
	 proc {MyLoui M}
	    {Send self.clientPort serv(M)}
	 end
	 SocketClass Init
      in
	 if self.agent then
	    SocketClass = {ByNeed fun{$} KqmlLisp2OzSocket end}
	    Init        = init(self.serv
			       inout:    self.inout
			       url:      self.brokerURL
			       ticket:   self.brokerTicket
			       agentName:'TpsAgent'
			       app:      App
			       myLoui:   MyLoui
			       display:  self.clientDisplay
			       mode:     self.serverMode
			       serverName: self.serverName
                               translate: self
			      )
	 else
	    SocketClass = Lisp2OzSocket
	    Init        = init(self.serv
			       url:    self.brokerURL
			       ticket: self.brokerTicket
			       app:    self.appManager
			       myLoui: MyLoui
			       display:self.clientDisplay
			       mode:   self.serverMode
			       serverName: self.serverName
                               translate: self                               
			      )
	 end
	 self.ServObj = {New SocketClass Init}
      end
      
      meth InstallWatcher
	 proc {Watch _ _}
	    {self.appManager.showInfo
	     "*** "#self.clientName#" died. Exiting..."}
	    if {HasFeature self serverProc}
            then {self.serverProc leave}
	    else {self leave}
	    end
         end
      in
	 FaultManager, installWatcher(Watch)
      end
      
      meth StartReader
         proc {Reader}
            Msg = {self.inout getS($)} in
            {self.appManager.showInfo "[ReadSocket]: "#Msg}
	    {Send self.clientPort msg({ByteString.make Msg})}
	    {Reader}
	 end
	 proc {DoThread}
	    try {Reader}
	    catch Err then
               case Err of system(kernel(terminate...)) then
		  try {Send self.clientPort false}
		  catch system(dp(...)...) then skip end
	       [] connectionClosed then
                  Msg = "Socket connection closed by LISP process." in
		  try {Send self.clientPort error(Msg)}
		  catch system(dp(...)...) then skip end
               else
		  Msg = "Error while reading from socket:\n"#{ErrToVS Err} in
		  try {Send self.clientPort error(Msg)}
		  catch system(dp(...)...) then skip end
	       end
	    end
         end
      in
	 ThreadManager, newThread(DoThread info:'socket-reader')
      end
      
      meth translate(Obj $)
         case Obj
         of 'OMOBJ'(...)|_ then
            {OM2Tps Obj}
         else
            {Browse 'no method to translate'#Obj}
            "please implement method in TpsService, translate for this result"
         end
      end
      
      meth send(ProblemString $ timeout: TimeOut <= 60)
	 
	 fun {ReadAll}
	    Rest Thr
	 in
	    thread
	       Thr = {Thread.this} 
	       Rest = {self.inout getS($)}
	    end
	    {WaitOr Rest {Alarm 50}}
	    try
	       {Thread.suspend Thr}
	       if {Thread.state Thr}==blocked then
		  {Thread.terminate Thr}   
		  else
		  {Thread.resume Thr}
	       end
	    catch error(kernel(deadThread...)...) then skip end
	    if {IsDet Rest} then
	       Rest#{ReadAll}
	    else
	       nil
	    end
	 end
	 Answer Thr 
      in
	 _ = {ReadAll}    %% first read everything that has been written so far
	 {self.appManager.showInfo "[WriteSocket]: "#ProblemString}
	 {self.inout putS(ProblemString)}
	 thread 
	    Thr = {Thread.this} 
	    try
	       Answer = {self.inout getS($)}
	    catch _ then
	       skip
	    end
	 end
	 {WaitOr Answer {Alarm (TimeOut * 1000)}}
	 if {IsFree Answer} then Answer = nil end
	 Answer
      end
      
      meth sig(Signal $<=_)
	 {Process.sigF Signal self.TpsPipe.pid}
      end

      meth load(Url ?BS)
	 try
	    case {Resolve.localize Url}
	    of old(F) then BS={File.loadBS F}
            [] new(F) then BS={File.loadBS F} {OS.unlink F}
            end
         catch _ then
            BS=false
         end
      end
      
      %%=============================================
      %% The TpsService can still call TPS in the old fashion
      %% with a problem on a file.
      %%=============================================
      meth submitProblem(ProblemBS)
         TmpDir  = @tmpdir
         Problem = {ByteString.toString ProblemBS}
         ProblemFile WorkFile
         {SplitString Problem "*TPS-SPLITTER*" ?WorkFile ?ProblemFile}
         TpsWork    = "tps.work"
         TpsProblem = TmpDir#"/tps.problem"
         TpsPrt     = TmpDir#"/tps.prt"
         TpsPrf     = TmpDir#"/tps.prf"
         {self saveVS(ProblemFile "tps.problem")}
         {self saveVS({MkWorkFile WorkFile TpsProblem TpsPrt TpsPrf} TpsWork)}
         CallTps = TPS#" -- -batch tps"
      in
         {self start(CallTps)}
      end
      
      meth analyzeResult($)
         Prt Proof ProofBS
         
         try Prt = {self loadVS("tps.prt" $)}
         catch system(...) then Prt = nil end
         
         try Proof = {self loadVS("tps.prf" $)}
         catch system(...) then Proof = nil end
         
         fun {FailedP Ls}
            fun {SearchForPlan Ls}
               case Ls of nil then nil
               [] &P|&L|&A|&N|Lr then Lr
               else _|Lr = Ls in {SearchForPlan Lr}
               end
            end
         in
            {SearchForPlan Ls} \= nil
         end
      in
         ProofBS = {ByteString.make Prt#"\n"#
                    "%%%%% THIS WAS THE PROOF FILE, NEXT IS THE "#
                    "FORMATTED-PROOF FILE %%%%%"#"\n"#Proof}
         atp_restul(state: if {FailedP Proof} then undetermined else proof end
                    name: self.name
                    host: self.host
                    output: ProofBS)
      end

      meth leave
	 {self.appManager.printInfo "Service "#ServerName#" is left...\n"}
	 FaultManager, deinstallWatcher
	 ThreadManager, tryTerminateAll
	 if {IsDet self.ServObj} then
	    {self.ServObj shutdown}
	 end
	 {self.TpsPipe close}
	 {ForAll [inout serv] proc {$ Sock} {self.Sock close} end}
	 {self.appManager.showInfo " done."}
	 ShellProver, leave
	 {Thread.terminate {Thread.this}}
      end
      
      meth kill
	 ShellProver, kill
         TpsService, leave
      end
      
      meth test
	 {App.showInfo 'I am a TPS service object.'}
      end
      
   end % of TpsService
   
   ServMan = {New ServiceManager init(ServerName TpsService
				      appName: AppName
				      appMan:  App
				      argSpec: ArgSpec 
				     )}
   {ServMan offerService}
end







