%%%
%%% MAPLE agent
%%%

functor
   
import
   System(printInfo) OS(getEnv)
   
   XService(cas:     CASService
            manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'maple'
   ServiceName = 'MAPLE'
   App         = {New XApplication.manager init}

   Cmd         = {OS.getEnv 'MAPLEHOME'}
   Args        = ['-q'] %['-w0']

   if Cmd==false then
      raise undefinedVariable('MAPLEHOME') end
   end
   
   local
      fun {DoRest Xs Sep}
         case Xs of nil then nil
         [] X|Xr then Sep#X#{DoRest Xr Sep} end
      end
   in
      fun {Enumerate First|Rest Sep}
         First#{DoRest Rest Sep}
      end
   end

   VS2BS = VirtualString.toByteString
   
   fun {DropSpaces S} {List.dropWhile S Char.isSpace} end
   %%fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end
   %%fun {NoNewlineQuotesOrBackslash C} {Not {Member C [&\n &" &\\]}} end
   %%fun {NoQuotes C} C\=&" end
   fun {MkError S} error({VS2BS S}) end
   
   class MapleService from CASService
      attr
         maxSols: 1
         
      meth startCAS
         CASService, startCAS(cmd:Cmd args:Args
                              suffix: ";\n"
                              prompt: "\n\n" %"\n> "
                              subst:  ["\\\n    "#nil "^\n"#"^\n\n"]
			      err:    ["Error, " "error, " "on line "]
                              hasPrompt: false
                              printInfo: System.printInfo)
      end
      meth preProcess(ExprVS $)
	 case @maxSols of 1 then
	    "_MaxSols:="#@maxSols#": convert("#ExprVS#",string)"
	 else
	    "convert({"#ExprVS#"},string)"
	 end
      end
      meth postProcess(S $)
	 case {String.token S &" _ $} of nil then {MkError S}
	 [] R1 then
	    case {DropSpaces {Reverse R1}} of &"|R2 then
               {VS2BS {Reverse R2}}
            else {MkError S}
	    end
	 end
         /*
         case {Reverse S} of &^|_ then {MkError S}
            {VS2BS {Clear {Filter S NoQuotes}}}
         end
         */
      end
      
      meth EvalCmd(Cmd Args $)
         MapleIn = case Args of nil then Cmd
                   else Cmd#"( "#{Enumerate Args " , "}#" )" end
      in
         {self eval(MapleIn $)}
      end

      meth compute(cmd:Cmd args:Args ?Result syntax:Syntax<=maple)
	 case Syntax
         of maple2maple then
            maxSols <- 1
	    {self EvalCmd(Cmd Args ?Result)}
	 [] maple2mapleallsolutions then
            maxSols <- 0
	    {self EvalCmd(Cmd Args ?Result)}
         [] maple2post then
            maxSols <- 1
            Translator = {Broker request('MAPLETRANS' $ client:ServiceName)}
            {Translator enter(client: ServiceName)}
	    MapleResult = {self EvalCmd(Cmd Args $)}
         in
	    case MapleResult of error(_) then 
	       Result = {ByteString.make "Error"} 
	       {self kill}
	    else
	       Result = {Translator maple2post(MapleResult $)}
	    end
            {Translator leave}
         [] post2maple then
            maxSols <- 1
            Translator = {Broker request('MAPLETRANS' $ client:ServiceName)}
            {Translator enter(client: ServiceName)}
	    MapleArgs   = {Map Args fun {$ E} {Translator post2maple(E $)} end}
	    MapleResult = {self EvalCmd(Cmd MapleArgs $)}
         in
	    case MapleResult of error(_) then 
	       Result = {ByteString.make "Error"} 
	       {self kill}
	    else
	       Result = MapleResult
	    end
            {Translator leave}
         [] post2post then
            maxSols <- 1
            Translator = {Broker request('MAPLETRANS' $ client:ServiceName)}
            {Translator enter(client: ServiceName)}
	    MapleArgs   = {Map Args fun {$ E} {Translator post2maple(E $)} end}
	    MapleResult = {self EvalCmd(Cmd MapleArgs $)}
         in
	    case MapleResult of error(_) then 
	       Result = {ByteString.make "Error"} 
	       {self kill}
	    else
	       Result = {Translator maple2post(MapleResult $)}
	    end
            {Translator leave}
         end
      end
   end
  
   ServMan = {New ServiceManager init(ServiceName MapleService
                                      appName: AppName
                                      appMan:  App)}
   Broker = ServMan.broker
   {ServMan offerService}
   {App.exit 0}
end

