%%%
%%% CASL toolset agent
%%%

functor
   
import
   OS(getEnv getDir)

   XService(generic: ServiceGeneric
            manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'cats'
   ServiceName = 'CATS'
   App         = {New XApplication.manager init}
   
   CATS = {OS.getEnv 'CATS'}

   if CATS==false
   then raise undefinedVariable('CATS') end
   else {App.showInfo  "$CATS = "#CATS}
   end

   %% CASLLIB dynamisch
   %% unsetenv HOME
   
   %%fun {IsSpace C} {Char.isSpace C} orelse C==&\n end
   %%fun {DropSpaces S} {List.dropWhile S IsSpace} end
   %%fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end

   local
      fun {DoEnum Xs Sep}
         case Xs of nil then ""
         [] X|Xr then
            Sep#X#{DoEnum Xr Sep}
         end
      end
   in
      fun {Enum Xs Sep}
         case Xs of nil then raise empty end
         [] X|Xr then
            X#{DoEnum Xr Sep} 
         end
      end
   end

   %% OptionsR: args(input: [casl static]
   %%                ...  : true)
     
   class CatsService from ServiceGeneric
                         
      meth CallCats(InputVS % inputfile
                    CaslLib % base URL for resolving library imports
                    OptionsR % record of command line options
                    ?OutputFilesRec % record, features = file extensions
                    ?ErrorsVS   
                   )
         Options = {Enum {Map {Record.toListInd OptionsR}
                          fun {$ Arg#Vs}
                             "-"#Arg#
                             if Vs==true then "" else "="#{Enum Vs ","} end
                          end}
                    " "}
         TmpDir = @tmpdir
	 {self saveVS(InputVS 'tmp.casl')}
	 {self exec(CATS#" "#Options#" tmp.casl > errors.txt"
		    nice:false)}
         OutFiles = {Filter {OS.getDir TmpDir}
                     fun {$ F}
                        {List.isPrefix "tmp." F} andthen
                        F\="tmp.casl"
                     end}
      in
         OutputFilesRec = {List.toRecord out
                           {Map OutFiles
                            fun {$ F}
                               {StringToAtom {List.drop F 4}} % extension
                               #{self loadS(F)}               % file content
                            end}}
         ErrorsVS = {self loadS('errors.txt')}
         %% remove all output files with dynamic extensions
         {ForAll OutFiles
          proc {$ F} {self rm(F)} end}
      end

      meth callCats(input:     InputVS
                    'CaslLib': CaslLib
                    options:   OptionsR
                    output:    ?OutputFilesRec
                    errors:    ?ErrorsVS)
         {self CallCats(InputVS CaslLib OptionsR ?OutputFilesRec ?ErrorsVS)}   
      end
      
   end
  
   ServMan = {New ServiceManager init(ServiceName CatsService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end
