%%%
%%% CASL toolset agent
%%%

functor
   
import
   OS(getEnv getDir)

   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'

   Base64 at 'x-ozlib://mathweb/share/Base64.ozf'
   
define
   AppName     = 'cats'
   ServiceName = 'CATS'
   App         = {New XApplication.manager init}
   
   CATSDIR = {OS.getEnv 'CATSDIR'}
   CATS    = CATSDIR#"/cats"
   
   if CATSDIR==false
   then raise undefinedVariable('CATSDIR') end
   else {App.showInfo  "$CATSDIR = "#CATSDIR}
   end
   
   %%fun {IsSpace C} {Char.isSpace C} orelse C==&\n end
   %%fun {DropSpaces S} {List.dropWhile S IsSpace} end
   %%fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end

   fun {DoEnum Xs Sep}
      case Xs of nil then ""
      [] X|Xr then
         Sep#X#{DoEnum Xr Sep}
      end
   end
   fun {Enum Xs Sep}
      case Xs of nil then raise empty end
      [] X|Xr then
         X#{DoEnum Xr Sep} 
      end
   end
   fun {Escape S}
      case S of nil then nil
      [] C|Sr then
         case C
         of &" then &\\|&"|{Escape Sr}
         [] &\\ then &\\|&\\|{Escape Sr}
         %% ... 
         else C|{Escape Sr}
         end
      end
   end
   
   %% OptionsR: args(input: [casl static]
   %%                ...  : true)
      
   class CatsService from ShellProver
                         
      meth CallCats(InputVS % inputfile
                    CaslLib % base URL for resolving library imports
                    OptionsR % record of command line options
                    ?OutputFilesRec % record, features = file extensions
                    ?ErrorsS
                   )
         Options = {Enum {Map {Record.toListInd OptionsR}
                          fun {$ Arg#Vs}
                             "-"#Arg#
                             if Vs==true then "" else "="#{Enum Vs ","} end
                          end}
                    " "}
         TmpDir = @tmpdir
	 {self saveVS(InputVS 'tmp.casl')}
         %% CASLLIB dynamisch
         %% unsetenv HOME
         {self exec("env CASLLIB="#CaslLib#" "#
                    "env HOME= "#" "#
                    "env PATH="#CATSDIR#":${PATH}"#" "#
                    CATS#" "#Options#" tmp.casl > errors.txt"
		    nice:false)}
         OutFiles = {Filter {OS.getDir TmpDir}
                     fun {$ F}
                        {List.isPrefix "tmp." F} andthen
                        F\="tmp.casl"
                     end}
      in
         OutputFilesRec = {List.toRecord out
                           {Map OutFiles
                            fun {$ F}
                               {StringToAtom {List.drop F 4}} % extension
			       #{self loadS(F $)}               % file content
                            end}}
         ErrorsS = {self loadS('errors.txt' $)}
         %% remove all output files with dynamic extensions
         {ForAll OutFiles
          proc {$ F} {self rm(F)} end}
      end

      meth callCats(InputsR $)
         InputVS  = InputsR.casl    % was: input
         CaslLib  = InputsR.lib     % was: 'CaslLib'
         OptionsR = InputsR.options % was: options
         ?OutputFilesRec
         ?ErrorsS
      in
         {self CallCats(InputVS CaslLib OptionsR ?OutputFilesRec ?ErrorsS)}
         "CATSresult([\n"
         #"errors(\""#{Escape ErrorsS}#"\")"
         #{DoEnum {Map {Record.toListInd OutputFilesRec}
		   fun {$ Ext#Content} Ext#'('#{Base64.encode Content}#')' end}
%                   fun {$ Ext#Content} Ext#'('#Content#')' end}
           ','}
         #"\n])"
      end
      
   end
  
   ServMan = {New ServiceManager init(ServiceName CatsService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end
