%%%
%%% MAPLE-TERM service
%%%

functor
   
import
   System(printInfo) OS(getEnv)
%   Browser(browse: Browse)
   
   CasService('class': CASService)
   at 'x-ozlib://mathweb/sys/CasService.ozf'
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'maple-term'
   ServiceName = 'MAPLETERM'
   App         = {New XApplication.manager init}

   Cmd         = {OS.getEnv 'MAPLEHOME'}
   Args        = ['-q' '-e0' '-w0']

   if Cmd==false then
      raise undefinedVariable('MAPLEHOME') end
   end
   
   local
      fun {DoRest Xs Sep}
         case Xs of nil then nil
         [] X|Xr then Sep#X#{DoRest Xr Sep} end
      end
   in
      fun {Enumerate First|Rest Sep}
         First#{DoRest Rest Sep}
      end
   end

   VS2S  = VirtualString.toString
%   VS2BS = VirtualString.toByteString
   
%   fun {DropSpaces S} {List.dropWhile S Char.isSpace} end
%   fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end
%   fun {NoNewlineQuotesOrBackslash C} {Not {Member C [&\n &" &\\]}} end
%   fun {NoQuotes C} C\=&" end
%   fun {MkError S} error({VS2BS S}) end
   
   fun {MapleDebug S Count} 
      case S of nil then nil
      [] C|Sr then
	 if {Char.isSpace C} andthen Count>=512 then
	    &\n|{MapleDebug Sr 0}
	 else
	    C|{MapleDebug Sr Count+1}
	 end
      end
   end

   local
      fun {FindAndDropOneOf Substring1 Substring2 S}
         %% if S = Prefix ++ Substring ++ Suffix
         %% then Prefix ++ Suffx
         %% else S
         %% end
         if {List.isPrefix Substring1 S} then
            {List.drop S {Length Substring1}}
         elseif {List.isPrefix Substring2 S} then
            {List.drop S {Length Substring2}}
         else
            case S of nil then nil
            [] C|Sr then C|{FindAndDropOneOf Substring1 Substring2 Sr} end
         end
      end
   in
      fun {StripEndmarker S}
         {FindAndDropOneOf "END_QWERTYUIOPASDFGHJKLZXCVBNM;"
          "                         END_QWERTYUIOPASDFGHJKLZXCVBNM\n\n" S}
      end
   end
   
   class MapleService from CASService
      attr
	 maxSols: 1
         
      meth startCAS
	 CASService, startCAS(cmd:Cmd
			      args:Args
			      suffix: ";END_QWERTYUIOPASDFGHJKLZXCVBNM;\n"
			      prompts: ["                         END_QWERTYUIOPASDFGHJKLZXCVBNM\n\n"] 
			      subst:  ["\\\n    "#nil "^\n"#"^\n\n"]
			      err:    ["Error, " "error, " "on line "]
			      hasPrompt: false
			      printInfo: System.printInfo)
      end
      
      meth preProcess(ExprVS $)
	 if {IsByteString ExprVS} andthen {ByteString.toString ExprVS} == "quit" then
	    "not quitted"
	 else
	    {MapleDebug {VS2S ExprVS} 0}
	 end
      end
      
      meth postProcess(S $)
% 	 case {String.token S &" _ $} of nil then {MkError S}
% 	 [] R1 then
% 	    case {DropSpaces {Reverse R1}} of &"|R2 then
% 	       {VS2BS {Reverse R2}}
% 	    else {MkError S}
% 	    end
% 	 end
         
	 case S
         of &o|&n|& |&l|&i|&n|&e|& |_ then
	    error(syntax({StripEndmarker S}))
         [] &E|&r|&r|&o|&r|&,|& |_ then
            error(runtime({StripEndmarker S}))
         [] &e|&r|&r|&o|&r|&,|& |_ then
            error(runtime({StripEndmarker S}))
	 else S end
	 
	 /*
	 case {Reverse S} of &^|_ then {MkError S}
            {VS2BS {Clear {Filter S NoQuotes}}}
         end
         */
      end
      
      meth eval(Expr ?Result)
	 case Expr
	 of nil then Result = nil
	 else
            /* afranke: commented out for complex maple scripts 12/6/00
	    ExprStr = {VS2S Expr}
	 in
	    if {Member &; ExprStr} then
	       Result = error(syntax("on line 1, syntax error, `;` unexpected."))
	    else
               */ local % replacement for if / else
	       Expr1 = {self preProcess(Expr $)}
	       proc {DoThread}
		  Result1 = {@cas eval(Expr1 $)} in
		  Result  = {self postProcess(Result1 $)}
	       end
	       Expr1VS = {Value.toVirtualString Expr1 5 20}
	    in
	       {self.appManager newThread(DoThread info:"eval("#Expr1VS#" $)")}
	       {Delay 10}
	    end
	 end
      end
      
      meth EvalCmd(Cmd Args $)
	 MapleIn = case Args of nil then Cmd
		   else Cmd#"( "#{Enumerate Args " , "}#" )" end
      in
	 {self eval(MapleIn $)}
      end

      meth compute(cmd:Cmd args:Args ?Result syntax:Syntax<=maple)
	 case Syntax
	    
	    %% maple2post
	 of maple2maple then
	    maxSols <- 1
	    {self EvalCmd(Cmd Args ?Result)}
	    
	    %% maple2mapleallsolutions then
	 [] maple2mapleallsolutions then
	    maxSols <- 0
	    {self EvalCmd(Cmd Args ?Result)}

	    %% maple2post
	 [] maple2post then
            maxSols <- 1
            Translator = {Broker request('MAPLETRANS' $ client:ServiceName)}
            {Translator enter(client: ServiceName)}
	    MapleResult = {self EvalCmd(Cmd Args $)}
         in
	    case MapleResult of error(_) then 
	       Result = {ByteString.make "Error"} 
	       {self kill}
	    else
	       Result = {Translator maple2post(MapleResult $)}
	    end
	    {Translator leave}

	    %% post2maple
         [] post2maple then
            maxSols <- 1
            Translator = {Broker request('MAPLETRANS' $ client:ServiceName)}
            {Translator enter(client: ServiceName)}
	    MapleArgs   = {Map Args fun {$ E} {Translator post2maple(E $)} end}
	    MapleResult = {self EvalCmd(Cmd MapleArgs $)}
         in
	    case MapleResult of error(_) then 
	       Result = {ByteString.make "Error"} 
	       {self kill}
	    else
	       Result = MapleResult
	    end
            {Translator leave}
	 [] post2mapleallsolutions then
	    maxSols <- 0
            Translator = {Broker request('MAPLETRANS' $ client:ServiceName)}
            {Translator enter(client: ServiceName)}
	    MapleArgs   = {Map Args fun {$ E} {Translator post2maple(E $)} end}
	    MapleResult = {self EvalCmd(Cmd MapleArgs $)}
         in
	    case MapleResult of error(_) then 
	       Result = {ByteString.make "Error"} 
	       {self kill}
	    else
	       Result = MapleResult
	    end
	    {Translator leave}

	    %% post2post
         [] post2post then
            maxSols <- 1
            Translator = {Broker request('MAPLETRANS' $ client:ServiceName)}
            {Translator enter(client: ServiceName)}
	    MapleArgs   = {Map Args fun {$ E} {Translator post2maple(E $)} end}
	    MapleResult = {self EvalCmd(Cmd MapleArgs $)}
         in
	    case MapleResult of error(_) then 
	       Result = {ByteString.make "Error"} 
	       {self kill}
	    else
	       Result = {Translator maple2post(MapleResult $)}
	    end
            {Translator leave}
         end
      end
   end
  
   ServMan = {New ServiceManager init(ServiceName MapleService
                                      appName: AppName
                                      appMan:  App)}
   Broker = ServMan.broker
   {ServMan offerService}
   {App.exit 0}
end

