%%%
%%% MAPLE service
%%%

functor
   
import
   System(printInfo showInfo show)
   OS(getEnv)
%   Browser(browse: Browse)
   
   CasService('class': CASService)
   at 'x-ozlib://mathweb/sys/CasService.ozf'
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'maple'
   ServiceName = 'MAPLE'
   App         = {New XApplication.manager init}

   Cmd         = {OS.getEnv 'MAPLEHOME'}
   Args        = ['-q' '-e0'] % '-w0']

   if Cmd==false then
      raise undefinedVariable('MAPLEHOME') end
   end
   
   local
      fun {DoRest Xs Sep}
	 case Xs of nil then nil
	 [] X|Xr then Sep#X#{DoRest Xr Sep} end
      end
   in
      fun {Enumerate First|Rest Sep}
	 First#{DoRest Rest Sep}
      end
   end

   VS2S  = VirtualString.toString
   VS2BS = VirtualString.toByteString
   
   fun {DropSpaces S} {List.dropWhile S Char.isSpace} end
   %%fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end
   %%fun {NoNewlineQuotesOrBackslash C} {Not {Member C [&\n &" &\\]}} end
   %%fun {NoQuotes C} C\=&" end
   fun {MkError S} error({VS2BS S}) end

   fun {MapleDebug S Count} 
      case S of nil then nil
      [] C|Sr then
	 if {Char.isSpace C} andthen Count>=512 then
	    &\n|{MapleDebug Sr 0}
	 else
	    C|{MapleDebug Sr Count+1}
	 end
      end
   end
   
   class Self=MapleService from CASService
      attr
	 maxSols: 1
	 translator: unit
         
      meth startCAS
	 CASService, startCAS(cmd:Cmd
			      args:Args
			      suffix: ";\n"
			      prompts: ["\n\n"] %"\n> "
			      subst:  ["\\\n    "#nil "^\n"#"^\n\n"]
			      err:    ["Error, " "error, " "on line "]
			      hasPrompt: false
			      printInfo: System.printInfo)
      end

      meth preProcess(ExprVS $)
	 VS = case @maxSols of 1 then
		 "_MaxSols:="#@maxSols#": convert("#ExprVS#",string)"
	      else
		 "convert({"#ExprVS#"},string)"
	      end
      in
	 {MapleDebug {VS2S VS} 0}
      end

      meth postProcess(S $)
	 case {String.token S &" _ $} of nil then {MkError S}
	 [] R1 then
	    case {DropSpaces {Reverse R1}} of &"|R2 then
	       {VS2BS {Reverse R2}}
	    else {MkError S}
	    end
	 end
	 /*
         case {Reverse S} of &^|_ then {MkError S}
            {VS2BS {Clear {Filter S NoQuotes}}}
         end
         */
      end
      
      meth EvalCmd(Cmd Args ?Result)
	 MapleInput = case Args of nil then Cmd
                      else Cmd#"( "#{Enumerate Args " , "}#" )" end
      in
	 Result = {self eval(MapleInput $)}
      end
      
      meth EvalCmdWithTranslation(Cmd Args ?Result
                                  inputSyntax:  InputSyntax  <=maple
                                  outputSyntax: OutputSyntax <=maple)
         case Self, GetTranslator($)
         of false then
            Result = "error"
         [] Transl then
            MapleArgs = case InputSyntax
                        of maple then Args
                        [] post then
                           {Map Args fun {$ E} {Transl post2maple(E $)} end}
                        end
            MapleResult = {self EvalCmd(Cmd MapleArgs $)}
         in
            case MapleResult of error(_) then 
               Result = {ByteString.make "Error"} 
               {self kill}
            else
               Result = case OutputSyntax
                        of maple then MapleResult
                        [] post then
                           {Transl maple2post(MapleResult $)}
                        end     
            end
         end
      end
              
      meth GetTranslator($)
         case @translator of unit then
            case {Broker request('MAPLETRANS' $ client:ServiceName)}
            of false then false
            [] T then
               {T enter(client: ServiceName)}
               translator <- T
               T
            end
         [] T then T
         end
      end
      
      meth LeaveTranslator
         case @translator of unit then skip
         [] T then
            {T leave}
            translator <- unit
         end
      end

      meth leave()
         Self, LeaveTranslator
         CASService, leave
      end
      
      meth compute(cmd:Cmd args:Args ?Result syntax:Syntax<=maple2maple)
	 {App.showInfo Syntax}
	 case Syntax
	    
	    %% maple2maple
	 of maple2maple then
	    maxSols <- 1
	    Self, EvalCmd(Cmd Args ?Result)
	    
	    %% maple2mapleallsolutions
	 [] maple2mapleallsolutions then
	    maxSols <- 0
	    Self, EvalCmd(Cmd Args ?Result)
            
	    %% maple2post
	 [] maple2post then
            maxSols <- 1
            Self, EvalCmdWithTranslation(Cmd Args ?Result outputSyntax:post)
            
	    %% post2maple
         [] post2maple then
            maxSols <- 1
            Self, EvalCmdWithTranslation(Cmd Args ?Result
                                         inputSyntax:post outputSyntax:maple)
            
	 [] post2mapleallsolutions then
	    maxSols <- 0
            Self, EvalCmdWithTranslation(Cmd Args ?Result
                                         inputSyntax:post outputSyntax:maple)
            
	    %% post2post
	 [] post2post then
	    maxSols <- 1
            Self, EvalCmdWithTranslation(Cmd Args ?Result
                                         inputSyntax:post outputSyntax:post)
         end
      end
   end
  
   ServMan = {New ServiceManager init(ServiceName MapleService
                                      appName: AppName
                                      appMan:  App)}
   Broker = ServMan.broker
   {ServMan offerService}
   {App.exit 0}
end

