%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000
%%%
%%% Last change:
%%%   $Date $ by $Author $
%%%   $Revision $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/*
<compute>
<cmd>simplify</cmd>
<arg>2+3</arg>
</compute>

<methodCall>
  <methodName>maple.compute</methodName>
  <params>
    <param>
      <value><i4>41</i4></value>
    </param>
  </params>
</methodCall>

*/

functor

import
   Foo(getCgiArgs:GetCgiArgs) at 'x-oz://system/Application.ozf'
   Connection(take)
   OS(getDir putEnv signal tmpnam unlink)
   Pickle(load save)
   Property(put)
   URL(make)

   XApplication at 'x-ozlib://mathweb/share/XApplication.ozf'
   File         at 'x-ozlib://mathweb/share/File.ozf'
   Internet     at 'x-ozlib://mathweb/share/Internet.ozf'

   Http(server:     BaseServer
        connection: BaseConnection)
   at 'x-ozlib://mathweb/httpd/Http.ozf'

   XmlParser
   at 'x-ozlib://mathweb/xmlparser/XmlParser.ozf'
   PModule
   at '/home/afranke/import/ilo/dtd/omdoc.parser.ozf'
   SModule
   at '/home/afranke/import/mathweb/xml/sem/omdoc.sem.ozf'

   %%Browser(browse:Browse)
define
   VS2A = VirtualString.toAtom
   VS2S = VirtualString.toString
   
   {Property.put 'errors.width' 100}
   {Property.put 'errors.depth' 100}

   ParserModule = {XmlParser.link PModule SModule}
   %%{XmlParser.setDebug true}
  
   AppName = 'mapled'
   App     = {New XApplication.manager init}
   
   Spec = record(port(single char:&p type:int default:8080)
                 hostname(single char:&h type:string default:Internet.hostname)
                 brokerticket(single char:[&t &b] type:string default:"http://www.ags.uni-sb.de/~omega/")
                 default(single type:string default:false)
                )
   Args
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {App.showError AppName#": "#S}
      {App.exit 1}
   end

   if Args.hostname\=Internet.hostname andthen
      {Internet.sameHosts Args.hostname Internet.hostname}==false
   then
      {App.showError "Sorry, "#Args.hostname#" and "#Internet.hostname#
       " seem to be different hosts :("}
      {App.exit 1}
   end
   HOSTNAME = Args.hostname 
   PORT     = Args.port
   PREFIX   = "http://"#HOSTNAME#":"#PORT
   
   Debug = App.show

   local          
      class Counter
         attr
            counter:0
            
         meth init(N0<=0)
            counter <- N0
         end         
         meth next(?N)
            N = @counter + 1
            counter <- N
         end
      end
      class ServRegistry
         prop
            locking
         feat
            broker
            counter
            dict
           
         meth init(broker:Broker)
            self.broker  = Broker
            self.counter = {New Counter init}
            self.dict    = {NewDictionary}
         end

	 meth open(ServiceName ?ID)
            lock
               case {Broker request(ServiceName $)}
               of false then
                  raise unknownService(ServiceName) end
               [] Obj then
                  {Obj enter}
                  ID = {self.counter next($)}
                  {Dictionary.put self.dict ID Obj}
               end
            end
         end
	 
	 meth send(ID Msg)
	    case {Dictionary.condGet self.dict ID false}
	    of false then
	       raise unknownID(ID) end
	    [] Obj then
	       {Obj Msg}
	    end
	 end

	 meth close(ID)
            lock
               case {Dictionary.condGet self.dict ID false}
               of false then
                  raise unknownID(ID) end
               [] Obj then
                  try {Obj leave}
                  catch _ then skip end
                  {Dictionary.remove self.dict ID}
               end
            end
         end
	 
	 meth closeAll
            lock
	       {ForAll {Dictionary.items self.dict}
		proc {$ Obj}
		   try {Obj leave}
		   catch _ then skip end
		end}
               {Dictionary.removeAll self.dict}
            end
         end
      end
   in
      {App.printInfo "Loading broker-ticket..."}
      BrokerTicket = {VS2A {XURL.loadVS Args.brokerticket}}
      {App.showInfo " done"}
      {App.printInfo "Connecting to broker..."}
      Broker       = {Connection.take BrokerTicket}
      {App.showInfo " done"}
      ServReg = {New ServRegistry init(broker: Broker)}
   end

   %% a session should go like this:
   %%      {ServReg open} --> ID
   %%      {ServReg send(ID Meth)}
   %%       ....
   %%      {ServReg close(ID)}

   
   local
      QueryLock = {NewLock}
   in
      fun {ParseQuery Query Spec}
         QueryString = if Query==unit then "" else Query end
      in
         lock QueryLock then
            {OS.putEnv "REQUEST_METHOD" "GET"}
            {OS.putEnv "QUERY_STRING" QueryString}
            {GetCgiArgs Spec}
         end
      end
   end
   
   fun {EscapeHtml S}
      case S of nil then nil
      [] C|R then
         case C
         of &<  then &&|&l|&t|&;|{EscapeHtml R}
         [] &>  then &&|&g|&t|&;|{EscapeHtml R}
         [] &&  then &&|&a|&m|&p|&;|{EscapeHtml R}
         [] &\\ then &\\|&\\|{EscapeHtml R}
         else C|{EscapeHtml R}
         end
      end
   end
   
   class ConnectionManager
      from
         BaseConnection
         
      meth beginResponseBase
         BaseConnection, beginResponse
         {self tag(base(href:PREFIX)#"\n")}
      end
      /*
      meth 'GET'(X...) = GET
         U = {URL.make X}
         P = U.path
      in
         case P of [nil] then
            BaseConnection, 'GET'("/index.html")
         else
            BaseConnection, GET
         end
      end
      */
      meth 'POST'(X Content...) = POST
         U = {URL.make X}
         P = U.path
      in
	 case P of ["maple.cgi"] then
	    {self ProcessRequest(U Content)}
         else
            BaseConnection, POST
         end
      end
      /*
      meth DumpValues(ContentR)
         {Record.forAllInd ContentR
          proc {$ Var Val}
             {self write(vs:Var#" = "#Val#"\n")}
          end}
      end
      */
      %%
      meth ProcessRequest(Request)
         /*
         Val Accept in
         {XmlParser.parseVS Request ParserModule
          'start_foopy'(?Val) ?Accept}
         if Accept==false then
            {self beginResponse(...)}
            {self doctype}
            {self tag(html(head(title("400 Parse Error"))
                           body(h1("Parse Error")
                                "Sorry, your request could not be parsed.")))}
            {self endResponse}
            {Delay 1000}
            raise close end
         end
         %%{Browse OMDocName#Val}
         {self }
         {self...}
         
         {self ok("Ok.")}
         {Delay 1000}
         */
         Expr  = Request
         {App.printInfo "Requesting MAPLE service..."}
         Maple = {Broker request('MAPLE' $)}
         {Wait Maple}
         {App.show Maple}
      in
         try
            {App.printInfo "Calling eval("#Expr#") ..."}
            Result = {Maple eval(Expr $)} in
            {Wait Result}
            {App.show Result}
            BaseConnection, ok({EscapeHtml {VS2S Result}})
         finally
            {Maple leave}
         end
         {Delay 1000}
      end
   end
   
   Server = {New BaseServer init(port:        PORT
                                 acceptClass: ConnectionManager)}
   
   {App.showInfo HOSTNAME#":"#PORT}
   
   local
      %%T = {Thread.this}
      proc {Term SIGNAL}
         {App.showError '*** '#SIGNAL#' ***'}
         %%{Thread.terminate T}
         {App.exit 0}
      end
   in
      {OS.signal 'SIGHUP' Term}
      {OS.signal 'SIGINT' Term}
      {OS.signal 'SIGTERM' Term}
   end
   
   try
      {Thread.suspend {Thread.this}}
   finally
      {Server close}
      {App.exit 0}
   end
end



