%%%
%%% rcas - Remote Computer Algebra Systems
%%%

functor
   
import
   Application Connection Pickle Open System 

   XURL(loadVS)
   at 'x-ozlib://mathweb/share/XURL.ozf'
   
define
   AppName = "rcas"

   PrintInfo = System.printInfo
   ShowInfo  = System.showInfo
   ShowError = System.showError
   Exit      = Application.exit

   DefaultBrokerURLs = ["http://www.ags.uni-sb.de/~omega/"
                        "http://www.ags.uni-sb.de/~omegadev/"]
   
   Spec = record(debug(single char:&d type:bool default:false)
                 casname(single type:atom/*('MAPLE' 'GAP' 'MAGMA' 'COCOA')*/)
                 maple(single alias:casname#'MAPLE')
                 magma(single alias:casname#'MAGMA')
                 gap(  single alias:casname#'GAP')
                 cocoa(single alias:casname#'COCOA')
                 pvs(  single alias:casname#'PVS') % well...
                 'MAPLE'(single alias:casname#'MAPLE')
                 'MAGMA'(single alias:casname#'MAGMA')
                 'GAP'(single alias:casname#'GAP')
                 'COCOA'(single alias:[casname#'COCOA' quit#["Quit" "quit"]])
                 'PVS'(single alias:casname#'PVS') %
                 test(single type:bool default:true)
                 quit(single type:list(string) default:["quit"])
                 linemode(single type:bool default:false)
                 tps(single alias:[casname#'TPS-INTERACTIVE' test#false
                                   quit#["exit"] linemode#true])
                 brokerurls(single char:[&b &u &t] type:list(string)
                            default:DefaultBrokerURLs)
                 inputfile(single char:&f type:string default:'stdin')
                )
   Args
   try 
      Args = {Application.getArgs Spec}
   catch error(ap(usage S) ...) then
      {ShowError AppName#": "#S}
      {Application.exit 1}
   end
  
   Debug = Args.debug
   
   proc {DebugPrintInfo VS}
      if Debug then {System.printInfo VS} end
   end
   proc {DebugShowInfo VS}
      if Debug then {System.showInfo VS} end
   end
   proc {DebugShow Val}
      if Debug then {System.show Val} end
   end

   VS2BS = VirtualString.toByteString
   VS2S  = VirtualString.toString
   
   %%
   %% determine CAS name
   %%
   FirstArg = case Args.1 of nil then nil
              [] Arg1|_ then Arg1 end
   
   ServiceName = if {HasFeature Args casname} then
                    Args.casname
                 else
                    case {Map FirstArg Char.toUpper}
                    of "MAPLE" then 'MAPLE'
                    [] "GAP"   then 'GAP'
                    [] "MAGMA" then 'MAGMA'
                    [] "PVS"   then 'PVS' % well...
                    [] X then
                       {ShowError "*** ERROR: Unknown CAS name `"#X#"'"}
                       {ShowError "    usage: rcas (--maple|--gap|--magma)"}
                       {ShowError ''}
                       {Exit 1} dummy
                    end
                 end

   %%
   %% connect to mathweb broker in Saarbruecken and ask for CAS-service
   %%
   proc {TryConnect URL ServiceName ?ServiceObj}
      try
         {DebugPrintInfo "Loading ticket from `"#URL#"'..."}
         Ticket = {XURL.loadVS URL}
         {DebugShowInfo " done."}
         %%
         {DebugPrintInfo "Connecting to `"#Ticket#"'..."}
         Broker = {Connection.take Ticket}
         {DebugShowInfo " done."}
      in
         {DebugPrintInfo "Asking for `"#ServiceName#"'-service..."}
         {Broker request(ServiceName ?ServiceObj)}
         if ServiceObj==false then
            {DebugShowInfo " not found."}
         else
            {DebugShowInfo " found."}
            %%
            {DebugPrintInfo "Entering service..."} 
            {ServiceObj enter(client:'rCAS' args:unit(entity:_))}
            {DebugShowInfo " done."}
         end
      catch _ then
         {DebugShowInfo " failed."}
         ServiceObj = false 
      end
   end
   fun {RecruitOne BrokerURLs ServiceName}
      case BrokerURLs of nil then
         {ShowError "*** ERROR: CAS-service `"#ServiceName#"' not available. "}
         {ShowError "    Please try again later."}
         {ShowError "    If the problem remains, mail to: omega@ags.uni-sb.de"}
         {ShowError ''}
         {Exit 1} dummy
      [] FirstURL|Rest then
         case {TryConnect FirstURL ServiceName}
         of false then {RecruitOne Rest ServiceName}
         [] Obj then Obj
         end
      end
   end

   BrokerURLs = {Map Args.brokerurls
                 fun {$ S}
                    if {List.last S}==&/ then {VS2S S#"broker-ticket.ozc"}
                    else S end
                 end}
   
   ServiceObj = {RecruitOne BrokerURLs ServiceName}

   if Args.test then
      proc {TestErr MsgVS}
         {ServiceObj leave}
         {ShowError "*** ERROR: CAS cannot compute 1+1; result was:"}
         {ShowError MsgVS}
         {Exit 1} 
      end
      {DebugPrintInfo "Testing..."}
      TestResult
   in
      {ServiceObj eval("1+1" ?TestResult)}
      {WaitOr TestResult {Alarm 60000}}
      if {IsFree TestResult} then
         {TestErr "*** TIMEOUT ***"}
      elsecase TestResult of error(BS) then
         {TestErr BS}
      [] VS then
         case {VS2S VS} of "2" then
            {DebugShowInfo " success."}
         [] S then
            {TestErr S}
         end
      end
   end
   
   %%
   %% read expressions and evaluate them
   %%
   %% `;' is used as separator
   %%
   class TextFile from Open.text Open.file end

   class Reader

      attr
         InFile: unit
         Buff: nil
         Line: nil
         
      meth open(FileName <= 'stdin')
         InFile <- {New TextFile init(name:FileName)}
         Buff <- nil
         Line <- nil
      end

      meth prompt
         {PrintInfo "> "}
      end

      meth readLine($)
         Reader, prompt
         {@InFile getS($)}
      end
      
      meth read($)
         Expr R in
         {List.takeDropWhile @Line fun {$ C} C\=&; end ?Expr ?R}
         case R of nil then
            Buff <- @Buff#@Line#"\n"
            Line <- nil
            if @InFile==unit then
               Reader,return($)
            else
               Reader,prompt
               case {@InFile getS($)}
               of false then
                  Reader,close
                  Reader,return($)
               [] NextLine then
                  Line <- NextLine
                  Reader,read($)     
               end
            end
         [] &;|R1 then
            Buff <- @Buff#Expr
            Line <- R1
            Reader, return($)
         end
      end
      
      meth return(?VS)
         VS = @Buff
         Buff <- nil
      end

      meth atEnd($)
         @InFile == unit
      end
      
      meth close
         {@InFile close}
         InFile <- unit
      end
   end
   
   ReaderObj = {New Reader open(Args.inputfile)}
 
   fun {DropSpaces S} {List.dropWhile S Char.isSpace} end
   fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end

   proc {LeaveExit}
      {ServiceObj leave}
      {Exit 0}
   end
   proc {Loop}
      if {ReaderObj atEnd($)} then {LeaveExit}
      else
         local
            ExprVS = if Args.linemode
                     then {ReaderObj readLine($)}
                     else {ReaderObj read($)}
                     end
            ExprS  = {Clear {VS2S ExprVS}}
            ExprBS = {VS2BS ExprS}
            {DebugPrintInfo "Expr: "} {DebugShow ExprBS}
         in
            if Args.linemode==false andthen ExprS==nil then skip
            elseif {Member ExprS Args.quit} then {LeaveExit}
            elsecase {ServiceObj eval(ExprBS $)}
            of error(BS) then
               {ShowError BS}
            [] ResultVS then
               ResultS  = {VS2S ResultVS}
               Result   = if {String.isInt ResultS}
                          then {String.toInt ResultS}
                          else ResultS
                          end
            in
               if Result\=nil then
                  {ShowInfo Result}
               end
            end
         end
         {Loop}
      end
   end

   {Loop}
end
