%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Manager.oz ; This file is part of the MathWeb system
%
%   major updates: 20.12.2000
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
import
%   Browser(browse:Browse)
%   System

%   Env(user: User)  at 'x-ozlib://mathweb/share/Env.ozf'

export

   Generals
   Hosts
   DummyHost
   DummyRemoteHost
   Services
   RemoteBrokers

define
   % service types:  general, cas, atp, modelchecker, csolver assistant, transformation, presentation
   DummyHost = 'dummy'(www: nil
		       ostype: nil
		       method: nil
		       services: nil)

   DummyRemoteHost = 'dummy'(www: nil
			     login: nil
			     method: nil
			     )
   Hosts = nil %[DummyHost]
%    ['leibniz'(www:"www.ags.uni-sb.de"
% 		      ostype: "solaris"
% 		      method:"rsh"
% 		      services: ["omega"
% 				 "maple"
% 				 "tps"
% 				])
% 	    'goedel'(www:"www.ags.uni-sb.de"
% 		     ostype: "linux"
% 		     method:"rsh"
% 		     services: ["spass"
% 				"bliksem"
% 				"cosie"
% 			       ]
% 		    )
% 	   ]
   RemoteBrokers = nil

%    ['top.coli.uni-sb.de'(login: "mathweb"
% 					 method: "rsh"
% 					 www: "www.coli.uni-sb.de")
% 		    'gromit.cs.bham.ac.uk'(login:"ceb"
% 					   method: "ssh"
% 					   www: firewall)	      
% 		   ]
   
   Generals = [
	       %% ===============================
	       %%          GENERAL
	       %% ===============================
	       'shell'(vars: ["SHELL"
			     ]
		       linux: ["/bin/tcsh"
			      ]
		       solaris: ["/bin/tcsh"
				]
		       ostypes: ["linux" "solaris"]
		       type: general
		       help: "General shell settings for MathWeb."
		      )
	       'emacs'(vars: ["EMACSGEOMETRY"
			     ]
		       linux:["138x53+0+0"
			     ]
		       solaris: ["138x53+0+0"
				]
		       ostypes: ["linux" "solaris"]
		       type: general
		       help: "Settings for emacs."
		      )
	       'Lisp'(vars: [
			     "ACLVERSION"
			     "ALLEGRO"
			     "INKALISP"
			    ]
		      linux: [
			      "5.0.1"
			      "/project/omega/lisp/linux/acl5.0.1/lisp"
			      "/project/vse/acl/bin/acl"
			     ]
		      solaris: [
				"5.0.1"
				"/project/omega/lisp/solaris/acl5.0.1/lisp"
				"/project/vse/acl/bin/acl"
			       ]
		      ostypes: ["linux" "solaris"]
		      type: general
		      help: "General settings for lisp processes."
		     )
	       'Linux'(vars: ["LINUXHOST" 
			      "LINUXSERVICES"
			     ]
		       linux: ["bledsoe"
			       "cosie magma pvs"			       
			      ]
		       solaris: ["bledsoe"
				 "cosie magma pvs"
				]
		       ostypes: ["linux" "solaris"]
		       type: general
		       help: "General settings for services only running on Linux."
		      )
	       
	       'ATPs'(vars: [
			     "CPULIMITEDRUN"
			     "CPUMAXDEFAULT"
			     "PROVERSHOME"
			     "DFG2TPTP"
			     "TPTP2X"       
			     "DFG2BLK"         
			     "DFG2OTTER"      
			    ]
		      linux: ["${MATHWEB_BIN}/CPULimitedRun.linux-i486"
			      "30"
			      "/project/omega/atp/bin/linux"
			      "${PROVERSHOME}/dfg2tptp"
			      "${PROVERSHOME}/tptp2X"
			      "${PROVERSHOME}/dfg2blk"
			      "${PROVERSHOME}/dfg2otter"
			     ]
		      solaris: [
				"${MATHWEB_BIN}/CPULimitedRun.solaris-sparc"
				"90"
				"/project/omega/atp/bin/solaris"
				"${PROVERSHOME}/dfg2tptp"
				"${PROVERSHOME}/tptp2X"
				"${PROVERSHOME}/dfg2blk"
				"${PROVERSHOME}/dfg2otter"
			       ]
		      ostypes: ["linux" "solaris"]
		      type: general
		      help: "General settings for ATPs."
		     )
	      ]
   
   Services = [
	       %% ===============================
	       %%          CAS
	       %% ===============================
	       'maple'(vars: ["MAPLEHOME"
			      "MAPLEPARSER"
			      "POST2MAPLE"     
			      "MAPLE2POST"
			     ]
		       linux: ["/project/omega/maple/bin/maple"
			       "/project/omega/maple/parser/maple-parser"
			       "/project/omega/maple/parser/post2maple"
			       "/project/omega/maple/parser/maple2post"
			      ]
		       solaris: ["/project/omega/maple/bin/maple"
				 "/project/omega/maple/parser/maple-parser"
				 "/project/omega/maple/parser/post2maple"
				 "/project/omega/maple/parser/maple2post"
				]
		       type: cas
		       ostypes: ["linux" "solaris"]
		       help: "The Maple-V Computer Algebra System."
		      )
	       'gap'(vars: ["GAPHOME"         
			   ]
		     linux: ["/project/omega/gap/bin/gap.sh"
			    ]
		     solaris: ["/project/omega/gap/bin/gap.sh"
			      ]
		     type: cas
		     ostypes: ["linux" "solaris"]
		     help: "The GAP Computer Algebra System V4."
		    )
	       
	       'magma'(vars: ["MAGMAHOME"
			     ]
		       linux: ["/project/omega/magma/magma"
			      ]
		       solaris: ["/project/omega/magma/magma"
				]
		       type: cas
		       ostypes: ["linux"]
		       help: "The Magma Computer Algebra System V2.3."
		      )
	       'cocoa'(vars: ["COCOA_LIBRARY"
			      "COCOA"
			     ]
		       linux: ["/project/omega/cas/cocoa/cocoa-3.7.linux/"
			       "/project/omega/cas/cocoa/cocoa-3.7.linux/cocoa"
			      ]
		       solaris: ["/project/omega/cas/cocoa/cocoa-3.7.solaris/"
				 "/project/omega/cas/cocoa/cocoa-3.7.solaris/cocoa"
				]
		       type: cas
		       ostypes: ["linux" "solaris"]
		       help: "The CoCoA Computer Algebra System V3.7."
		      )
	       %% ===============================
	       %%          OMEGA
	       %% ===============================
	       'omega'(vars: ["OMEGADUMPEMACS"
			      "OMEGADUMPQUIET"
			      "LOUIEMACS"
			      "LOUIEMACS_INIT"
			     ]
		       linux: ["/home/${USER}/omega/omega-3/bin/omega-3-emacs-linux.dxl"
			       "/home/${USER}/omega/omega-3/bin/omega-3-quiet-linux.dxl"
			       "/usr/bin/emacs"
			       "/home/${USER}/omega/omega-3/bin/ilisp.emacs"
			      ]
		       solaris: ["/home/${USER}/omega/omega-3/bin/omega-3-emacs-solaris.dxl"
				 "/home/${USER}/omega/omega-3/bin/omega-3-quiet-solaris.dxl"
				 "/opt/emacs/bin/emacs"
				 "/home/${USER}/omega/omega-3/bin/ilisp.emacs"
				]
		       type: assistant
		       ostypes: ["linux" "solaris"]
		       help: "The OMEGA system."
		      )
	       'prex'(vars: ["PREXDUMPEMACS"
			     "PREXDUMPQUIET"
			    ]
		      linux: ["/home/${USER}/omega/prex/bin/prex-3-emacs-linux.dxl"
			      "/home/${USER}/omega/prex/bin/prex-3-quiet-linux.dxl"
			     ]
		      solaris: ["/home/${USER}/omega/prex/bin/prex-3-emacs-solaris.dxl"
				"/home/${USER}/omega/prex/bin/prex-3-quiet-solaris.dxl"
			       ]
		      type: presentation
		      ostypes: ["linux" "solaris"]
		      help: "The P.rex proof presentation system."
		     )
	       'davinci'(vars: ["DAVINCIHOME"
			       ]
			 linux: ["/project/omega/daVinci_V2.1"
				]
			 solaris: ["/project/omega/daVinci_V2.1"
				  ]
			 ostypes: ["linux" "solaris"]
			 help: "DaVinci graph drawing tool V2.1")
	       
	       %% ===============================
	       %%          ATPs
	       %% ===============================
	       'spass'(vars: ["SPASSHOME"
			     ]
		       linux: ["${PROVERSHOME}/spass"
			      ]
		       solaris: ["${PROVERSHOME}/spass"
				]
		       type: atp
		       ostypes: ["linux" "solaris"]
		       help: "The theorem prover SPASS V0.97."
		      )
	       
	       'otter'(vars: ["OTTERHOME"
			     ]
		       linux: ["${PROVERSHOME}/otter"
			      ]
		       solaris: ["${PROVERSHOME}/otter"]
		       type: atp
		       ostypes: ["linux" "solaris"]
		       help: "The theorem prover Otter V3.0.6."
		      )

	       'bliksem'(vars: ["BLIKSEMHOME"
			       ]
			 linux: ["${PROVERSHOME}/bliksem"
				]
			 solaris: ["${PROVERSHOME}/bliksem"
				  ]
			 type: atp
			 ostypes: ["linux" "solaris"]
			 help: "The theorem prover Bliksem V1.10."
			)

	       'protein'(vars: ["PROTEINHOME"
				"PROTEINHEADER"
			       ]
			 linux: ["${PROVERSHOME}/protein"
				 "/project/omega4/atp/protein"
				]
			 solaris: ["${PROVERSHOME}/protein"
				   "/project/omega4/atp/protein"
				  ]
			 type: atp
			 ostypes: ["linux" "solaris"]
			 help: "The theorem prover ProTEin."
			)
	       'fdpll'(vars: ["FDPLLHOME"
			      "FDPLLSTRATEGY"
			     ]
		       linux: [nil
			       nil
			      ]
		       solaris: ["/project/omega4/atp/fdpll/bin/fdpll"
				 "/project/omega4/atp/fdpll/lib/fdpll/bottom_up.fdpll				     "
				]
		       type: atp
		       ostypes: ["linux" "solaris"]
		       help: "The theorem prover FDPLL V0.9.14."
		      )

	       %% ===============================
	       %%          MODEL CHECKERS
	       %% ===============================
	       'kimba'(vars: ["KIMBAHOME"
			     ]
		       linux: ["/project/omega4/atp/kimba/akimba"
			      ]
		       solaris: ["/project/omega4/atp/kimba/akimba"
				]
		       type: modelchecker
		       ostypes: ["linux" "solaris"]
		       help: "The model checker kimba."
		      )

	       'mace'(vars: ["MACEHOME"
			    ]
		      linux: ["${PROVERSHOME}/mace"
			     ]
		      solaris: ["${PROVERSHOME}/mace"
			       ]
		      type: modelchecker
		      ostypes: ["linux" "solaris"]
		      help: "The model checker MACE V1.2.0."
		     )
	       'satchmo'(vars: ["SATCHMOHOME"
			       ]
			 linux: ["${PROVERSHOME}/satchmo"
				]
			 solaris: ["${PROVERSHOME}/satchmo"
				  ]
			 type: modelchecker
			 ostypes: ["linux" "solaris"]
			 help: "The model checker satchmo."
			)

	       'finder'(vars: ["FINDERHOME"
			       ]
			 linux: ["${PROVERSHOME}/finder-pc"
				]
			 solaris: ["${PROVERSHOME}/finder-ultra"
				  ]
			 type: modelchecker
			 ostypes: ["linux" "solaris"]
			 help: "The model checker finder."
			)


	       'eqp'(vars: ["EQPHOME"
			   ]
		     linux: ["${PROVERSHOME}/eqp09c"
			    ]
		     solaris: ["${PROVERSHOME}/eqp09c"
			      ]
		     type: atp
		     ostypes: ["linux" "solaris"]
		     help: "The theorem prover."
		    )

	       'pvs'(vars: ["PVSHOME"
			   ]
		     linux: ["/project/omega/pvs"
			    ]
		     solaris: [""
			      ]
		     type: atp
		     ostypes: ["linux"]
		     help: "The PVS Specification and Verification System V2.3."
		    )
	       'eprover'(vars: ["EPROVERHOME"
			       ]
			 linux: ["${PROVERSHOME}/eprover"
				]
			 solaris: ["${PROVERSHOME}/eprover"
				  ]
			 type: atp
			 ostypes: ["linux" "solaris"]
			 help: "The Equational theorem prover E V0.6."
			)
	       'tps'(vars: ["TPSPROGRAM"
			    "TPSFONTS"
			   ]
		     linux: ["${PROVERSHOME}/tps3"
			     "/project/omega3/tps/fonts"
			    ]
		     solaris: ["${PROVERSHOME}/tps3"
			       "/project/omega3/tps/fonts"
			      ]
		     type: atp
		     ostypes: ["linux" "solaris"]
		     help: "The higher order theorem prover TPS V3."
		    )
	       'cosie'(vars: nil
		       linux: nil		
		       solaris: nil
		       type: csolver
		       ostypes: ["linux"]
		       help: "CoSIE, a Constraint Solver for Inequalities and Equations over the real numbers V2.0."
		      )
	       'doris'(vars: nil
		       linux: nil		
		       solaris: nil
		       type: assistant
		       ostypes: ["linux" "solaris"]
		       daemon: true
		       help: "The DORIS service."
		      )
	       'atp'(vars: nil
		     linux: nil		
		     solaris: nil
		     type: assistant
		     ostypes: ["linux" "solaris"]
		     help: "The concurrent ATPs service for DORIS."
		    )

	       
	      ]
      
   
end
