%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% XService.oz ; This file is part of the MathWeb system
%
%   major updates: 20.12.2000
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
import
   Tk
%   Browser(browse: Browse)
   XTkDialog                 at 'x-ozlib://mathweb/share/XTkDialog.ozf'
   Env(home:Home)  at 'x-ozlib://mathweb/share/Env.ozf'
   
   Dialogs(uRLDialog:  URLDialog
	   hostDialog: HostDialog
	   serviceDialog: ServiceDialog
	  )
   at  'x-ozlib://mathweb/config/Dialogs.ozf'
export
   'class': DialogManager
define
   local
      AboutText  = 'MathWeb Configuration Tool V1.0b\n'#
      'from Juergen Zimmer (jzimmer@ags.uni-sb.de)\n'#
      'Acknowledgements to Abdelhakim Freihat (hakim@ags.uni-sb.de)\n'#
      'Also look at http://www.mathweb.org/mathweb/'
      AboutGifFile =  'http://www.mathweb.org/images/MathWeb-SB.gif'
   in
      class DialogManager
	 feat
	    symbolDir
	 attr
	    CurPath: nil
	    CurURL: nil

	    hostDialog: nil
	    serviceDialog: nil
	    
	 meth init(SymbolDir URL)
	    self.symbolDir = SymbolDir
	    CurPath <- Env.home
	    CurURL <- URL
	    hostDialog <- nil
	    serviceDialog <- nil
	 end

	 meth loadFilebox(Title Type Suffix ?File)
	    case {Tk.return
		  tk_getOpenFile(filetypes:  q(q(Type q(Suffix))
					       q('All Files' '*'))
				 initialdir: @CurPath
				 parent:     self.toplevel
				 title:      Title)}
	    of nil then
	       File = false
	    elseof Filename then
	       Path = {Reverse {List.dropWhile {Reverse Filename}
				fun {$ C} C\=&/ end}}
	    in
	       CurPath <- Path
	       File = Filename
	    end
	 end
      
	 meth saveFilebox(Title Type Suffix ?File)
	    case {Tk.return
		  tk_getSaveFile(filetypes:  q(q(Type q(Suffix))
					       q('All Files' '*'))
				 initialdir: @CurPath
				 parent:     self.toplevel
				 title:      Title)}
	    of nil then
	       File = false
	    elseof Filename then
	       Path = {Reverse {List.dropWhile {Reverse Filename}
				fun {$ C} C\=&/ end}}
	    in
	       CurPath <- Path
	       File = Filename
	    end
	 end
      
	 meth urlFilebox(Title Info Suffix Load Confirm ?File)
	    {Wait {New URLDialog init(master : self.toplevel
				      status : self.status
				      manager: self
				      title  : Title
				      info   : Info
				      suffix : Suffix
				      load   : Load
				      confirm: Confirm
				      initurl: @CurURL
				      file   : ?File)}.tkClosed}
	    case File of false then
	       skip
	    else
	       CurURL <- File
	    end
	 end
      
	 meth hostDialog(Title Info Change ?Result)
	    hostDialog <- {New HostDialog init(master : self.toplevel
					       status : self.status
					       manager: self
					       title  : Title
					       symbolDir: self.symbolDir
					       info   : Info
					       change : Change
					       result : ?Result)}
	 end
      
	 meth closeHostDialog
	    case @hostDialog of nil then skip
	    else {@hostDialog tkClose}
	       hostDialog <- nil
	    end
	 end

	 meth serviceDialog(Title Info Heading Change ?Result)
	    serviceDialog <- {New ServiceDialog init(master : self.toplevel
						     status : self.status
						     manager: self
						     title  : Title
						     heading: Heading
						     symbolDir: self.symbolDir
						     info   : Info
						     change : Change
						     result : ?Result)}
	 end
	 
	 meth closeServiceDialog
	    case @serviceDialog of nil then skip
	    else {@serviceDialog tkClose}
	       serviceDialog <- nil
	    end
	 end
	 
	 meth exit
	    Result in 
	    {Wait
	     {New XTkDialog.exit init(master : self.toplevel
				      aspect : 500 
				      text   : 'Do you really want to exit?'
				      result : ?Result)}.tkClosed}
	    if Result==true then
	       {self.main exit}
	    end
	 end
	 
	 meth about
	    {Wait
	     {New XTkDialog.about
	      init(master:self.toplevel
		   urlGif: AboutGifFile
		   text: AboutText)}.tkClosed}
	 end
	 
      end      
   end
end














