%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Manager.oz ; This file is part of the MathWeb system
%
%   major updates: 20.12.2000
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
import
   OS
   Pickle
%   Browser(browse:Browse)
   System
   
   DialogManager    at 'x-ozlib://mathweb/config/DialogManager.ozf'
   MenuManager      at 'x-ozlib://mathweb/config/MenuManager.ozf'
   SymbolManager    at 'x-ozlib://mathweb/config/SymbolManager.ozf'
   ToplevelManager  at 'x-ozlib://mathweb/config/ToplevelManager.ozf'
   StatusManager    at 'x-ozlib://mathweb/config/StatusManager.ozf'
   ConfigData       at 'x-ozlib://mathweb/config/ConfigData.ozf'
   Misc(virtualStringToValue: VS2V lookup: Lookup)
   at 'x-ozlib://mathweb/share/Misc.ozf'
   Env(home:Home user:User)  at 'x-ozlib://mathweb/share/Env.ozf'
   File
   at 'x-ozlib://mathweb/share/File.ozf'

export
   'class': Manager%

define
   DefaultConfigFile = Home#'/.mathweb/Config.ozc'

   VS2A = VirtualString.toAtom
   A2S = Atom.toString

   class Manager
      from
	 BaseObject
         DialogManager.'class'
	 MenuManager.'class'
	 SymbolManager.'class'
	 ToplevelManager.'class'
	 StatusManager.'class'
      
      prop
         locking
         final
      attr
	 unsaved
	 initialized
	 started
	 
	 currentFile

	 currentHost
	 currentService
	 currentBroker

	 hostsDict
	 generalsDict
	 servicesDict
	 remoteBrokersDict
      feat
	 main
         vs2v

	 symbolDir
	 
	 hostsFrame
	 generalsFrame
	 servicesFrame
	 remoteBrokersFrame

      meth init(Main  SymbolDir SymbolbarConfig)
         lock
	    self.main    = Main
	    self.vs2v    = VS2V
	    self.symbolDir = SymbolDir
	    
	    unsaved      <- false
	    initialized  <- false
	    started      <- false

	    currentFile  <- DefaultConfigFile
	    currentHost  <- nil
	    currentBroker<- nil
	    hostsDict    <- {Dictionary.new}
	    generalsDict <- {Dictionary.new}
	    servicesDict <- {Dictionary.new}
	    remoteBrokersDict  <- {Dictionary.new}
	    
	    PackAll = ToplevelManager.'class',init($)
	 in
	    
	    DialogManager.'class'   ,init(SymbolDir @currentFile)
	    MenuManager.'class'     ,init
	    SymbolManager.'class'   ,init(SymbolDir SymbolbarConfig self.vs2v)
	    StatusManager.'class'   ,init
	    {PackAll}
	    Manager,ConfigurePointer(idle)
	    self.hostsFrame = {{self.toplevel frameBook($)} hostsFrame($)}
	    self.generalsFrame = {{self.toplevel frameBook($)} generalsFrame($)}
	    self.servicesFrame = {{self.toplevel frameBook($)} servicesFrame($)}
	    self.remoteBrokersFrame = {{self.toplevel frameBook($)} remoteBrokersFrame($)}
	 end
      end
      
      meth ConfigurePointer(Status)
         ToplevelManager.'class',configurePointer(Status)
      end
      meth Idle
         MenuManager.'class',idle
         Manager,ConfigurePointer(idle)
      end
%%%
%%%   Hier beginen die Methoden der Applikation:      
%%%
      meth replyToMethod(Dict Method Entry)
	 case Method
	 of "Edit new"
	 then {self editEntry(Dict "---")} 
	 elsecase Entry 
	 of "" then skip
	 elsecase Method
	 of "Modify this" then {self editEntry(self.Dict Entry)}
	 [] "Delete this" then {self deleteEntry(self.Dict Entry)}
	 else skip
	 end
      end
      
      meth showAvailableServices
	 ServiceEntries = {Dictionary.entries @servicesDict}
	 RightServices = if @currentHost == nil then
			    {Map ServiceEntries fun{$ N#_} {A2S N} end}
			 else
			    {Map 
			     {Filter ServiceEntries
			      fun{$ _#Entry}
				 HostInfo = {Dictionary.condGet @hostsDict @currentHost nil}
			      in
				 (HostInfo \= nil) andthen
				 {Member HostInfo.ostype Entry.ostypes}
			      end}
			     fun{$ N#_} {A2S N} end}
			 end
      in
	 thread
	    {self.hostsFrame showAvailableServices(RightServices)}
	 end
      end

      meth showHostInfo(HostName)
	 Info = {Dictionary.condGet @hostsDict {VS2A HostName} nil}
      in
	 if Info\=nil then
	    currentHost <- {VS2A HostName}
	    thread
	       {self.hostsFrame showHostInfo(Info)}
	    end
	    Manager, showAvailableServices
	 end
      end
      meth showServiceHelp(ServiceName)
	 Info = {Dictionary.condGet @servicesDict {VS2A ServiceName} nil}
      in
	 if Info\=nil then
	    {self setMessage(Info.help)}
	 end
      end
      meth showServiceInfo(ServiceName)
	 Info = {Dictionary.condGet @servicesDict {VS2A ServiceName} nil}
      in
	 if Info\=nil then
	    thread
	       {self.servicesFrame showServiceInfo(Info)}
	    end
	 end
      end
      meth showGeneralInfo(GeneralName)
	 Info = {Dictionary.condGet @generalsDict {VS2A GeneralName} nil}
      in
	 if Info\=nil then
	    thread
	       {self.generalsFrame showGeneralInfo(Info)}
	    end
	 end
      end
      meth showRemoteBrokerInfo(BrokerName)
	 Info = {Dictionary.condGet @remoteBrokersDict {VS2A BrokerName} nil}
      in
	 if Info\=nil then
	    thread
	       {self.remoteBrokersFrame showRemoteBrokerInfo(Info)}
	    end
	 else
	    thread
	       {self.remoteBrokersFrame clear}
	    end
	 end
      end      
      meth showAll
	 Manager, showGenerals
	 Manager, showServices
	 Manager, showHosts
	 Manager, showRemoteBrokers
      end
      meth showGenerals
	 case {Dictionary.keys @generalsDict}
	 of nil then skip
	 [] H|Hr then
	    thread
%	       {self.generalsFrame updateGeneralInfo({A2S {Label H}})}
	       {self.generalsFrame insertGenerals({Map H|Hr fun{$ Info}
							       {A2S {Label Info}}
							    end})}
	    end
	 end
      end
      meth showServices
	 case {Dictionary.keys @servicesDict}
	 of nil then skip
	 [] H|Hr then
	    thread
	       %{self.servicesFrame updateServiceInfo({A2S {Label H}})}
	       {self.servicesFrame insertServices({Map H|Hr fun{$ Info}
							       {A2S {Label Info}}
							    end})}
	    end
	 end
      end
      meth showHosts
	 Hosts = {Dictionary.keys @hostsDict}
      in
	 thread
	    {self.hostsFrame insertHosts({Map Hosts fun{$ Info}
						       {A2S {Label Info}}
						    end})}
	 end
	 Manager, showHostInfo(@currentHost)
	 Manager, showAvailableServices
      end
      meth showRemoteBrokers
	 Brokers = {Dictionary.keys @remoteBrokersDict}
      in
	 thread
	    {self.remoteBrokersFrame insertRemoteBrokers({Map Brokers fun{$ Info}
									 {A2S {Label Info}}
								      end})}
	 end
	 Manager, showRemoteBrokerInfo(@currentBroker)
      end      
      
      meth editHostInfo(Host)
	 HostName = {VS2A Host}
	 Title = "Edit Host Entry"
	 Info = {Dictionary.condGet @hostsDict {VS2A HostName} nil}
      in
	 if Info \= nil then
	    GuessWWW = local
			  Hosts = {Dictionary.entries @hostsDict}
			  RestHosts = {Filter Hosts fun{$ H} {Label H} \= HostName end}
		       in
			  case RestHosts of nil then nil
			  [] _#H|_ then
			     H.www
			 end
		       end
	    NewInfo = if Info.www == nil then
			 {AdjoinAt Info www GuessWWW}
		      else Info end
	    Result
	 in
	    MenuManager.'class', busy
	    DialogManager.'class', hostDialog(Title NewInfo  false ?Result)
	    if Result \= nil then
	       {Dictionary.put @hostsDict {Label Result} Result}
	       currentHost <- {Label Result}
	       Manager, showHosts
	       unsaved <- true
	    end
	    MenuManager.'class', idle
	 end
      end
      
      meth addHost
	 if {Not @initialized} then
	    Manager, newConfigsFile
	 end
	 Title = "Add Host Entry"
	 DummyHost = case {Dictionary.items @hostsDict}
		     of nil then ConfigData.dummyHost
		     [] H|_ then
			H
		     end
	 Result
      in
	 MenuManager.'class', busy
	 DialogManager.'class', hostDialog(Title DummyHost true ?Result)
	 if Result \= nil then
	    Info = {Dictionary.condGet @hostsDict {Label Result} nil}
	 in
	    if Info \= nil
	    then
	       if {self confirm("Override entry for host "#{Label Info}#"?" $)} then
		  {Dictionary.put @hostsDict {Label Result} Result}
		  currentHost <- {Label Result}
		  Manager, showHosts
	       end
	    else
	       {Dictionary.put @hostsDict {Label Result} {AdjoinAt Result services nil}}
	       currentHost <- {Label Result}
	       Manager, showHosts
	    end
	    unsaved <- true 
	 end
	 MenuManager.'class', idle
      end
      
      meth deleteHost(Host)
	 HostName = {VS2A Host}
      in
	 if {self confirm("Delete entry for host "#HostName#"?" $)} then
	    {Dictionary.remove @hostsDict HostName}
	    currentHost <- nil
	    unsaved <- true
	    Manager, showHosts
	 end
      end
      
      meth editServiceInfo(Service)
	 ServiceName = {VS2A Service}
	 Title = "Edit Service Entry"
	 Info = {Dictionary.condGet @servicesDict {VS2A ServiceName} nil}
      in
	 if Info \= nil andthen {Not (Info.vars == nil)} then
	    Result
	 in
	    MenuManager.'class', busy
	    DialogManager.'class', serviceDialog(Title Info "Service"  false ?Result)
	    if Result \= nil then
	       {Dictionary.put @servicesDict {Label Result} Result}
	       Manager, showServiceInfo(Service)
	       unsaved <- true
	    end
	    MenuManager.'class', idle
	 end
      end
      meth editGeneralInfo(General)
	 GeneralName = {VS2A General}
	 Title = "Edit General Entry"
	 Info = {Dictionary.condGet @generalsDict {VS2A GeneralName} nil}
      in
	 if Info \= nil andthen {Not (Info.vars == nil)} then
	    Result
	 in
 	    MenuManager.'class', busy
	    DialogManager.'class', serviceDialog(Title Info "Settings for" false ?Result)
	    if Result \= nil then
	       {Dictionary.put @generalsDict {Label Result} Result}
	       Manager, showGeneralInfo(General)
	       unsaved <- true
	    end
	    MenuManager.'class', idle
	 end
      end
      
      meth addService(Element)
	 Info = {Dictionary.get @hostsDict @currentHost}
	 Services = Info.services
	 NewInfo = {AdjoinAt Info services {Append Services [Element]}}
      in
	 unsaved <- true
	 {Dictionary.put @hostsDict @currentHost NewInfo}
	 thread
	    {self.hostsFrame showHostInfo(NewInfo)}
	 end
      end
      
      meth subService(Element)
	 Info = {Dictionary.get @hostsDict @currentHost}
	 Services = Info.services
      in
	 unsaved <- true
	 if {Member Element Services} then
	    NewInfo = {AdjoinAt Info services {List.subtract Services Element}}
	 in
	    {Dictionary.put @hostsDict @currentHost NewInfo}
	    thread
	       {self.hostsFrame showHostInfo(NewInfo)}
	    end
	 end
      end
      meth subAllServices
	 if @currentHost \= nil then
	    Info = {Dictionary.get @hostsDict @currentHost}
	    NewInfo = {AdjoinAt Info services nil}
	 in
	    {Dictionary.put @hostsDict @currentHost NewInfo}
	    unsaved <- true
	    thread
	       {self.hostsFrame showHostInfo(NewInfo)}
	    end
	 end
      end
      meth addAllServices
	 if @currentHost \= nil then
	    Info = {Dictionary.get @hostsDict @currentHost}
	    Services = Info.services
	    AllServices = {Map {Dictionary.keys @servicesDict} A2S}
	    NewInfo = {AdjoinAt Info services {Append Services AllServices}}
	 in
	    unsaved <- true
	    {Dictionary.put @hostsDict @currentHost NewInfo}
	    thread
	       {self.hostsFrame showHostInfo(NewInfo)}
	    end
	 end
       end


      meth editRemoteHostInfo(Host)
% 	 ['top.coli.uni-sb.de'(login: "mathweb"
% 			       method: "rsh"
% 			       www: "www.coli.uni-sb.de")
	 HostName = {VS2A Host}
	 Title = "Edit Host Entry"
	 Info = {Dictionary.condGet @remoteBrokersDict {VS2A HostName} nil}
      in
	 {System.show Info}
	 if Info \= nil then
	    GuessWWW = local
			  Brokers = {Dictionary.entries @remoteBrokersDict}
			  RestBrokers = {Filter Brokers fun{$ H} {Label H} \= HostName end}
		       in
			  case RestBrokers of nil then nil
			  [] _#H|_ then
			     H.www
			  end
		       end
	    NewInfo = if Info.www == nil then
			 {AdjoinAt Info www GuessWWW}
		      else Info end
	    Result
	 in
	    MenuManager.'class', busy
	    DialogManager.'class', hostDialog(Title NewInfo  false ?Result)
	    if Result \= nil then
	       {Dictionary.put @remoteBrokersDict {Label Result} Result}
	       currentBroker <- {Label Result}
	       Manager, showRemoteBrokers
	       unsaved <- true
	    end
	    MenuManager.'class', idle
	 end
      end
      
      meth addRemoteHost
	 if {Not @initialized} then
	    Manager, newConfigsFile
	 end
	 Title = "Add Remote Host"
	 DummyHost = case {Dictionary.items @remoteBrokersDict}
		     of nil then ConfigData.dummyRemoteHost
		     [] H|_ then
			H
		     end
	 Result
      in
	 MenuManager.'class', busy
	 DialogManager.'class', hostDialog(Title DummyHost true ?Result)
	 if Result \= nil then
	    Info = {Dictionary.condGet @remoteBrokersDict {Label Result} nil}
	 in
	    if Info \= nil
	    then
	       if {self confirm("Override entry for host "#{Label Info}#"?" $)} then
		  {Dictionary.put @remoteBrokersDict {Label Result} Result}
		  currentBroker <- {Label Result}
		  Manager, showRemoteBrokers
	       end
	    else
	       {Dictionary.put @remoteBrokersDict {Label Result} if {HasFeature Result services} then
								    {AdjoinAt Result services nil}
								 else
								    Result
								 end}
	       currentBroker <- {Label Result}
	       Manager, showRemoteBrokers
	    end
	    unsaved <- true 
	 end
	 MenuManager.'class', idle
      end
      
      meth deleteRemoteHost(Host)
	 HostName = {VS2A Host}
      in
	 if {self confirm("Delete entry for host "#HostName#"?" $)} then
	    {Dictionary.remove @remoteBrokersDict HostName}
	    currentBroker <- nil
	    unsaved <- true
	    Manager, showRemoteBrokers
	 end
      end
      
       meth editEntry(Dict Entry)
	  unsaved <- true
	  skip
       end
       meth deleteEntry(Dict Entry)
	  skip
       end
 %%% 3 Methoden zum Verwalten von ConfigsDateien
 %%%      
       meth newConfigsFile
	  currentHost  <- nil
	  currentFile  <- nil
	  hostsDict    <- {Dictionary.new}
	  generalsDict <- {Dictionary.new}
	  servicesDict <- {Dictionary.new}
	  remoteBrokersDict  <- {Dictionary.new}
	  Manager, initConfigData(ConfigData)  %% fill up the dictionaries!
	  {self.hostsFrame clear}
	  {self.generalsFrame clear}
	  {self.servicesFrame clear}
	  Manager, showAll
	  unsaved <- true
	  initialized <- true
       end
       meth initConfigData(DataFunctor)
	  %% 1. ConfigsDatein sollen gelesen Werden und schliesslich in Dictiomnaries gespeichert.
% 	  {ForAll DataFunctor.hosts
% 	   proc{$ Info}
% 	      {Dictionary.put @hostsDict {Label Info} Info}
% 	   end}
	  {ForAll DataFunctor.generals
	   proc{$ Info}
	      {Dictionary.put @generalsDict {Label Info} Info}
	   end}
	  {ForAll DataFunctor.services
	   proc{$ Info}
	      {Dictionary.put @servicesDict {Label Info} Info}
	   end}
% 	  {ForAll DataFunctor.remoteBrokers
% 	   proc{$ Info}
% 	      {Dictionary.put @remoteBrokersDict {Label Info} Info}
% 	   end}
       end

       meth loadConfigsFile(file: File <= nil)
	  lock
	     if{Not @unsaved}
		orelse
		{self confirm("The current configuration is not saved.\nProceed?" $)}
	     then
		FileName
	     in
		if File \=nil then
		   FileName = File
		else
		   MenuManager.'class', busy
		   DialogManager.'class', urlFilebox('Load Configuration'
							 'Config Files' '.ozc' true 'Open' ?FileName)
		end
		if FileName==false then
		   Manager,Idle	    
		else
		   try
		      Data = {Pickle.load FileName}
		   in
		      if {IsRecord Data}
			 andthen {All [hostsDict servicesDict generalsDict remoteBrokersDict]
				  fun{$ F} {HasFeature Data F} end}
		      then
			 {ForAll [hostsDict servicesDict generalsDict remoteBrokersDict]
			  proc{$ F}
			     F <- {Record.toDictionary Data.F}
			  end}
			 unsaved <- false
			 initialized <- true
			 currentFile <- FileName
			 Manager, showAll
			 MenuManager.'class', idle
		      end
		   catch _ then
		      MenuManager.'class', idle
		   end
		end
	     end
	  end
       end


       meth saveConfigsFile(file: DefaultName <= nil)
	  if @unsaved then
	     FileName
	  in
	     MenuManager.'class', busy
	     if DefaultName \= nil then
		FileName = DefaultName
	     else
		DialogManager.'class', urlFilebox('Save Configuration'
						  'Config Files' '.ozc' false 'Save' ?FileName)
	     end
	     if FileName==false then
		Manager,Idle	    
	     else
		Content = {Record.mapInd mathWebConfig(hostsDict: @hostsDict
						       servicesDict: @servicesDict
						       generalsDict: @generalsDict
						       remoteBrokersDict: @remoteBrokersDict)
			   fun{$ Label Dict}
			      {Dictionary.toRecord Label Dict}
			   end}
	     in
		try
		   if {File.exists FileName} then
		      {File.backup FileName}
		      {File.removeForce FileName}
		      {Delay 1000}
		   end
		   {Pickle.save Content FileName}
		   unsaved <- false
		   currentFile <- FileName
		   MenuManager.'class', idle
		catch E then
		   {System.show E}
		   MenuManager.'class', idle
		end
	     end
	  else
	     {self.status setMessage("There are no changes to be saved!")}
	  end
       end
       
 %%%% 2 methoden zum Generieren und Testen von Config-Files
       meth generateConfigs
	  %% ~/.mathweb/config
	  %% config.linux
	  %% config.solaris
	  %% config.tcshrc
	  %% local-brokers.config
	  %% remote-brokers.config
	  %% host1.config ... hostn.config

	  if {self confirm("Do you really want to write the configuration files?" $)}
	  then
	     Manager, generateConfig
	     Manager, generateConfigLinux
	     Manager, generateConfigSolaris
	     Manager, generateConfigTcshrc
	     Manager, generateConfigLocal
	     Manager, generateConfigRemote
	     Manager, generateConfigHosts
	  end
       end	       
       meth generateConfig
	  FileName = Home#'/.mathweb/config'
	  Content = ""
       in
	  Manager, writeFile(Content FileName)
       end
       meth generateConfigLinux
	  FileName = Home#'/.mathweb/config.linux'
	  Content1 = Manager, generateSetenv({Dictionary.items @generalsDict} nil nil linux $)
	  Content2 = Manager, generateSetenv({Dictionary.items @servicesDict} nil nil linux $)
	  Content = "# This is the shell configuration for all Linux machines\n\n"#Content1#Content2
       in
	  Manager, writeFile(Content FileName)
       end
       meth generateConfigSolaris
	  FileName = Home#'/.mathweb/config.solaris'
	  Content1 = Manager, generateSetenv({Dictionary.items @generalsDict} nil nil solaris $)
	  Content2 = Manager, generateSetenv({Dictionary.items @servicesDict} nil nil solaris $)
	  Content = "# This is the shell configuration for all Solaris machines\n\n"#Content1#Content2
       in
	  Manager, writeFile(Content FileName)
       end
       meth generateConfigTcshrc
	  FileName = Home#'/.mathweb/config.tcshrc'
	  Content1 = case {Dictionary.items @hostsDict}
		    of I|_ then if {HasFeature I www} then I.www else nil end
		    else nil end
	  Content = "# This is the configuration for tcsh and Mosh\n\n"#
	  "#######################################\n"#
	  "# set $BROKER_TICKET_URL for remote brokers\n"#
	  "#######################################\n\n"#
	  Content1
       in
	  Manager, writeFile(Content FileName)
       end
       meth generateConfigLocal
	  FileName = Home#'/.mathweb/local-brokers.config'
	  LocalHostNames = {Dictionary.keys @hostsDict}
	  
	  Content =  "# This file contains the names of all local MathWeb-brokers\n\n"
	  #Manager, generateList(LocalHostNames $)
       in
	  Manager, writeFile(Content FileName)
       end
       
       meth generateConfigRemote
	  FileName = Home#'/.mathweb/remote-brokers.config'
	  LocalHostURLs = {Map {Dictionary.entries @hostsDict}
			   fun{$ Host#Entry}
			      'http://'#Entry.www#'/~'#User#'/broker@'#Host
			   end}
	  RemoteHostURLs = {Map {Dictionary.entries @remoteBrokersDict}
			    fun{$ Host#Entry}
			       HostName = {List.takeWhile {A2S Host} fun{$ C}
									C \= &.
								     end}
			    in
			       if Entry.www == firewall then
				  ""
			       else
				  'http://'#Entry.www#'/~'#Entry.login#'/broker@'#HostName
			       end
			    end}
	  Content =  "# This file contains the URLs to possibly available remote MathWeb-brokers\n\n"
	  #Manager, generateList({Append LocalHostURLs RemoteHostURLs} $)
       in
	  Manager, writeFile(Content FileName)
       end
       meth generateConfigHosts
	  {ForAll {Dictionary.entries @hostsDict}
	   proc{$ Host#Entry}
	      FileName = Home#'/.mathweb/'#Host#'.config'
	      Content = "# Services running on "#Host#"\n\n"
	      #Manager, generateServiceList(Entry.services $)
	      %% should be more in the future ToDo !
	   in
	      Manager, writeFile(Content FileName)
	   end}
       end

       meth generateServiceList(Es $)
	  case Es of nil then nil
	  [] E|Er then
	     Info = {Dictionary.get @servicesDict {VS2A E}}
	  in
	     (if {HasFeature Info daemon} then
		 'start '
	      else nil end)
	     #E#'\n'#Manager, generateServiceList(Er $)
	  else nil end
       end
       meth generateList(Es $)
	  case Es of nil then nil
	  [] E|Er then
	     	    E#'\n'
	    #Manager, generateList(Er $)
	 else nil end
      end
	    
      meth generateSetenv(Entries Vars Ls Feature $)
	 case Vars of nil then
	    case Entries of nil then nil 
	    [] E|Er then
	       Manager, generateSetenv(Er E.vars E.Feature Feature $)
	    else nil end
	 [] V|Vr then
	    case Ls of nil then  Manager, generateSetenv(Entries nil nil Feature $)
	    [] L|Lr then 
	       if L == nil then
		  Manager, generateSetenv(Entries Vr Lr Feature $)
	       else
		  (if {Length V} > 8 then
		      "setenv "#V#" \t\""#L#"\"\n"
		   else
		      "setenv "#V#"     \t\""#L#"\"\n"
		   end)
		  #Manager, generateSetenv(Entries Vr Lr Feature $)
	       end
	    else
	       Manager, generateSetenv(Entries nil nil Feature $)
	    end
	 end
      end
      
      meth writeFile(Content FileName)
	 try
	    {File.backup FileName}
	    {File.saveVS Content FileName}
	 catch _  then
	    {System.showInfo 'Config File not saved due to NFS failure.'}
	 end
      end
      
      meth testConfigs
         skip
      end

      %%  start MathWeb via shell scripts
      meth startMathweb
	 FirewallTicket = {Lookup 'FIREWALL_TICKET' unit}
	 Port =           {Lookup 'PORT' 30000}
      in
	 MenuManager.'class', busy
	 try
	    {OS.system 'cd '#Home#'/.oz/bin/' _}
	    {OS.system 'mathweb-down > /dev/null' _}
	    if FirewallTicket \= unit then
	       {OS.system 'sleep 2; xterm -fn 8x13 -geometry 70x30+0+5 -e mathweb-up --firewallticket '#FirewallTicket#' --port '#Port#"\&" _}
	    else
	       {OS.system 'sleep 2; xterm -fn 8x13 -geometry 70x30+0+5 -e mathweb-up '#"\&" _}
	    end
	 catch E then
	    {System.show E}
	 end
	 Manager,Idle
      end

            %%  start MathWeb via shell scripts
      meth startMathwebAll
	 if @started == false then
	    FirewallTicket = {Lookup 'FIREWALL_TICKET' unit}
	    Port =           {Lookup 'PORT' 30000}
	 in
	    MenuManager.'class', busy
	    try
% 	    {OS.system 'cd '#Home#'/.oz/bin/' _}
% 	    {OS.system 'mathweb-down-all > /dev/null' _}
	       if FirewallTicket \= unit then
		  {OS.system 'sleep 2; xterm -fn 8x13 -geometry 70x30+0+5 -e mathweb-up-all --firewallticket '#FirewallTicket#' --port '#Port#"\&" _}
	       else
		  {OS.system 'sleep 2; xterm -fn 8x13 -geometry 70x30+0+5 -e mathweb-up-all '#"\&" _}
	       end
	    catch E then
	       {System.show E}
	    end
	    started <- true
	    Manager,Idle
	 end
      end

      %%  shutdownMathWeb via shell scripts
      meth shutdownMathweb
	 MenuManager.'class', busy
	 try
	    {OS.system 'cd '#Home#'/.oz/bin/' _}
	    {OS.system 'sleep 2; xterm -fn 8x13 -geometry 70x30+0+5 -e mathweb-down \&' _}
	 catch E then
	    {System.show E}
	 end
	 Manager,Idle
      end
      
      %%  shutdown MathWeb on all hosts
      meth shutdownMathwebAll
	 MenuManager.'class', busy
	 try
	    {OS.system 'cd '#Home#'/.oz/bin/' _}
	    {OS.system 'sleep 2; xterm -fn 8x13 -geometry 70x30+0+5 -e mathweb-down-all \&' _}
	 catch E then
	    {System.show E}
	 end
	 started <- false
	 Manager,Idle
      end

      meth about
	 lock
	    MenuManager.'class',busy
	    DialogManager.'class',about
            Manager,Idle
         end
      end
   
      meth exit
	 lock
	    MenuManager.'class',busy
	    if @unsaved andthen 
	       {self confirm("The current configuration is not saved.\nSave it now?" $)}
	    then
	       Manager, saveConfigsFile(file: @currentFile)
	    end
	    DialogManager.'class',exit % von hier gibts event. kein zurueck!!!
	    Manager,Idle
	 end
      end

      meth otherwise(M ...)
	 {System.show 'Sorry, unknown method for Config Manager '#M}
      end
   end %of class
end

