%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MenuManager.oz ; This file is part of the MathWeb system
%
%   major updates: 20.12.2000
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   
import
   %%Browser(browse:Browse)
   Tk
   XManagers   at 'x-ozlib://mathweb/config/XManagers.ozf'
   Misc(printRepresentation  : PrintRepresentation
	doEntries            : DoEntries
	makeMenu             : MakeMenu
	mapMenu              : MapMenu)
   at 'x-ozlib://mathweb/share/Misc.ozf'
   
export
   'class': MenuManager
define
   MenuSpec = [
	       'File.Exit'                  # true
               'File.New'                   # true
               'File.Load...'               # true                         
               'File.Save...'               # false
               'Configs.Generate Config'    # false
               'Configs.Test Config'        # false
	       'MathWeb.Start MathWeb'      # false
	       'MathWeb.Start MathWeb All'  # false
	       'MathWeb.Shutdown MathWeb'      # false
	       'MathWeb.Shutdown MathWeb All'  # false
               'Help.About'                 # true
               
               
              ]
   
   FileMenuSpec = [command(label: 'New'     key: ctrl(n) action:newConfigsFile)
                   command(label: 'Load...' key: ctrl(l) action:loadConfigsFile)
                   command(label: 'Save...' key: ctrl(s) action:saveConfigsFile)
                   command(label: 'Exit'    key: ctrl(x) action:exit)

                  ]
   ConfigsMenuSpec = [
                      command(label: 'Generate Config' key: ctrl(g) action: generateConfigs)
                      command(label: 'Test Config'     key: ctrl(t) action: testConfigs)
                     ]
   MathwebMenuSpec =[
                     command(label: 'Start MathWeb'        action: startMathweb)
		     command(label: 'Start MathWeb All'    action: startMathwebAll)
		     command(label: 'Shutdown MathWeb'     action: shutdownMathweb)
		     command(label: 'Shutdown MathWeb All' action: shutdownMathwebAll)
                    ]
   
   HelpMenuSpec =[command(label: 'About' action: about)]    
   
      
   class MenuManager
      feat
	 menu
	 items    % dictionary for all menu items
	 dynamics % dictionary for dynamic menu(-buttons)
         preparemenu      
      meth init
         
	 self.items       = {NewDictionary}
	 self.dynamics    = {NewDictionary}
         self.preparemenu = {New XManagers.prepareMenu init(MenuSpec)}
         Menu          = self.menu
	 ObjMap        = objs(default:self)
      in
	 local
	    fun {MakeMenubutton Text#Menu#Feature}
	       M = {MapMenu ObjMap unit Menu} in
	       menubutton(text:Text menu:M feature:Feature)
	    end
	 in   
	    Menu = {MakeMenu self.toplevel self.toplevel
		    { Map
		      ['File'   # FileMenuSpec    # file
		       'Configs'# ConfigsMenuSpec # configs
		       'MathWeb'# MathwebMenuSpec # mathweb
		      ]
		      MakeMenubutton}
		    {Map[
			 'Help' # HelpMenuSpec   # help
			] MakeMenubutton }
		             
		    self.items}
	 end
	 {Menu.file.menu tk(conf tearoff:false)}
	 {Menu.help.menu tk(conf tearoff:false)}	 	 
	 {self.menu tk(conf cursor:arrow)}
      end

      meth addCommand(menu:Menu submenus:Submenus command:Command action:Action state:State)
	 MenuKey = {String.toAtom Menu}
	 Dynamic
      in
	 case {Dictionary.condGet self.dynamics MenuKey false} of false then
	    MenuButton = {New Tk.menubutton tkInit(parent: self.menu
						   text  : {PrintRepresentation Menu})}
	 in
	    Dynamic = {New XManagers.dynamicMenu init(parent:MenuButton)}
	    {MenuButton tk(conf menu:Dynamic)}
	    {Dictionary.put self.dynamics MenuKey Dynamic}
	    {Tk.send pack(MenuButton side:left)}
	 elseof M then
	    Dynamic = M
	 end
	 case Submenus of nil then % no submenu
	    {Dynamic addCommand(command: Command 
				action : Action
				state  : State)}
	 else
	    {ForAll Submenus proc {$ C}
				CascadeKey = {String.toAtom C}
				CascadeMenu
			     in
				case {Dictionary.condGet self.dynamics CascadeKey false} of false then
				   CascadeMenu = {Dynamic addCascade(cascade:C state:State $)}
				   {Dictionary.put self.dynamics CascadeKey CascadeMenu}
				elseof M then
				   CascadeMenu = M
				end
				{CascadeMenu addCommand(command: Command
							action : Action
							state  : State)}
			     end}
	 end
         
      end
            
      meth new
	 {DoEntries self.items self.buttons
          {self.preparemenu newNormal($)}
          normal}
      end
      %%
      meth busy
	 {DoEntries self.items self.buttons
          %BusyDisabled
          {self.preparemenu busyDisabled($)}
          disabled}
      end      
      meth idle 
	 {DoEntries self.items self.buttons
          %IdleNormal
          {self.preparemenu idleNormal($)}
          normal}
      end
      %%
      meth normal(Es)
	 {DoEntries self.items self.buttons Es normal}
      end
      meth disable(Es)
	 {DoEntries self.items self.buttons Es disabled}
      end
      meth state(B Es)
	 if B then
	    {DoEntries self.items self.buttons Es normal}
	 else
	    {DoEntries self.items self.buttons Es disabled}
	 end
      end
      %%
      meth normalPopup(E)
	 {self.E tk(entryconf state:normal)}
      end
      meth disablePopup(E)
	 {self.E tk(entryconf state:disabled)}
      end
      meth statePopup(B E)
	 if B then {self.E tk(entryconf state:normal)}
	 else {self.E tk(entryconf state:disabled)}
	 end
      end
   end %of class
end













