%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% XManagers.oz ; This file is part of the MathWeb system
%
%   major updates: 20.12.2000
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% support for "managers"
%%

functor

import
   Tk

%   System
   
   XTkTools(helpButton /*animlabel*/)
   at 'x-ozlib://mathweb/share/XTkTools.ozf'
   File(exists loadVS)
   at 'x-ozlib://mathweb/share/File.ozf'
   
   Misc(isFeature:    IsFeature
        errorMessage: ErrorMessage
        stringLess:   StringLess)
   at 'x-ozlib://mathweb/share/Misc.ozf'

export
   status      : Status
   symbol      : Symbol
   prepareMenu :PrepareMenu
   dynamicMenu : DynamicMenu

define
   
   %%
   %% status
   %%
   local
      local
         FontFamily      = '-*-helvetica-medium-r-normal--*-'
         FontMatch       = '-*-*-*-*-*-*'
      in 
         StatusFont      = FontFamily#100#FontMatch
      end
   in  
      class Status
         from Tk.frame            
        %prop final		
         
         feat
            Message
            
         meth init(Parent)
            Tk.frame,tkInit(parent: Parent
                            relief: flat)
            
            MessageFrame = {New Tk.frame tkInit(parent: self
                                                height: 20
                                                relief: flat)}
	 
            MessageLabel = {New Tk.label tkInit(parent: MessageFrame
                                                text:   ''
                                                font:   StatusFont)}
         in
            {Tk.batch [pack(MessageLabel side:left)
                       pack(MessageFrame side:top fill:x)]}
            self.Message = MessageLabel
         end
      
         meth setMessage(M)
            {self.Message tk(conf text:M)}
         end
      
         meth clrMessage
            {self.Message tk(conf text:'')}
         end
      
      end %of class   
   end

   %%
   %% symbol
   %%
   class Symbol
      from Tk.frame
     %prop
     %   locking
     %   final

      feat
         Items
         configDir
         symbolDir
      meth init(Parent Status ItemDict ConfigDir  SymbolDir  VirtualStringToValue ?ButtonDict)
         Tk.frame,tkInit(parent: Parent
                         border: 1
                         relief: flat)
         self.symbolDir = SymbolDir
         self.configDir = ConfigDir
	    
         Buttons = Symbol,GetConfiguration($)
      in
         if Buttons==false then
            ButtonDict = {NewDictionary} % WICHTIG!!!
         else
%	 try
            ButtonList = {VirtualStringToValue '['#Buttons#']'}
         in
            ButtonDict = {NewDictionary}
            self.Items = ItemDict
		  
            {ForAll ButtonList
             proc {$ ?Button}
                Type = {Label Button}
             in
                case Type of button then
                   Key = Button.action
                in
                   if {IsFeature Key} then
                      Object
                      Command = {Dictionary.condGet self.Items Key false}
                      Already = {Dictionary.condGet ButtonDict Key false}
                   in
                      if Command\=false then % menu entry exists
                         if Already==false then 
                            Object = {New XTkTools.helpButton
                                      init(parent: self
                                           status: Status
                                           action : self # command(Key)
                                           file   : Button.file
                                           help   : Button.help
                                           message: Button.message
                                           url:self.symbolDir
                                          )}
                            {Dictionary.put ButtonDict Key Object}
                            {Tk.send pack(Object side:left)}
                         else
                            {ErrorMessage 'Same button already exists!'}
                         end
                      else
                         {ErrorMessage 'Unknown menu entry!!!'}
                      end
                   else
                      {ErrorMessage 'Wrong type of menu entry!!!'}
                   end
                [] space then
                   Object = {New Tk.frame
                             tkInit(parent: self
                                    width : 14)}
                in
                   {Tk.send pack(Object side:left)}
                else {ErrorMessage 'Unknown type of symbol!'} end
             end}

         end
      end
	 
      meth GetConfiguration($)
	 if {File.exists self.configDir#'symbolbar.mathweb'} then
            {File.loadVS self.configDir#'symbolbar.mathweb'}
         else
            {ErrorMessage 'File error, no configuration available!'}
            false
         end
      end
	 
      meth command(Key)
         Command = {Dictionary.get self.Items Key} % entry exists!!!
      in
         {Command tk(invoke)}  % invoke corresponding menu entry
      end

   end %of class

   %%
   %% dynamic menu
   %%
   local
      fun {PrintRepresentation String}
         fun {PrintRepresentationRecursive Cs}
            case Cs of nil then nil
            elseof &-|C|Cr then 32|C|{PrintRepresentationRecursive Cr}
            elseof C|Cr then {Char.toLower C}|{PrintRepresentationRecursive Cr}
            end
         end
      in
         case String of nil then nil
         elseof C|Cr then
            C|{PrintRepresentationRecursive Cr}
         end
      end
      
      local
         proc {InsertEntryR Xs E I ?Ys ?P}
            case Xs of nil then Ys=[E] P=I
            [] X|Xr then
               if X==E then % entry already exists
                  Ys=X|Xr P=false
               elseif {StringLess X E} then
                  Ys=X|{InsertEntryR Xr E I+1 $ P}
               else
                  Ys=E|X|Xr P=I
               end
            end
         end
      in
         proc {InsertEntry Xs E ?Ys ?P}
            {InsertEntryR Xs E 0 ?Ys ?P}
         end
      end
   in
      class DynamicMenu
         from
            Tk.menu
	    
         prop
            final
	    
         attr
            EntryList

         feat
            EntryDict
	    
         meth init(parent:Parent)
            Tk.menu,tkInit(parent : Parent
                           tearoff: false)
            EntryList <- nil
            self.EntryDict = {Dictionary.new}
         end
	 
         meth addCommand(command:Command action:Action state:State)
            SortedEntries
            Index
         in
            {InsertEntry @EntryList Command ?SortedEntries ?Index}
            EntryList <- SortedEntries
            if Index\=false then
               Entry = {New Tk.menuentry.command tkInit(parent: self
                                                        before: Index
                                                        label : {PrintRepresentation Command}
                                                        action: Action
                                                        state : State)}
            in
               {Dictionary.put self.EntryDict {String.toAtom Command} Entry}
            else
               skip
            end
         end
	 
         meth addSeparator(before: Index)
            _ =  {New Tk.menuentry.separator tkInit(parent: self
                                                    before: Index)}
         end
		 
         meth addCascade(cascade:Cascade state:State $)
            SortedEntries
            Index
         in
            {InsertEntry @EntryList Cascade ?SortedEntries ?Index}
            EntryList <- SortedEntries
            if Index\=false then
               Menu  = {New DynamicMenu init(parent: self)}
               Entry = {New Tk.menuentry.cascade tkInit(parent: self
                                                        before: Index
                                                        label : {PrintRepresentation Cascade}
                                                        menu  : Menu
                                                        state : State)}
            in
               {Dictionary.put self.EntryDict {String.toAtom Cascade} Entry}
               Menu
            else
               false
            end
         end

         meth enableAll
            {ForAll {Dictionary.items self.EntryDict}
             proc {$ E}
                {E tk(entryconf state:normal)}
             end}
         end

         meth disableAll
            {ForAll {Dictionary.items self.EntryDict}
             proc {$ E}
                {E tk(entryconf state:disabled)}
             end}
         end

      end %of class
   end

   %%
   %% PrepareMenu (I should chane the name)
   %%
    local         
      
      fun {First X#_} X end
   
      fun {Those MenuSpec F}
         {Map {Filter MenuSpec
               fun {$ _#Spec} {F Spec} end}
          First}
      end
      fun {Conj Fs}
         fun {$ X} {All Fs fun {$ F} {F X} end} end
      end
      fun {NonQuietExists X}
         {CondSelect X quiet unit} \= only
      end
      fun {ThoseNonQ MenuSpec  F}
         {Those MenuSpec {Conj [F NonQuietExists]}} 
      end
      fun {LabelIn Ls}
         fun {$ X} {Member {Label X} Ls} end
      end
   in
      class PrepareMenu
         feat menuSpec
         meth init(MenuSpec)
            self.menuSpec = MenuSpec
         end
         meth newNormal($)
         
            F = {LabelIn [always true]}
         in
            {ThoseNonQ self.menuSpec F}
         end
         meth busyDisabled($)
            F = {LabelIn [never true false]}
         in
            {ThoseNonQ self.menuSpec  F}
         end
         meth idleNormal($)
            F = {LabelIn [always true false]}
         in
            {ThoseNonQ self.menuSpec F}
         end
      end
    end

   %%
   %% DoEntries
   %%
   proc {DoEntries ID BD Es W}
      case Es of nil then skip
      [] E|Er then {DoEntries ID BD E W} {DoEntries ID BD Er W}
      else
         if {IsFeature Es} then
            Entry  = {Dictionary.condGet ID Es false}
            Button = {Dictionary.condGet BD Es false}
         in
            if Entry\=false then
               case Entry of En#Ks then
                  {En tk(entryconf state:W)} % change state of menu entry
                  {DoEntries ID BD Ks W}     % and recursively subentries
               else
                  {Entry tk(entryconf state:W)} % change state of menu entry
               end
               if Button\=false then
                  {Button tk(conf state:W)}  % change state of button 
               else
                  skip
               end
            else
               {ErrorMessage 'Unknown menu entry to change state:\n'#Es}
            end
         else
            {ErrorMessage 'Wrong type of menu entry to change state:\n'#Es}
         end
      end
   end

end










