%%
%% doris service
%%

functor

import
   Connection
   Open(socket)
   OS(system unlink tmpnam getEnv)
   Property
   
   XApplication          at 'x-ozlib://mathweb/share/XApplication.ozf'
   XURL                  at 'x-ozlib://mathweb/share/XURL.ozf'
   XCompiler             at 'x-ozlib://mathweb/share/XCompiler.ozf'
   XOpen(socket:XSocket) at 'x-ozlib://mathweb/share/XOpen.ozf'
   XTime                 at 'x-ozlib://mathweb/share/XTime.ozf'
   File(saveVS:SaveVS)   at 'x-ozlib://mathweb/share/File.ozf'
   Buffer(fifo:FIFO)     at 'x-ozlib://mathweb/share/Buffer.ozf'
   Env(host:Host display:Display)
   at 'x-ozlib://mathweb/share/Env.ozf'
   Internet(hostname:Hostname)
   at 'x-ozlib://mathweb/share/Internet.ozf'
   BrokerAccess(getBroker)
   at 'x-ozlib://mathweb/sys/BrokerAccess.ozf'
   
   %%Browser(browse:Browse)
   
define
   AppName = 'doris'
   App     = {New XApplication.manager init}
   VS2V    = {XCompiler.new}
   VS2A = VirtualString.toAtom
   
   SavePrefix  = "/tmp/tptp/problem" %% additionaly feature for saving all problems
   %%SaveSuffix  = ".dfg"
   SaveCounter = {NewCell 0}
   %%{Property.put 'time.detailed' true}
   
   %% args
   Spec = record(brokerticket(single  char:[&t &d] type:string)
		 portticketurl(single char:&p      type:string)
		 name(single char:&n               type:atom default:'doris'))
   Args
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {App.showError AppName#': '#S}
      {App.exit 1}
   end
   %% end of args
   
   PortTicketURL = if {HasFeature Args portticketurl} then
		      Args.portticketurl
		   else
		      {OS.getEnv "PORT_TICKET_URL"}
		   end
   		      
   {App.showInfo "This is `"#AppName#"' on `"#Host#"'."}
   
   %% connect to broker
   Broker = {BrokerAccess.getBroker App Args}.send
%   {App.printInfo "Broker = "} {App.show Broker}
   
   proc {NewATP Return ?ATP} 
      {App.printInfo "Requesting ATP-service..."}
      ATP = {Broker request('ATP' $ client:'DORIS')}
      {App.showInfo ' done.'}
      {Wait ATP}
      {App.printInfo "ATP = "} {App.show ATP}
      {App.printInfo "Entering ATP..."}
      {Wait {ATP enter(return:Return client:'DORIS' display:Display
                       done:$)}}
      {App.showInfo ' done.'}
   end
   AvailableProvers = {NewCell nil}
   
   MainThread = {Thread.this}
   local
      proc {DoEvent} {App.showInfo event} end
      proc {DoAlarm} {Thread.terminate MainThread} end
   in
      WatchDog = {New XTime.watchDog init(timeOut: 5*60*1000
                                          onEvent: DoEvent
                                          onAlarm: DoAlarm
                                          start:   false)}
   end

   class Counter
      attr N
      meth init N<-0 end
      meth inc N<-@N+1 end
      meth dec N<-@N-1 end
      meth get($) @N end
   end

   fun {GetEmail Provers}
      case Provers of nil then false#0
      [] Prover|Rest then
         case Prover of email(...)
         then Prover.1#{Value.condSelect Prover max unit}
         else {GetEmail Rest}
         end
      end
   end

   proc {MailTo _ _ _} skip end
   /*
   proc {MailTo Addr Subj Body}
      Tmp = {OS.tmpnam} in
      {SaveVS Body Tmp}
      %%{OS.system "mail -s \""#Subj#"\" "#Addr#" <"#Tmp _}
      {OS.system "mail "#Addr#" <"#Tmp _}
      {OS.unlink Tmp}
   end
   */
   
   fun {CountArgs S}
      L = {List.last S} in
      if {Char.isDigit L}
      then L-&0
      else 1
      end
   end
   
   local
      DefaultMode       = first
      DefaultProvers    = nil
      DefaultTimeout    = 30
      DefaultDomainsize = 10
      Init         = {NewName}
      InitDefaults = {NewName}
      Put          = {NewName}
      Get          = {NewName}
      Clone        = {NewName}
      
      PutAxioms    = {NewName}
      GetAxioms    = {NewName}

      PutFile      = {NewName}
      
      proc {UnknownMethod M}
         {App.showInfo "*** WARNING *** Ignoring following message:"}
         {App.show M}
      end
   in      
      class GeneralAttr
         attr
            dict
         meth !Init(Dict)
            dict <- Dict
         end
         meth !Put(Attrib Val)
            {Dictionary.put @dict Attrib Val}
         end
         meth !Get($)
            {Dictionary.toRecord unit @dict}
         end
         meth !InitDefaults
            dict <- {NewDictionary}
            GeneralAttr,Put(mode       DefaultMode)
            GeneralAttr,Put(provers    DefaultProvers)
            GeneralAttr,Put(timeout    DefaultTimeout)
            GeneralAttr,Put(domainsize DefaultDomainsize)
         end
         meth !Clone($)
            {Dictionary.clone @dict}
         end
         %%
         meth mode(ModeS)
            GeneralAttr,Put(mode {StringToAtom ModeS})
%            {App.showInfo "mode:       "#ModeS}
         end
         meth provers(ProversS)
            Provers               = {VS2V ProversS}
            EmailAddress#EmailMax = {GetEmail Provers}
         in
            GeneralAttr,Put(provers Provers)
            GeneralAttr,Put(email EmailAddress#EmailMax)
%            {App.showInfo  "provers:    "#ProversS}
%            {App.printInfo "email:      "} {App.show EmailAddress}
%            {App.showInfo  "            (max: "#EmailMax#")"}
         end
         meth timeout(TimeoutS)
            GeneralAttr,Put(timeout {StringToInt TimeoutS})
%            {App.showInfo "timeout:    "#TimeoutS}
         end
         meth domainsize(DomainsizeS)
            GeneralAttr,Put(domainsize {StringToInt DomainsizeS})
%            {App.showInfo "domainsize: "#DomainsizeS}
         end       
         %%
         meth otherwise(M)
            {UnknownMethod M}
         end
      end
      class Axioms
         attr axioms
         meth !Init
            axioms <- {NewDictionary}
         end
         meth !PutAxioms(Format AxiomsS)
            {Dictionary.put @axioms Format AxiomsS} 
         end
         meth !GetAxioms(Format $)
            {Dictionary.condGet @axioms Format unit}
         end
         meth dfg(AxiomsS)
            Axioms, PutAxioms(dfg    {ByteString.make AxiomsS})
         end
         meth otter(AxiomsS)
            Axioms, PutAxioms(otter  {ByteString.make AxiomsS})
         end
         meth kimba(AxiomsS)
            Axioms, PutAxioms(kimba  {ByteString.make AxiomsS})
         end
         meth chorus(AxiomsS)
            Axioms, PutAxioms(chorus {ByteString.make Axioms})
         end
      end
      class Pack from GeneralAttr
         attr ax
         meth !Init(Ax)
            ax <- Ax
            GeneralAttr,InitDefaults
         end
         meth !GetAxioms(Format $)
            {@ax GetAxioms(Format $)}
         end
      end
      class Prob from GeneralAttr
         feat pack
         attr files
         meth !Init(PackageObj)
            self.pack = PackageObj
            files    <- {NewDictionary}
            GeneralAttr, Init({PackageObj Clone($)})
         end
         meth !PutFile(Format FileVS)
            {Dictionary.put @files Format {ByteString.make FileVS}}
         end
         meth !Get($)
            GeneralAttr, Put(files {Dictionary.toRecord unit @files})
            GeneralAttr,Get($)
         end
         meth problem_id(ID)
            GeneralAttr, Put(id {StringToInt ID})
         end
         meth dfg(FileS)
            Prob, PutFile(dfg FileS)          
         end
         meth otter(FileS)
            Prob, PutFile(otter FileS)
         end
         meth kimba(FileS)
            Prob, PutFile(kimba FileS)
         end
         meth chorus(FileS)
            Prob, PutFile(chorus FileS)
         end
         meth dfg2(PrefixS SuffixS)
            AxiomsS = {self.pack GetAxioms(dfg $)} in
            {self dfg(PrefixS#AxiomsS#SuffixS)}
         end
         meth otter2(PrefixS SuffixS)
            AxiomsS = {self.pack GetAxioms(otter $)} in
            {self otter(PrefixS#AxiomsS#SuffixS)}
         end
         meth kimba2(PrefixS SuffixS)
            AxiomsS = {self.pack GetAxioms(kimba $)} in
            {self kimba(PrefixS#AxiomsS#SuffixS)}
         end
         meth chorus2(PrefixS SuffixS)
            AxiomsS = {self.pack GetAxioms(chorus $)} in
            {self chorus(PrefixS#AxiomsS#SuffixS)}
         end
      end

      class AnswersPipe
         feat
            counter % Obj
            fifo    % Obj
            max     % Int
            onExit  % Proc/0
         attr
            send     % Proc/1
            senderOn % Bool
            
         meth init(send:     DoSend
                   senderOn: SenderOn <= true
                   onExit:   DoExit   <= proc {$} skip end)
            self.counter = {New Counter init}
            self.fifo    = {New FIFO init}
            {self setSend(DoSend)}
            senderOn <- SenderOn
            self.onExit = DoExit
         end
         meth setSend(DoSend)
            send <- DoSend
         end
         meth flush(DoSend<=@send)
            {ForAll {self.fifo getAll($)} DoSend}
            {self terminateIfAllSent}
         end
         meth senderOn
            senderOn <- true
            {self flush}
         end
         meth senderOff
            senderOn <- false
         end
	 meth total(Total)
            self.max = Total
            {self senderOn}
            {self terminateIfAllSent}
         end
	 meth receive(Msg)
%            {App.showInfo {XTime.localTimeHHMMSS}#"  "#Msg}
	    {self.counter inc}
	    if @senderOn then
	       {@send Msg}
               {self terminateIfAllSent}
	    else
	       {self.fifo put(Msg)}
	    end
	 end
         meth terminateIfAllSent
            if {IsDet self.max} andthen
               {self.counter get($)}==self.max then
%               {App.showInfo "All answers sent."}
               {self terminate}
            end
         end
         meth terminate
            {self.onExit}
            {Thread.terminate {Thread.this}}
         end
         meth ping($)
            unit
         end
      end

      class StatusStack
         attr
            stack:  nil
            status: unit
            
         meth push(X)
            stack  <- @status|@stack
            status <- X
         end
         meth pop(?T<=_)
            R in
            T|R = @stack
            stack <- R
         end           
      end
      
      class SocketManager from StatusStack     
         feat
            socket
            registry
         attr
            axioms
            package
            problem
            numberProblems
            doSend
            atp
            receiverObj
            available
            'ID'
         %%
         %% connection loop
         %%
         meth init(Sock LocalRegistry)
            self.socket   = Sock
            self.registry = LocalRegistry 
            status  <- init
            axioms  <- {New Axioms Init}
            package <- {New Pack   Init(@axioms)}
            problem <- {New Prob   Init(@package)}
            SocketManager, Loop
         end
         meth Loop
            Str = {self.socket getS($)} in
%            {App.showInfo "[Read]: "#Str}
            case Str %{self.socket getS($)}
            of "info"               then SocketManager, info
            [] "available_services" then SocketManager, available_services
            [] "begin_package"      then SocketManager, begin_package
            [] S then
	       {App.showError "Unknown message: "#S}
               {self.socket putS("unknown_command.")}
               {Thread.terminate {Thread.this}}
            end
            SocketManager, Loop
         end
         meth Exec(S)
            N = {CountArgs S}
            M = {Tuple.make {StringToAtom S} N}
         in
            {For 1 N 1 proc {$ I}
                          M.I = {self.socket getS($)}
%                          {App.showInfo "[Arg"#I#"]: "#M.I}
                       end}
            SocketManager, M
         end
         meth Exit(X<=unit)
            {self.registry.parent terminate(self.registry.id exit:X)}
         end
         %%
         %% div. commands
         %%
         meth info
            {self.socket putS("this_is_an_experimental_version.")}
            {Thread.terminate {Thread.this}}
         end
         meth available_services
            case {Access AvailableProvers} %{@ availableProvers($)}
            of nil then
               {self.socket putS("[].")}
            [] First|Rest then
               fun {Enumerate Xs}
                  case Xs of nil then ""
                  [] X|Xr then ","#X#{Enumerate Xr} end
               end
            in {self.socket putS("["#First#{Enumerate Rest}#"].")}
            end
            {Thread.terminate {Thread.this}}
         end
         %%
         %% package loop
         %%
         meth begin_package
            try
               SocketManager, Begin_package
            finally
               if {IsDet @available} then
                  {Assign AvailableProvers @available}
               end
               try {@atp leave}
               catch _ then skip end
               {self.registry tryTerminateAll}
               {self.socket putS("eos.")}
            end
            {Thread.terminate {Thread.this}}
         end
         meth Begin_package
            {@package InitDefaults}
%            {App.showInfo {XTime.localTimeStamp}#"  "#"begin_package"}
%            {App.show {Property.get time}}
            numberProblems <- {New Counter init}
            proc {DoSend AnswerVS}
               {self.socket putS(AnswerVS#".")}
            end
            proc {DoExit}
               {self.registry.parent terminate(self.registry.id exit:done)}
            end
         in
            doSend      <- DoSend
            receiverObj <- {self.registry newServer(AnswersPipe
                                                    init(send:   DoSend
                                                         onExit: DoExit
                                                        ) $
                                                    detach: false
                                                    info:   "receiver")}
            atp         <- {NewATP proc {$ VS} {@receiverObj receive(VS)} end}
            local AvProv  = {@atp availableProvers($)}
            in	
              {Wait AvProv}
              available <- AvProv
            end	
            {self push(package)}
            SocketManager, PackageLoop
         end
         meth PackageLoop
            Str = {self.socket getS($)} in
%            {App.showInfo "[PackageRead]: "#Str}
            case Str %{self.socket getS($)}
            of "end_package" then
               SocketManager, end_package
            [] "begin_axioms" then
               SocketManager, begin_axioms
               SocketManager, PackageLoop
            [] S then
               SocketManager, Exec(S)
               SocketManager, PackageLoop
            end
         end
         meth end_package
%            {App.showInfo "***************** end *******************"}
%            {App.showInfo {XTime.localTimeHHMMSS}#"  "#"All problems read."}
%            {App.show {Property.get time}}
            {@receiverObj total({@numberProblems get($)})}
            
            if {self.socket getS($)} == "stop" then
               {App.showInfo "****************** stop *****************"}
%               {App.showInfo "Writing partial answers..."}
               {Wait {@atp returnRestAndLeave('proc':@doSend done:$)}}
%               {App.showInfo "Writing partial answers... done"}
               {self Exit(stop)}
            else
               {App.showError stop_expected}
               try {self.socket putS("stop_expected.")}
               catch _ then skip end
               {self Exit(error(stop_expected))}
            end
            {self pop}
         end
         %%
         %% axioms loop
         %%
         meth begin_axioms
            {self push(axioms)}
            SocketManager, AxiomsLoop
         end
         meth AxiomsLoop
            Str = {self.socket getS($)} in
%            {App.showInfo "[AxiomsRead]: "#Str}
            case Str %{self.socket getS($)}
            of "end_axioms" then
               SocketManager, end_axioms
            [] S then
               SocketManager, Exec(S)
               SocketManager, AxiomsLoop
            end
         end
         meth end_axioms
            {self pop}
         end
         %%
         %% problem loop
         %%
         meth begin_problem(ID)
            'ID' <- if ID == "problem_id"
                    then {self.socket getInt($)}     
                    else {StringToInt ID}
                    end
            {@problem Init(@package)}
            {self push(problem)}
            SocketManager, ProblemLoop
         end
         meth ProblemLoop
            Str = {self.socket getS($)} in
%	    {App.showInfo "[ProblemRead]: "#Str}
            case Str %{self.socket getS($)}
            of "end_problem" then
               SocketManager, end_problem
            [] S then
               SocketManager, Exec(S)
               SocketManager, ProblemLoop
            end
         end
         meth end_problem
	    P  = {@problem Get($)}
            ID = @'ID'
         in
            {@numberProblems inc}
            
            local
               EmailAddress#EmailMax = {Value.condSelect P email false#0}
            in
               if (EmailAddress\=false) andthen
                  (EmailMax==unit orelse ID=<EmailMax)
               then
                  {Record.forAllInd P.files
                   proc {$ Syntax ProblemBS}
                      {MailTo EmailAddress Syntax ProblemBS}
                   end}
               end
            end
            
%            {App.printInfo {XTime.localTimeHHMMSS}#"  "}
%            {App.showInfo "Starting "#ID#{XTime.ord ID}#" problem... "}

	    %% Save all problems
	    if {Property.condGet 'doris_save' false}==true andthen
               {Arity P.files}\=nil then
               NewCounter = {Access SaveCounter}+1 in
	       {Assign SaveCounter NewCounter}               
               {ForAll {Arity P.files}
                proc {$ Format}
                   {File.saveVS P.files.Format
                    SavePrefix#NewCounter#"."#Format}
                end}
	    end
            {@atp prove(id:ID
			problem: P.files
			provers: P.provers
			mode:    P.mode
			timeOut: P.timeout)}
	    
	    {self pop}
	 end
	 %%
	 meth otherwise(M)
	    case @status
	    of package then {@package M}
	    [] axioms  then {@axioms  M}
	    [] problem then {@problem M}
	    else {UnknownMethod M}
            end
	 end
      end
   end
      
%   ATP = {NewATP...}
%   try
   %% install new port server
   {App.showInfo "Creating new server socket."}
   S = {New Open.socket init}
   {App.printInfo "Binding socket..."} 
   ServicePort
   {S bind(port:?ServicePort)}
   {App.showInfo " done."}
   {S listen}
   {App.showInfo "Listening..."}
   %% port server installed

   proc {Available Sock}
      {Sock write(vs:"host_port('"#Hostname#"',"#ServicePort#").\n")}
   end
   /*
   proc {NotAvailable Sock}
      {Sock write(vs:"service_not_available.\n")}
   end
   */
   class AcceptClass
      from XSocket
      attr format: fun {$ VS} VS end
      meth setFormat(Format)
         format<-Format
      end
      meth putS(VS)
         {App.showInfo {@format VS}}
         XSocket, putS(VS) 
      end
   end
      
   proc {AcceptLoop}
      Sock P H ThrID Format StartSignal
      proc {DoThread LocalRegistry}
         {Wait StartSignal}
         _ = {New SocketManager init(Sock LocalRegistry)}
      end
      proc {DoError Err}
         {App.showError {Format "Error in thread "#ThrID#":"}}
         {App printExc(Err)}
      end
      proc {DoExit ExitStatus}
         try
            {App.printInfo {Format "Shutting down socket [send]..."}}
            {Sock shutDown(how:[send])}
            {App.showInfo " done."}
            {App.printInfo {Format "Shutting down socket [receive]..."}}
            {Sock shutDown(how:[receive])}
            {App.showInfo " done."}
         catch system(os(os "shutdown" 134 ...)...) then
            {App.showInfo " done."} %Transport endpoint is not connected
         [] system(os(os "shutdown" 131 ...)...) then
            {App.showInfo " done."} %Connection reset by peer
	 [] _ then
	    {App.showInfo {Format "Error while shutting down socket:"}}
%            {App printExc(Err)}
         end
         {App.printInfo {Format "Closing socket..."}}
         {Sock close}
         {App.showInfo " done."}
      end
   in
      {S accept(acceptClass:AcceptClass accepted:?Sock host:?H port:?P)}
      {WatchDog event}
      {Sock doInit(readStopC:     0
		   writeAppendVS: "\n")}
      {App newThread(DoThread ?ThrID info:'doris('#H#','#P#')'
                     onError:DoError onExit:DoExit)}
      {Wait ThrID}
      Format = fun {$ VS}
		  {XTime.localTimeHHMMSS}#" ["#AppName#" "#ThrID#"]: "#VS
               end
      {Sock setFormat(Format)}
      {App.showInfo {Format "Connection accepted from port "#P#" at "#H#"."}}  
      StartSignal = unit
      {AcceptLoop}
   end
in
   try
      ServiceID PortTicket
      {App.printInfo "Loading `"#PortTicketURL#"'..."}
      PortTicket = {VS2A {XURL.loadVS PortTicketURL}}
      {App.showInfo ' done.'}      
      
      {App.printInfo "Connecting to `"#PortTicket#"'..."}
      PortPoint = {Connection.take PortTicket} 
      {App.showInfo ' done.'}
      {Wait PortPoint}
      {App.printInfo "PortPoint = "} {App.show PortPoint}
      {PortPoint offer(Args.name Available ?ServiceID host:Host)}
   in
      try
	 /*
	 thread
            {Delay 5*60*1000}
	    {App replicate}
	    {WatchDog start}
         end
	 */
	 {AcceptLoop}
      finally
	 try
	    {PortPoint retire(ServiceID)}
	 catch Err then
	    {App.showError 'Error in "{PortPoint retire(...)}":'}
            {App printExc(Err)}
	 end
      end
   catch system(kernel(terminate...)...) then
      {App.showInfo '\n*****************************************************************'}
      {App.showInfo 'DORIS is going down.'}
      {App.showInfo '*****************************************************************'}
   [] error(connection...) then
      {App.showInfo '\n*****************************************************************'}
      {App.showInfo 'DORIS: The port-daemon is not running, please start port-up first.'}
      {App.showInfo '*****************************************************************'}
   [] error(connection(illegalTicket ...)...) then
      {App.showInfo '\n*****************************************************************'}
      {App.showInfo 'DORIS: The port-daemon runs with an old Mozart version :-('}
      {App.showInfo '*****************************************************************'}
   [] Err then
      {App.showInfo '*****************************************************************'}
      {App printExc(Err)}
      {App.showInfo '*****************************************************************'}
   finally
   try {S shutDown}
      catch _ then skip end
      {S close}
   end
%   finally
%      try {ATP leave}
%      catch _ then skip end
%   end
end






