%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LispSocket.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%  An Allegro Common Lisp socket access to MathWeb.
%%%
functor
   
import 
   Connection Error OS
   Browser(browse: Browse)
   
   XCompiler at 'x-ozlib://mathweb/share/XCompiler.ozf'
   BrokerAccess(getBroker)
   at 'x-ozlib://mathweb/sys/BrokerAccess.ozf'
   
export
   manager: OzService
   
define
   
   VS2A = VirtualString.toAtom
   VS2S = VirtualString.toString
   VS2V = {XCompiler.new}
   
   fun {ExceptionToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end
   
   %%
   %% An auxiliary class for Debug-Output
   %%
   
   class Output

      attr
         debugFlag
         
      meth setDebugFlag(DebugFlag)
         debugFlag <- DebugFlag
      end

      meth debug(X)
         if @debugFlag == true then
            {Browser.browse X}
         end
      end
      
   end
 
   %%
   %% OzService
   %%
   
   local
      fun {DoQuote S}
         case S of nil then nil
         [] C|Sr then
            if {Member C [&" &\\]}
            then &\\|C|{DoQuote Sr}
            else C|{DoQuote Sr}
            end
         end
      end
   in
      fun {Quote VS}
         '"'#{DoQuote {VS2S VS}}#'"'
      end
   end
   
   class OzService from Output
      
%       prop
% 	 final
         
      feat
	 stream
	 openServices
	 broker
         myLoui
	 display
	 mode
	 serverName
         translator
         
      attr
	 debug: false
	 browseResult: true
         LoopThread:_
         
      meth init(Stream
		url:     BrokerUrl
		ticket:  BrokerTicket
		debug:   Debug    <= false
		app:     App
		myLoui:  MyLoui   <= false
		display: Display  <= ":0.0"  % hack for PREX
		mode:    Mode     <= emacs   % hack for PREX
		browseResult: BrowseResult <= false
		serverName: ServerName     <= "SomeServer"
                translate: Translator      <= unit
	       )
         self.display    = Display
	 self.mode       = Mode
	 self.serverName = ServerName
         self.translator = Translator
         debug <- Debug
         Output, setDebugFlag(Debug)
         browseResult <- BrowseResult
         Output, debug('OzService, init')
         thread
            try
	       self.stream = Stream
	       self.openServices = {NewDictionary}
	       Output, debug('Connecting to Broker...')
	       if BrokerTicket \= false then
		  self.broker = {Connection.take BrokerTicket}
	       else
		  self.broker = {BrokerAccess.getBroker App a(brokerticket: BrokerUrl)}.send
	       end
	       Output, debug('connected.')
	       self.myLoui = MyLoui
	       Output, debug('myLoui is '#self.myLoui)
	       %%
	       LoopThread <- {Thread.this}
	       OzService, Loop
	    catch Err then
	       try
		  if Err \= closeException then
                     Output, debug('Exception catched. Closing...')
                     Output, debug(Err)
		  end
		  OzService, close
               catch _ then skip end
	    end
	 end
      end
     
      meth shutdown
         Output, debug('shutdown...')
	 if {IsFree @LoopThread}==false then
            try {Thread.injectException @LoopThread closeException}
            catch _ then skip end
            LoopThread <- _
         end
         Output, debug('shutdown done.')
      end
      
      meth sendAnswer(Answer)
	 Output, debug('sendAnswer'#Answer)
	 {self.stream putS('(0 '#{Quote Answer}#')')}
      end      
      meth sendError(ErrNr ErrMsg)
	 Output, debug('sendError'#ErrMsg)
	 {self.stream putS('('#{IntToString ErrNr}#' '
			   #{Quote ErrMsg#"\n"}#')')}
      end
     
      meth Loop
	 Output, debug('Waiting for message...')
	 Message = {Map {self.stream getS($)}
		    fun {$ C}
		       %% this is tricky and dirty: the '$'-char is
		       %% supposed to be used correctly and only once;
		       %% it is reinserted later 
		       if C==&$ then &  else C end
		    end}
      in
	 Output, debug('Message read: '#Message)
	 try
	    Method = {VS2V Message}
	 in
	    Output, debug('compiled: '#Method)
	    if {IsDet Method} andthen {IsRecord Method}
	    then
	       if {IsAtom Method}
	       then
		  case Method   
		  of debugOn         then OzService, SimpleTry(debugOn)
		  [] debugOff        then OzService, SimpleTry(debugOff)
		  [] browseResultOn  then OzService, SimpleTry(browseResultOn)
		  [] browseResultOff then OzService, SimpleTry(browseResultOff)
		  else OzService, UnknownCommand
		  end
               else
		  case {Label Method}
		  of available    then OzService, TryAvailable(Method)
                  [] enter        then OzService, TryEnter(Method)  
                  [] leave        then OzService, TryLeave(Method)
		  [] close        then OzService, TryClose(Method)
		  [] applyMethod  then OzService, TryApplyMethod(Method)
		  [] sendKqml     then OzService, TrySendKqml(Method)
		  [] sendKqmlAndWait then OzService, TrySendKqmlAndWait(Method)
                  else OzService, UnknownCommand
                  end
               end
            else OzService, sendError(2 'Message must be a record.')
            end
         catch Err then
	    OzService, sendError(1 'Compiler cannot read the message.'#
                                 {ExceptionToVS Err})
	 end
	 OzService, Loop
      end
      
      meth UnknownCommand
         OzService, sendError(3 'Unknown Command. '
                              #'Allowed Commands are:'#"\n"
                              #'available(ServiceName),'#"\n"
                              #'enter(erviceName),'#"\n"
                              #'applyMethod(Method ServiceName),'#"\n"
                              #'leave(ServiceName),'#"\n"
%			      #'restart(ServiceName),'#"\n"
                              #'close,'#"\n"
                              #'debugOn,'#"\n"#'debugOff,'#"\n"
			      #'browseResultOn,'#"\n"#'browseResultOff.'#"\n")
      end
      
      %% Table of Error Codes:
      %%
      %%   0 <This is not an error code. Everything went allright.>
      %%
      %%   1 Compiler cannot read the message.         (method Loop)
      %%   2 Message must be a record.                 (method Loop)
      %%   3 Unknown Command.                          (method Loop)
      %%
      %%  10 Error using command 'enter'.              (method TryEnter)
      %%  11 Service already open.                     (method enter)
      %%  12 Time out while creating ServiceObject.    (method enter)
      %%  13 Service unknown.                          (method enter)
      %%  14 Error while creating ServiceObject.       (method enter)
      %%
      %%  20 Error using command 'leave'.              (method TryLeave)
      %%  21 Service not open.                         (method leave)
      %%  22 Time out while leaving service.           (method leave)
      %%
      %%  30 Error using simple command.               (method SimpleTry)
      %%
      %%  40 Time out while leaving service.           (method close)
      %%
      %%  50 Error using command 'available'.          (method TryAvailable)
      %%  51 Timeout in 'available.'                   (method available)

      %%  80 Error using command 'sendKqml'            (method TrySendKqml)
      %%
      %%  90 Error using command 'applyMethod'.        (method TryApplyMethod)
      %%  91 Service not open.                         (method applyMethod)
      %%  92 Error while executing 'applyMethod'.      (method applyMethod)
      %%  93 Time out while applying method.           (method applyMethod)
      %%  94 Result is not a det value.                (method applyMethod)
      %%  95 Result is not a VirtualString.            (method applyMethod)
      %%  96 User defined error has illegal structure. (method applyMethod)
      %%
      %% 100-...: User defined error returned by the service.
      
      meth TryAvailable(Method)
         try
            Output, debug('TryAvailable'#Method)
            OzService, Method
         catch Err then
            OzService, sendError(50 'Error using command \'available\'.'#"\n"
                                 #'Usage: available(ServiceName)'#
                                 {ExceptionToVS Err})
         end
      end
      
      meth available(ServiceName atURL:RemoteURL<=unit timeout:Timeout<=60)
         Output, debug('available:'#ServiceName)
         Result = {self.broker available(ServiceName atURL:RemoteURL $)}
      in
         {WaitOr Result {Alarm Timeout*1000}}
         if {IsFree Result} then
            OzService, sendError(51 "Timeout in 'available' ("#Timeout#"s).")
         elseif Result
         then OzService, sendAnswer("t")
         else OzService, sendAnswer("nil")
         end
      end
      
      meth TryEnter(Method)
         try
            Output, debug('TryEnter'#Method)
	    OzService, Method
	 catch Err then
	    OzService, sendError(10 'Error using command \'enter\'.'#"\n"
				 #'Usage: enter(ServiceName)'#
				 {ExceptionToVS Err})
	 end
      end
      
      meth enter(ServiceName                  %% Service name (atom)
		 atURL  : RemoteURL<= unit    %% URL of the broker
		 timeout: Timeout  <= 20      %% timeout for starting service
		 mode   : Mode     <= emacs   %% the mode you want to enter the service in
		 unique : Unique   <= false)  %% if you want a unique service name (e.g. "MAPLE23433")
	 
	 Output, debug('enter:'#ServiceName)
	 if {Dictionary.member self.openServices ServiceName} andthen
	    {Not Unique} then 
	    %%OzService, sendError(11 'Service '#ServiceName#' already open.')
	    %%To avoid the MAPLE error (ask cullrich & jzimmer), replaced by:
	    OzService, sendAnswer('Service '#ServiceName#' opened.')
	 else
	    Output, debug('enter: creating ServiceObject...')
	    ServiceObject Done
	 in
	    case ServiceName of 'MYLOUI' then
	       %% this is for loading local theory files over Loui
	       %% (loadURL method) if Loui and Omega run on different file systems
	       if self.myLoui==false then
		  ServiceObject=false Done=false
	       else
		  MYLOUI = self.myLoui in
		  ServiceObject = {New class
					  meth init(...)  skip end
					  meth enter(...) skip end
					  meth leave(...) skip end
					  meth otherwise(M) {MYLOUI M} end
				       end init}
		  {ServiceObject ping(?Done)}
	       end
	    else
	       {self.broker request(ServiceName ?ServiceObject
				    atURL:  RemoteURL
				    client: 'Allegro'
				    done:   ?Done)}
	    end
	    {WaitOr Done {Alarm Timeout*1000}}
	    if {IsFree Done} then
	       OzService, sendError(12 'Timeout while creating ServiceObject ('#Timeout#'s).')
	    else
	       case Done of unit then
		  Output, debug('ServiceObject '#ServiceObject#' created.')
		  if {Member ServiceName ['LPLAN' 'PREX' 'OMEGA' 'TPS']} then
		     {ServiceObject enter(client: unit(name:    self.serverName
						       display: self.display)
					  args:   unit(mode:    Mode)
					 )}
		  else
		     {ServiceObject enter(client: 'Allegro')}
		  end
		  %% store the service object under the ServiceName
		  UniqueName = if Unique==false then
				  ServiceName
			       elsecase {Dictionary.condGet self.openServices ServiceName unit}
			       of unit then
				  ServiceName
			       else
				  OzService, NewID(ServiceName $)
			       end
	       in
		  {Dictionary.put self.openServices UniqueName ServiceObject}
		  OzService, sendAnswer(UniqueName)
	       [] false then
		  OzService, sendError(13 'Service '#ServiceName#' unknown.')
	       [] timeout(create) then
		  OzService, sendError(12 'Timeout while creating ServiceObject. ('#Timeout#'s)')
	       elseof Err then
		  OzService, sendError(14 "Error while creating ServiceObject:"
				       #{ExceptionToVS Err})
               end
            end
	 end
      end

      meth NewID(ServiceName $)
	 NewName =  {VS2A ServiceName#{OS.rand}}
      in
	 if {Dictionary.member self.openServices NewName} then
	    {self NewID(ServiceName $)}
	 else
	    NewName
	 end
      end
      
      meth TrySendKqml(Method)
	 try
	    {self Method}
	 catch Err then
	    OzService, sendError(90 'Error using command \'sendKqml\'.'#
				 "\n"#
				 'Usage: sendKqml(Receiver Performative Content)'#
				 {ExceptionToVS Err})
	 end
      end

      meth TrySendKqmlAndWait(Method)
	 try
	    {self Method}
	 catch Err then
	    OzService, sendError(90 'Error using command \'sendKqmlAndWait\'.'#
				 "\n"#
				 'Usage: sendKqml(Performative Service Language Content)'#
				 {ExceptionToVS Err})
	 end
      end

      
      meth Apply(Obj Meth ?Result)
	 fun {GetFirstUnusedIntFeature R Start}
	    if {HasFeature R Start}
            then {GetFirstUnusedIntFeature R Start+1}
            else Start
	    end
	 end
	 I = {GetFirstUnusedIntFeature Meth 1}
	 Ar= {Arity Meth}
	 M = {MakeRecord {Label Meth} I|Ar}
      in
	 {ForAll Ar proc {$ X} M.X=Meth.X end}
         M.I=Result
         Output, debug('Apply: "{'#Obj#' '#M#'}"')
	 {Obj M}
      end
      
      meth TryApplyMethod(Method)
         try
            Output, debug('TryApplyMethod'#Method)
	    OzService, Method
	 catch Err = _ then
	    OzService, sendError(90 'Error using command \'applyMethod\'.'#
				 "\n"#
				 'Usage: applyMethod(Method ServiceName)'#
				 {ExceptionToVS Err})
	 end
      end

      meth HandleResult(Data id: MessageID <= unit)
         Output, debug('Result is present.')
%         {Browse 'Result is: '(Data MessageID)}
         if {IsDet Data}
         then
            %% Virtual strings are sent directly to the LISP process
            if {IsVirtualString Data} then
               case MessageID
               of unit then
                  OzService, sendAnswer(Data)
               [] ID then
                  OzService, sendAnswer("'("#ID#" "#Data#")")
               end
               %% Records are either an error or they must be translated by the @translator
            elseif {IsRecord Data} then
               %% send errors with sendError
               if {Label Data} == 'ERROR' then
                  try
                     OzService, sendError(100+Data.1 Data.2)
                  catch Err then
                     OzService, sendError(96 'User-defined error has illegal structure.'#{ExceptionToVS Err})
                  end
               else
                  %% browse result, ONLY FOR DEBUGGING
                  if @browseResult then
                     {Browser.browse Data}
                     OzService, sendAnswer('The result is shown in the \'Oz Browser\' window.')
                  else
                     %% translate the result with service specific translators.
                     case self.translator
                     of unit then
                        try
                           OzService, sendError(95 'Result is not a VirtualString.'
                                                #"\n"#{Value.toVirtualString
                                                       {Value.status Data} 3 3})
                        catch Err then
                           OzService, sendError(96 'User-defined error has illegal structure.'#{ExceptionToVS Err})
                        end
                     [] T then
                        Result = {T translate(Data $)}     %% use the service specific translator
                     in
                        case MessageID
                        of unit then
                           OzService, sendAnswer(Result)
                        [] ID then
                           OzService, sendAnswer("'("#ID#" "#Result#")")
                        end
                     end
                  end
               end
            elseif @browseResult then
               {Browser.browse Data}
               OzService, sendAnswer('The result is shown in the \'Oz Browser\' window.')
            else
               OzService, sendError(95 'Result is not a VirtualString.'#"\n"
                                    #{Value.toVirtualString {Value.status Data} 3 3})
            end   
         elseif @browseResult
         then 
            {Browser.browse Data}
            OzService, sendAnswer('The result is shown in the \'Oz Browser\' window.')
         else
            OzService, sendError(94 'Result is not a det value.')
         end
      end
      
      
      meth applyMethod(ServiceName
                       Method
                       id:      MessageID <= unit
                       timeout: Timeout   <= 60
                      )
         Output, debug('applyMethod'#Method#ServiceName)
         if {Dictionary.member self.openServices ServiceName}
         then      
	    ServiceObject = {Dictionary.get self.openServices ServiceName}
            Data
	 in
            try
	       Output, debug('calling Apply...')
	       OzService, Apply(ServiceObject Method Data)
	       Output, debug('Apply done.')
	       try
		  %% Wait for data, but only for a limited time.
		  ThreadMarker = {Thread.this}
		  Ok
	       in
                  thread
		     try
			{Delay Timeout*1000}
			if {IsFree Ok} then
			   {Thread.injectException ThreadMarker timeOut}
                        end
                     catch _ then skip end
                  end
		  case MessageID of unit
                  then
                     {Wait Data}
                     Ok = unit
                     %% waiting done.
                     OzService, HandleResult(Data)
                  [] ID then
%                     OzService, sendAnswer("T")  % we send nothing! (ASYNCHRONOUS!)
                     thread
                        {Wait Data}
                        OzService, HandleResult(Data id: ID)
                     end
                  end
               catch
                  timeOut then
                  OzService, sendError(93 'Time out while applying method ('#Timeout#'s).'#"\n"
                                       #'Probably the specified method does not exist.'#"\n"
                                       #'If the service hangs, you have to restart it.')        
               end
            catch Err then
               OzService, sendError(92 'Error while executing applyMethod.'#
                                    {ExceptionToVS Err})
            end
         else
            OzService, sendError(91 'Service '#ServiceName#' not open.')
         end
      end
      
      meth TryLeave(Method)
         try
            Output, debug('TryLeave'#Method)
            OzService, Method
         catch Err then
            OzService, sendError(20 'Error using command \'leave\'.'#"\n"#
                                 'Usage: leave(ServiceName)'#
                                 {ExceptionToVS Err})
         end
      end
      
      meth leave(ServiceName timeout:Timeout<=60)
         Output, debug('leave'#ServiceName)
	 if {Dictionary.member self.openServices ServiceName}
	 then
	    ServiceObject = {Dictionary.get self.openServices ServiceName}
         in
            Output, debug('leave: leaving '#ServiceName)
	    try
	       %% Try leaving the service, but only for a limited time.
	       ThreadMarker = {Thread.this}
	       Ok
               thread
                  try
                     {Delay Timeout*1000}
                     if {IsFree Ok} then
                        {Thread.injectException ThreadMarker timeOut}
                     end
                  catch _ then skip end
	       end
	       {ServiceObject leave}
	    in
               Ok = unit
               %% leaving done.
               Output, debug('leaving done.')
	       {Dictionary.remove self.openServices ServiceName}
	       OzService, sendAnswer('Service '#ServiceName#' closed.')
	    catch timeOut then
               OzService, sendError(22 'Time out while leaving service '#ServiceName#' ('#Timeout#'s).')
	    end
	 else
	    OzService, sendError(21 'Service '#ServiceName#' not open.')
	 end
      end

      meth TryClose(Method)
         try
            Output, debug('TryClose'#Method)
            OzService, Method
         catch Err then
            OzService, sendError(20 'Error using command \'close\'.'#"\n"#
                                 'Usage: close'#
                                 {ExceptionToVS Err})
         end
      end
            
      meth close(timeout:Timeout<=60)
         Output, debug('closing...')
         try
            {ForAll {Dictionary.entries self.openServices}
             proc {$ ServiceName#ServiceObject}
                Output, debug('leave: leaving '#ServiceName)
                %% Try leaving the service, but only for a limited time.
                ThreadMarker = {Thread.this}
                Ok
             in
                thread
                   try
                      {Delay Timeout*1000}
                      if {IsFree Ok} then
                         {Thread.injectException ThreadMarker
                          timeOut(ServiceName)}
                      end
                   catch _ then skip end
                end
                {ServiceObject leave}
                Ok = unit
                %% leaving done.
                Output, debug('leaving done.')
                {Dictionary.remove self.openServices ServiceName}
             end}
            OzService, sendAnswer('All services closed.')
            Output, debug('closing done.')
         catch timeOut(ServiceName) then
            OzService, sendError(40 'Time out while leaving service '#ServiceName#' ('#Timeout#'s).')
         end
      end
      
      meth SimpleTry(Method)
         try
            Output, debug('SimpleTry'#Method)
	    OzService, Method
	 catch Err then
	    OzService, sendError(30 'Error using command \''#Method#'\'.'#
				 {ExceptionToVS Err})
	 end
      end
      
      meth debugOn
         Output, debug('debugOn...')
         debug <- true
         Output, setDebugFlag(true)
	 OzService, sendAnswer('Debug switched on.')
	 Output, debug('debugOn... done.')
      end
      meth debugOff
         Output, debug('debugOff...')
         debug <- false
         Output, setDebugFlag(false)
         OzService, sendAnswer('Debug switched off.')
         Output, debug('debug done.')
      end
      meth browseResultOn
         Output, debug('browseResultOn...')
	 browseResult <- true
	 OzService, sendAnswer('BrowseResult switched on.')
         Output, debug('browseResultOn done.')
      end
      meth browseResultOff
         Output, debug('browseResultOff...')
	 browseResult <- false
	 OzService, sendAnswer('BrowseResult switched off.')
	 Output, debug('browseResultOff done.')
      end
   end % class OzService

end
