%%%
%%% INKA service
%%%

functor
   
import               
   Error(exceptionToMessage messageToVirtualString)
   OS(getEnv)
   
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XOpen(server:Stream pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   LispSocket(manager: OzService)
   at 'x-ozlib://mathweb/sys/LispSocket.ozf'
   XService(threadManager: ThreadManager
            clientManager: ClientManager
%           faultManager: FaultManager
            manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   
define
   AppName     = 'inka'
   ServerName = 'INKA'
   App         = {New XApplication.manager init}
   Host        = {OS.getEnv 'HOST'}

   INKALISP = {OS.getEnv "INKALISP"}
   INKAHOME = {OS.getEnv "INKAHOME"}
   
   LoadBoot = '(load "'#INKAHOME#'/lisp/inka/loui-boot.lisp")' 
   
   EOF = 128 
   
   fun {ErrToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end
   
   %%
   %% xterm mode
   %%

   local
      proc {LocalShowInfo Line}
         {App.showInfo "[XTERM]: "#Line}
      end
   in
      class InkaXterm
         from PipeGeneric
         prop final
            
         meth start(display : Display
                    host    : Host
                    showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end
                    ports   : Ports)
            
            PipeGeneric, init(cmd:  'xterm'
                              args: {Append
                                     ["-display" Display
                                      /* Adapted for INKA */
                                      "-title" "Inka Service@"#Host#" (xterm)"
                                      "-geometry" "80x25"
                                      "-fn" "8x13"
                                      "-e" INKALISP
                                      "-qq" "-e" LoadBoot "--" Host]
                                     Ports}
                              showInfo:ShowInfo onExit:ExitProc)
         end
            
         meth close
            PipeGeneric, close
         end
      end
   end

   %%
   %% quiet mode
   %%
   local skip
      proc {LocalShowInfo Line}
         {App.showInfo "[LISP]: "#Line}
      end
   in
      class InkaQuiet
         from PipeGeneric
         prop final
            
         meth start(showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end
                    host    : Host
                    ports   : Ports ...)
            
            PipeGeneric, init(cmd:  INKALISP
                              args: {Append ["-qq" "-e" LoadBoot "--" Host]
                                     Ports}
                              showInfo:ShowInfo onExit:ExitProc)
         end
             
         meth close
            PipeGeneric , close(kill:true)
         end
      end
   end
      
   class InkaService from ThreadManager ClientManager
      
      feat
	 serverName
         brokerURL
         brokerTicket
         args
         
	 inout    % input/output stream
         serv     % mathweb stream

	 ServObj  % the Object providing the Oz-Services 
         InkaPipe
            
      meth init(url:    URL    <=false
                ticket: Ticket <=false
		cmdArgs: CmdArgs <= unit)
	 self.serverName   = ServerName
         self.brokerURL    = URL
         self.brokerTicket = Ticket
         ThreadManager, init
      end
      
      /**
      %% NOTE: ClientInfo = ClientName(name: user: host: display: port:)
      %%       ServerArgs = unit(mode: remoteURL: sessionID:)
      */
      meth enter(client : ClientInfo <= unit
                 args   : ArgRec     <= unit)
         
         ClientManager, init(ClientInfo)

	 {self.appManager.showInfo "Hallo"}

         ClientManager, showEnter(self.appManager.showInfo)
         ClientManager, startThreads
        
         self.args = ArgRec

         InOut Serv

         self.inout = {New Stream init(port:?InOut readStopC:EOF)}
	 self.serv  = {New Stream init(port:?Serv  readStopC:EOF)} 

         Mode = {Value.condSelect ArgRec mode xterm}

         proc {SendClient Msg}
            try {Send self.clientPort Msg}
            catch _ then skip end
         end

         ExitSignal
         proc {DoExit} {SendClient exit} ExitSignal=unit end

	 PipeClass = case Mode
                     of emacs then InkaXterm
                     [] xterm then InkaXterm
                     [] quiet then InkaQuiet
                     else
                        {SendClient error("Mode `"#Mode#"' unknown.")}
                        {DoExit}
                        class
                           meth start(...) skip end
                           meth close(...) skip end
                        end
                     end
         
         DefaultInit = start(display: self.clientDisplay
                             host   : Host
                             onExit : DoExit
                             ports  : [InOut Serv])

         Init = if {HasFeature ArgRec outputLabel} then
                   OutLabel = ArgRec.outputLabel
                   proc {ShowInfo Line} {SendClient OutLabel(Line)} end
                in
                   {Adjoin x(showInfo:ShowInfo) DefaultInit}
                else
                   DefaultInit
                end
      in
         self.InkaPipe = {New PipeClass Init}

	 {Delay 2500} % wichtig !!!

         if {IsFree ExitSignal} then
            {ForAll [inout serv] proc {$ Sock} {self.Sock accept} end}
            
            InkaService, StartReader
            
            self.ServObj = {New OzService init(self.serv
                                               url:    self.brokerURL
					       ticket: self.brokerTicket
					       app: App
					       display:  self.clientDisplay
					       serverName: self.serverName
					      )}
         end
      end

      %%
      %% socket communication between LISP and LOUI
      %%
      
      meth StartReader
	 proc {Reader}
            Msg = {self.inout getS($)} in
            %%{self.appManager.showInfo "[ReadSocket]: "#Msg}
	    {Send self.clientPort msg({ByteString.make Msg})}
            {Reader}
         end
         proc {DoThread}
	    try {Reader}
            catch Err then
               case Err of system(kernel(terminate...)) then
                  {Send self.clientPort false}
               [] connectionClosed then
                  Msg = "Socket connection closed by LISP process." in
                  {Send self.clientPort error(Msg)}
               else
                  Msg = "Error while reading from socket:\n"#{ErrToVS Err} in
                  {Send self.clientPort error(Msg)}
               end
            end
         end
      in
	 ThreadManager, newThread(DoThread info:'socket-reader')
      end
      
      meth send(String)
         %%{self.appManager.showInfo "[WriteSocket]: "#String}
	 {self.inout putS(String)}
      end
      
      meth leave
	 {self.appManager.printInfo "Service "#ServerName#" is left..."}
	 ThreadManager, tryTerminateAll

         if {IsDet self.ServObj} then
            {self.ServObj shutdown}
         end
         {self.InkaPipe close}
         {ForAll [inout serv] proc {$ Sock} {self.Sock close} end}
         {self.appManager.showInfo " done."}
         {Thread.terminate {Thread.this}}
      end
      
   end % of InkaService
   
   ServMan = {New ServiceManager init(ServerName InkaService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
end
