%%%
%%% LCLAM service
%%%

functor
   
import
   Error(exceptionToMessage messageToVirtualString)
   OS(getEnv unlink)
   Resolve(localize)
   
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XOpen(server:Stream pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   File
   at 'x-ozlib://mathweb/share/File.ozf'
   Process(sigF)
   at 'x-ozlib://mathweb/share/Process.ozf'
   LispSocket(manager: OzService)
   at 'x-ozlib://mathweb/sys/LispSocket.ozf'
   XService(threadManager: ThreadManager
            clientManager: ClientManager
            faultManager:  FaultManager
            manager: ServerManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   
define
   AppName    = 'lclam'
   ServerName = 'LCLAM'
   App        = {New XApplication.manager init}
   Host       = {OS.getEnv 'HOST'}
   
   LCLAM_BIN = case {OS.getEnv "LCLAM_BIN"} of false then
		  case {OS.getEnv "LCLAM_BIN_DIR"} of false then
		     raise undefined('$LCLAM_BIN' '$LCLAM_BIN_DIR') end
		  [] D then D#"/lclam" end
               [] X then X end
   
   EOF = 128
   
   fun {ErrToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end

   %%
   %% xterm mode
   %%
   local
      proc {LocalShowInfo Line}
         {App.showInfo "[XTERM]: "#Line}
      end
   in 
      class LClamXterm
         from PipeGeneric
         prop final
            
         meth start(display : Display
                    host    : Host
                    showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end
                    ports   : Ports)
            
            PipeGeneric, init(cmd:  'xterm'
                              args: {Append
                                     ["-display" Display
                                      /* Adapted for LCLAM */
                                                    "-title" "LClam Service@"#Host#" (xterm)"
                                                    "-geometry" "80x25"
                                                    "-fn" "8x13"
                                                    "-e" LCLAM_BIN "-mw" "-s" Host "-t" EOF]
                                     Ports}
                              showInfo:ShowInfo onExit:ExitProc)
         end
             
         meth close
            PipeGeneric, close
         end
      end
   end

   %%
   %% quiet mode
   %%
   local skip
      proc {LocalShowInfo Line}
         {App.showInfo "[LCLAM_BIN]: "#Line}
      end
   in
      class LClamQuiet
         from PipeGeneric
         prop final
            
         meth start(showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end
                    host    : Host
                    ports   : Ports ...)
            
            PipeGeneric, init(cmd:  LCLAM_BIN
                              args: {Append ["-s" Host "-t" EOF] Ports}
                              showInfo:ShowInfo onExit:ExitProc)
         end
         
         meth close
            PipeGeneric , close(kill:true)
         end
      end   
   end

   class LClamServer from ThreadManager ClientManager FaultManager
      
      feat
         serverName
         brokerURL
         brokerTicket

         inout     % input/output stream
         serv      % mathweb stream
         
         LClamPipe
         ServObj   % the Object providing the Oz-Services 
         
      meth init(url:    URL    <=false
                ticket: Ticket <=false)
         self.serverName   = ServerName
         self.brokerURL    = URL
         self.brokerTicket = Ticket
         ThreadManager, init
      end
      
      meth StartLClamPipe(ServerArgs exit:?ExitSignal)

         %% Oz <-> LClam stream connections
         
         InOut %%Serv
         
         self.inout   = {New Stream init(port:?InOut readStopC:EOF)}
         %%self.serv    = {New Stream init(port:?Serv  readStopC:EOF)} 

         %% auxiliuary procedures
         proc {SendClient Msg}
            try {self.clientPort Msg}
            catch _ then skip end
         end
         proc {SendError VS}
            {SendClient error(VS)}
         end
         proc {DoExit}
            {SendClient exit}
            ExitSignal=unit
         end

         %% pipe class
         PipeClass = case {CondSelect ServerArgs mode xterm}
                     of emacs then LClamXterm
                     [] xterm then LClamXterm
                     [] quiet then LClamQuiet
                     [] Mode then
                        {SendError "Mode `"#Mode#"' unknown."}
                        {DoExit}
                        class
                           meth start(...) skip end
                           meth close(...) skip end
                        end
                     end
         %% init method
         Init = local
                   DefaultInit = start(display: self.clientDisplay
                                       host   : Host
                                       onExit : DoExit
                                       ports  : [InOut]) %[InOut Serv])
                in
                   if {HasFeature ServerArgs outputLabel} then
                      OutLabel = ServerArgs.outputLabel
                      proc {ShowInfo Line} {SendClient OutLabel(Line)} end
                   in
                      {Adjoin x(showInfo:ShowInfo) DefaultInit}
                   else
                      DefaultInit
                   end
                end
      in
         self.LClamPipe <- {New PipeClass Init}
      end
      
      meth enter(client: ClientInfo <= unit
                 args:   ServerArgs <= unit)

         ClientManager, init(ClientInfo)
         ClientManager, showEnter(self.appManager.showInfo)
         ClientManager, startThreads

         ExitSignal
      in
         LClamServer, StartLClamPipe(ServerArgs exit:?ExitSignal)
         
         {Delay 2500} % wichtig !!!

         if {IsFree ExitSignal} then
            {ForAll [inout /*serv*/] proc {$ Sock} {self.Sock accept} end}
            
            LClamServer, StartReader
            %%LClamServer, StartServObj
            LClamServer, InstallWatcher   
         end
      end

      meth StartServObj
         proc {MyLoui M}
            {Send self.clientPort serv(M)}
         end
      in
         self.ServObj = {New OzService init(self.serv
                                            url:    self.brokerURL
                                            ticket: self.brokerTicket
                                            myLoui: MyLoui)}
      end
      meth InstallWatcher
         proc {Watch _ _}
            {self.appManager.showInfo
             "*** "#self.clientName#" died. Exiting..."}
            if {HasFeature self serverProc}
            then {self.serverProc leave}
            else {self leave}
            end
         end
      in
         FaultManager, installWatcher(Watch)
      end
       
      %%
      %% socket communication between LClam and LOUI
      %%
      
      meth StartReader
         proc {Reader}
            Msg = {self.inout getS($)} in
            {self.appManager.showInfo "[ReadSocket]: "#Msg}
            {Send self.clientPort msg({ByteString.make Msg})}
            {Reader}
         end
         proc {DoThread}
            try {Reader}
            catch Err then
               case Err of system(kernel(terminate...)) then
                  try {Send self.clientPort false}
                  catch system(dp(...)...) then skip end
               [] connectionClosed then
                  Msg = "Socket connection closed by LISP process." in
                  try {Send self.clientPort error(Msg)}
                  catch system(dp(...)...) then skip end
               else
                  Msg = "Error while reading from socket:\n"#{ErrToVS Err} in
                  try {Send self.clientPort error(Msg)}
                  catch system(dp(...)...) then skip end
               end
            end
         end
      in
         ThreadManager, newThread(DoThread info:'socket-reader')
      end
      
      meth send(ProblemString $ timeout: TimeOut <= 60)
	 
	 fun {ReadAll}
	    Rest Thr
	 in
	    thread
	       Thr = {Thread.this} 
	       Rest = {self.inout getS($)}
	    end
	    {WaitOr Rest {Alarm 50}}
	    try
	       {Thread.suspend Thr}
	       if {Thread.state Thr}==blocked then
		  {Thread.terminate Thr}   
               else
		  {Thread.resume Thr}
	       end
	    catch error(kernel(deadThread...)...) then skip end
	    if {IsDet Rest} then
	       Rest#{ReadAll}
	    else
	       nil
	    end
	 end
	 Answer Thr Rest
      in
	 Rest = {ReadAll}    %% first read everything that has been written so far
	 {self.appManager.showInfo "[WriteSocket]: "#ProblemString}
	 {self.inout putS(ProblemString)}
	 thread 
	    Thr = {Thread.this} 
	    try
	       Answer = {self.inout getS($)}
	    catch _ then
	       skip
	    end
	 end
	 {WaitOr Answer {Alarm (TimeOut * 1000)}}
	 if {IsFree Answer} then Answer = nil end
	 Answer
      end
%       meth send(String)
%          {self.appManager.showInfo "[WriteSocket]: "#String}
%          {self.inout putS(String)}
%       end
      
      meth sig(Signal $<=_)
	 {Process.sigF Signal self.LClamPipe.pid}
      end
    
      meth load(Url ?BS)
         try
            case {Resolve.localize Url}
            of old(F) then BS={File.loadBS F}
            [] new(F) then BS={File.loadBS F} {OS.unlink F}
            end
         catch _ then
            BS=false
         end
      end
        
      meth leave
         {self.appManager.printInfo "Service "#ServerName#" is left..."}
         FaultManager, deinstallWatcher
         ThreadManager, tryTerminateAll
         if {IsDet self.ServObj} then
            {self.ServObj shutdown}
         end
         {self.LClamPipe close}
         {ForAll [inout /*serv*/] proc {$ Sock} {self.Sock close} end}
         {self.appManager.showInfo " done."}
         {Thread.terminate {Thread.this}}
      end

   end % of LClamService
   
   ServMan = {New ServerManager init(ServerName LClamServer
                                     appName: AppName
                                     appMan:  App)}

   {ServMan offerService}
end
