%%%
%%% LCLAM test program
%%%

functor
   
import
   Error(exceptionToMessage messageToVirtualString printException)
   Open(socket:Socket)
   OS(getEnv unlink)
   Type

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XOpen(/*server:Stream*/ pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   Internet
   at 'x-ozlib://mathweb/share/Internet.ozf'
   XService(threadManager: ThreadManager
            clientManager: ClientManager
            faultManager:  FaultManager           
            /*manager: ServerManager*/)
   at 'x-ozlib://mathweb/sys/XService.ozf'

define
   AppName    = 'lclam'
   ServerName = 'LCLAM'
   App        = {New XApplication.manager init}
   Host       = {OS.getEnv 'HOST'}

   LCLAM_BIN = case {OS.getEnv "LCLAM_BIN"} of false then
		  case {OS.getEnv "LCLAM_BIN_DIR"} of false then
		     raise undefined('$LCLAM_BIN' '$LCLAM_BIN_DIR') end
		  [] D then D#"/lclam" end
               [] X then X end

   Spec = record(port(single char:&p type:int default:9999)
                )
   Args
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {App.showError AppName#": "#S}
      {App.exit 1}
   end

   EOF = 128
   
   fun {ErrToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end

   %%
   %% quiet mode
   %%
   local skip
      proc {LocalShowInfo Line}
         {App.showInfo "[stdout]: "#Line}
      end
   in
      class LClamQuiet
         from PipeGeneric
         prop final
            
         meth start(showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end
                    host    : Host
                    ports   : Ports ...)
	    
            PipeGeneric, init(cmd:  LCLAM_BIN
                              args: {Append ["-mw" "-s" Host "-t" EOF] Ports}
                              showInfo:ShowInfo onExit:ExitProc)
	 end
         
         meth close
            PipeGeneric , close(kill:true)
         end
      end   
   end

   class LClamServer from ThreadManager ClientManager FaultManager
      
      feat
         serverName
         brokerURL
         brokerTicket

         inout     % input/output stream
         serv      % mathweb stream
         
         LClamPipe
         ServObj   % the Object providing the Oz-Services 
         
      meth init(url:    URL    <=false
		ticket: Ticket <=false)
         self.serverName   = ServerName
         self.brokerURL    = URL
         self.brokerTicket = Ticket
         ThreadManager, init
      end
      
      meth StartLClamPipe(ServerArgs exit:?ExitSignal)

         %% Oz <-> LClam stream connections
         
         InOut %%Serv
         
         self.inout   = {New Socket init}
         {self.inout bind(port:?InOut)}
         {self.inout listen}
         %{New Stream init(port:?InOut readStopC:EOF)}
         %%self.serv    = {New Stream init(port:?Serv  readStopC:EOF)} 

         %% auxiliuary procedures
         proc {SendClient Msg}
	    if {IsDet self.clientPort} then
	       try {self.clientPort Msg}
	       catch _ then skip end
	    end
         end
         proc {SendError VS}
            {SendClient error(VS)}
         end
         proc {DoExit}
            {SendClient exit}
            ExitSignal=unit
         end

         %% pipe class
         PipeClass = case {CondSelect ServerArgs mode quiet} % was: xterm
                     of quiet then LClamQuiet
                     [] Mode then
                        {SendError "Mode `"#Mode#"' unknown."}
                        {DoExit}
                        class
                           meth start(...) skip end
                           meth close(...) skip end
                        end
                     end
         %% init method
         Init = local
                   DefaultInit = start(display: self.clientDisplay
                                       host   : Host
                                       onExit : DoExit
                                       ports  : [InOut]) %[InOut Serv])
                in
                   if {HasFeature ServerArgs outputLabel} then
                      OutLabel = ServerArgs.outputLabel
                      proc {ShowInfo Line} {SendClient OutLabel(Line)} end
                   in
                      {Adjoin x(showInfo:ShowInfo) DefaultInit}
                   else
                      DefaultInit
                   end
                end
      in
	 self.LClamPipe = {New PipeClass Init}
      end
      
      meth enter(client: ClientInfo <= unit
                 args:   ServerArgs <= unit)

         ClientManager, init(ClientInfo)
	 ClientManager, showEnter(App.showInfo) %was: self.appManager.showInfo
	 ClientManager, startThreads
         ExitSignal
      in
	 LClamServer, StartLClamPipe(ServerArgs exit:?ExitSignal)
	 
	 {Delay 2500} % wichtig !!!
	 
	 if {IsFree ExitSignal} then
	    
            {ForAll [inout /*serv*/] proc {$ Sock} {self.Sock accept} end}
	    
	    %% LClamServer, StartReader
         end
      end

      %%
      %% socket communication between LClam and LOUI
      %%
      
      meth StartReader
         proc {Reader}
            Msg = {self.inout getS($)} in
            {App.showInfo "[ReadSocket]: "#Msg}
            {Send self.clientPort msg({ByteString.make Msg})}
            {Reader}
         end
         proc {DoThread}
            try {Reader}
            catch Err then
               case Err of system(kernel(terminate...)) then
                  try {Send self.clientPort false}
                  catch system(dp(...)...) then skip end
               [] connectionClosed then
                  Msg = "Socket connection closed by LISP process." in
                  try {Send self.clientPort error(Msg)}
                  catch system(dp(...)...) then skip end
               else
                  Msg = "Error while reading from socket:\n"#{ErrToVS Err} in
                  try {Send self.clientPort error(Msg)}
                  catch system(dp(...)...) then skip end
               end
            end
         end
      in
         ThreadManager, newThread(DoThread info:'socket-reader')
      end
      
      meth send1(String)
         {App.showInfo "[WriteSocket]: "#String}
         {self.inout putS(String)}
      end
      
      meth leave
         {App.printInfo "Service "#ServerName#" is left..."}
         ThreadManager, tryTerminateAll
         {self.LClamPipe close}
         {ForAll [inout /*serv*/] proc {$ Sock} {self.Sock close} end}
         {App.showInfo " done."}
         {Thread.terminate {Thread.this}}
      end

   end % of LClamService
   
   %%
   %% connection manager
   %%
   fun {WriteEscape S}
      case S of nil then nil
      [] C|Sr then
	 if {Char.isPrint C} then
	    C|{WriteEscape Sr}
	 else
	    Digit0 = (C mod 8) + &0
	    C1     = C div 8
	    Digit1 = (C1 mod 8) + &0
	    Digit2 = (C1 div 8) + &0
	 in
	    &\\|Digit2|Digit1|Digit0|{WriteEscape Sr}
	 end
      end
   end
   class ConnectionManager
      from LClamServer Socket
         
      feat
	 localRegistry
	 remoteHost
         remotePort
	 server
	 connectionManagerThread
	 inoutListenerThread
	 
      meth init(...)=M
	 LClamServer, M
      end
      meth doConnection(server:S remoteHost:H remotePort:P)
         self.server     = S
         self.remoteHost = H
         self.remotePort = P
         try
            {App.showInfo "*** connection accepted from "#H#", port "#P}
            ConnectionManager, initConnection
            ConnectionManager, ReadLoop1
	 catch E then
	    case E of system(kernel(terminate...)...) then skip
	    [] connectionClosed then skip
	    [] close then skip
	    else
	       {Error.printException E}
            end
         finally
            try Socket, shutDown
            catch Err then {Error.printException Err} end
            Socket,close
            {App.showInfo "*** connection closed ("#H#", port "#P#")"}
	    {App tryTerminate(self.inoutListenerThread)}
	 end
      end
      meth GetLine($)
	 case {self read(list:$ size:1)} of nil then nil
	 [] [C] then
	    case C
	    of &\n then nil
	    [] &\r then % discard next char after linefeed
	       {self read(list:_ size:1)} nil
	    else 
	       C|{self GetLine($)}
	    end
	 end
      end
      meth ReadLoop1
	 case {self GetLine($)} of nil then
	    raise close end
	 [] Line then
	    {App.showInfo {WriteEscape Line}}
	    {self.inout write(vs:Line)}
	    ConnectionManager, ReadLoop1
	 end
      end
      meth initConnection
	 LClamServer, init
	 LClamServer, enter
	 local
	    proc {DoThread}
	       try
		  ConnectionManager, ReadLoop2
	       catch close then
		  try
		     {self leave}
		  finally 
		     {App terminate(self.connectionManagerThread)}
		     {Thread.terminate {Thread.this}}
		  end
	       end
	    end
	 in
	    {App newThread(DoThread ?self.inoutListenerThread
			   info:"readloop2")}
	 end
      end
      meth GetInOut($)
	 case {self.inout read(list:$ size:1)} of nil then nil
	 [] [C] then
	    case C of &\200 then nil
	    else 
	       C|{self GetInOut($)}
	    end
	 end
      end
      meth ReadLoop2
	 case {self GetInOut($)} of nil then
	    raise close end
	 [] Msg then
	    {App.showInfo Msg#"\200"}
	    {self write(vs:Msg#"\200")}
	    ConnectionManager, ReadLoop2
	 end
      end
   end


   %%
   %% server
   %%
   class Server from Socket

      feat
         host
         port
         acceptClass
                   
      meth init(port:P exit:?Exit<=_
                acceptClass: AcceptClass <= ConnectionManager)
         {Type.ask.'class' AcceptClass}
         self.acceptClass = AcceptClass

         Socket, init
         Socket, bind(takePort:P)
         Socket, listen
         self.host        = Internet.hostname
         self.port        = P
	 local
	    proc {DoThread}
	       try
		  Server, AcceptLoop
	       catch Exc then
		  Exit=Exc
	       finally
		  Server, close
		  if {IsFree Exit} then Exit=unit end
	       end
	    end
	 in
	    {App newThread(DoThread info:ServerName#".server")}
	 end
      end
      
      meth AcceptLoop
         H P M in
         Socket, accept(acceptClass: self.acceptClass
                        accepted:    ?M
			host:?H port:?P)
	 local
	    proc {DoThread LocalRegistry}
	       M.localRegistry = LocalRegistry
	       {M doConnection(server:     self
			       remoteHost: H
			       remotePort: P)}
	    end
	 in
	    {App newThread(DoThread ?M.connectionManagerThread
			   info:H#":"#P)}
	 end
	 Server, AcceptLoop
      end
      
      meth close
         try Socket, shutDown
         catch _ then skip end
         Socket, close
      end
   end

   Exit
   Serv = {New Server init(port:Args.port exit:?Exit)}
   try
      {Wait Exit}
   catch error(kernel(terminate...)...) then
      skip
   [] Err then
      {App printException(Err)}
   end
   {Serv close}
   {Delay 1000}
   {App.exit 0}
end
