%%
%% LClam test client
%%

functor

import
   Connection XURL            at 'x-ozlib://mathweb/share/XURL.ozf' Property OS

   XApplication       at 'x-ozlib://mathweb/share/XApplication.ozf'
   Dir                at 'x-ozlib://mathweb/share/Dir.ozf'
   Env                at 'x-ozlib://mathweb/share/Env.ozf'
   Internet(hostname) at 'x-ozlib://mathweb/share/Internet.ozf'
   BrokerAccess(getBroker) at 'x-ozlib://mathweb/sys/BrokerAccess.ozf'
   
define
   AppName    = 'lclamclient'
   ClientName = 'LCLAM-TEST-CLIENT'
   App        = {New XApplication.manager init}
   
   fun {Lookup VarA Default}
      try {Property.get VarA}
      catch system(system(getProperty...)...) then
         case {OS.getEnv VarA} of false then Default
         [] Val then Val end 
      end
   end
   
   DefaultTICKET  = {Lookup 'DUMMY_TICKET'  "http://www.dai.ed.ac.uk/~michaelk/"}
   DefaultSERVICE = {Lookup 'DUMMY_SERVICE' "LCLAM"}
   DefaultMODE    = {Lookup 'DUMMY_MODE'    "xterm"}
   
   Spec = record(home(   single char:&h type:string optional:true) %ignored
                 ticketfile( single char:&t type:string default:DefaultTICKET)
                 service(single char:&s type:string default:DefaultSERVICE)
                 mode(   single char:&m type:string default:DefaultMODE)
                 emacs(  single char:&e alias:mode#"emacs")
                 xterm(  single char:&x alias:mode#"xterm"))
   Args
   try 
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S) ...) then
      {App.showError AppName#": "#S}
      {App.exit 1}
   end
   
   BrokerURL  = Args.ticketfile
   Service    = {VirtualString.toAtom Args.service}
   Mode       = {VirtualString.toAtom Args.mode}
   
   {App.showInfo "Requesting for service "#Service#" at "#BrokerURL#"."}
   Object
   try
      Broker = {BrokerAccess.getBroker App a(brokerticket: Args.ticketfile)}.send
   in
      Object = {Broker request(Service $ client:ClientName)}
      if Object==false then
         {App.showError "Service "#Service#" is not available! Sorry!"}
         {App.exit 1}
      else
         MainThread = {Thread.this}
         class MyPort
            attr prefix
            meth init(label:L<="pipe") prefix<-"["#L#"]: " end
            meth otherwise(_) skip end
            meth msg(Msg)
               {App.printInfo "[lclam]: "}
               {App.show Msg}
            end
            meth pipeOutput(VS)
               {App.showInfo @prefix#VS}
            end
            meth error(ErrVS)
               {App.showError ErrVS}
            end
            meth exit
               {App.showInfo "Leaving service "#Service#" ..."}
               try
                  {Object leave} {Delay 1000}
               finally
                  {Thread.terminate MainThread}
                  MyPort,terminate
               end
            end
            meth terminate
               {Thread.terminate {Thread.this}}
            end
         end
         ThrID = {App newServer(MyPort init(label:Mode) _ $
                                detach: false
                                info:   dummyport)}
         ClientInfo = ClientName(name:    ClientName
                                 user:    Env.user
                                 host:    Internet.hostname
                                 display: Env.display
                                 port:    {App getValue(ThrID $)}.serverPort
                                )
         ServerArgs = unit(mode:        Mode
                           outputLabel: pipeOutput
                          )
      in
         {Object enter(client:ClientInfo args:ServerArgs)}
         {App.showInfo "Service "#Service#" started."}
      end
   catch Err then
      {App.showError "Service "#Service#" at \n\n"#BrokerURL#"\n\n"#
       "is unreachable!"}
      {App.showError "Try again later..."}
      {App printExc(Err)}
      {App.exit 1}
   end

   case Args.1 of nil then
      {Object send("dplan.\n")}
      {Delay 30000}
      {Object send("assp.\n")}
      {Delay 30000}
      {Object send("quit.\n")}
      try
         {Thread.suspend {Thread.this}}
      catch system(kernel(terminate...)...) then
         {App.exit 0}
      end
   else
      {App.showError AppName#": "#"illegal number of arguments"}
      {Object leave}
      {Delay 1000}
      {App.exit 1}
   end
end




