%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% omega.oz ; This file is part of the MathWeb system
%
%   major updates: 20.01.2001
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% This file contains OMEGA server
%%%

functor

import 
   Error(exceptionToMessage messageToVirtualString)
   OS(getEnv tmpnam unlink)
   Resolve(localize)
   Property
   Browser(browse:Browse)
   
   
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XOpen(server:Stream pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   File
   at 'x-ozlib://mathweb/share/File.ozf'
   Process(sigF)
   at 'x-ozlib://mathweb/share/Process.ozf'
   LispSocket(manager: Lisp2OzSocket)
   at 'x-ozlib://mathweb/sys/LispSocket.ozf'
   
   XService(threadManager: ThreadManager
	    clientManager: ClientManager
	    faultManager:  FaultManager           
	    manager:       ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   
   %%----------------------------------
   %% KQML agent functors
   %%----------------------------------
   KqmlAgent('class': KQMLAgent)
   at 'x-ozlib://mathweb/kqml/KqmlAgent.ozf'
   KqmlLispSocket('class': KqmlLisp2OzSocket)
   at 'x-ozlib://mathweb/kqml/KqmlLispSocket.ozf'
   
define
   VS2S = VirtualString.toString
%   VS2A = VirtualString.toAtom
   
   {Property.put 'print.depth' 100}
   {Property.put 'print.width' 100}

   AppName    = 'omega'
   ServerName = 'OMEGA'
   App        = {New XApplication.manager init}
   Host       = {OS.getEnv 'HOST'}

   %  Special CMD-Args because of Kqml agent mode
   ArgSpec = record(kqmlagent(single char:[&a] type:bool default: false))

   LOUIEMACS      = {OS.getEnv "LOUIEMACS"}
   LOUIEMACS_INIT = {OS.getEnv "LOUIEMACS_INIT"} % 'ilisp.emacs'
   LOUIXTERM      = {OS.getEnv "LOUIXTERM"}
   OMEGADUMPEMACS = {OS.getEnv "OMEGADUMPEMACS"}
   OMEGADUMPQUIET = {OS.getEnv "OMEGADUMPQUIET"}
   ACLVERSION     = case {OS.getEnv "ACLVERSION"}
                    of false then false
                    [] S     then {StringToAtom S}
                    end
   ALLEGRO        = {OS.getEnv "ALLEGRO"}
   GEOMETRY       = case {OS.getEnv "EMACSGEOMETRY"}  %  for Volker :-)
		    of false then
		        "138x53+0+0"
		    [] Geometry then
		       Geometry
		    end
   local
      proc {ShowVar Name Val}
         if Val==false
	 then {App.showInfo "$"#Name#" is undefined."}
	 else {App.showInfo "$"#Name#" = "#Val}
	 end
      end
   in
      {ShowVar "LOUIEMACS"      LOUIEMACS}
      {ShowVar "LOUIEMACS_INIT" LOUIEMACS_INIT}
      {ShowVar "LOUIXTERM"      LOUIXTERM}
      {ShowVar "OMEGADUMPEMACS" OMEGADUMPEMACS}
      {ShowVar "OMEGADUMPQUIET" OMEGADUMPQUIET}
      case ACLVERSION of false then skip else
         {ShowVar "ACLVERSION" ACLVERSION}
         {ShowVar "ALLEGRO"    ALLEGRO}
      end
   end
   
   EOF = 128
   
   fun {ErrToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end
   
   fun {Concat Ls}
      case Ls of nil then nil
      [] L1|Lr then {Append L1 {Concat Lr}} end
   end
   
   %%
   %% emacs mode
   %%
   local
      proc {LocalShowInfo Line}
         {App.showInfo "[EMACS]: "#Line}
      end
   in
      class OmegaEmacs
         from PipeGeneric 
            
         feat error: {NewCell nil}
            
         meth start(display : Display
                    host    : Host
                    cmds    : CmdsVS
                    showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end)
                    
            IlispTmp = {OS.tmpnam}
            InitTmp  = {OS.tmpnam}#'.lisp'
            
            local %% create temporary init file
               InitTmpVS =
               '(delete-file "'#IlispTmp#'")\n'#
               '(delete-file "'#InitTmp#'")\n'#
               CmdsVS
            in
               {File.saveVS InitTmpVS InitTmp}
            end
            local %% create temporary ilisp file
               DumpArgs   = if ACLVERSION\=false andthen ACLVERSION>='5'
                            then ALLEGRO#" -I "#OMEGADUMPEMACS
                            else OMEGADUMPEMACS
                            end
	       IlispTmpVS = {File.loadVS LOUIEMACS_INIT}#
               '(setq loui-program "'#DumpArgs#' -qq -e (load\\"'#InitTmp#'\\")")'
            in
               {File.saveVS IlispTmpVS IlispTmp}
            end
            
         in
            PipeGeneric, init(cmd:  "env"
                              args: ["DISPLAY="#Display
                                     LOUIEMACS %"-display" Display
                                     "-debug-init" "-name" "Omega Service@"#Host#" (Emacs)"
                                     "-geometry" GEOMETRY "-font" "8x13"
                                     "-q" "-l" IlispTmp "-f" "loui"]
                              showInfo:ShowInfo onExit:ExitProc)
         end
         
         meth close
            PipeGeneric, close(kill:true)
         end      
      end
      
      OmegaEmacsAgent = {ByNeed
			 fun {$}
			    class $
			       from OmegaEmacs KQMLAgent
			       meth init(...) = Init
				  OmegaEmacs, Init
				  KQMLAgent, init(app: App
                                                  name: 'OmegaAgent')
			       end
			       meth test
				  {App.showInfo 'Hi, I can speak KQML!'}
			       end
			    end % class ..Agent
			 end}
      
   end
   
   %%
   %% xterm mode
   %%
   local
      proc {LocalShowInfo Line}
         {App.showInfo "[XTERM]: "#Line}
      end
   in
      class OmegaXterm
         from PipeGeneric

         meth start(display : Display
                    host    : Host
                    cmds    : CmdsVS
                    showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end)
            
            InitTmp  = {OS.tmpnam}#'.lisp'

            %% create temporary init file
            {File.saveVS '(delete-file "'#InitTmp#'")\n'#CmdsVS InitTmp}

            DumpArgs = if ACLVERSION\=false andthen ACLVERSION>='5'
                       then [ALLEGRO "-I" OMEGADUMPEMACS]
                       else [OMEGADUMPEMACS]
                       end    
         in
	    PipeGeneric, init(cmd:  "env"
			      args: {Concat
                                     [["DISPLAY="#Display
				       LOUIXTERM %"-display" Display
				       "-title" "Omega Service@"#Host#" (xterm)"
				       "-geometry" "80x25"
				       "-e"] DumpArgs  
				      ["-qq" "-e" "(load \""#InitTmp#"\")"]]}
                              showInfo:ShowInfo onExit:ExitProc)
         end

         meth close
            PipeGeneric, close(kill:true)
         end
      end
      
      OmegaXtermAgent = {ByNeed
			 fun {$}
			    class $
			       from OmegaXterm KQMLAgent
			       meth init(...) = Init
				  OmegaXterm, Init
				  KQMLAgent, init(app: App
                                                  name: 'OmegaAgent')
			       end
			    end % class ..Agent
			 end}
   end

   %%
   %% quiet mode
   %%
   local
      skip
      %%proc {LocalShowInfo Line}
      %%   {App.showInfo "[LISP]: "#Line}
      %%end
   in
      class OmegaQuiet
         from PipeGeneric
	    
	 meth start(display : Display
		    cmds    : CmdsVS
		    showInfo: ShowInfo <= /*LocalShowInfo*/ proc {$ _} skip end
		    onExit  : ExitProc <= proc {$} {ShowInfo "Process terminated."} end
                    ...)

            InitTmp = {OS.tmpnam}#'.lisp'

            %% create temporary init file
            {File.saveVS '(delete-file "'#InitTmp#'")\n'#CmdsVS InitTmp}

%             DumpArgs = if ACLVERSION\=false andthen ACLVERSION>='5'
%                        then [ALLEGRO "-I" OMEGADUMPQUIET]
%                        else [OMEGADUMPQUIET]
%                        end
% changed jzimmer 9.3.2001 we don't need two dumps anymore...
            DumpArgs = if ACLVERSION\=false andthen ACLVERSION>='5'
                       then [ALLEGRO "-I" OMEGADUMPEMACS]
                       else [OMEGADUMPEMACS]
                       end
         in
            PipeGeneric, init(cmd:  "env"
                              args: {Concat
                                     [["DISPLAY="#Display]
                                      DumpArgs
%                                       ["-qq" "-e" "(load \""#InitTmp#"\")"]]}  %% changed jzimmer 9.3.2001
				      ["-qq" "-e" "(progn (load \""#InitTmp#"\")(setf excl::*restart-app-function* (lambda () (omega::omega~loui))))"]]}
                              showInfo:ShowInfo onExit:ExitProc)
         end
         
         meth close
	    PipeGeneric, close(kill:true)
	 end
      end

      OmegaQuietAgent = {ByNeed
			 fun {$}
			    class $
			       from OmegaQuiet KQMLAgent
			       meth init(...) = Init
				  OmegaQuiet, Init
				  KQMLAgent, init(app: App
                                                  name: 'OmegaAgent')
			       end
			    end % class ..Agent
			 end}
   end % local
   


   %%
   %% main
   %%
   class OmegaService from ThreadManager ClientManager FaultManager
                        
      feat
         serverName
         brokerURL
         brokerTicket
	 agent
	 
	 inout % input/output stream
         serv  % mathweb stream

         OmegaPipe 
         ServObj   % the Object providing the Oz-Services 

	 serverMode % emacs, quiet, ...
	 
      meth init(url:     URL     <= false
		ticket:  Ticket  <= false
		cmdArgs: CmdArgs <= unit
	       )
	 self.agent        = {HasFeature CmdArgs kqmlagent} andthen CmdArgs.kqmlagent
	 self.serverName   = if self.agent then
				{VS2S ServerName#'-agent'}
			     else ServerName end
	 self.brokerURL    = URL
	 self.brokerTicket = Ticket
	 ThreadManager, init
	 if self.agent then
	    {App.showInfo 'I am a KQML agent!'}
	 end

      end
      
      meth StartOmegaPipe(ServerArgs exit:?ExitSignal)
    
         %% Oz <-> Lisp stream connections
         local
	    RemoteURL = {CondSelect ServerArgs remoteURL unit}
	    SessionID = {CondSelect ServerArgs sessionID unit}
            MethodName = 'iLouiIsAlive'

            InOut Serv
         in
	    self.inout = {New Stream init(port:?InOut readStopC:EOF)}
	    self.serv  = {New Stream init(port:?Serv  readStopC:EOF)}

	    %% stuff added for new interface by VS
	    CmdsVS =
	    '(ags::socket~reset)\n'#
            '(ags::socket~define :inout)\n'#
	    '(ags::socket~connect "'#Host#'" '#InOut#' :inout)\n'#
            '(ags::socket~define :service)\n'#
	    '(ags::socket~connect "'#Host#'" '#Serv #' :service)\n'#
	    if RemoteURL\=unit andthen SessionID\=unit then
	       '(ags::rpc~start-server '#
               '\'(\"'#RemoteURL#'\" \"'#MethodName#'\" \"'#SessionID#'\"))\n'
            else '' end
            %% oz  ->lisp: (rpc~start-server '(URL MethodName activityID))
            %% lisp->java: iLouiIsAlive(string activityID, string urlToOmega)
         end

         %% auxiliuary procedures
         proc {SendClient Msg}
            try {Send self.clientPort Msg}
            catch _ then skip end
         end
         proc {SendError VS}
            {SendClient error(VS)}
         end
         proc {DoExit}
            {SendClient exit}
            ExitSignal=unit
         end

         self.serverMode = {CondSelect ServerArgs mode emacs} 

	 %% pipe class
	 PipeClass = case self.serverMode
		     of emacs then
			if self.agent then OmegaEmacsAgent else OmegaEmacs end
		     [] xterm then
			if self.agent then OmegaXtermAgent else OmegaXterm end
		     [] quiet then
			if self.agent then OmegaQuietAgent else OmegaQuiet end
		     [] Mode then
			{SendError "Mode `"#Mode#"' unknown."}
			{DoExit}
			class
			   meth start(...) skip end
                           meth close(...) skip end
                        end
                     end

         %% init method
         Init = local
		   DefaultInit = start(display : self.clientDisplay
				       host    : Host
				       cmds    : CmdsVS
                                       onExit  : DoExit) 
                in
                   if {HasFeature ServerArgs outputLabel} then
                      OutLabel = ServerArgs.outputLabel
                      proc {ShowInfo Line} {SendClient OutLabel(Line)} end
                   in
                      {Adjoin x(showInfo:ShowInfo) DefaultInit}
                   else
                      DefaultInit
                   end
                end
      in
	 self.OmegaPipe = {New PipeClass Init}        
      end
      
      /** This method enters an omega service */
      /** NOTE: ClientInfo = ClientName(name: user: host: display: port:) */
      /**       ServerArgs = unit(mode: remoteURL: sessionID:) */
      meth enter(client: ClientInfo <= unit
		 args:   ServerArgs <= unit)
	 
	 if {HasFeature ClientInfo display} andthen
	    {Not {IsVirtualString ClientInfo.display}} then
	    {Browse error(corruptDisplay(ClientInfo))}
	    OmegaService, leave
	 else
	    ClientManager, init(ClientInfo)
	    ClientManager, showEnter(self.appManager.showInfo)
	    ClientManager, startThreads
	    
	    ExitSignal
	 in
	    OmegaService, StartOmegaPipe(ServerArgs exit:?ExitSignal)
	    
	    {Delay 2500} % important !!!
	    
	    if {IsFree ExitSignal} then
	       {ForAll [inout serv] proc {$ Sock} {self.Sock accept} end}
	       OmegaService, StartReader
	       OmegaService, StartServObj
	       OmegaService, InstallWatcher    
	    end
	 end
      end
  
      meth StartServObj
	 proc {MyLoui M}
	    {Send self.clientPort serv(M)}
	 end
	 SocketClass Init
      in
	 if self.agent then
	    SocketClass = {ByNeed fun{$} KqmlLisp2OzSocket end}
	    Init        = init(self.serv
                               inout:    self.inout
			       url:      self.brokerURL
			       ticket:   self.brokerTicket
			       agentName:'OmegaAgent'
			       app:      App
			       myLoui:   MyLoui
			       display:  self.clientDisplay
			       mode:     self.serverMode
			       serverName: self.serverName
			      )
	 else
	    SocketClass = Lisp2OzSocket
	    Init        = init(self.serv
			       url:    self.brokerURL
			       ticket: self.brokerTicket
			       app:    App
			       myLoui: MyLoui
			       display:self.clientDisplay
			       mode:   self.serverMode
			       serverName: self.serverName
			      )
	 end
	 self.ServObj = {New SocketClass Init}
      end
      
      meth InstallWatcher
	 proc {Watch _ _}
	    {self.appManager.showInfo
	     "*** "#self.clientName#" died. Exiting..."}
	    if {HasFeature self serverProc}
            then {self.serverProc leave}
	    else {self leave}
	    end
         end
      in
	 FaultManager, installWatcher(Watch)
      end
      
      %%
      %% socket communication between LISP and LOUI
      %%
      
      meth StartReader
         proc {Reader}
            Msg = {self.inout getS($)} in
            {self.appManager.showInfo "[ReadSocket]: "#Msg}
            {Send self.clientPort msg({ByteString.make Msg})}
            {Reader}
         end
         proc {DoThread}
	    try {Reader}
            catch Err then
               case Err of system(kernel(terminate...)) then
                  try {Send self.clientPort false}
                  catch system(dp(...)...) then skip end
               [] connectionClosed then
                  Msg = "Socket connection closed by LISP process." in
                  try {Send self.clientPort error(Msg)}
                  catch system(dp(...)...) then skip end
               else
                  Msg = "Error while reading from socket:\n"#{ErrToVS Err} in
                  try {Send self.clientPort error(Msg)}
                  catch system(dp(...)...) then skip end
               end
            end
         end
      in
         ThreadManager, newThread(DoThread info:'socket-reader')
      end
      
      meth send(String)
         {self.appManager.showInfo "[WriteSocket]: "#String}
         {self.inout putS(String)}
      end
      
      meth sig(Signal $<=_)
	 {Process.sigF Signal self.OmegaPipe.pid}
      end

      meth load(Url ?BS)
         try
            case {Resolve.localize Url}
            of old(F) then BS={File.loadBS F}
            [] new(F) then BS={File.loadBS F} {OS.unlink F}
            end
         catch _ then
            BS=false
         end
      end
      
      meth leave
	 {self.appManager.printInfo "Service "#ServerName#" is left...\n"}
         FaultManager, deinstallWatcher
         ThreadManager, tryTerminateAll
         if {IsDet self.ServObj} then
            {self.ServObj shutdown}
         end
         {self.OmegaPipe close}
         {ForAll [inout serv] proc {$ Sock} {self.Sock close} end}
         {self.appManager.showInfo " done."}
         {Thread.terminate {Thread.this}}
      end
      
      meth test
	 {App.showInfo 'I am the omega server.'}
      end
      
   end % of OmegaService
   
   ServMan = {New ServiceManager init(ServerName
				      OmegaService
				      appName: AppName
				      appMan:  App
				      argSpec: ArgSpec 
				     )}
   {ServMan offerService}
end







