%%%
%%% PREX server
%%%

functor

import 
   Error(exceptionToMessage messageToVirtualString)
   OS(getEnv tmpnam unlink)
   Resolve(localize)
   
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XOpen(server:Stream pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   File
   at 'x-ozlib://mathweb/share/File.ozf'
   Process(sigF)
   at 'x-ozlib://mathweb/share/Process.ozf'
   LispSocket(manager: OzService)
   at 'x-ozlib://mathweb/sys/LispSocket.ozf'
   XService(threadManager: ThreadManager
            clientManager: ClientManager
            faultManager:  FaultManager           
            manager: ServerManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'

define
   AppName    = 'prex'
   ServerName = 'PREX'
   App        = {New XApplication.manager init}
   Host       = {OS.getEnv 'HOST'}

   LOUIEMACS      = {OS.getEnv "LOUIEMACS"}
   LOUIEMACS_INIT = {OS.getEnv "LOUIEMACS_INIT"} % 'ilisp.emacs'
   LOUIXTERM      = {OS.getEnv "LOUIXTERM"}
   PREXDUMPEMACS = {OS.getEnv "PREXDUMPEMACS"}
   PREXDUMPQUIET = {OS.getEnv "PREXDUMPQUIET"}
   ACLVERSION     = case {OS.getEnv "ACLVERSION"}
                    of false then false
                    [] S     then {StringToAtom S}
		    end
   ALLEGRO        = {OS.getEnv "ALLEGRO"}
   GEOMETRY       = case {OS.getEnv "EMACSGEOMETRY"}  %  for Volker :-)
		    of false then
		        "138x53+0+0"
		    [] Geometry then
		       Geometry
		    end
   
   local
      proc {ShowVar Name Val}
         if Val==false
         then {App.showInfo "$"#Name#" is undefined."}
         else {App.showInfo "$"#Name#" = "#Val}
         end
      end
   in
      {ShowVar "LOUIEMACS"      LOUIEMACS}
      {ShowVar "LOUIEMACS_INIT" LOUIEMACS_INIT}
      {ShowVar "LOUIXTERM"      LOUIXTERM}
      {ShowVar "PREXDUMPEMACS" PREXDUMPEMACS}
      {ShowVar "PREXDUMPQUIET" PREXDUMPQUIET}
      case ACLVERSION of false then skip else
	 {ShowVar "ACLVERSION" ACLVERSION}
         {ShowVar "ALLEGRO"    ALLEGRO}
      end
   end
   
   EOF = 128
   
%    fun {ErrToVS Exc}
%       {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
%    end
   
   fun {Concat Ls}
      case Ls of nil then nil
      [] L1|Lr then {Append L1 {Concat Lr}} end
   end
     
   %%
   %% emacs mode
   %%
   local
      proc {LocalShowInfo Line}
         {App.showInfo "[EMACS]: "#Line}
      end
   in
      class PrexEmacs
         from PipeGeneric 
         prop final
            
         feat error: {NewCell nil}
            
         meth start(display : Display
                    host    : Host
                    cmds    : CmdsVS
                    showInfo: ShowInfo <= LocalShowInfo
		    onExit  : ExitProc <= proc {$} {ShowInfo "Process terminated."} end)
	    
	    IlispTmp = {OS.tmpnam}
	    InitTmp  = {OS.tmpnam}#'.lisp'
	    
	    local %% create temporary init file
	       InitTmpVS =
	       '(delete-file "'#IlispTmp#'")\n'#
	       '(delete-file "'#InitTmp#'")\n'#
	       CmdsVS
	    in
	       {File.saveVS InitTmpVS InitTmp}
	    end
	    local %% create temporary ilisp file
	       Dump       = if ACLVERSION\=false andthen ACLVERSION>='5'
			    then ALLEGRO#" -I "#PREXDUMPEMACS
			    else PREXDUMPEMACS
			    end
	       IlispTmpVS = {File.loadVS LOUIEMACS_INIT}#
	       '(setq loui-program "'#Dump#' -qq -e (load\\"'#InitTmp#'\\")")'
	    in
	       {File.saveVS IlispTmpVS IlispTmp}
	    end
            
         in
            PipeGeneric, init(cmd:  "env"
                              args: ["DISPLAY="#Display
                                     LOUIEMACS %"-display" Display
                                     "-name" "Prex Service@"#Host#" (Emacs)"
				     "-geometry" GEOMETRY "-font" "8x13"
				     "-q" "-l" IlispTmp "-f" "loui"]
			      showInfo: ShowInfo
			      onExit:   ExitProc)
         end
         
         meth close
            PipeGeneric, close
         end      
      end
   end

   %%
   %% xterm mode
   %%
   local
      proc {LocalShowInfo Line}
         {App.showInfo "[XTERM]: "#Line}
      end
   in
      class PrexXterm
         from PipeGeneric
         prop final

         meth start(display : Display
                    host    : Host
                    cmds    : CmdsVS
                    showInfo: ShowInfo <= LocalShowInfo
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end)
            
            InitTmp  = {OS.tmpnam}#'.lisp'

            %% create temporary init file
            {File.saveVS '(delete-file "'#InitTmp#'")\n'#CmdsVS InitTmp}

            DumpArgs = if ACLVERSION\=false andthen ACLVERSION>='5'
                       then [ALLEGRO "-I" PREXDUMPEMACS]
                       else [PREXDUMPEMACS]
                       end    
         in
            PipeGeneric, init(cmd:  "env"
                              args: {Concat
                                     [["DISPLAY="#Display
                                       LOUIXTERM %"-display" Display
                                       "-title" "Prex Service@"#Host#" (xterm)"
                                       "-geometry" "80x25"
                                       "-e"] DumpArgs  
				      ["-qq" "-e" "(load \""#InitTmp#"\")"]]}
                              showInfo:ShowInfo onExit:ExitProc)
         end

         meth close
            PipeGeneric, close
         end
      end
   end

   %%
   %% quiet mode
   %%
   local
      skip
      %%proc {LocalShowInfo Line}
      %%   {App.showInfo "[LISP]: "#Line}
      %%end
   in
      class PrexQuiet
         from PipeGeneric
         prop final

         meth start(display : Display
                    cmds    : CmdsVS
                    showInfo: ShowInfo <= /*LocalShowInfo*/ proc {$ _} skip end
                    onExit  : ExitProc <=
                    proc {$} {ShowInfo "Process terminated."} end
                    ...)

            InitTmp = {OS.tmpnam}#'.lisp'

            %% create temporary init file
            {File.saveVS '(delete-file "'#InitTmp#'")\n'#CmdsVS InitTmp}

            DumpArgs = if ACLVERSION\=false andthen ACLVERSION>='5'
                       then [ALLEGRO "-I" PREXDUMPQUIET]
                       else [PREXDUMPQUIET]
                       end
         in
            PipeGeneric, init(cmd:  "env"
                              args: {Concat
                                     [["DISPLAY="#Display]
                                      DumpArgs
                                      ["-qq" "-e" "(load \""#InitTmp#"\")"]]}
                              showInfo:ShowInfo onExit:ExitProc)
         end

         meth close
            PipeGeneric, close(kill:true)
         end
      end
   end

   %%
   %% main
   %%
   class PrexServer from ThreadManager ClientManager FaultManager
                        
      feat
         serverName
         brokerURL
         brokerTicket
         
         inout % input/output stream
         serv  % mathweb stream

	 PrexPipe  % the pipe to the lisp- or emacs-process
         ServObj   % the Object providing the Oz-Services 

      meth init(url:    URL    <=false
		ticket: Ticket <=false
		cmdArgs: CmdArgs <= unit
	       )
         self.serverName   = ServerName
         self.brokerURL    = URL
         self.brokerTicket = Ticket
         ThreadManager, init
      end
      
      meth StartPrexPipe(ServerArgs exit:?ExitSignal)
    
         %% Oz <-> Lisp stream connections
         local
            RemoteURL = {CondSelect ServerArgs remoteURL unit}
            SessionID = {CondSelect ServerArgs sessionID unit}
            MethodName = 'iLouiIsAlive'

	    InOut Serv
         in
	    self.inout = {New Stream init(port:?InOut readStopC:EOF)}
            self.serv  = {New Stream init(port:?Serv  readStopC:EOF)}

	    %% stuff added for new interface by VS
            CmdsVS =
            '(ags::socket~reset)\n'#
	    '(ags::socket~define :inout)\n'#
            '(ags::socket~connect "'#Host#'" '#InOut#' :inout)\n'#
            '(ags::socket~define :service)\n'#
            '(ags::socket~connect "'#Host#'" '#Serv #' :service)\n'#
	    if RemoteURL\=unit andthen SessionID\=unit then
               '(ags::rpc~start-server '#
               '\'(\"'#RemoteURL#'\" \"'#MethodName#'\" \"'#SessionID#'\"))\n'
            else '' end
	    %% oz  ->lisp: (rpc~start-server '(URL MethodName activityID))
            %% lisp->java: iLouiIsAlive(string activityID, string urlToPrex)
         end

         %% auxiliuary procedures
         proc {SendClient Msg}
            try {Send self.clientPort Msg}
            catch _ then skip end
         end
         proc {SendError VS}
            {SendClient error(VS)}
         end
         proc {DoExit}
            {SendClient exit}
            ExitSignal=unit
         end

         %% pipe class
         PipeClass = case {CondSelect ServerArgs mode emacs}
		     of emacs then PrexEmacs
                     [] xterm then PrexXterm
                     [] quiet then PrexQuiet
                     [] Mode then
                        {SendError "Mode `"#Mode#"' unknown."}
                        {DoExit}
                        class
                           meth start(...) skip end
                           meth close(...) skip end
                        end
                     end

         %% init method
         Init = local
                   DefaultInit = start(display : self.clientDisplay
                                       host    : Host
                                       cmds    : CmdsVS
                                       onExit  : DoExit) 
                in
                   if {HasFeature ServerArgs outputLabel} then
                      OutLabel = ServerArgs.outputLabel
                      proc {ShowInfo Line} {SendClient OutLabel(Line)} end
                   in
                      {Adjoin x(showInfo:ShowInfo) DefaultInit}
                   else
                      DefaultInit
                   end
                end
      in
	 self.PrexPipe = {New PipeClass Init}        
      end
      
      /** This method enters an prex service */
      /** NOTE: ClientInfo = ClientName(name: user: host: display: port:) */
      /**       ServerArgs = unit(mode: remoteURL: sessionID:) */
      meth enter(client: ClientInfo <= unit
		 args:   ServerArgs <= unit)

	 ClientManager, init(ClientInfo)
         ClientManager, showEnter(self.appManager.showInfo)
	 ClientManager, startThreads

         ExitSignal
      in
	 PrexServer, StartPrexPipe(ServerArgs exit:?ExitSignal)
	 
         {Delay 2500} % important !!!
         
         if {IsFree ExitSignal} then
            {ForAll [inout serv] proc {$ Sock} {self.Sock accept} end}
%            PrexServer, StartReader
            PrexServer, StartServObj
            PrexServer, InstallWatcher    
         end
      end
  
      meth StartServObj
         proc {MyLoui M}
            {Send self.clientPort serv(M)}
         end
      in
         self.ServObj = {New OzService init(self.serv
                                            url:    self.brokerURL
                                            ticket: self.brokerTicket
                                            myLoui: MyLoui)}
      end
      meth InstallWatcher
         proc {Watch _ _}
            {self.appManager.showInfo
             "*** "#self.clientName#" died. Exiting..."}
            if {HasFeature self serverProc}
            then {self.serverProc leave}
            else {self leave}
            end
         end
      in
         FaultManager, installWatcher(Watch)
      end
      
      %%
      %% socket communication 
      %%

      meth poll(?Msg)
	 try
	    Msg = {self.inout getS($)}
	    {self.appManager.showInfo "[ReadSocket]: "#Msg}
	 catch _ then
	    try Msg="nil"
	    catch _ then skip end
	 end
      end
      
      meth send(String ?Done<=_)
	 {self.appManager.showInfo "[WriteSocket]: "#String}
	 {self.inout putS(String)}
	 Done="t"
      end

      meth sig(Signal $<=_)
	 {Process.sigF Signal self.PrexPipe.pid}
      end

      meth load(Url ?BS)
         try
            case {Resolve.localize Url}
            of old(F) then BS={File.loadBS F}
            [] new(F) then BS={File.loadBS F} {OS.unlink F}
            end
         catch _ then
            BS=false
         end
      end
      
      meth leave
	 {self.appManager.printInfo "Service "#ServerName#" is left...\n"}
	 FaultManager, deinstallWatcher
	 ThreadManager, tryTerminateAll
         if {IsDet self.ServObj} then
	    {self.ServObj shutdown}
	 end
	 {ForAll [inout serv] proc {$ Sock} {self.Sock close} end}  %% order changed...
	 {self.PrexPipe close}                                      %% which one is correct?
	 {self.appManager.showInfo " done."}
	 {Thread.terminate {Thread.this}}
      end

   end % of PrexServer

   ServMan = {New ServerManager init(ServerName PrexServer
                                     appName: AppName
                                     appMan:  App)}
   {ServMan offerService}
end







