%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% mbase-agent.oz ; This file is part of the MathWeb system
%
%   major updates: 31.05.2001
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% This file contains a first MBase Agent
%%%
functor

import
   Error
   Connection
   OS(getEnv tmpnam unlink)
   Resolve(localize)
   Browser(browse:Browse)

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'

   TermDB('class': MBaseConnection)
   at 'x-ozlib://mathweb/mbase/TermDB.ozf'
   Mbase('class':CommandClass)
   at 'x-ozlib://mathweb/mbase/Mbase.ozf'
   
   
   XOpen(server:Stream pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   XURL     at 'x-ozlib://mathweb/share/XURL.ozf'
   File
   at 'x-ozlib://mathweb/share/File.ozf'
   Process(sigF)
   at 'x-ozlib://mathweb/share/Process.ozf'
   
   %%----------------------------------
   %% KQML agent functors
   %%----------------------------------
   KqmlAgent('class': KQMLAgent)
   at 'x-ozlib://mathweb/kqml/KqmlAgent.ozf'
   KqmlLispSocket('class': KqmlLisp2OzSocket)
   at 'x-ozlib://mathweb/kqml/KqmlLispSocket.ozf'


define
   AppName     = 'mbase'
   ServiceName = 'MBaseAgent'
   App         = {New XApplication.manager init}
   
   VS2A = VirtualString.toAtom
   VS2S = VirtualString.toString
   
   ThreadID

   fun {ExceptionToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end
   
   StdPrefix = '/home/afranke/omega/mbase/db/'
   
%  Special CMD-Args because of Kqml agent mode
   ArgSpec = record(kqmlagent(single char:[&a] type:bool default: false))
   Args    
   try
      Args = {App.getCmdArgs ArgSpec}
   catch error(ap(usage S)...) then
      {App.showError 'mbase: '#S}
      {App.exit 1}
   end

   THEOREMS = 'OMS'(cd: "reasys" name: "theorems")
   
   LIST   = 'OMS'(cd: "list1" name: "list")
   
   FORALL = 'OMS'(cd: "quant1" name: "forall")
   LEQ    = 'OMS'(cd:"relation1" name:"leq")

   ABS    = 'OMS'(cd: "arith1" name: "abs")
   MINUS  = 'OMS'(cd: "arith1" name: "minus")
   TIMES  = 'OMS'(cd: "arith1" name: "times")
   
   SIN    = 'OMS'(cd: "transc1" name: "sin")
   COS    = 'OMS'(cd: "transc1" name: "cos")
   
   class MBaseAgent from MBaseConnection KQMLAgent
      attr
         locking
      feat
         facilitator
         
      meth init(ticket:  BrokerTicket
                url:     BrokerURL
               )
         
% Facilitator,  advertise(sender: Sender content: ServiceContent url: URL ...) = M
         try
            self.facilitator = {Connection.take BrokerTicket}
         catch Exc then
            {App.showInfo {ExceptionToVS Exc}}
            {App.showInfo 'Broker is not reachable :-('}
            {App.exit 1}
         end
         try
            ID
	    Content = 'ask-all'(content: 'OMOBJ'(
					    'OMA'('OMS'(name: "theorems" cd: "reasys")
						  ['OMV'(name: "X")])))
	    RealName Symbol
         in
            {self.facilitator registerAgent(ServiceName self ?ID)}
	    RealName = {VS2S ServiceName#ID}
	    self.id = ID
            
            {self.facilitator advertise(sender:  ServiceName %RealName ToDo
                                        content: Content
                                        agent:   self)}
            KQMLAgent, init(app:  App
                            name: ServiceName
                            facilitator: self.facilitator
                           )
            
%            MBaseConnection, connectToDatabase()
            
%            MBaseConnection, lookupSymbol(cd: "arith1" name: "plus" Symbol)
         catch Exc then
            {Browse Exc}
            {App.showInfo 'Broker is not a facilitator :-('}
         end
      end

      meth showInfo(VS)
         {App.showInfo VS}
      end
      
      meth test
         {App.showInfo 'Hi, ich bin ein KQML MBase agent'}
      end
      
      meth handleMessage(Message ?Accept) = M
%         {Browse M}
         case Message
         of 'ask-all'(...) then
            {self Message}
         else
            {App.showInfo 'Message not understood '}
         end
      end
                       
      meth 'ask-all'(sender: Sender content: Content ...) = M
         lock
            Conv
            Message
            ResultContent
         in
%            {Browse M}
            {self.convModule newInConversation(M
                                               type:   'query'
                                               wait:   true
                                               Conv)}
	    case Content
%%% LOWER bounds
	    of 'OMOBJ'('OMA'(!THEOREMS [Formula ResultMetaVar]))
	    then
%	       {Browse mbase(Content)}
	       MBaseConnection, connectToDatabase(host:"mbase.mathweb.org")
	       
	       Result = MBaseConnection, getMatchingTheorems(Formula nil $)
	       /*
	       case Formula
%%% LOWER bounds
	       of 'OMBIND'(!FORALL _ 'OMA'(!LEQ [_ 'OMA'(!SIN _)]))
	       then
		  X = 'OMV'(name:"x")
		  Term =
		  'OMBIND'(FORALL [X] %% -1 * abs x <= sin x
			   'OMA'(LEQ
				 ['OMA'(TIMES ['OMI'(~1) 'OMA'(ABS [X])])
				  'OMA'(SIN [X])
				 ]))
	       in
		  'OMA'(LIST [Term])
		  		  
	       [] 'OMBIND'(!FORALL _ 'OMA'(!LEQ [_ 'OMA'(!COS _)]))
	       then
		  X = 'OMV'(name:"x")
		  Term =
		  'OMBIND'(FORALL [X] %% 1 - x*x <= cos x
			   'OMA'(LEQ
				 ['OMA'(MINUS ['OMI'(1) 'OMA'(TIMES [X X])])
				  'OMA'(COS [X])]))
	       in
		  'OMA'(LIST [Term])
%%% UPPER bounds 
	       [] 'OMBIND'(!FORALL _ 'OMA'(!LEQ ['OMA'(!SIN _) _]))
	       then
		  X = 'OMV'(name:"x")
		  Term =
		  'OMBIND'(FORALL [X] %% sin x <= abs x
			   'OMA'(LEQ ['OMA'(SIN [X]) 'OMA'(ABS [X])]))
	       in
		  'OMA'(LIST [Term])
		  
	       [] 'OMBIND'(!FORALL _ 'OMA'(!LEQ ['OMA'(!COS _) _]))
	       then
		  X = 'OMV'(name:"x")
		  Term =
		  'OMBIND'(FORALL [X] %% cos x <= 1
			   'OMA'(LEQ ['OMA'(COS [X]) 'OMI'(1)]))
	       in
		  'OMA'(LIST [Term])
	       else
		  X = 'OMV'(name:"x") 
		  Term = 
		  'OMBIND'(FORALL [X] %% sin x <= abs x
			   'OMA'(LEQ ['OMA'(SIN [X]) 'OMA'(ABS [X])]))
	       in
		  'OMA'(LIST [Term])
	       end
	       */
	       ResultContent = 'OMOBJ'('OMA'('OMS'(name: "theorems" cd: "reasys")
					     [Formula                                              
					      Result])
				      )
	    in
%               {Browse result(Result)}
%             KQMLAgent, send(Receiver Performative Content ?OK
	       KQMLAgent, send(Sender 'tell' ResultContent _
			       message: Message)
	       {Conv addOutMessage(Message)}
               {self.facilitator showConvs}
%      KqmlAgent, send(Receiver Performative Content ?OK
            end
         end
      end
      
      meth close
         {App.showInfo "Agent is dying!"}
%         MBaseConnection, close
         {self.facilitator unregisterAgent(ServiceName self.id)}
         {App tryTerminate(ThreadID info: "Trying to terminate agent")}
      end
      
      
      
   end % class MBaseAgent
   
   BrokerTicket BrokerURL
   case {App.getEnv 'BROKER_TICKET'} of false then
      BrokerURL = "http://www.ags.uni-sb.de/~mathweb/broker@leibniz"
      {App.printInfo "Loading broker-ticket from `"#BrokerURL#"'..."}
      BrokerTicket = {VS2A {XURL.loadVS BrokerURL}}
      {App.showInfo " done."}
   elseof BROKER_TICKET then
      BrokerURL = false
      BrokerTicket = {VS2A BROKER_TICKET}
   end
   
   Server
   DoExit = proc {$ _}
%               {Thread.terminate {Thread.this}}
               {Server close}
            end
   
   Server = {App newServer(MBaseAgent
			   init(ticket:  BrokerTicket
                                url:     BrokerURL
			       )
			   $
			   ThreadID
                           detach: true
                           info:   ServiceName
                           onExit: DoExit)}

   {Broker offer(adfasf
		 mode: pass_on
		)
%          ServMan = {New ServiceManager  init(ServiceName
%                                              MBaseAgent
%                                              appName: AppName
%                                              appMan:  App
%                                              argSpec: ArgSpec
%                                             )}
   
end  %define




