%% -*- indent-tabs-mode: nil -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% mbase.oz ; This file is part of the MathWeb system
%
%   major updates: 22.05.2001
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%% a first Mbase service
%%

functor

import
   Error
   Connection
   OS(getEnv tmpnam unlink)
   Resolve(localize)
   Browser(browse:Browse)

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   BrokerAccess(getBroker)
   at 'x-ozlib://mathweb/sys/BrokerAccess.ozf'

   TermDB('class': MBaseConnection)
   at 'x-ozlib://mathweb/mbase/TermDB.ozf'
   Mbase('class':CommandClass)
   at 'x-ozlib://mathweb/mbase/Mbase.ozf'
   
   
   XOpen(server:Stream pipe:PipeGeneric)
   at 'x-ozlib://mathweb/share/XOpen.ozf'
   File
   at 'x-ozlib://mathweb/share/File.ozf'
   Process(sigF)
   at 'x-ozlib://mathweb/share/Process.ozf'
   
   %%----------------------------------
   %% KQML agent functors
   %%----------------------------------
   KqmlAgent('class': KQMLAgent)
   at 'x-ozlib://mathweb/kqml/KqmlAgent.ozf'
   KqmlLispSocket('class': KqmlLisp2OzSocket)
   at 'x-ozlib://mathweb/kqml/KqmlLispSocket.ozf'

define
   AppName     = 'mbase'
   ServiceName = 'MBASE'
   App         = {New XApplication.manager init}
   
   VS2S = VirtualString.toString
   
   fun {ExceptionToVS Exc}
      {Error.messageToVirtualString {Error.exceptionToMessage Exc}}
   end

   StdPrefix = '/home/afranke/omega/mbase/db/'
   
%  Special CMD-Args because of Kqml agent mode
   ArgSpec = record(kqmlagent(single char:[&a] type:bool default: false))
   Args    
   try
      Args = {App.getCmdArgs ArgSpec}
   catch error(ap(usage S)...) then
      {App.showError 'mbase: '#S}
      {App.exit 1}
   end

%    class CommandClass from MB
                         
%       % feat db
%       /*
%       meth init
%          self.db = {New MB noop}
%       end
%       meth open(DBPrefix<=StdPrefix)
%          {self.db openRead(DBPrefix)}
%       end
%       meth close
%          {self.db close}
%       end
%       */
      
%       meth init
%          skip
%       end
%       meth open
%          MB, connectToDatabase
%       end
%       meth close
%          MB, disconnect
%       end
%       meth showInfo(VS)
%          {App.showInfo VS}
%       end
      
%       /*
%       meth ping($)
%          'MBASE is alive.'
%       end
%       meth testQuote($)
%          'this "is" a test'
%       end
      
%       meth otherwise(M)
%          {self.db M}
%       end
%       */
%    end
   
   class MBaseService from MBaseConnection
      feat
         BrokerURL
         BrokerTicket
         broker
         client
         args
         
         mbase

         agent
         
      meth init(url:    BrokerUrl    <=false
                ticket: BrokerTicket <=false
                cmdArgs: CmdArgs     <= unit
               )
         {Browse ticket(BrokerTicket)}
         try
            if BrokerTicket \= false then
               self.broker = {Connection.take BrokerTicket}
            else
               self.broker = {BrokerAccess.getBroker App a(brokerticket: BrokerUrl)}
            end
         catch Exc then
            {App.showInfo {ExceptionToVS Exc}}
            {App.exit 1}
         end
         
         self.agent        = {HasFeature CmdArgs kqmlagent} andthen CmdArgs.kqmlagent
         
         if self.agent then
            {App.showInfo 'I am a KQML agent!'}
         end
      end
      
      meth enter(client : ClientName
                 display: Display <= _
                 args   : ArgRec  <= unit)
         
         self.client = ClientName
         self.args   = ArgRec
         
         MBaseConnection, connectToDatabase(...)
      end

      meth Apply(Meth ?Result)
         fun {GetFirstUnusedIntFeature R Start}
            if {HasFeature R Start}
            then {GetFirstUnusedIntFeature R Start+1}
            else Start
            end
         end
         I = {GetFirstUnusedIntFeature Meth 1}
         Ar= {Arity Meth}
         M = {MakeRecord {Label Meth} I|Ar}
      in
         {ForAll Ar proc {$ X} M.X=Meth.X end}
         M.I=Result
         {self  M}
      end
      
      meth exec(Command ?Result)
         if {IsDet Command} andthen {IsRecord Command} then
            try
               LocalResult in
	       MBaseService, Apply(Command ?LocalResult)
               Result = LocalResult
            catch Err then
               Result = 'ERROR'(2 'MBASE: Error while executing command.'#
                                {ExceptionToVS Err})
            end
         else Result = 'ERROR'(1 'MBASE: Specified command-method is not a record.')
         end
      end

      meth getDefinitions(
      %% meth kill
      %%    skip
      %% end
      
      meth leave
         try {self close}
         catch _ then skip end
         {Thread.terminate {Thread.this}}
      end

   end 

   THEOREMS    = 'OMS'(cd: "reasys"    name: "theorems")
   LIST        = 'OMS'(cd: "list1"     name: "list")
   FORALL      = 'OMS'(cd: "quant1"    name: "forall")
   LEQ         = 'OMS'(cd: "relation1" name: "leq")
   SIN         = 'OMS'(cd: "transc1"   name: "sin")
   COS         = 'OMS'(cd: "transc1"   name: "cos")
   ABS         = 'OMS'(cd: "arith1"    name: "abs")
   TIMES       = 'OMS'(cd: "arith1"    name: "times")
   MINUS       = 'OMS'(cd: "arith1"    name: "minus")
   UNARY_MINUS = 'OMS'(cd: "arith1"    name: "unary_minus")   

   MBaseAgent =
   {ByNeed
    fun {$}
       class $
          from MBaseService KQMLAgent
          meth init(...) = Init
             MBaseService, Init
% Facilitator,  advertise(sender: Sender content: ServiceContent url: URL ...) = M
             try
                ID
                Content = 'ask-all'(content: 'OMOBJ'('OMA'(THEOREMS
                                                           ['OMV'(name: "X")
                                                            'OMV'(name: "Y")
                                                           ])))
                RealName
             in
                {self.broker registerAgent(ServiceName self ?ID)}
                RealName = {VS2S ServiceName#ID}
                
                {self.broker advertise(sender:  RealName
                                       content: Content
                                       agent:   self)}
                KQMLAgent, init(app:  App
                                name: RealName)
             catch _ then
                {App.showInfo 'Broker is not a facilitator :-('}
             end
          end
          
          meth test
             {App.showInfo 'Hi, ich bin ein KQML MBase agent'}
          end
          
          meth handleMessage(Message ?Accept)
             case Message
             of 'ask-all'(...) then
                {self Message}
             else
                {App.showInfo 'Message not understood '}
             end
          end
          
          meth 'ask-all'(sender: Sender content: Content ...)
             case Content
             of 'OMOBJ'('OMA'(!THEOREMS Formula ResultMetaVar)) then
                %% Result = MBaseService, getMatchingTheorems(Formula $)
                X = 'OMV'(name:"x")
                Result =
                'OMA'(LIST
                      [
                       'OMBIND'(FORALL 'OMBVAR'(X)
                                'OMA'(LEQ ['OMA'(UNARY_MINUS ['OMA'(ABS [X])])
                                           'OMA'(SIN [X])]))
                       'OMBIND'(FORALL 'OMBVAR'(X)
                                'OMA'(LEQ ['OMA'(MINUS
                                                 ['OMI'(1) 'OMA'(TIMES [X X])])
                                           'OMA'(COS [X])
                                          ]))])
                
                ResultContent = 'OMOBJ'('OMA'(THEOREMS Formula Result))
             in
                KQMLAgent, send(Sender 'tell' ResultContent)
                %% KqmlAgent, send(Receiver Performative Content ?OK
             else
                skip      %% ToDo
             end
          end
          
       end % class ..Agent
    end}
   
   ServiceClass = if Args.kqmlagent then
                     MBaseAgent
                  else
                     MBaseService
                  end
   
   ServMan = {New ServiceManager  init(ServiceName
                                       ServiceClass
                                       appName: AppName
                                       appMan:  App
                                       argSpec: ArgSpec
                                      )}
   {ServMan offerService}
   {App.exit 0}
end




