%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% kimba.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% The Kimba service
%%%

functor
   
import
   OS(getEnv)
   Browser(browse: Browse)
   
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'kimba'
   ServiceName = 'KIMBA'
   App         = {New XApplication.manager init}

   KIMBA       = {OS.getEnv "KIMBAHOME"}

   SatisfiableLine = "Model found!"
   RefutationLine  = "No model found."

   fun {FirstLine S}
      {List.takeWhile S fun {$ C} C\=&\n end}
   end
   
   class KimbaService from ShellProver
      feat
         name: kimba
         
      meth submitProblem(ProblemBS syntax: Syntax<=kimba)
         
         case Syntax of kimba then
            {self saveVS(ProblemBS "kimba.in")}
            {self start(KIMBA#" --in kimba.in --out kimba.out")}
         else
            raise unknownSyntax(Syntax) end
         end
      end

      meth analyzeResult($)
         %%{self exec("tail +2 kimba.out >kimba.model")}
         FL = {FirstLine {self loadS("kimba.out" $)}}
         {Browse FL}
         State = case FL 
                 of !RefutationLine  then proof
                 [] !SatisfiableLine then satisfiable
                 else undetermined
                 end
         
         %% cleanup
         {self exec("rm -f kimba.in kimba.out kimba.model")}
      in
         atp_result(state: State
                    host:  self.host
                    name:  self.name
                    output: "dummy output")
      end       
   end
   
   ServMan = {New ServiceManager init(ServiceName KimbaService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end

%       meth analyzeResult($)
%          %%{self exec("tail +2 kimba.out >kimba.model")}
         
%          ProofOk = case {FirstLine {self loadS("kimba.out" $)}}
%                    of !RefutationLine  then true
%                    [] !SatisfiableLine then false
%                    else unknown
%                    end

%          Proof = case ProofOk
%                  of true    then "proof found\n"
%                  [] false   then "-------------------------------\n"
%                  [] unknown then "unknown result\n"
%                  end
         
%          %% cleanup
%          {self exec("rm -f kimba.in kimba.out kimba.model")}
%       in
%          s(string: {ByteString.make Proof}
%            bool  : ProofOk\=true
%            host  : self.host)
%       end       
