%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% mace.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% The MACE service
%%%


functor
   
import
   OS(getEnv)

   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'mace'
   ServiceName = 'MACE'
   App         = {New XApplication.manager init}

   MACE        = {OS.getEnv "MACEHOME"}
   TIME_LIMIT  = 300
   DOMAINSIZE  = 10
   
   %% to use MACE for DORIS I set MAX_SYMBOLS in generate.c to 100 (was: 50):
   %%
   %%    #define MAX_SYMBOLS   100  number of functors

   %% Problem is satisfiable iff Mace's output file contains
   %%
   %% "    Models found                      1"
   
%   SatisfiableLine = "    Models found
%   SatisfiableLine = "Models found"     %% changed jz. 3.9.2001
   SatisfiableLine = "======================= Model"  %% also compatible with Mace 2.0
   
   fun {FirstLine S}
      {List.takeWhile S fun {$ C} C\=&\n end}
   end
   
   class MaceService from ShellProver
      feat
         name: mace
         
      meth enter(...) = M
         ShellProver, M
         %%{self resetSaver("/home/afranke/mace")}
      end
      
% solaris:      /project/omega/atp/bin/solaris/mace -n 10 -p -t 300 -m1 < prob.otter
      
      meth submitProblem(ProblemBS syntax:Syntax<=otter)

         case Syntax of otter then
            
            {self saveVS(ProblemBS "mace.in")}
            
            %% ./mace -nDomainSize -p -tTimeOut -m1 < temp.in > temp.out
            %%
            %% -n: domain size 
            %% -p: prints models in temp.out
            %% -t: for timeout
            %% -m1 for finding at most one model

            {self start(MACE#" -n"#DOMAINSIZE#" -p -t"#TIME_LIMIT
                        #" -m1 < mace.in > mace.out")}
         else
            raise unknownSyntax(Syntax) end
         end
      end

      meth analyzeResult($)
         %% old mace:
         %% ======================= Model #1 at 0.00 seconds:
         %%     Models found                      1

         %% new mace:
         %% ======================= Model #1 at 0.00 seconds:
         {self exec("grep '"#SatisfiableLine#"' mace.out > mace.result")}
         
         State = case {FirstLine {self loadS("mace.result" $)}}
                 of &=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|&=|& |&M|&o|&d|&e|&l|_ then
                    satisfiable
                 else
                    undetermined
                 end
         
         %% cleanup
         %%{self save}
         {self exec("rm -f mace.in mace.out mace.result")}
      in
         atp_result(state: State
                    host: self.host
                    name: self.name
                    output: "dummy output")
      end       
   end
   
   ServMan = {New ServiceManager init(ServiceName MaceService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end


%       meth analyzeResult($)
         
%          {self exec("grep '"#SatisfiableLine#"' mace.out > mace.result")}
         
%          ProofOk = case {FirstLine {self loadS("mace.result" $)}}
%                    of !SatisfiableLine then false
%                    else unknown
%                    end

%          Proof = case ProofOk
%                  of true    then "proof found\n"
%                  [] false   then "-------------------------------\n"
%                  [] unknown then "unknown result\n"
%                  end
         
%          %% cleanup
%          %%{self save}
%          {self exec("rm -f mace.in mace.out mace.result")}
%       in
%          s(string: {ByteString.make Proof}
%            bool  : ProofOk\=true
%            host  : self.host)
%       end       
