%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% satchmo.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% The SATCHMO service
%%%

functor
   
import
   OS(getEnv)
   
   XService(manager: ServiceManager)
   at 'x-ozlib://mathweb/sys/XService.ozf'
   ShellProverClass('class': ShellProver)
   at 'x-ozlib://mathweb/sys/ShellProver.ozf'

   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
define
   AppName     = 'satchmo'
   ServiceName = 'SATCHMO'
   App         = {New XApplication.manager init}

   SATCHMO     = {OS.getEnv "SATCHMOHOME"}

   SatisfiableLine = "%% Model found:"
   RefutationLine  = "%% No model"
   
   class SatchmoService from ShellProver
      feat
         name: satchmo
         
      meth enter(...) = M
         ShellProver, M
         %%{self resetSaver("/home/afranke/satchmo")}
      end
      
      meth submitProblem(ProblemBS syntax: Syntax<=otter)
         
         case Syntax of otter then
            {self saveVS(ProblemBS "satchmo.in")}
            
            {self start(SATCHMO#" +RTS -H32M -K4M -RTS "      % memory settings
                        #"--complement-splitting --models=1 " % general options
                        #"satchmo.in >satchmo.out.tmp ;"
                        #"mv satchmo.out.tmp satchmo.out")}
         else
            raise unknownSyntax(Syntax) end
         end
      end

      meth analyzeResult($)
         
         {self exec("grep '"#SatisfiableLine#"' satchmo.out > satchmo.sat")}
         {self exec("grep '"#RefutationLine#"' satchmo.out > satchmo.prf")}
         
         State = case {self loadS("satchmo.sat" $)}
                 of !SatisfiableLine then
                    satisfiable
                 elsecase {self loadS("satchmo.prf" $)}
                 of !RefutationLine then
                    proof
                 else
                    undetermined
                 end
         
         %% cleanup
         %%{self save}
         {self exec("rm -f satchmo.in satchmo.out.tmp satchmo.out"
                    #" satchmo.sat satchmo.prf")}
      in
         atp_result(state: State
                    host:  self.host
                    name:  self.name
                    output: "dummy output")
      end       
   end
   
   ServMan = {New ServiceManager init(ServiceName SatchmoService
                                      appName: AppName
                                      appMan:  App)}
   {ServMan offerService}
   {App.exit 0}
end

%       meth analyzeResult($)

%          {self exec("grep '"#SatisfiableLine#"' satchmo.out > satchmo.sat")}
%          {self exec("grep '"#RefutationLine#"' satchmo.out > satchmo.prf")}
         
%          ProofOk = case {self loadS("satchmo.sat" $)}
%                    of !SatisfiableLine then false
%                    elsecase {self loadS("satchmo.prf" $)}
%                    of !RefutationLine then true
%                    else unknown
%                    end

%          Proof = case ProofOk
%                  of true    then "proof found\n"
%                  [] false   then "-------------------------------\n"
%                  [] unknown then "unknown result\n"
%                  end
         
%          %% cleanup
%          %%{self save}
%          {self exec("rm -f satchmo.in satchmo.out.tmp satchmo.out"
%                     #" satchmo.sat satchmo.prf")}
%       in
%          s(string: {ByteString.make Proof}
%            bool  : ProofOk\=true
%            host  : self.host)
%       end       
