/*---------------------------------------------------------------------------*/
/*----Program to limit the CPU time available to a job                       */
/*----Usage: CPULimitedRun <TimeLimit> <Job>                                 */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/times.h>
#include <limits.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
/*---------------------------------------------------------------------------*/
#define STRING_LENGTH 80
#define MAX_ARGUMENTS 30

#define READ_INDEX 0
#define WRITE_INDEX 1
#define STDIN 0
#define STDOUT 1
/*---------------------------------------------------------------------------*/
typedef char String[STRING_LENGTH];
typedef char *StringName;
/*---------------------------------------------------------------------------*/
void XCPUHandler(int TheSignal)
{
if (TheSignal != SIGXCPU)
    printf("Error, %d is not SIGXCPU\n",TheSignal);
else {
    printf("SGIXCPU caught\n");
    exit(-1);
    }
}
/*---------------------------------------------------------------------------*/
void SetCPUTimeLimit(int CPUTimeLimit) {

struct rlimit ResourceLimits;

/*----Set the signal handler                                                 */
/*----No point - signals are reset on exec                                   */
/*----For some reason this is still needed, or the signal doesn't work       */
if (signal(SIGXCPU,XCPUHandler) == SIG_ERR)
    perror("CPULimitedRun - Setting signal handler:");

/*----Limit the CPU time. Need to get old one for hard limit field           */
if (getrlimit(RLIMIT_CPU,&ResourceLimits) == -1)
    perror("CPULimitedRun - Getting resource limit:");
/*----Set new CPU limit in ms (sent in secs)                                 */
ResourceLimits.rlim_cur = CPUTimeLimit-1; /*---- -1 because its weird */
if (setrlimit(RLIMIT_CPU,&ResourceLimits) == -1)
    perror("CPULimitedRun - Setting resource limit:");
if (getrlimit(RLIMIT_CPU,&ResourceLimits) == -1)
    perror("CPULimitedRun - Getting resource limit:");
}
/*---------------------------------------------------------------------------*/
/*----Prevent core dumps that occur on timeout                               */
void SetNoCoreDump(void) {

struct rlimit ResourceLimits;

/*----Get old resource limits                                                */
if (getrlimit(RLIMIT_CORE,&ResourceLimits) == -1)
    perror("CPULimitedRun - Getting resource limit:");
/*----Set new core limit to 0                                                */
ResourceLimits.rlim_cur = 0;
if (setrlimit(RLIMIT_CORE,&ResourceLimits) == -1)
    perror("CPULimitedRun - Setting resource limit:");
}
/*---------------------------------------------------------------------------*/
void main(int NumberOfArguments,char *Arguments[])
{
int TimeLimit;
int ArgumentNumber;

printf(
"CPULimitedRun: ----------------------------------------------------------\n");
TimeLimit = atoi(Arguments[1]); 
/*----Check the time limit is one                                            */
if (TimeLimit > 0) {
    printf("CPULimitedRun: %s ",Arguments[2]);
    for (ArgumentNumber=3;ArgumentNumber<NumberOfArguments;ArgumentNumber++)
        printf("%s ",Arguments[ArgumentNumber]);
    printf("\n");
    printf("CPULimitedRun: Time limit is %ds\n",TimeLimit);
/*----Output the PID for possible later use                                  */
    printf("CPULimitedRun: PID is %d\n",getpid());
    printf(
"CPULimitedRun: ----------------------------------------------------------\n");
    fflush(stdout);
    SetCPUTimeLimit(TimeLimit);
    SetNoCoreDump();

/*----Execute the ATP system                                                 */
    execv(Arguments[2],Arguments+2);
    perror("CPULimitedRun - Cannot exec program:");
    }
else printf("Invalid time limit as first argument\n");
}
/*---------------------------------------------------------------------------*/
