%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Cas.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Defines the superclass of all CAS systems
%%    - 'CAS' for Computer Algebra Systems
%%
functor  

import
   Error  Open OS(uName system tmpnam unlink getEnv)
   System
%   Browser(browse: Browse)

   File            at 'x-ozlib://mathweb/share/File.ozf'   

export
   cas: CAS
   
define
   
%   VS2A = VirtualString.toAtom
   VS2S = VirtualString.toString
   
   fun {ExceptionToVS Err}
      {Error.messageToVirtualString {Error.exceptionToMessage Err}}
   end

   proc {Skip1 _} skip end

   fun {IsSubstring Xs Ys}
      {List.isPrefix Xs Ys} orelse
      case Ys of nil then false
      [] _|Yr then {IsSubstring Xs Yr} end
   end

   local
      fun {Substitute S S1 S2}
	 if {List.isPrefix S1 S}
	 then {Append S2 {Substitute {List.drop S {Length S1}} S1 S2}}
	 elsecase S of nil then nil
	 [] C|Sr then C|{Substitute Sr S1 S2}
	 end
      end
      fun {MultiSubst S SubstPairs}
	 {FoldL SubstPairs fun {$ S S1#S2} {Substitute S S1 S2} end S}
      end

      fun {MySome Xs Test}
	 %%Authors = [afranke]
	 %%Input   = "A list and a Test, which may return an _arbitrary_ value or false." 
	 %%Effect  = "None."
	 %%Value   = "Test is applied sequentially to the elements of the list. "#
	 %%          "The first Test result \\= false is returned, otherwise false."
	 %%in
	 case Xs
	 of nil then false
	 [] X|Xr then
	    case {Test X}
	    of false then {MySome Xr Test}
	    [] Result then Result
	    end
	 end
      end
      
      class OutputBuffer
	 feat
	    Prompts  %% the reverse of the prompts
	    RevSubsts
	    RevErrors
	 attr
	    RevS: nil
            
	 meth init(PromptVSs subst:SubstPairs<=nil err:Errors<=nil)
	    self.Prompts = {Map PromptVSs
			    fun{$ P}
			       VS#TimeOut = case P
					    of _#T andthen {IsInt T} then P
					    else P#unit
					    end
			       PS = {VS2S VS}
			    in
			       prompt(reverse: {Reverse PS}
				      len: {Length PS}
				      timeOut: TimeOut)
			    end}
	    self.RevSubsts = {Map {Map SubstPairs
				   fun {$ VS1#VS2} {VS2S VS1}#{VS2S VS2} end}
			      fun {$ S1#S2} {Reverse S1}#{Reverse S2} end}
	    self.RevErrors = {Map Errors fun {$ VS} {Reverse {VS2S VS}} end}
	    RevS <- nil
	 end
	 meth put(S)
	    RevS <- {Append {Reverse S} @RevS}
	    RevS <- {MultiSubst @RevS self.RevSubsts} %slow but simple
	 end
	 meth prompt(?P)
	    P = {MySome self.Prompts
		 fun {$ Prompt}
		    if {List.isPrefix Prompt.reverse @RevS} then
		       Prompt
		    else false end
		 end}
	    if P \= false then {System.printInfo "[prompt]"} end
	 end
	 meth err(?B)
	    B={Some self.RevErrors
	       fun {$ RevErr} {IsSubstring RevErr @RevS} end}
	    if B then
	       {System.printInfo "[err]"}
	       {ForAll self.RevErrors
		proc {$ RevErr}
		   if {IsSubstring RevErr @RevS} then
		      {System.printInfo "[["#RevErr#"]]"}
		   end
		end}
	    end
	 end

	 meth dropPrompt(Prompt)
	    RevS <- {List.drop @RevS Prompt.len}
	 end

	 meth getAll(?S)
	    S = {Reverse @RevS}
	    RevS <- nil
	 end
      end

      class SafePipeReader from Open.pipe
	 meth safeRead(list:?S
		       timeOut: TimeOut<=10
		       default: Default<=nil)
	    Thr Exc in
	    thread
	       Thr = {Thread.this} 
	       try
		  Open.pipe, read(list:?S)
		  raise unit end
	       catch E then Exc=E end
	    end
	    {WaitOr S {Alarm TimeOut}}
	    try
	       {Thread.suspend Thr}
	       if {Thread.state Thr}==blocked then
		  {Thread.injectException Thr unit}   
	       else
		  {Thread.resume Thr}
	       end
	    catch error(kernel(deadThread...)...) then skip end
	    {Wait Exc}
	    if Exc==unit then
	       if {IsFree S} then S=Default end
	    else
	       raise Exc end
	    end
	 end
      end
      ShortTimeOut = 100
   in
      class CAS from Open.pipe SafePipeReader OutputBuffer

	 prop
	    final
	    locking

	 feat
	    pid
	    printInfo
	    suffix
	    logFile

	 attr
	    status: unit

	 meth status($)
	    @status
	 end

	 meth init(cmd:Cmd args:Args
		   suffix:    SuffixVS  <= ";\n"
		   prompts:   PromptVSs <= ["\n> "]
		   subst:     SubstPairs<= nil
		   err:       Errors    <= nil
		   hasPrompt: HasPrompt <= true
		   prelude:   Prelude   <= nil
		   logFile:   LogFile   <= unit
		   printInfo: PrintInfo <= Skip1)
	    self.pid       = Open.pipe, init(cmd:Cmd args:Args pid:$)
	    self.printInfo = PrintInfo
	    self.suffix    = {VirtualString.toString SuffixVS}

	    self.logFile = LogFile
	    OutputBuffer, init(PromptVSs subst:SubstPairs err:Errors)
	    {self.printInfo Prelude}
	    case self.logFile
	    of unit then skip
	    [] Name then
	       {File.appendVS Prelude Name}
	    end
	    Open.pipe, write(vs: Prelude)
	    if HasPrompt then
	       CAS, ReadTillPrompt(_)
	    end
	    status <- idle
	 end

	 meth eval(Expr ?Result<=_
		   prompt: ?Prompt <= _
		   started:?Started<= _)
	    TmpResult
	 in
	    lock
	       status <- busy
	       Started = unit
	       CAS, write(Expr)
	       CAS, Read(?TmpResult prompt: ?Prompt)
	       status <- idle
	    end
	    Result = TmpResult
	 end

	 meth write(ExprVS)
	    L
	    Suffix = self.suffix in
	    {System.showInfo "\n--"#ExprVS#Suffix}
	    case self.logFile
	    of unit then skip
	    [] Name then
	       {File.appendVS ExprVS#Suffix Name}
	    end
	    Open.pipe, write(vs: ExprVS#Suffix len:?L)
	    if L\={Length {VS2S ExprVS#Suffix}} then
	       {self.printInfo "**********************AAAAAAAAAAAAAAAAARgh!!!!!!!!!!!!!!!!!!!\n"}
	       {self.printInfo "Only "#L#" characters written out of "#{Length {VS2S ExprVS#Suffix}}#"\n"}
	    end
	    Open.pipe, flush(how:[send])
	    {self.printInfo "[flushed]"}
	 end

	 meth Read($ prompt: ?Prompt <= _)
	    try CAS, ReadTillPrompt($ prompt: ?Prompt)
	    catch Err then
	       {self.printInfo "*** ERROR IN CAS, Read($):"}
	       {self.printInfo {ExceptionToVS Err}#"\n"} 
	       nil 
	    end
	 end
	 meth readRest($)
	    lock
	       try CAS, ReadTillTimeOut($)
	       catch Err then 
		  {self.printInfo "*** ERROR IN CAS, readRest($):"}
		  {self.printInfo {ExceptionToVS Err}#"\n"} 
		  nil 
	       end
	    end
	 end


	 meth ReadTillPrompt(?ReadString prompt:?Prompt<=_)
	    {System.printInfo "[wait]"}
	    S = Open.pipe, read(list: $
				/*tail: TailX   <= nil*/   
				/*size: +SizeAI <= 1024*/   
				/*len:  ?LenI   <= _*/)
	 in
	    if S==nil then
	       %% pipe dead
	       {self.printInfo '\ndead\n'}
	       {self close}
	       {self.printInfo '\nclosed\n'}
	       ReadString = nil
	       Prompt = false
	    else
	       {self.printInfo S}
	       OutputBuffer, put(S)
	       CAS, ProcessString(?ReadString prompt:?Prompt)
	    end
	 end

	 meth ProcessString(?ReadString prompt:?Prompt<=_)
	    if OutputBuffer, err($) then
	       %% substring error indicator found 
	       ReadString = {VS2S (OutputBuffer,getAll($))#(CAS,ReadTillTimeOut($))}
	       Prompt = false
	    elsecase OutputBuffer, prompt($)
	    of false then
	       %% no prompt, continue reading (recursion)
	       ReadString = CAS, ReadTillPrompt($ prompt: ?Prompt)
	    [] P then
	       %% possible prompt found at end of string
	       More = case {CondSelect P timeOut unit}
		      of unit then nil
		      [] TimeOut then
			 SafePipeReader, safeRead(list:$
						  timeOut: TimeOut
						  default: nil)
		      end
	    in
	       if More \= nil then
		  OutputBuffer, put(More)
		  CAS, ProcessString(?ReadString prompt: ?Prompt)
	       else
		  %% prompt found at end of string
		  Prompt = P
		  OutputBuffer, dropPrompt(Prompt)
		  ReadString = OutputBuffer, getAll($)
	       end
	    end
	 end
         
	 meth ReadTillTimeOut($)
	    S = SafePipeReader, safeRead(list:$
					 timeOut: ShortTimeOut
					 default: nil)
	 in
	    if S==nil then
	       OutputBuffer, getAll($)
	    else
	       OutputBuffer, put(S)
	       CAS, ReadTillTimeOut($)
	    end
	 end
         
	 meth close
	    {self.printInfo '^D\n'}
	    %% to do: all dependent processes have to be killed separately
	    Open.pipe, close(true) % not supported by older Oz-versions
	 end
      end
   end

end %define
