%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CasService.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Defines the superclass of all CAS MathWeb services,
%%  namely the Class 
%%    - 'CASService' for Computer Algebra Systems
%%
functor  

import
   Fault
   OS(uName system tmpnam unlink getEnv)
   System
%   Browser(browse: Browse)

   XService(server:        Server
	   )
   at 'x-ozlib://mathweb/sys/XService.ozf'
   CASClass(cas: CAS)
   at 'x-ozlib://mathweb/sys/Cas.ozf'

export   

   'class': CASService

define
   
   %% =================================
   %%
   %% generic class for CAS servers
   %%
   %% =================================
   class CASService from Server
                       
      attr
         cas:     unit
         watcher: unit
         
      feat
         entity
 
      meth doInit
	 {self startCAS}
	 self.entity = {CondSelect self.serverArgs entity unit}
	 watcher <- proc {$ _ _} {self.serverProc leave} end
	 {Fault.installWatcher self.entity [permFail] @watcher _}         
      end
      
      meth startCAS(...) = M
	 cas <- {New CAS {Adjoin M init}}
	 {System.showInfo "[CASService]: CAS started."}
      end

      meth preProcess(ExprVS $)  ExprVS end
      meth postProcess(Result $) Result end
      
      meth eval(Expr ?Result
                prompt: ?Prompt <= _
                done:   ?Done   <= _)
	 case Expr
	 of nil then
            Result = nil
            Done = unit
	 else
	    Started
	    Expr1   = {self preProcess(Expr $)}
	    proc {DoThread}
               try
                  Result1 = {@cas eval(Expr1 $
                                       prompt:  ?Prompt
                                       started: ?Started)}
               in
                  Result = {self postProcess(Result1 $)}
                  Done = unit
               catch E then
                  Done = E
               end
	    end
	    Expr1VS = {Value.toVirtualString Expr1 5 20}
	 in
	    {self.appManager newThread(DoThread info:"eval("#Expr1VS#" $)")}
	    {Wait Started}
	 end
      end
      
      %% Currently input(...) seems to be a duplicate of eval(...).
      %% I don't remember if there was a reason for this or if this
      %% was just an attempt to rewrite & cleanup that never got finished.
      %% /afranke 11.9.00
      meth input(Expr ?Result)
	 case Expr
	 of nil then Result = nil
	 else 
	    Expr1   = {self preProcess(Expr $)}
	    Expr1VS = {Value.toVirtualString Expr1 5 20}
	    proc {DoThread}
	       Result1 = {@cas eval(Expr1 $)} in
	       Result  = {self postProcess(Result1 $)} 
	    end
	 in
	    {self.appManager newThread(DoThread info:"eval("#Expr1VS#" $)")}
	    {Delay 10}
	 end
      end

      meth readRest($)
	 {@cas readRest($)}
      end
      
      meth kill
         {System.showInfo "[CASService]: entering method 'kill'"}
         OldCAS = @cas in
         thread
            {OldCAS close}
            {System.showInfo "[CASGeneric]: CAS terminated."}
         end
         {self startCAS}
         {System.showInfo "[CASService]: leaving 'kill'"}
      end
      
      meth leave
	 {Fault.deInstallWatcher self.entity @watcher _}
	 {@cas close}
         {System.showInfo "[CASService]: CAS terminated. terminating now"}
         Server, terminate
      end
   end


end   %define
