%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ServiceRegistry.oz ; This file is part of the MathWeb system
%
%   major updates: 18.10.2000
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
%%% This is the functor for service registries

import
   XApplication(counter)  at 'x-ozlib://mathweb/share/XApplication.ozf'
   
export
   'class': ServiceRegistry

define
   
   %%
   %% service registry
   %%
      
   class ServiceRegistry

      prop
	 locking
         
      feat
	 app
	 counter
	 registry
	 nameIndex
         
      attr
	 empty:    unit
	 emptyVal: unit
         
      meth init(Application)
	 self.app = Application
	 self.counter   = {New XApplication.counter init(0)}
	 self.registry  = {NewDictionary}
	 self.nameIndex = {NewDictionary}
	 empty <- unit
      end
      meth EmptyVal(V)
	 emptyVal <- V
	 if {IsDet @empty} then
	    empty <- V
	 end
      end
      meth lockEmpty
	 lock ServiceRegistry, EmptyVal(locked) end
      end
      meth unlockEmpty
	 lock ServiceRegistry, EmptyVal(unit) end
      end
      meth checkAccess
	 lock
	    if {IsDet @empty} andthen @empty==locked then
	       raise registry(locked) end
	    end
	 end
      end
   
      %% dictionary operations for (ID -> Service) registry
      
      meth keysID($)
	 lock {Dictionary.keys self.registry} end
      end
      meth itemsID($)
	 lock {Dictionary.items self.registry} end
      end
      meth entriesID($)
	 lock {Dictionary.entries self.registry} end
      end
      meth toRecordID($)
	 lock {Dictionary.toRecord services self.registry} end
      end
      meth memberID(ServiceID $)
	 lock {Dictionary.member self.registry ServiceID} end
      end
      meth condGetID(ServiceID Default $)
	 lock {Dictionary.condGet self.registry ServiceID Default} end
      end
      meth getID(ServiceID $)
	 lock {Dictionary.get self.registry ServiceID} end
      end
      meth PutID(ServiceID Entry)
	 lock {Dictionary.put self.registry ServiceID Entry} end
      end
      meth RemoveID(ServiceID)
	 lock {Dictionary.remove self.registry ServiceID} end
      end
      meth forAllIDs(P)
	 lock {ForAll {Dictionary.keys self.registry} P} end
      end
      meth forAllIndIDs(P)
	 lock
	    {ForAll {Dictionary.entries self.registry}
	     proc {$ ID#Entry} {P ID Entry} end}
	 end
      end
         
      %% dictionary operations for name index
      
      meth names($)
	 lock {Dictionary.keys self.nameIndex} end
      end
      meth availableNames($)
	 lock
	    {Filter {Dictionary.keys self.nameIndex} fun {$ X} X\=nil end}
	 end
      end
      meth toRecordNames($)
	 lock {Dictionary.toRecord names self.nameIndex} end
      end
      meth memberName(ServiceName $)
	 lock {Dictionary.member self.nameIndex ServiceName} end
      end
      meth getIDsByName(ServiceName $)
	 lock {Dictionary.condGet self.nameIndex ServiceName nil} end
      end
      meth AddNameID(ServiceName ServiceID)
	 lock
	    OldIDs = {Dictionary.condGet self.nameIndex ServiceName nil}
	    NewIDs = ServiceID | OldIDs
	 in
	    {Dictionary.put self.nameIndex ServiceName NewIDs}
	 end
      end
      meth RemNameID(ServiceName ServiceID remove:R<=false)
	 lock
	    OldIDs = {Dictionary.condGet self.nameIndex ServiceName nil}
	    NewIDs = {List.subtract OldIDs ServiceID}
	 in
	    if R andthen NewIDs == nil
	    then {Dictionary.remove self.nameIndex ServiceName}
	    else {Dictionary.put self.nameIndex ServiceName NewIDs}
	    end
	 end
      end
      meth forAllNames(P)
	 lock {ForAll {Dictionary.keys self.nameIndex} P} end
      end
      meth forAllIndNames(P)
	 lock
	    {ForAll {Dictionary.entries self.nameIndex}
	     proc {$ Name#IDs} {P Name IDs} end}
	 end
      end
   
      %% isEmpty
   
      meth isEmpty($)
	 ServiceRegistry, keysID($) == nil
      end
   
      %% register, unregister, emptyFlag
      
      meth register(ServiceName Entry ?ServiceID)
	 lock
	    ServiceRegistry, checkAccess
	    ServiceID = {self.counter next($)}
	    ServiceRegistry, PutID(ServiceID Entry)
	    ServiceRegistry, AddNameID(ServiceName ServiceID)
	    if {IsFree @empty}==false then
	       empty <- _
	    end
	 end
%      ServiceRegistry, trace
      end
      meth unregister(ServiceName ServiceID)
	 lock
	    ServiceRegistry, checkAccess
	    ServiceRegistry, RemoveID(ServiceID)
	    ServiceRegistry, RemNameID(ServiceName ServiceID)
	    if ServiceRegistry, isEmpty($) then
	       @empty=@emptyVal
	    end
	 end
%      ServiceRegistry, trace
      end
      meth empty($)
	 lock @empty end
      end
   
      %% trace
      meth trace
	 proc {ShowEntry Key Val}
	    {self.app printInfo Key#': '}
	    {self.app show Val}
	 end
      in
	 {self.app showInfo 'ServiceRegistry:'}
	 {self.app showInfo 'services:'}
	 ServiceRegistry, forAllIndIDs(ShowEntry)
	 {self.app showInfo 'names:'}
	 ServiceRegistry, forAllIndNames(ShowEntry)
      end
   end

end
