%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ShellProver.oz ; This file is part of the MathWeb system
%
%   major updates: 09.08.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Defines the superclasses of shell prover MathWeb services,
%%  namely the class
%%    - 'ShellProver' for ATPs which are started once on each problem
%%
functor
import
   Open OS(uName system tmpnam unlink getEnv)
   System
   Browser(browse: Browse)
   
   File     at 'x-ozlib://mathweb/share/File.ozf'
   XService(server:   Server
	    mailer:   Mailer
	    timeDict: TimeDict
	   )
   at 'x-ozlib://mathweb/sys/XService.ozf' 
  
export
   'class': ShellProver

define
   
   CPULimitedRun = {OS.getEnv "CPULIMITEDRUN"}
   CPUMaxDefault = {OS.getEnv "CPUMAXDEFAULT"}
   
   VS2A = VirtualString.toAtom
%   VS2S = VirtualString.toString
   
   proc {Skip1 _} skip end
   
   fun {IsSubstring Xs Ys}
      {List.isPrefix Xs Ys} orelse
      case Ys of nil then false
      [] _|Yr then {IsSubstring Xs Yr} end
   end
   
   fun {Convert FloatS}
      {Float.toInt {String.toFloat FloatS}*1000.0}
   end
      
   class Shell from Open.pipe Open.text
      prop
	 final
	 
      feat
	 pid
	 showInfo
	 
      attr 
	 real:  unit
	 user:  unit
	 sys:   unit
	 
	 times: unit
	    
      meth init(showInfo: ShowInfo <= Skip1)
	 self.pid      = Open.pipe, init(cmd:"sh" args:["-s"] pid:$)
	 self.showInfo = ShowInfo
	 times <- unit
      end

      meth exec(Cmd ?Result<=_ times:Times<=_
		timer:Timer<=false nice:Nice<=false)
	 Shell, start(Cmd timer:Timer nice:Nice)
	 Shell, waitReady(?Result times:?Times)
      end
      
      /**
	 % does <what>
	 % @arg 1 Cmd foo(feat1 feat2 ...)
	 % @arg 2 
      */
      meth start(Cmd
		 times: ?Times <= _
		 timer: Timer<=true
		 nice:Nice<=false)
	 NicePrefix = if Nice\=false then "nice -"#1#" " else "" end
	 TimePrefix = if Timer       then "time -p "        else "" end
	 Prefix = TimePrefix#NicePrefix
	 Suffix = ';echo \"READY.\"'
      in
	 if {IsFree @times} then
	    raise shell(not_ready) end
	 else
	    Open.text, putS(Prefix#Cmd#Suffix)
	    times <- Times
	 end
      end
	 
      meth waitReady($ times:Times<=_)
	 Times = @times
	 try
            Shell,ReadLoop
            done
         catch timeout then
            Shell,Ready
            timeout
         [] _ then
            Shell,Ready
            killed
         end
      end
      
      meth ReadLoop
         R = Open.text, getS($)
      in
	 case R
         of false            then raise brokenPipe end
	 [] &r|&e|&a|&l|& |T then real <- {Convert T}   Shell,ReadLoop 
	 [] &u|&s|&e|&r|& |T then user <- {Convert T}   Shell,ReadLoop
	 [] &s|&y|&s|& |T    then sys  <- {Convert T}   Shell,ReadLoop
	 [] "READY."         then {self.showInfo ready} Shell,Ready
	 [] ""               then                       Shell,ReadLoop
         [] &s|&h|&:|& |Rest then
%            {Browse Rest#{IsSubstring "CPU time limit exceeded" Rest}}
            if {IsSubstring "CPU time limit exceeded" Rest} then
               {self.showInfo "sh: "#Rest}
               raise timeout end
            else
               if {IsSubstring "Segmentation fault" Rest} then
                  {self.showInfo "sh: "#Rest}
                  raise error end
               else 
                  {self.showInfo "sh: "#Rest}
                  Shell,ReadLoop
               end
            end
	 elseof Line         then {self.showInfo Line}  Shell,ReadLoop
	 end
      end
      
      meth Ready
	 @times = times(real:@real user:@user sys:@sys)
	 real <- 0
	 user <- 0
	 sys  <- 0
      end
    
      meth close
	 %% to do: all dependent processes have to be killed separately
	 Open.pipe, close%(true) , not supported by older Oz-versions
      end
   end %class  Shell
      
   %%
   %% copy directories for later evaluation
   %%
   class FileSaver
      attr
	 destdir: "/tmp"
	 counter: 0
	    
      meth resetSaver(DestDir<=@destdir)
	 counter <- 0
	 destdir <- DestDir
      end
      meth save(Dir to:DestDir<=@destdir)
	 counter <- @counter + 1
	 {OS.system "\\rm -rf "#DestDir#"/"#@counter _}
	 {OS.system "\\cp -r "#Dir#" "#DestDir#"/"#@counter _}
      end
   end %class FileSaver
   
   %% =================================
   %%
   %% support for a private tmp directory
   %%
   %% =================================
   class TextFile from Open.file Open.text end
   class TmpDir from FileSaver
		      
      attr
	 tmpdir:   unit
	    
      meth mkTmpDir(?TmpDir<=_)
	 if @tmpdir == unit then
	    TmpDir = {OS.tmpnam}
	    {OS.system 'mkdir '#TmpDir _}
	    tmpdir <- TmpDir 
	 else
	    raise tmpdir('attrib_not_free' @tmpdir) end
	 end
      end
      meth rmTmpDir(TmpDir <= @tmpdir)
	 if TmpDir \= unit
	 then {OS.system "\\rm -rf "#TmpDir _}
	 else raise tmpdir('attrib_not_set') end
	 end
      end
         
      meth saveVS(VS FileName)
	 {File.saveVS VS @tmpdir#"/"#FileName}
      end
      meth loadVS(FileName $)
	 {File.loadVS @tmpdir#"/"#FileName}
      end
      meth loadS(FileName $)
	 {File.loadS @tmpdir#"/"#FileName}
      end
         
      meth rm(FileName)
	 {OS.unlink @tmpdir#"/"#FileName}
      end
         
      meth openRead(FileName ?F)
	 F = {New TextFile init(name:@tmpdir#"/"#FileName)}
      end
      meth readAll(F $)
	 {F seek}
	 {F read(list:$ size:all)}
      end
         
      meth save
	 FileSaver,save(@tmpdir)
      end
   end %class TmpDir
   
   %%
   %% support for automatic removal of tmp-files
   %%
   class TmpFiles from TmpDir
      attr
	 tmpfiles: nil
	 
      meth registerTmpFile(FileName)
	 tmpfiles <- (@tmpdir#'/'#FileName)|@tmpfiles
      end
      meth rmTmpFiles
	 {ForAll @tmpfiles
	  proc {$ F}
	     if F \= unit then
		thread
		   try {OS.unlink F} {System.showInfo F#" removed."}
		   catch _ then {System.showInfo F#" not removed."} end
		end
	     end
	  end}
	 tmpfiles <- nil
      end
   end % class TmpFiles
   
   %% =================================
   %%
   %% service with a private system shell
   %%
   %% =================================
   class ShellProver
      from Server TmpDir TmpFiles Mailer
	 
      attr
	 outfile:  unit
	 shell:    unit
         
      feat
	 times
	 client
         
      meth doInit
	 %%{self startThreads}
	 local
	    ClientPort = self.clientPort in
	    self.client = proc {$ M} {Send ClientPort M} end
	 end
	 self.times = {New TimeDict init}
	 local
	    Tmp = TmpDir, mkTmpDir($) in
	    shell <- {New Shell init(showInfo:System.showInfo)}
	    {@shell exec('cd '#Tmp)}
	 end
	 FileSaver, resetSaver
      end
      
      meth exec(Cmd timer:Timer<=true nice:Nice<=15)
	 {@shell exec(Cmd timer:Timer nice:Nice)}
      end
      
      meth start(Cmd
		 times:  Times  <= _
		 nice:   Nice   <= 15
		 maxCPU: CPUMax <= CPUMaxDefault
		)
	 CMD = if CPUMax==false then Cmd
	       else CPULimitedRun#" "#CPUMax#" "#Cmd end
      in
	 {self.times putTime(start1)}
	 {@shell start(CMD times:Times nice:Nice)}
	 {self.times putTime(start2)}
      end
      
      meth saveVS(VS FileName)
	 {self.times putTime(saveVS1)}
	 TmpDir, saveVS(VS FileName)
	 {self.times putTime(saveVS2)}
      end
      
      meth submitProblem(ProblemBS
                         syntax: Syntax   <= dfg
                         timeout: TimeOut <= unit)
	 {Browse {VS2A 'You should redefine the method submitProblem!'}}
      end
      
      meth analyzeResult($)
         {Browse {VS2A 'You should redefine the method analyzeResult!'}}
         atp_result(state: error      % can be proof, timeout, ...
                    name:  'prover'
                    host:  'DummyHost'
                    runtime: 0
                    ouput: "Foo"
                    models: "[]"
                   )                    
      end
         
      meth prove(ProblemBS  ?Result
		 syntax:    Syntax    <= unit
                 replyWith: ReplyWith <= unit(state: _)
                 timeout:   Timeout   <= unit
		) 
%         {Browse M}

	 {self.times putTime(send1)}      %% store start time
         
	 Submit = if Syntax == unit
                  then if Timeout == unit then
                          submitProblem(ProblemBS)
                       else
                          submitProblem(ProblemBS timeout: Timeout)
                       end
                  else if Timeout == unit then
                          submitProblem(ProblemBS syntax:Syntax)
                       else
                          submitProblem(ProblemBS syntax:Syntax timeout: Timeout)
                       end
                  end
         {self Submit}
	 {self.times putTime(send2)}
	 proc {DoThread}
	    Status ProverResult T Times
         in 
	    {self.times putTime(return1)}      
	    %% blocks until the prover has finished his job
	    {self.times putTime(waiting1)}
	    {@shell waitReady(?Status
                              times: ?T)}
	    {self.times putTime(waiting2)}
	    case Status
            of killed then
	       {self.times putTime(result1)}
	       ProverResult = atp_result(state: error)
               /*Result = KillDefault*/
	       {self.times putTime(result2)}
	    [] timeout then
	       {self.times putTime(result1)}
               ProverResult = atp_result(state: timeout)
	       {self.times putTime(result2)}
            else
	       {self.times putTime(result1)} 
	       ProverResult = {self analyzeResult($)}
	       {self.times putTime(result2)}
            end
%            {Browse ProverResult}
%            {Browse result(Result)}
	    %%{self rmTmpFiles}
	    {self.times putTime(return2)}    %% store end time
	    Times = {Adjoin T {self getTimes($)}}
            Result = {Record.mapInd ReplyWith
                      fun{$ F _}
                         case {CondSelect ProverResult F unit}
                         of unit then
                            case F
                            of time then Times
                            [] host then self.host
                            [] name then {CondSelect self name unknown}
                            else unit
                            end
                         [] V then V
                         end                                                  
                      end}
%            {Browse times(Times)}
	 end
      in
	 {self.appManager newThread(DoThread info:'return(...)')}
      end
      
%             meth prove(ProblemBS ?Result
% 		 syntax: Syntax <= unit
% 		 times:  ?Times <= _
% 		) 
% 	 {self.times putTime(send1)}      %% store start time
% 	 if Syntax == unit
% 	 then {self submitProblem(ProblemBS)}
% 	 else {self submitProblem(ProblemBS syntax:Syntax)}
% 	 end
% 	 {self.times putTime(send2)}
% 	 proc {DoThread}
% 	    Status T in 
% 	    {self.times putTime(return1)}      
% 	    %% blocks until the prover has finished his job
% 	    {self.times putTime(waiting1)}
% 	    {@shell waitReady(?Status times:?T)}
% 	    {self.times putTime(waiting2)}
% 	    if Status == 'killed' then
% 	       {self.times putTime(result1)}
% 	       /*Result = KillDefault*/
% 	       {self.times putTime(result2)}
% 	    else
% 	       {self.times putTime(result1)} 
% 	       Result = {self analyzeResult($)}
% 	       {self.times putTime(result2)}
% 	    end
% 	    %%{self rmTmpFiles}
% 	    {self.times putTime(return2)}    %% store end time
% 	    Times = {Adjoin T {self getTimes($)}}
% 	 end
%       in
% 	 {self.appManager newThread(DoThread info:'return(...)')}         
%       end

      meth getTimes(?Times)
	 T = {self.times toRecord($)}
      in
	 try
	    Times = {Adjoin T msecs(saveVS:   T.saveVS2   - T.saveVS1
				    start:    T.start2    - T.start1
				    send:     T.send2     - T.send1
				    waiting:  T.waiting2  - T.waiting1
				    result:   T.result2   - T.result1
				    return:   T.return2   - T.return1)}
	    {self.times removeAll}
	 catch _ then
	    Times = msecs(saveVS:  0
			  start:   0
			  send:    0
			  waiting: 0
			  result:  0
			  return:  0)
	 end
      end
      
      meth kill
	 {System.showInfo "[ProverService]: entering method 'kill'"}
	 OldShell = @shell in
	 thread
	    {OldShell close}
	    {System.showInfo "[ProverService]: shell closed."}
	 end
	 shell <- {New Shell init}
	 {System.showInfo "[ProverService]: new shell opened. leaving 'kill'"}
      end
      
      meth leave
	 {System.showInfo "[ProverService]: entering method 'leave'"}
	 {@shell close}
	 {System.showInfo "[ProverService]: shell closed."}
	 {self rmTmpDir}
	 {System.showInfo "[ProverService]: tmp-dir removed, terminating now"}
	 Server, terminate
      end
   end %class ShellProver
end %define


   
   
   
