%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SimpleService.oz ; This file is part of the MathWeb system
%
%   major updates: 09.08.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer, afranke}@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Defines the superclasses of shell prover MathWeb services,
%%  namely the class
%%    - 'SimpleService' for oz-services which don't need an external shell
%%
functor
import
   OS(uName system tmpnam unlink getEnv)
   System
%   Browser(browse: Browse)
   
   XService(server:   Server
	    mailer:   Mailer
	    timeDict: TimeDict
	   )
   at 'x-ozlib://mathweb/sys/XService.ozf'
   
export
   'class': SimpleService

define
   class SimpleService
      from Server Mailer
	 
      attr
         result: unit
         killP:  unit
         idle:   unit
         
      feat
         times

      meth doInit
         self.times = {New TimeDict init}
         result <- unit
         idle   <- unit
      end
      
      meth send(...)=Meth
         result <- _
         idle   <- _
         killP  <- _
         {self.times putTime(send1)}
         {self {Adjoin Meth submitProblem(result:@result killProc:@killP)}}
         {self.times putTime(send2)}
      end
      
      meth analyzeResult($)
         s(string: {ByteString.make "error: method not redefined"}
           bool:   true)
      end
      
      meth return(?Result
                  killDefault: KillDefault <=_
                  times:       Times       <=_)
         
         proc {DoThread}
            T in
            {self.times putTime(return1)}
            
            %% blocks until the result is determined
            {self.times putTime(waiting1)}
            {WaitOr @result @idle}
            @idle = unit
            T = times(user:0 system:0 real:0)
            {self.times putTime(waiting2)}
            if {IsFree @result} then
               {self.times putTime(result1)}
               Result = KillDefault
               {self.times putTime(result2)}
            else
               {self.times putTime(result1)} 
               Result = {self analyzeResult($ result:@result)}
               {self.times putTime(result2)}
            end
            {self.times putTime(return2)}
            Times = {Adjoin T {self getTimes($)}}
         end
      in
         {self.appManager newThread(DoThread info:'return(...)')}
      end
      
      meth getTimes(?Times)
         T = {self.times toRecord($)} in
         Times = {Adjoin T msecs(saveVS:   0
                                 start:    0
                                 send:     T.send2     - T.send1
                                 waiting:  T.waiting2  - T.waiting1
                                 result:   T.result2   - T.result1
                                 return:   T.return2   - T.return1)}
         {self.times removeAll}
      end

      meth Kill
         if {IsFree @idle} andthen
            {IsFree @result} andthen
            {IsDet @killP} then
            try {@killP}
            catch Err then {System.show Err} end
         end
         @idle = unit
      end
      meth kill
         {System.showInfo "[SimpleService]: entering method 'kill'"}        
         {self Kill}
         {System.showInfo "[SimpleService]: leaving method 'kill'"}
      end
      
      meth leave
         {System.showInfo "[SimpleService]: entering method 'leave'"}
         {self Kill}
         Server, terminate
      end
      
      meth idle($)
         @idle
      end
   end
end
