% -*- indent-tabs-mode: nil -*-
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% XService.oz ; This file is part of the MathWeb system
%
%   major updates: 01.02.2001
%
%   Authors: Juergen Zimmer, Andreas Franke
%
%   email: {jzimmer,afranke}@ags.uni-sb.de
%
% For information about this program, write to:
%   MathWeb-Project
%   AG Siekmann/FB nformatik
%   Universitaet des Saarlandes
%   Bau 36, 4. Stock
%   D-66041 Saarbruecken
%   Germany
%
% For information about the newest version of MathWeb, see
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Defines the superclasses of all MathWeb services,
%%  namely the Classes
%%
functor

import
   Connection /*Error*/ Fault /*Open*/ OS(uName system tmpnam unlink getEnv)
   Property /*System*/ Type(ask)
   %%Browser(browse: Browse)

   %%File     at 'x-ozlib://mathweb/share/File.ozf'
   Internet at 'x-ozlib://mathweb/share/Internet.ozf'
   XURL     at 'x-ozlib://mathweb/share/XURL.ozf'

export
   manager:     ServiceManager
   metaService: MetaService
   server:      Server

   threadManager: ThreadManager
   clientManager: ClientManager
   faultManager:  FaultManager

   timeDict: TimeDict
   mailer:   Mailer

define

   VS2A = VirtualString.toAtom
   %%VS2S = VirtualString.toString
   /*
   fun {ExceptionToVS Err}
      {Error.messageToVirtualString {Error.exceptionToMessage Err}}
   end

   proc {Skip1 _} skip end
   */
   fun {IsSubstring Xs Ys}
      {List.isPrefix Xs Ys} orelse
      case Ys of nil then false
      [] _|Yr then {IsSubstring Xs Yr} end
   end



   %%
   %% create serviceobjects on demand
   %%
   class MetaService

      feat
         ServiceName
         ServiceClass
         BrokerURL
         BrokerTicket
         AppMan
         CmdArgs

      meth init(name:     Name
                'class':  Class
                url:      URL      <= false
                ticket:   Ticket   <= false
                appMan:   App
                autoexit: Autoexit <= false
                cmdArgs:  Args     <= record
               )
         self.ServiceName  = Name
         self.ServiceClass = Class
         self.BrokerURL    = URL
         self.BrokerTicket = Ticket
         self.AppMan       = App
         self.CmdArgs      = Args
         if Autoexit\=false then
            proc {DoAutoexit}
               {Delay Autoexit*1000}
               {self.serverProc shutdown}
            end
         in
            {App newThread(DoAutoexit detach:true info:Name#'.autoexit')}
         end
      end

      meth createInstance(?PortSend
                          brokerURL:    URL    <= self.BrokerURL
                          brokerTicket: Ticket <= self.BrokerTicket
                          done:         ?Done)
         Result
      in
         {self.AppMan newServer(self.ServiceClass
                                init(url:    URL
                                     ticket: Ticket
                                     cmdArgs: self.CmdArgs
                                     done:    ?Done
                                    )  %% all services must be adjusted first...
                                ?PortSend
                                detach: true
                                info:   self.ServiceName#'.object')}
         if Done \= unit then
            if {IsFree PortSend} then
               PortSend = Done
            end
            {self.serverProc shutdown}
         end
      end

      meth ping($)
         unit
      end

      meth shutdown
         {Thread.terminate {Thread.this}}
      end
   end

   %%
   %% threads
   %%
   local
      proc {Suspension}
         {Thread.suspend {Thread.this}}
      end
   in
      class ThreadManager
         %% relies on the feature 'localRegistry'
         %% which is added when the main class is created with newServer(...)

         attr
            ThreadIDs: nil

         meth init
            ThreadIDs <- nil
         end

         meth newThread(Proc0 info:Info)
            ID = {self.localRegistry newThread(Proc0 $ info:Info)} in
            ThreadIDs <- ID | @ThreadIDs
         end

         meth infoThread(Info)
            ThreadManager, newThread(Suspension info:Info)
         end

         meth tryTerminateAll
            {ForAll @ThreadIDs
             proc {$ ID} {self.localRegistry tryTerminate(ID)} end}
         end
      end
   end

   %%
   %% client data
   %%
   class ClientManager from ThreadManager

      feat
         clientName
         clientUser
         clientHost
         clientDisplay
         clientPort

      meth init(ClientInfo)
         ClientName    = case {CondSelect ClientInfo name unit} of unit then
                            N = {Label ClientInfo} in
                            if {IsAtom N} then N
                            else '<unknown>' end
                         [] N then N
                         end
      in
         self.clientName    = ClientName
         self.clientUser    = {CondSelect ClientInfo user    '<unknown>'}
         self.clientHost    = {CondSelect ClientInfo host    '<unknown>'}
         self.clientDisplay = {CondSelect ClientInfo display '<unknown>'}
         self.clientPort    = {CondSelect ClientInfo port unit}
         {Type.ask.virtualString self.clientDisplay}
      end

      meth showEnter(ShowInfo)
         ServerName = self.serverName in
         {ShowInfo "Service "#ServerName#" is entered by "#self.clientName}
         {ShowInfo "USER    = "#self.clientUser}
         {ShowInfo "HOST    = "#self.clientHost}
         {ShowInfo "DISPLAY = "#self.clientDisplay}
      end

      %% info-threads: Who is using the server just now?
      meth startThreads
         ThreadManager, infoThread("CLIENT  = "#self.clientName)
         ThreadManager, infoThread("USER    = "#self.clientUser)
         ThreadManager, infoThread("HOST    = "#self.clientHost)
         ThreadManager, infoThread("DISPLAY = "#self.clientDisplay)
      end

      meth showLeave(ShowInfo)
         skip %% unimplemented
      end
   end

   %%
   %% distribution faults
   %%
   class FaultManager from ClientManager

      feat
         clientWatcher

         %% Watch is a procedure ...
      meth installWatcher(Watch conditions:Conditions<=[permFail])
         self.clientWatcher = Watch
         {Fault.installWatcher self.clientPort Conditions Watch _}
      end

      meth deinstallWatcher
         if {IsDet self.clientPort} andthen
            {IsDet self.clientWatcher} then
            P = self.clientPort
            W = self.clientWatcher
         in
            {Fault.deInstallWatcher P W _}
         end
      end
   end

   %%
   %% generic server
   %%
   class Server from ThreadManager ClientManager FaultManager

      feat
         brokerURL
         brokerTicket
         host
         kqmlagent
         ident
         serverArgs

      meth get(host:?FeatHost <=_
               name:?FeatName <=_)
         FeatHost = self.host
         FeatName = self.ident
      end

      meth init(url:    BrokerURL
                ticket: BrokerTicket
                cmdArgs: CmdArgs <= record
                done:    ?Done   <= _
               )
         self.brokerURL    = BrokerURL
         self.brokerTicket = BrokerTicket
         self.host  = Internet.hostname
         self.ident = {NewName}
         self.kqmlagent    = {HasFeature CmdArgs kqmlagent} andthen CmdArgs.kqmlagent
         Done = unit
      end

      meth enter(client: ClientInfo <= unit
                 args:   ServerArgs <= unit)
         
         ClientManager, init(ClientInfo)
         self.serverArgs = ServerArgs
         {self doInit}
      end

      meth terminate
         {Thread.terminate {Thread.this}}
      end
   end


   %% =================================
   %%
   %% support for collecting times
   %%
   %% =================================
   local
      class DictObj
         feat D
         meth init
            self.D = {NewDictionary}
         end
         meth put(Key Val)
            {Dictionary.put self.D Key Val}
         end
         meth toRecord($ label:L<=unit)
            {Dictionary.toRecord L self.D}
         end
         meth removeAll
            {Dictionary.removeAll self.D}
         end
      end
      fun {GetTime}
         {Property.get 'time.total'}
      end
   in
      class TimeDict from DictObj
         meth putTime(Key)
            DictObj, put(Key {GetTime})
         end
      end
   end


   %%
   %% service which does not need an external shell
   %%
   class Mailer

      meth mail(FileVS to:AddressVS)
         {OS.system "mail "#AddressVS#" <"#FileVS _}
      end
   end


   %%
   %% manager
   %%
   class ServiceManager

      feat
         appName       %% the application name
         appManager    %% the XApplication.manager
         serviceName   %% the service name
         serviceClass  %% the service class, e.g., OmegaServer
         args          %% the command line args
         kqmlagent     %% true, iff the service is a KQML speaking agent

         host          %% the local host
         broker        %% the broker to offer the service/agent to
         brokerURL     %% URL of the broker ticket
         brokerTicket  %% the broker ticket itself
         
      meth getArgs(ArgSpec ?Args)
         App     = self.appManager
         Spec = {Adjoin
                 record(ticketfile(single char:&t type:string)
                        autoexit(single type:int default:false)
                        'require'(single type:bool default:false))
                 ArgSpec}
      in
%        {App.show Spec}
         try
            Args = {App.getCmdArgs Spec}
         catch error(ap(usage S)...) then
            {App.showError self.serviceName#': '#S}
            {App.exit 1}
         end
      end

      meth getBroker(port:   ?Broker
                     url:    ?BrokerURL    <=_
                     ticket: ?BrokerTicket <=_)
         App = self.appManager
      in
         if {HasFeature self.args ticketfile} then
            BrokerURL = (self.args).ticketfile
            {App.printInfo "Loading broker-ticket from `"#BrokerURL#"'..."}
            %%{App.printInfo 'Connecting to `'#BrokerTicket#'\'...'}
            BrokerTicket = {VS2A {XURL.loadVS BrokerURL}}
            {App.showInfo " done."}
         else
            case {App.getEnv 'BROKER_TICKET'} of false then
               BrokerURL = "http://www.ags.uni-sb.de/~omega/broker"
               {App.printInfo "Loading broker-ticket from `"#BrokerURL#"'..."}
               BrokerTicket = {VS2A {XURL.loadVS BrokerURL}}
               {App.showInfo " done."}
            elseof BROKER_TICKET then
               BrokerURL    = false
               BrokerTicket = {VS2A BROKER_TICKET}
            end
         end

         Broker = {Connection.take BrokerTicket}
         %%{App.showInfo ' done.'}
      end


      meth init(ServiceName ServiceClass
                appName: AppName <= ServiceName
                appMan:  App
                argSpec: ArgSpec <= record
               )
%        self.appName      = AppName
         self.appManager   = App
         self.serviceClass = ServiceClass
         self.host         = {OS.uName}.nodename
         self.args         = ServiceManager, getArgs(ArgSpec $)  % uses self.appXXX
         self.kqmlagent    = {HasFeature self.args kqmlagent} andthen self.args.kqmlagent
         self.appName      = if self.kqmlagent then
                                {VS2A AppName#'-agent'}
                             else AppName end
         self.serviceName  = if self.kqmlagent then
                                {VS2A ServiceName#'-AGENT'}
                             else ServiceName end
         
         {App.showInfo "This is "#self.appName#" on "#self.host#"."}
         
         try
            ServiceManager, getBroker(port:   ?self.broker
                                      url:    ?self.brokerURL
                                      ticket: ?self.brokerTicket)
         catch _ then
            {App.showInfo 'failed.'}
            {App.showInfo 'Sorry, could not load Brokerticket...'}
            {App.exit 0}
         end
         if self.args.'require' andthen
            {self.broker available(ServiceName $)} then
            {App.printInfo "Service "#ServiceName#" is already available. "}
            {App.showInfo "Exiting..."}
            {App.exit 0}
         end
      end

      meth offerService
         App         = self.appManager
         Broker      = self.broker
         ServiceName = self.serviceName
         Host        = self.host
         ServiceID
         Server
         proc {DoExit _}
            try {Broker retire(ServiceID)}
            catch _ then skip end
         end
      in
         if {Not {Broker isAgent($)}} andthen self.kqmlagent then
            {App.showInfo 'Sorry, cannot offer agent because broker is not a facilitator.'}
         else
            %%{App.printInfo 'Creating MetaServer...'}
            Server = {App newServer(MetaService
                                    init(name:     self.serviceName
                                         'class':  self.serviceClass
                                         url:      self.brokerURL
                                         ticket:   self.brokerTicket
                                         appMan:   self.appManager
                                         autoexit: self.args.autoexit
                                         cmdArgs:  self.args)
                                    $
                                    detach: true
                                    info:   ServiceName#'.metaservice'
                                    onExit: DoExit)}
            %%{App.showInfo ' done.'}
            {App.printInfo "Offering service `"#ServiceName#"'..."}
            {Broker offer(ServiceName
                          Server
                          ?ServiceID
                          host:      Host
                         )}
            {App.showInfo " done."}
         end
      end
   end

end


% local
%    fun {Substitute S S1 S2}
%       if {List.isPrefix S1 S}
%       then {Append S2 {Substitute {List.drop S {Length S1}} S1 S2}}
%       elsecase S of nil then nil
%       [] C|Sr then C|{Substitute Sr S1 S2}
%       end
%    end
%    fun {MultiSubst S SubstPairs}
%       {FoldL SubstPairs fun {$ S S1#S2} {Substitute S S1 S2} end S}
%    end

%    class OutputBuffer
%       feat
%          RevPrompt
%          LenPrompt
%          RevSubsts
%          RevErrors
%       attr
%          RevS: nil

%       meth init(Prompt subst:SubstPairs<=nil err:Errors<=nil)
%          self.RevPrompt = {Reverse {VirtualString.toString Prompt}}
%          self.LenPrompt = {Length self.RevPrompt}
%          self.RevSubsts = {Map {Map SubstPairs
%                                 fun {$ VS1#VS2} {VS2S VS1}#{VS2S VS2} end}
%                            fun {$ S1#S2} {Reverse S1}#{Reverse S2} end}
%          self.RevErrors = {Map Errors fun {$ VS} {Reverse {VS2S VS}} end}
%          RevS <- nil
%       end
%       meth put(S)
%          RevS <- {Append {Reverse S} @RevS}
%          RevS <- {MultiSubst @RevS self.RevSubsts} %slow but simple
%       end
%       meth prompt(?B)
%          B={List.isPrefix self.RevPrompt @RevS}
%          if B then {System.printInfo "[prompt]"} end
%       end
%       meth err(?B)
%          B={Some self.RevErrors
%             fun {$ RevErr} {IsSubstring RevErr @RevS} end}
%          if B then
%             {System.printInfo "[err]"}
%             {ForAll self.RevErrors
%              proc {$ RevErr}
%                 if {IsSubstring RevErr @RevS} then
%                    {System.printInfo "[["#RevErr#"]]"}
%                 end
%              end}
%          end
%       end
%       meth get($)
%          RevS <- {List.drop @RevS self.LenPrompt}
%          OutputBuffer, getAll($)
%       end
%       meth getAll(?S)
%          S = {Reverse @RevS}
%          RevS <- nil
%       end
%    end

%    class SafePipeReader from Open.pipe
%       meth safeRead(list:?S
%                     timeOut: TimeOut<=10
%                     default: Default<=nil)
%          Thr Exc in
%          thread
%             Thr = {Thread.this}
%             try
%                Open.pipe, read(list:?S)
%                raise unit end
%             catch E then Exc=E end
%          end
%          {WaitOr S {Alarm TimeOut}}
%          try
%             {Thread.suspend Thr}
%             if {Thread.state Thr}==blocked then
%                {Thread.injectException Thr unit}
%             else
%                {Thread.resume Thr}
%             end
%          catch error(kernel(deadThread...)...) then skip end
%          {Wait Exc}
%          if Exc==unit then
%             if {IsFree S} then S=Default end
%          else
%             raise Exc end
%          end
%       end
%    end
%    ShortTimeOut = 100
% in
%    class CAS from Open.pipe SafePipeReader OutputBuffer

%       prop
%          final
%          locking

%       feat
%          pid
%          printInfo
%          suffix

%       attr
%          status: unit

%       meth status($)
%          @status
%       end

%       meth init(cmd:Cmd args:Args
%                 suffix:    SuffixVS  <= ";\n"
%                 prompt:    PromptVS  <= "\n> "
%                 subst:     SubstPairs<= nil
%                 err:       Errors    <= nil
%                 hasPrompt: HasPrompt <= true
%                 prelude:   Prelude   <= nil
%                 printInfo: PrintInfo <= Skip1)
%          self.pid       = Open.pipe, init(cmd:Cmd args:Args pid:$)
%          self.printInfo = PrintInfo
%          self.suffix    = {VirtualString.toString SuffixVS}
%          OutputBuffer, init(PromptVS subst:SubstPairs err:Errors)
%          {self.printInfo Prelude}
%          Open.pipe, write(vs: Prelude)
%          if HasPrompt then
%             CAS, ReadTillPrompt(_)
%          end
%          status <- idle
%       end

%       meth eval(Expr ?Result<=_)
%          TmpResult in
%          lock
%             status <- busy
%             CAS, Write(Expr)
%             CAS, Read(?TmpResult)
%             status <- idle
%          end
%          Result = TmpResult
%       end

%       meth Write(ExprVS)
%          Suffix = self.suffix in
%          Open.pipe, write(vs: ExprVS#Suffix)
%       end

%       meth Read($)
%          try CAS, ReadTillPrompt($)
%          catch Err then
%             {self.printInfo "*** ERROR IN CAS, Read($):"}
%             {self.printInfo {ExceptionToVS Err}#"\n"}
%             nil
%          end
%       end
%       meth readRest($)
%          lock
%             try CAS, ReadTillTimeOut($)
%             catch Err then
%                {self.printInfo "*** ERROR IN CAS, readRest($):"}
%                {self.printInfo {ExceptionToVS Err}#"\n"}
%                nil
%             end
%          end
%       end

%       meth ReadTillPrompt(?ReadString)
%          {System.printInfo "[wait]"}
%          S = Open.pipe, read(list: $
%                              /*tail: TailX   <= nil*/
%                              /*size: +SizeAI <= 1024*/
%                              /*len:  ?LenI   <= _*/)
%       in
%          if S==nil then
%             %% pipe dead
%             {self.printInfo '\ndead\n'}
%             {self close}
%             {self.printInfo '\nclosed\n'}
%             ReadString = nil
%          else
%             {self.printInfo S}
%             OutputBuffer, put(S)
%             if OutputBuffer, err($) then
%                {System.printInfo a}
%                %% substring error indicator found
%                ReadString = {VS2S (OutputBuffer,getAll($))#(CAS,ReadTillTimeOut($))}
%                {System.printInfo b}
%             elseif OutputBuffer, prompt($) then
%                {System.printInfo c}
%                %% prompt found at end of string
%                ReadString = OutputBuffer, get($)
%                {System.printInfo "##### ReadString: #####"}
%                {System.printInfo ReadString}
%                {System.printInfo "#####\n"}
%                {System.printInfo d}
%             else
%                {System.printInfo e}
%                ReadString = CAS, ReadTillPrompt($)
%                {System.printInfo f}
%             end
%          end
%       end

%       meth ReadTillTimeOut($)
%          S = SafePipeReader, safeRead(list:$
%                                       timeOut: ShortTimeOut
%                                       default: nil)
%       in
%          if S==nil then
%             OutputBuffer, getAll($)
%          else
%             OutputBuffer, put(S)
%             CAS, ReadTillTimeOut($)
%          end
%       end

%       meth close
%          {self.printInfo '^D\n'}
%          %% to do: all dependent processes have to be killed separately
%          Open.pipe, close(true) % not supported by older Oz-versions
%       end
%    end
% end


% local
%    fun {Convert FloatS}
%       {Float.toInt {String.toFloat FloatS}*1000.0}
%    end
% in
%    class Shell from Open.pipe Open.text

%       prop
%          final

%       feat
%          pid
%          showInfo

%       attr
%          real:  unit
%          user:  unit
%          sys:   unit

%          times: unit

%       meth init(showInfo: ShowInfo <= Skip1)
%          self.pid      = Open.pipe, init(cmd:"sh" args:["-s"] pid:$)
%          self.showInfo = ShowInfo
%          times <- unit
%       end

%       meth exec(Cmd ?Result<=_ times:Times<=_
%                 timer:Timer<=false nice:Nice<=false)
%          Shell, start(Cmd timer:Timer nice:Nice)
%          Shell, waitReady(?Result times:?Times)
%       end

%       /**
%        % does <what>
%        % @arg 1 Cmd foo(feat1 feat2 ...)
%        % @arg 2
%       */
%       meth start(Cmd
%                  times: ?Times <= _
%                  timer: Timer<=true
%                  nice:Nice<=false)
%          NicePrefix = if Nice\=false then "nice -"#1#" " else "" end
%          TimePrefix = if Timer       then "time -p "        else "" end
%          Prefix = TimePrefix#NicePrefix
%          Suffix = ';echo \"READY.\"'
%       in
%          if {IsFree @times} then
%             raise shell(not_ready) end
%          else
%             Open.text, putS(Prefix#Cmd#Suffix)
%             times <- Times
%          end
%       end

%       meth waitReady($ times:Times<=_)
%          Times = @times
%          try Shell,ReadLoop done
%          catch _ then killed end
%       end

%       meth ReadLoop
%          case Open.text, getS($)
%          of &r|&e|&a|&l|& |T then real <- {Convert T}   Shell,ReadLoop
%          [] &u|&s|&e|&r|& |T then user <- {Convert T}   Shell,ReadLoop
%          [] &s|&y|&s|& |T    then sys  <- {Convert T}   Shell,ReadLoop
%          [] "READY."         then {self.showInfo ready} Shell,Ready
%          [] ""               then                       Shell,ReadLoop
%          elseof Line         then {self.showInfo Line}  Shell,ReadLoop
%          end
%       end

%       meth Ready
%          @times = times(real:@real user:@user sys:@sys)
%          real <- unit
%          user <- unit
%          sys  <- unit
%       end

%       meth close
%          %% to do: all dependent processes have to be killed separately
%          Open.pipe, close%(true) , not supported by older Oz-versions
%       end

%    end
% end

% %% =================================
% %%
% %% support for a private tmp directory
% %%
% %% =================================
% local
%    class TextFile from Open.file Open.text end
% in
%    class TmpDir from FileSaver

%       attr
%          tmpdir:   unit

%       meth mkTmpDir(?TmpDir<=_)
%          if @tmpdir == unit then
%             TmpDir = {OS.tmpnam}
%             {OS.system 'mkdir '#TmpDir _}
%             tmpdir <- TmpDir
%          else
%             raise tmpdir('attrib_not_free' @tmpdir) end
%          end
%       end
%       meth rmTmpDir(TmpDir <= @tmpdir)
%          if TmpDir \= unit
%          then {OS.system "\\rm -rf "#TmpDir _}
%          else raise tmpdir('attrib_not_set') end
%          end
%       end

%       meth saveVS(VS FileName)
%          {File.saveVS VS @tmpdir#"/"#FileName}
%       end
%       meth loadVS(FileName $)
%          {File.loadVS @tmpdir#"/"#FileName}
%       end
%       meth loadS(FileName $)
%          {File.loadS @tmpdir#"/"#FileName}
%       end

%       meth rm(FileName)
%          {OS.unlink @tmpdir#"/"#FileName}
%       end

%       meth openRead(FileName ?F)
%          F = {New TextFile init(name:@tmpdir#"/"#FileName)}
%       end
%       meth readAll(F $)
%          {F seek}
%          {F read(list:$ size:all)}
%       end

%       meth save
%          FileSaver,save(@tmpdir)
%       end
%    end
% end

% %%
% %% support for automatic removal of tmp-files
% %%
% class TmpFiles from TmpDir
%    attr
%       tmpfiles: nil

%    meth registerTmpFile(FileName)
%       tmpfiles <- (@tmpdir#'/'#FileName)|@tmpfiles
%    end
%    meth rmTmpFiles
%       {ForAll @tmpfiles
%        proc {$ F}
%           if F \= unit then
%              thread
%                 try {OS.unlink F} {System.showInfo F#" removed."}
%                 catch _ then {System.showInfo F#" not removed."} end
%              end
%           end
%        end}
%       tmpfiles <- nil
%    end
% end

% %
% % support for sending mails
% %

% %% =================================
% %%
% %% service with a private system shell
% %%
% %% =================================
% class ShellProver
%    from Server TmpDir TmpFiles Mailer

%    attr
%       outfile:  unit
%       shell:    unit

%    feat
%       times
%       client

%    meth doInit
%       %%{self startThreads}
%       local
%          ClientPort = self.clientPort in
%          self.client = proc {$ M} {Send ClientPort M} end
%       end
%       self.times = {New TimeDict init}
%       local
%          Tmp = TmpDir, mkTmpDir($) in
%          shell <- {New Shell init(showInfo:System.showInfo)}
%          {@shell exec('cd '#Tmp)}
%       end
%       FileSaver, resetSaver
%    end

%    meth exec(Cmd timer:Timer<=true nice:Nice<=15)
%       {@shell exec(Cmd timer:Timer nice:Nice)}
%    end

%    meth start(Cmd
%               times:  Times  <= _
%               nice:   Nice   <= 15
%               maxCPU: CPUMax <= CPUMaxDefault
%              )
%       CMD = if CPUMax==false then Cmd
%             else CPULimitedRun#" "#CPUMax#" "#Cmd end
%    in
%       {self.times putTime(start1)}
%       {@shell start(CMD times:Times nice:Nice)}
%       {self.times putTime(start2)}
%    end

%    meth saveVS(VS FileName)
%       {self.times putTime(saveVS1)}
%       TmpDir, saveVS(VS FileName)
%       {self.times putTime(saveVS2)}
%    end

%    meth submitProblem(ProblemBS)
%       {Browse 'You should overwrite this method'}
%    end

%    meth prove(ProblemBS ?Result
%               syntax: Syntax <= unit
%               times:  ?Times <= _
%              )
%       {self.times putTime(send1)}      %% store start time
%       if Syntax == unit
%       then {self submitProblem(ProblemBS)}
%       else {self submitProblem(ProblemBS syntax:Syntax)}
%       end
%       {self.times putTime(send2)}
%       proc {DoThread}
%          Status T in
%          {self.times putTime(return1)}
%          %% blocks until the prover has finished his job
%          {self.times putTime(waiting1)}
%          {@shell waitReady(?Status times:?T)}
%          {self.times putTime(waiting2)}
%          if Status == 'killed' then
%             {self.times putTime(result1)}
%             /*Result = KillDefault*/
%             {self.times putTime(result2)}
%          else
%             {self.times putTime(result1)}
%             Result = {self analyzeResult($)}
%             {self.times putTime(result2)}
%          end
%          %%{self rmTmpFiles}
%          {self.times putTime(return2)}    %% store end time
%          Times = {Adjoin T {self getTimes($)}}
%       end
%    in
%       {self.appManager newThread(DoThread info:'return(...)')}
%    end

%    meth getTimes(?Times)
%       T = {self.times toRecord($)}
%    in
%       try
%          Times = {Adjoin T msecs(saveVS:   T.saveVS2   - T.saveVS1
%                                  start:    T.start2    - T.start1
%                                  send:     T.send2     - T.send1
%                                  waiting:  T.waiting2  - T.waiting1
%                                  result:   T.result2   - T.result1
%                                  return:   T.return2   - T.return1)}
%          {self.times removeAll}
%       catch _ then
%          Times = msecs(saveVS:  0
%                        start:   0
%                        send:    0
%                        waiting: 0
%                        result:  0
%                        return:  0)
%       end
%    end

%    meth kill
%       {System.showInfo "[ProverService]: entering method 'kill'"}
%       OldShell = @shell in
%       thread
%          {OldShell close}
%          {System.showInfo "[ProverService]: shell closed."}
%       end
%       shell <- {New Shell init}
%       {System.showInfo "[ProverService]: new shell opened. leaving 'kill'"}
%    end

%    meth leave
%       {System.showInfo "[ProverService]: entering method 'leave'"}
%       {@shell close}
%       {System.showInfo "[ProverService]: shell closed."}
%       {self rmTmpDir}
%       {System.showInfo "[ProverService]: tmp-dir removed, terminating now"}
%       Server, terminate
%    end
% end

% %%
% %% copy directories for later evaluation
% %%
% class FileSaver
%    attr
%       destdir: "/tmp"
%       counter: 0

%    meth resetSaver(DestDir<=@destdir)
%       counter <- 0
%       destdir <- DestDir
%    end
%    meth save(Dir to:DestDir<=@destdir)
%       counter <- @counter + 1
%       {OS.system "\\rm -rf "#DestDir#"/"#@counter _}
%       {OS.system "\\cp -r "#Dir#" "#DestDir#"/"#@counter _}
%    end
% end
% class SimpleService
%    from Server Mailer

%    attr
%       result: unit
%       killP:  unit
%       idle:   unit

%    feat
%       times

%    meth doInit
%       self.times = {New TimeDict init}
%       result <- unit
%       idle   <- unit
%    end

%    meth send(...)=Meth
%       result <- _
%       idle   <- _
%       killP  <- _
%       {self.times putTime(send1)}
%       {self {Adjoin Meth submitProblem(result:@result killProc:@killP)}}
%       {self.times putTime(send2)}
%    end

%    meth analyzeResult($)
%       s(string: {ByteString.make "error: method not redefined"}
%         bool:   true)
%    end

%    meth return(?Result
%                killDefault: KillDefault <=_
%                times:       Times       <=_)

%       proc {DoThread}
%          T in
%          {self.times putTime(return1)}

%          %% blocks until the result is determined
%          {self.times putTime(waiting1)}
%          {WaitOr @result @idle}
%          @idle = unit
%          T = times(user:0 system:0 real:0)
%          {self.times putTime(waiting2)}
%          if {IsFree @result} then
%             {self.times putTime(result1)}
%             Result = KillDefault
%             {self.times putTime(result2)}
%          else
%             {self.times putTime(result1)}
%             Result = {self analyzeResult($ result:@result)}
%             {self.times putTime(result2)}
%          end
%          {self.times putTime(return2)}
%          Times = {Adjoin T {self getTimes($)}}
%       end
%    in
%       {self.appManager newThread(DoThread info:'return(...)')}
%    end

%    meth getTimes(?Times)
%       T = {self.times toRecord($)} in
%       Times = {Adjoin T msecs(saveVS:   0
%                               start:    0
%                               send:     T.send2     - T.send1
%                               waiting:  T.waiting2  - T.waiting1
%                               result:   T.result2   - T.result1
%                               return:   T.return2   - T.return1)}
%       {self.times removeAll}
%    end

%    meth Kill
%       if {IsFree @idle} andthen
%          {IsFree @result} andthen
%          {IsDet @killP} then
%          try {@killP}
%          catch Err then {System.show Err} end
%       end
%       @idle = unit
%    end
%    meth kill
%       {System.showInfo "[SimpleService]: entering method 'kill'"}
%       {self Kill}
%       {System.showInfo "[SimpleService]: leaving method 'kill'"}
%    end

%    meth leave
%       {System.showInfo "[SimpleService]: entering method 'leave'"}
%       {self Kill}
%       Server, terminate
%    end

%       meth idle($)
%          @idle
%       end
%    end

