%%% -*- indent-tabs-mode: nil -*-
%%%
%%% Author:
%%%   Juergen Zimmer <jzimmer@mathweb.org>
%%%
%%% Copyright:
%%%   Juergen Zimmer, 2001
%%%
%%% Last change:
%%%   $Date: 2001/09/14 08:11:59 $ by $Author: jzimmer $
%%%   $Revision: 1.2 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
 This is a server that handles XML-remote-procedure-calls and
 forwards them to the MathWeb meta-services.
*/

functor

import
   System
   Browser(browse: Browse)
   
   XApplication
   at 'x-ozlib://mathweb/share/XApplication.ozf' 

   XService(server: Server)
   at 'x-ozlib://mathweb/sys/XService.ozf'

   XmlRpcClient
   at 'x-ozlib://mathweb/xmlrpc/XmlRpcClient.ozf'


export
   'class': XmlRpcService
   
define
   VS2A = VirtualString.toAtom
   VS2S = VirtualString.toString
   
   proc {Print X}
      if {IsVirtualString X}
      then {System.printInfo X}
      else {System.print X}
      end
   end
   proc {Show X}
      {Print X}
      {System.showInfo ""}
   end
   
   proc {Call Url MethodName Params ?Result}
      {Browse call(Url  MethodName Params)}
      {ForAll Params
       proc {$ P}
	  {System.printInfo " "}
	  {Print P}
       end}
      {System.printInfo " ... "}
      {XmlRpcClient.execute Url MethodName Params ?Result}
      {Show Result}  
   end

   class XmlRpcService from Server
      feat
         serverUrl
         serviceObjectName
         
      meth init(url:     URL
                ticket:  MetaService
                cmdArgs: CmdArgs
                done:    ?Done)  %% all services must be adjusted first...

         self.serverUrl = URL 
         try
            self.serviceObjectName = {Call URL MetaService#".createInstance" nil}
            Done = unit
         catch E then
            Done = E
            Server, terminate 
         end
      end
      
      meth enter(client: ClientInfo <= unit
		 args:   ServerArgs <= unit) = M
	 
	 Server, M
      end
      
      meth leave
         {Call self.serverUrl self.serviceObjectName#".leave" nil _}
	 Server, terminate
      end
      
      meth 'SHUTDOWN'
         {Call self.serverUrl self.serviceObjectName#".leave" nil _}
         Server, terminate
      end
      
      meth otherwise(M)
         {Browse xmlrpcCall(M)}
         MethodName = {Label M}
         Params Return
      in
         {List.partition {Record.toList M}
          fun{$ X}
             {IsDet X}
          end
          Params
          Return
         }
         case Return
         of nil then
            skip
         [] R|_ then
            {Call self.serverUrl self.serviceObjectName#"."#MethodName Params ?R}
         else
            skip
         end
      end
   end % class
end % define
   
