%%% -*- indent-tabs-mode: nil -*-
%%%
%%% broker.oz ; This file is part of the MathWeb system
%%%
%%% Author:
%%%   Juergen Zimmer <jzimmer@mathweb.org>
%%%
%%% Copyright:
%%%   Juergen Zimmer, 2001
%%%
%%% Last change:
%%%   $Date: 2001/09/13 16:18:41 $ by $Author: jzimmer $
%%%   $Revision: 1.102 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

/**
The broker offers mathematical services and connects to other brokers.
It can also act as a KQML facilitator.
*/

functor
   
import
   OS(unlink getEnv)
   Connection(gate take)
   DPInit %%(getSettings)       [for Mozart 1.2.0]
   DPMisc %%(initIPConnection)  [for Mozart 1.1.0]
   Property(get)
   Browser(browse: Browse)
   
   XApplication    at 'x-ozlib://mathweb/share/XApplication.ozf'
   XURL            at 'x-ozlib://mathweb/share/XURL.ozf'
   Internet        at 'x-ozlib://mathweb/share/Internet.ozf'
   ServiceRegistry at 'x-ozlib://mathweb/sys/ServiceRegistry.ozf'
   Env(home:Home)  at 'x-ozlib://mathweb/share/Env.ozf'
   File            at 'x-ozlib://mathweb/share/File.ozf'
   FireWall        at 'x-ozlib://mathweb/sys/FireWall.ozf'

   %%-------------------------------------
   %%   Functors for KQML communication
   %%-------------------------------------
   KqmlFacilitator('class': KQMLFacilitator)
   at 'x-ozlib://mathweb/kqml/KqmlFacilitator.ozf'

define

   AppName = 'broker'
   DefaultTicketName = 'broker-ticket'
   FireWallAppletName = 'mathweb-fw.oza'
   
   MyTicketFile      %% the local broker-ticket
                     % e.g. /home/mathweb/public_html/broker@leibniz
   MyTicketURL       %% the URL to the broker-ticket
                     % e.g. http://www.ags.uni-sb.de/~mathweb/broker@leibniz
   
   DelayBeforeRegistering = 7000  % miliseconds
   DelayToGetServicesFromRegisteringBroker = 5000

   App = {New XApplication.manager init}
   
   %% The command line arguments
   Spec = record(ticketfile(    single char:[&t &d]  type:string)
                 ticketurl(     single char:[&u] type:string       default: unit)
                 portticketurl( single           type:string)
                 firewallticket(single           type:string       default: nil)
                 facilitator(   single           type:bool         default: false)
                 kqmlport(      single           type:int          default: 20000)
                 writeapplet(   single           type:bool         default: false)
                 remotetickets( single char:[&r] type:list(string) default: nil)
                 autoexit(      single           type:int          default: false))
   Args
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {App.showError AppName#': '#S}
      {App.exit 1}
   end
   %% end of args

   VS2A = VirtualString.toAtom
   Port = if {Property.get 'oz.version'} < '1.2.0' 
          then {DPMisc.initIPConnection info}.port
          else {DPInit.init info _} {DPInit.getSettings}.port
          end

   BehindFirewall = (Args.firewallticket \= nil)
   
   PortTicketURL = if {HasFeature Args portticketurl} then
                      Args.portticketurl
                   else
                      {OS.getEnv "PORT_TICKET_URL"}
                   end

   if PortTicketURL \= false then
      {App.showInfo "portticket: "#PortTicketURL}
   end
   Host = Internet.hostname
   
   KqmlFacilities = Args.facilitator
   KqmlPort       = Args.kqmlport
   
   %% --------------------------------------------
   local
      proc {MkObject Entry ?Object ?Done}
         case Entry.mode
         of create then  %%      create a new object
            MetaService = Entry.1 in
            try
               ServerDone
               {App.printInfo 'Creating ServiceObject...'}
               SObject = {MetaService createInstance($ done: ?ServerDone)}  %% gives a port-send proc
               %% ToDo:  should give a port to enable installation of watchers
            in
               {WaitOr SObject {Alarm 50000}}
               if {IsDet SObject} then
                  if ServerDone == unit then
                     Object = SObject
                     Done   = unit
                  else
                     {Browse broker(serverDone(ServerDone SObject))}
                     Object = SObject
                     Done = ServerDone
                  end
               else
                  %%{App.showInfo ' failed.'}
                  Object = false
                  Done   = timeout(create)
               end
            catch Err then
               Object = false
               Done   = Err
            end
         [] port then    %%   return a port-send procedure to an object
            P SObject in
            P = Entry.1
            proc {SObject M} {Port.send P M} end
            try
               Ping = {SObject ping($)}
            in
               {WaitOr Ping {Alarm 50000}}
               if {IsDet Ping} then
                  %%{App.showInfo ' alive.'}
                  Object = SObject
                  Done   = unit
               else
                  %%{App.showInfo ' no answer.'}
                  Object = false
                  Done   = timeout(ping)
               end
            catch Err then
               Object = false
               Done   = Err
            end
         [] pass_on then  %%  pass an existing object
            {App.showInfo "mode: pass_on"}
            SObject = Entry.1 in
            try
               {App.showInfo "sending ping($) ..."}
               Ping = {SObject ping($)}
            in
               {WaitOr Ping {Alarm 50000}}
               if {IsDet Ping} then
                  {App.showInfo ' alive.'}
                  Object = SObject
                  Done   = unit
               else
                  {App.showInfo ' no answer.'}
                  Object = false
                  Done   = timeout(ping)
               end
            catch Err then
               {App.show Err}
               Object = false
               Done   = Err
            end
         end
      end

      proc {GetObjects SName Services ?SObjects ?Done}
         ServiceIDs = {Services getIDsByName(SName $)}
      in
         {App.printInfo "Service '"#SName#"' is requested..."}
         case ServiceIDs
         of nil then
            {App.showInfo " not found."}
            SObjects = nil
         else
            {App.showInfo " found: "#{Length ServiceIDs}}
            fun {ID2Object ID}
               thread
                  {MkObject {Services getID(ID $)} $ _}
               end
            end
         in
            SObjects = {Filter {Map ServiceIDs ID2Object}
                        fun {$ X} X\=false end}
         end
         {App.showInfo "'"#SName#"': "#
          {Length SObjects}#' objects created.'}
         Done = unit
      end
      
   in
      %%=====================================================
      %%
      %% the class of all MathWeb brokers
      %%
      %%=====================================================
      
      class Broker
         %% this class is extended (at least) with self.serverProc
         %% using the method {App newServer}
         prop
            locking
         attr
            requestResult
         feat
            Services           % a registry with the local services
            remoteBrokers      % dictionary with all known remote brokers
                               %  entries: BrokerTicket#PortSendProcedure
            remoteServices     % dictionary with all remote services
                               %  entries: Servicename#ListOfBrokers
    
         meth init(autoexit:         Autoexit <= false
                   remoteTickets:    RemoteTickets <= nil
                  )
            self.Services = {New ServiceRegistry.'class' init(App)}
            self.remoteServices = {NewDictionary}
            self.remoteBrokers = {NewDictionary}

            if Autoexit\=false then
               %% only used by Serg Autexier so far.... not very useful
               proc {DoAutoexit}
                  {Delay Autoexit*1000}
                  %% no new services can be started anymore...
                  {self.Services lockEmpty}
                  %% wait until all running services are dead
                  {Wait {self.Services empty($)}}
%                 {self.serverProc shutdown} %
               end
            in
               {App newThread(DoAutoexit detach:true info:'broker.autoexit')}
            end
            
            thread
               %% register new Broker to all known brokers
               %% after it is completely initialized 
               {Wait MyTicketURL}
               {Delay DelayBeforeRegistering}
               
               if BehindFirewall then
                  RemoteBroker = Broker, ConnectThroughFireWall(Args.firewallticket $)
               in
                  if RemoteBroker \= false then
                     lock
                        {Dictionary.put self.remoteBrokers
                         'RemoteBroker' RemoteBroker}
                        %%... then and get the remote services
                        {self getRemoteServices('RemoteBroker'
                                                remoteBroker: RemoteBroker
                                                getAll: BehindFirewall)}
                     end
                  end
               else
                  %% connect to a single remote-broker beyond the firewall
                  %% (one-way connection)
                  {ForAll RemoteTickets
                   %% try to connect to the given remote brokers...
                   proc{$ RemoteTicket}
                      Broker, ContactBroker(RemoteTicket)
                   end}
               end
            end
         end
         
         meth ContactBroker(RemoteTicket)
            RemoteBroker = Broker, ConnectToBroker(RemoteTicket $)
         in
            if RemoteBroker \= false then
               lock
                  {Dictionary.put self.remoteBrokers {VS2A RemoteTicket} RemoteBroker}
               end
               %%... then and get the remote services
               {self getRemoteServices(RemoteTicket remoteBroker: RemoteBroker)}
               {RemoteBroker registerBroker(MyTicketURL)}
            else
               {App.showInfo 'Broker at '#RemoteTicket#' not available.'}
            end
         end
         
         %%----------------------------------------------
         %%  methods for connecting to other brokers
         %%----------------------------------------------
         meth ConnectToBroker(TicketURL $)
            case TicketURL
            of false then
               {App.showInfo 'Invalid broker ticket '#{VS2A TicketURL}#'.'}
               false
            [] !MyTicketURL then
%              {App.showInfo 'Detected a cycle...'}
%              {App.showInfo 'Broker: I don\'t accept connection to myself!'}
               false
            else
               {App.printInfo ' Trying to get ticket at '#TicketURL#'...'}
               try
                  %% gives a port-send procedure!
                  Ticket = {VS2A {XURL.loadVS TicketURL}}
                  RemoteBroker = {Connection.take Ticket}
                  Signal = {RemoteBroker ping($)}
               in
                  {WaitOr Signal {Alarm 10000}}
                  if {IsFree Signal} then
                     {App.showInfo ' timeout (10 sec.)'}
                     {App.showInfo 'The remote broker does not respond,'
                      #' maybe crashed...'}
                     raise timeout end
                  else
                     {App.showInfo ' done.'}
                     RemoteBroker 
                  end
               catch Err then
                  {App.showInfo ' failed.'}
                  case Err
                  of error(url(localize ...)...) then
                     {App.showInfo ' Broker ticket not found.'}
                     if {IsDet PortPoint} then
                        {App.printInfo ' Trying to get Ticket from port-daemon...'}
                        try
                           StoredTicket Done
                        in
                           {PortPoint request({VS2A TicketURL} ?StoredTicket
                                              done: ?Done)}
                           {WaitOr Done {Alarm 3000}}
                           if {IsFree Done} then
                              {App.showInfo ' timeout (10 sec.)'}
                              {App.showInfo 'The port-daemon does not respond, maybe crashed...'}
                              raise timeout end
                           elseif Done \= unit then
                              {App.showInfo ' it does not know the ticket!'}
                              false
                           else
                              {App.showInfo ' found.'}
                              {App.printInfo 'Trying to connect to '#StoredTicket#'...'}
                              try
                                 RemoteBroker = {Connection.take StoredTicket}
                                 Signal = {RemoteBroker ping($)}
                              in
                                 {WaitOr Signal {Alarm 10000}}
                                 if {IsFree Signal} then
                                    {App.showInfo ' timeout (10 sec.)'}
                                    {App.showInfo 'The remote broker does not respond, maybe crashed...'}
                                    raise timeout end
                                 else
                                    {App.showInfo ' done.'}
                                    RemoteBroker
                                 end
                              catch _ then
                                 false
                              end
                           end
                        catch _ then false end
                     else
                        false
                     end  %% if {IsDet Port...
                  else  %% case Err of...
%                    {App.show Err}
                     false
                  end
               end
            end
         end
         
         meth ConnectThroughFireWall(Ticket $)
            try
               {App.showInfo 'Connecting through firewall to ticket:'}
               {App.printInfo ' '#Ticket#'...'}
               %% gives a port-send procedure!
               RemoteBroker = {Connection.take Ticket}
               Signal = {RemoteBroker ping($)}
            in
               {WaitOr Signal {Alarm 10000}}
               if {IsFree Signal} then
                  {App.showInfo ' timeout (10 sec.)'}
                  {App.showInfo 'The remote broker does not respond,'#
                   ' maybe crashed...'}
                  raise timeout end
               else
                  {App.showInfo ' done.'}
                  RemoteBroker 
               end
            catch _ then
               {App.showInfo ' failed.'}
               false
            end
         end
                  
         %%----------------------------------------------
         %%  methods for (un)registering of brokers
         %%----------------------------------------------
         meth registerBroker(TicketURL)
            TicketURLAtom = {VS2A TicketURL}
         in
            if {Member TicketURLAtom {Dictionary.keys self.remoteBrokers}} then
               {App.printInfo
                'The broker '#TicketURLAtom#' is already registered...'}
               try
                  RemoteBroker = {Dictionary.get self.remoteBrokers
                                  TicketURLAtom}
                  {RemoteBroker ping(_)}
               in
                  {App.showInfo 'he is still alive. '}
               catch _ then
                  {App.showInfo 'he seems to have died.'}
                  {Dictionary.remove self.remoteBrokers TicketURLAtom}
                  Broker, registerBroker(TicketURL)
               end
            else 
               RemoteBroker = Broker, ConnectToBroker(TicketURLAtom $)
            in
               {Dictionary.remove self.remoteBrokers TicketURLAtom}
               if RemoteBroker == false then
                  {App.showInfo 'The broker is not reachable.'}
               else
%  ToDo: sp"ater hier:
%  {Fault.installWatcher self.clientPort Conditions Watch _}
                  lock
                     {ForAll {Dictionary.items self.remoteBrokers}
                      %% tell it to all known brokers
                      proc{$ RemoteBroker}
                         {RemoteBroker registerBroker(TicketURL)}
                      end}
                     {Dictionary.put self.remoteBrokers TicketURLAtom RemoteBroker}
                     {RemoteBroker registerBroker(MyTicketURL)}
                     Broker, showRemoteBrokers
                  end
                  thread
                     {Delay DelayToGetServicesFromRegisteringBroker}
                     {self getRemoteServices(TicketURL
                                             remoteBroker: RemoteBroker
                                             getAll: false)}
                     %% warum war hier getAll: true?
                  end
               end
            end
         end
         
         meth unregisterBroker(Ticket)
            if Ticket == false then 
               {App.showInfo
                'Broker '#Ticket#' unknown => cannot be unregistered.'}
            elseif {Dictionary.member self.remoteBrokers {VS2A Ticket}} then
               {Dictionary.remove self.remoteBrokers  {VS2A Ticket}}
               {ForAll {Dictionary.entries self.remoteServices}
                proc{$ ServiceName#URLList}
                   if {Member Ticket URLList} then
                      case URLList
                      of [_] then
                         {Dictionary.remove self.remoteServices ServiceName}
                      else
                         {Dictionary.put self.remoteServices
                          ServiceName {List.subtract URLList Ticket}}
                      end
                   end
                end}
               {ForAll {Dictionary.items self.remoteBrokers}
                proc{$ RemoteBroker}
                   {RemoteBroker unregisterBroker(Ticket)}
                end}
               {App.showInfo 'Broker '#Ticket#' is unregistering.x'}
               Broker, showRemoteBrokers
            end
         end
         
         meth remoteBrokers($)
            {Dictionary.keys self.remoteBrokers}
         end
         
         %%----------------------------------------------
         %%  methods for getting the services of other brokers
         %%----------------------------------------------
         meth getRemoteServices(RemoteTicket
                                remoteBroker: RemoteBroker <= unit
                                getAll:       GetAll       <= false)
            lock
               LocalServices = {self.Services names($)}
               RBroker = if RemoteBroker == unit then
                            Broker, ConnectToBroker(RemoteTicket $)
                         else
                            RemoteBroker
                         end
            in
               case RBroker
               of false then skip
               else 
                  RemoteServices = {RBroker services(getAll: GetAll $)}
               in
                  {Wait RemoteServices}
                  {ForAll RemoteServices
                   proc{$ RemoteService}
                      if {Not {Member RemoteService LocalServices}} then
                         OtherTickets = {Dictionary.condGet self.remoteServices
                                         RemoteService nil}
                      in
                         if {Not {Member RemoteTicket OtherTickets}} then
                            %% should be sorted by domain in future...
                            {Dictionary.put self.remoteServices
                             RemoteService RemoteTicket|OtherTickets}
                         end
                      end
                   end}
               end
               {self showRemoteServices}
            end
         end
         
         meth showRemoteBrokers
            OtherBrokers =  {Dictionary.keys self.remoteBrokers}
         in
            if OtherBrokers == nil then
               {App.showInfo
                'Broker at '#MyTicketURL#' knows no other brokers.'}
            else
               {App.showInfo 'Broker at '#MyTicketURL#' now knows brokers:'}
               {ForAll  {Dictionary.keys self.remoteBrokers}
                proc{$ Key}
                   {App.showInfo '   '#Key}
                end}
            end
         end
         
         meth showLocalServices
            {App.showInfo 'Local services:'}
            {ForAll {self.Services names($)}
             proc{$ S}
                {App.showInfo '  '#S}
             end}
         end

         meth showRemoteServices
            {App.showInfo 'Broker at '#MyTicketURL#' knows remoteServices:'}
            {ForAll  {Dictionary.entries self.remoteServices}
             proc{$ Name#TicketList}
                {App.showInfo '   '#Name#' at:'}
                {ForAll TicketList
                 proc{$ Ticket}
                    {App.showInfo '      '#Ticket}
                 end}
             end}
         end
         
         %%----------------------------------------------
         %%  methods for handling of (local) serivces
         %%----------------------------------------------
         meth offer(ServiceName Service ?ServiceID
                    mode:      Mode       <= create
                    host:      Host       <= "<unknown>"
                    signature: Signature  <= unit
                    done:      Done       <= _
                   )
            try
               Entry = entry(Service
                             name:      ServiceName
                             mode:      Mode
                             host:      Host
                             signature: Signature)
            in
               {self.Services register(ServiceName Entry ?ServiceID)}
               {App.showInfo ' done.'}
               Done = unit
            catch Err then
               {App printExc(Err)}
               Done = Err
            end
         end
         
         meth retire(ServiceID
                     signature: Signature <= unit
                     done:      Done      <= _)
            try
               Entry             = {self.Services getID(ServiceID $)}
               ServiceName       = Entry.name
               OriginalSignature = Entry.signature
            in
               {App.printInfo 'Service \''#ServiceName#'\' is retired...'}
               if Signature == OriginalSignature
               then
                  {self.Services unregister(ServiceName ServiceID)}
                  {App.showInfo ' done.'}
                  Done = unit
               else
                  {App.showInfo ' Signature mismatch!'}
                  Done = signature(mismatch)
               end
            catch Err then
               {App printExc(Err)}
               Done = Err
            end
         end
         
         meth services(getAll: GetAll <= false ?Servs)
            if GetAll then
               Servs = {Append {self.Services names($)}
                        {Dictionary.keys self.remoteServices}}
            else
               Servs = {self.Services names($)}
            end
         end
         
         meth request(ServiceName ?Object
                      atURL:    RemoteURL  <= unit
                      client:   ClientName <= anonymous
                      timeOut:  TimeOut    <= 0
                      done:     Done       <= _
                      password: Passwd     <= unit
                      locally:  Locally    <= false) = M
            try
               if {IsDet RemoteURL} andthen
                  {IsVirtualString RemoteURL} andthen
                  Locally \=true 
               then
                  RemoteBroker = {Connection.take
                                  {VS2A {XURL.loadVS RemoteURL}}}
               in
                  {RemoteBroker {Record.subtract M atURL}}
               else
                  ServiceIDs = {self.Services getIDsByName({VS2A ServiceName}
                                                           $)}
               in
                  {App.printInfo 'Service \''#ServiceName#'\' is requested...'}
                  case ServiceIDs
                  of nil then
                     {App.showInfo ' not found locally...'}
                     if Locally \= true then
                        {App.printInfo
                         ' Trying to get service from external brokers...'}
                        try
                           RemoteTickets = {Dictionary.get self.remoteServices
                                            {VS2A ServiceName}}
                        in
                           {ForAll RemoteTickets
                            proc{$ RemoteTicket}
                               try
                                  RemoteBroker =
                                  {Dictionary.get self.remoteBrokers
                                   {VS2A RemoteTicket}}
%                                 {Browse remoteBroker(RemoteBroker)}
%                                 {App.show {Record.subtract M atURL}}
%                                 RDone
                                  RemoteObject =
                                  {RemoteBroker request(ServiceName $
                                                        client:  ClientName 
                                                        done:    _       %RDone
                                                        timeOut: TimeOut    
                                                       )}
                               in
                                  if RemoteObject \= false then
                                     raise found(RemoteTicket RemoteObject) end
                                  else
                                     raise notfound end
                                  end
                               catch E then
                                  case E
                                  of found(...) then
                                     raise E end
                                  [] system(dp(...)) then 
                                     {App.showInfo 'Remote broker '#RemoteTicket#' died...'}
                                     Broker, unregisterBroker(RemoteTicket)
                                     
                                     {App.showInfo 'Trying to reconnect later'}
                                     thread
                                        {Delay 2000}
                                        Broker, ContactBroker(RemoteTicket)
                                     end
                                  else
                                     raise notfound end
                                  end
                               end
                            end}
                           raise notfound end
                        catch E then
                           case E
                           of found(Ticket Obj) then
                              {App.showInfo ' found at '#Ticket#'.'}
                              Object = Obj
                              Done = unit
                           else 
                              {App.showInfo ' failed.'}
                              Object = false
                              Done = false
                           end
                        end
                     else
                        {App.showInfo ' failed.'}
                        Object = false
                        Done = false
                     end
                  [] FirstID|_ then
                     {App.showInfo ' found.'}
                     {MkObject {self.Services getID(FirstID $)} ?Object ?Done}
                  end
               end
            catch Err then
               {App printExc(Err)}
               Done = Err
            end
         end
         
         meth isAgent($)
            KqmlFacilities
         end
         
         meth test
            Query = provers(bliksem: 'BLIKSEM'
                            spass:   'SPASS'
                            fdpll:   'FDPLL'
                            otter:   'OTTER'
                            mace:    'MACE'
                            satchmo: 'SATCHMO'
                            kimba:   'KIMBA'
                            chorus:  'CHORUS'
                           )
         in
            {App.show {self request('SPASS' $)}}
            {App.show {self requestAll(Query $)}}
         end
         
         meth requestAll(Query ?Result
                         atURL:   RemoteURL  <= unit
                         remotely:Remotely   <= true
                         client:  ClientName <= anonymous
                         timeOut: TimeOut    <= 0
                         done:    Done       <= _) = M
            requestResult <- nil
            try
               if {IsDet RemoteURL} andthen
                  {IsVirtualString RemoteURL}
               then
                  RemoteBroker = {Connection.take
                                  {VS2A {XURL.loadVS RemoteURL}}}
               in
                  {RemoteBroker {Record.subtract M atURL}}
               else
                  if {IsRecord Query} andthen
                     {IsVirtualString Query}==false then
                     Sync = {NewCell unit}
                  in
                     requestResult <- {Record.map Query
                                       proc {$ SName ?SObjects}
                                          OldSync NewSync
                                       in
                                          {Exchange Sync OldSync NewSync}
                                          thread
                                             {Wait
                                              {GetObjects SName self.Services ?SObjects $}}
                                             OldSync = NewSync
                                          end
                                       end}
                     if Remotely then
                        {ForAll {Dictionary.keys self.remoteBrokers}
                         proc{$ RemoteTicket}
                            try
                               RemoteBroker =
                               {Dictionary.get self.remoteBrokers
                                {VS2A RemoteTicket}}
                               RemoteResult = {RemoteBroker
                                               requestAll(Query $
                                                          client:ClientName 
                                                          timeOut:TimeOut
                                                          remotely: false)}
                            in
                               {Record.forAll RemoteResult Wait}
                               requestResult <- {Record.zip
                                                 @requestResult
                                                 RemoteResult
                                                 fun{$ X Y} {Append X Y} end}
                            catch E then
                               case E
                               of system(dp(...)) then 
                                  {App.showInfo 'Remote broker '#RemoteTicket#' died...'}
                                  Broker, unregisterBroker(RemoteTicket)
                                  
                                  {App.showInfo 'Trying to reconnect later'}
                                  thread
                                     {Delay 2000}
                                     Broker, ContactBroker(RemoteTicket)
                                  end
                               else
                                  skip
                               end
                            end
                         end}
                     end
                     {App.show @requestResult}
                     Result = @requestResult
                     Done = {Access Sync}
                     {Wait Done}
                  else
                     {GetObjects Query self.Services ?Result ?Done}
                  end
               end
            catch Err then
               {App printExc(Err)}
               Done = Err
            end
         end
         
         meth available(ServiceName $
                        ping: Ping       <= true
                        atURL: RemoteURL <= unit
                        locally: Locally <= false)=M
            try
               if {IsDet RemoteURL} andthen
                  {IsVirtualString RemoteURL} andthen
                  Locally \= true 
               then
                  RemoteBroker = {Connection.take
                                  {VS2A {XURL.loadVS RemoteURL}}}
               in
                  {RemoteBroker {Record.subtract M atURL}}
               else
                  ServiceIDs =
                  {self.Services getIDsByName({VS2A ServiceName} $)}
                  fun {ServiceOk ID}
                     Entry = {self.Services getID(ID $)}
                  in
                     case Entry.mode
                     of create then
                        MetaService = Entry.1
                        Signal = try {MetaService ping($)}
                                 catch _ then error end
                     in
                        {WaitOr Signal {Alarm 50000}}
                        {IsDet Signal} andthen Signal\=error
                     end
                  end   
               in
                  if ServiceIDs\=nil andthen
                     (Ping==false orelse
                      {Filter ServiceIDs ServiceOk} \= nil) then
                     true
                  else
                     if Locally \= true then
                        BrokerTickets = {Dictionary.get self.remoteServices
                                         {VS2A ServiceName}}
                     in
                        {App.printInfo ' Check availability of service '
                         #ServiceName#' from external brokers...'}
                        {ForAll BrokerTickets App.showInfo}
                        {Some BrokerTickets
                         fun{$ RemoteTicket}
                            RemoteBroker = {Dictionary.get self.remoteBrokers
                                            {VS2A RemoteTicket}}
                            {App.show remoteBroker(RemoteBroker)}
                            SResult = {RemoteBroker available(ServiceName $
                                                              ping: Ping
                                                             )}
                         in
                            if SResult \= false then
                               true
                            else false end
                         end}
                     else true end 
                  end
               end
            catch Err then
               {App printExc(Err)}
               false
            end
         end
         
         meth ping($)
%           {App.showInfo 'hallo ping'}
            'Broker is alive'
         end
         
         meth 'SHUTDOWN'
            %% say goodbye to all remote brokers
            {ForAll {Dictionary.entries self.remoteBrokers}
             proc{$ RemoteTicket#RemoteBroker}
                {App.printInfo 'Unregistering at broker '#RemoteTicket#"..."}
                try
                   {RemoteBroker unregisterBroker(MyTicketURL)}
                   {App.showInfo 'done.'}
                catch _ then
                   {App.showInfo ' failed.'}
                end
             end}
            %{Thread.terminate {Thread.this}}
         end
      end   %% end of class Broker
   end  
   
   %%==================================================
   %% The class of KQML speaking brokers, i.e. facilitators
   %%--------------------------------------------------
   KqmlBroker = {ByNeed
                 fun {$}
                    class $
                       from Broker KQMLFacilitator
                          
                       meth init(autoexit: AutoExit
                                 remoteTickets: RemoteTickets) = Init
                          Broker, Init
                          {App.showInfo 'creating KqmlBroker...'}
                          KQMLFacilitator, init(app: App
                                                name: {VS2A 'Broker'#Host}
                                                port: KqmlPort)
                       end
                       
                       meth 'SHUTDOWN'
                          KQMLFacilitator, close
                          Broker, 'SHUTDOWN'
                       end
                       
                       meth test
                          {App.show 'hallo, ich kann KQML!'}
%                         {self handleMessage(kqml('ask-if'(foo bar)))}
                       end
       
                    end % class KqmlBroker
                 end}
   
   %%==================================================
   %% create new broker server and handle ticket
   %%--------------------------------------------------
   try
      {App.showInfo 'Looking for an existing broker-ticket:'}
      TicketFile OldTicket BrokerProc  Signal
   in
      if {HasFeature Args ticketfile} then
         MyTicketFile = Args.ticketfile
         {App.showInfo MyTicketFile}
         {App.printInfo 'Loading \`'#MyTicketFile#'\`...'}
         OldTicket = {VS2A {XURL.loadVS MyTicketFile}}
         {App.showInfo ' done.'}
      else
         case {App getEnv('BROKER_TICKET' $)} of false then
            TicketFile = Home#"/public_html/"#DefaultTicketName
            {App.printInfo 'Loading `'#TicketFile#'\'...'}
            OldTicket = {VS2A {XURL.loadVS TicketFile}}
            {App.showInfo ' done.'}
            MyTicketFile= TicketFile
         elseof BROKER_TICKET then
            TicketFile = false 
            OldTicket  = {VirtualString.toAtom BROKER_TICKET}
            MyTicketFile = OldTicket
         end
      end
      
      {App.printInfo 'Connecting to \''#OldTicket#'\'...'}
      BrokerProc = {Connection.take OldTicket}
      {App.showInfo ' done.'}
      {App.printInfo 'Sending ping($)...'}
      Signal = {BrokerProc ping($)}
      {WaitOr Signal {Alarm 50000}}
      if {IsFree Signal} then
         {App.showInfo ' timeout (50 sec.)'}
         {App.showInfo 'The broker does not respond, maybe crashed...'}
         raise timeout end
      end
      {App.printInfo ' received: '} {App.show Signal}
      {App.showInfo 'A broker is already running. Exiting...'}
      {App.exit 0}
   catch _ then
      {Delay 1000}
      {App.showInfo ' failed.'}
   end
   
   %% Set the URL where the broker-ticket can be found by other brokers.
   
   if {HasFeature Args ticketurl} then %% ToDo neu machen
      MyTicketURL = case Args.ticketurl
                    of unit then
                       MyTicketFile
                    [] X then X
                    end
   else
      if {HasFeature Args remotetickets} then 
         {App.showInfo 'You should specify the URL'
          #' of the broker-ticket for remote brokers!'}
      else 
         MyTicketURL = MyTicketFile
      end
   end

   PortPoint
%   PortTicket
   PortServiceID 
   %%  Create a new broker if there is no one running:
   proc {DoExit _}
      if {HasFeature Args ticketfile} then
         {App.printInfo 'Removing broker-ticket...'}
         {OS.unlink Args.ticketfile}
         {App.showInfo ' done.'}
      end
      {App.printInfo 'Unsetting $BROKER_TICKET...'}
      {App putEnv('BROKER_TICKET' 'false')}
      {App.showInfo ' done.'}
      try
         if {IsDet PortPoint} andthen {IsDet PortServiceID} then 
            {PortPoint retire(PortServiceID)}
         end
      catch Err then
         {App.showError 'Error in "{PortPoint retire(...)}":'}
         {App printExc(Err)}
      end
      
   end

   Server
   if KqmlFacilities then
      {App.printInfo 'Creating new KQML communication broker...'}
      Server = {App newServer(KqmlBroker
                              init(autoexit:         Args.autoexit
                                   remoteTickets:    Args.remotetickets
                                  )
                              $
                              detach: true
                              info:   'KQMLBroker.server'
                              onExit: DoExit)}
   else
      {App.printInfo 'Creating new broker...'}
      Server = {App newServer(Broker
                              init(autoexit:         Args.autoexit
                                   remoteTickets:    Args.remotetickets
                                  )
                              $
                              detach: true
                              info:   'broker.server'
                              onExit: DoExit)}
   end
   {App.showInfo ' done.'}
   {App.showInfo 'Creating new broker-ticket...'}
   Gate      = {New Connection.gate init(Server)}
   MyTicket    = {Gate getTicket($)}
   {App.showInfo 'New ticket is \''#MyTicket#'\'.'}
in
   {App.printInfo 'Setting $BROKER_TICKET...'}
   {App putEnv('BROKER_TICKET' MyTicket)}
   {App.showInfo ' done.'}

   if {HasFeature Args ticketfile} then
      if BehindFirewall then
         {App.showInfo 'Running behind firewall.'}
      else
         FileName = Args.ticketfile
      in
         {App.printInfo "Writing ticket to `"#FileName#"'..."}
         {File.saveVS MyTicket FileName}
         {App.showInfo ' done.'}
         {App.showInfo "Ticket available at `"#MyTicketURL}
         try
            if PortTicketURL \= false then
               {App.showInfo "Registering ticket at port-daemon..."}
               {App.printInfo "Loading `"#PortTicketURL#"'..."}
               PortTicket = {VS2A {XURL.loadVS PortTicketURL}}
               {App.showInfo ' done.'}      
            in
               {App.printInfo "Connecting to port-daemon..."}
               PortPoint = {Connection.take PortTicket} 
               {WaitOr PortPoint {Alarm 10000}}
               if {IsFree PortPoint} then
                  {App.showInfo ' timeout (10 sec.)'}
                  {App.showInfo 'The port-daemon does not respond, maybe crashed...'}
                  raise timeout end
               else
                  {App.showInfo ' done.'}
               end
               {App.printInfo "PortPoint = "} {App.show PortPoint}
               {PortPoint offer({VS2A MyTicketURL} MyTicket ?PortServiceID mode: atom host:Host)}
            end
         catch _ then
            {App.showInfo' failed'}
%           {App.showInfo 'Warning: There is no port-daemon!'}
         end
         if Args.writeapplet then
            {App.printInfo 'Writing firewall applet...'}
            {FireWall.makeApplet
             {VS2A MyTicket}
             Port
             'x-ozlib://mathweb/mathweb-up.exe'
             Home#"/public_html/"#FireWallAppletName
            }
            {App.showInfo 'done.'}
         end
      end
   end
   
   {App.exit 0}
end


%% in registerBroker()
%    {App.showInfo 'The broker '#TicketAtom#' is not answering...'}
%    {App.showInfo '...maybe because of network latency... trying again...'}
%    {Delay 3000}
%    RemoteBroker2 = Broker, ConnectToBroker(Ticket $)
% in
%    if RemoteBroker == false then
