%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% test-broker.oz ; This file is part of the MathWeb system
%
%   major updates: 18.10.2000
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Test whether a broker is running. Prints 'alive', if it is running.
%% Prints 'dead' if the broker has crashed.
%%
functor
   
import
   Connection
   System
   XApplication    at 'x-ozlib://mathweb/share/XApplication.ozf'
   XURL            at 'x-ozlib://mathweb/share/XURL.ozf'


define
   %% The command line arguments
   Spec = record(ticketurl( single char:[&u] type:string)
                )
   App = {New XApplication.manager init}
   Args
   
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage _)...) then
      {App.exit 1}
   end
   %% end of args
   
   VS2A = VirtualString.toAtom
   
   if {HasFeature Args ticketurl} then
      case Args.ticketurl
      of false then
         {System.showInfo 'alive'}
      [] TicketURL then
         try
            %% gives a port-send procedure!
            Ticket = {VS2A {XURL.loadVS TicketURL}}
            RemoteBroker = {Connection.take Ticket}
            Signal = {RemoteBroker ping($)}
         in
            {WaitOr Signal {Alarm 10000}}
            if {IsFree Signal} then
               raise timeout end
            else
               {System.showInfo 'alive'}
               {App.exit 0}
            end
         catch _ then
            {System.showInfo 'dead'}
            {App.exit 1}
         end
      end
   else
      {System.showInfo 'alive'}
      {App.exit 0}
   end
   

end  %% define 
