%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% portdaemon.oz ; This file is part of the MathWeb system
%
%   major updates: 09.02.2001
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   MathWeb-Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of MathWeb, see 
%   http://www.mathweb.org/mathweb
%
% See the file "LICENSE" or
%    http://www.mathweb.org/mathweb/LICENSE.html
% for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL
% WARRANTIES.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor     
   %%
   %% The portdaemon was mainly develop to act as a simple nameserver for DORIS.
   %% it is based on simple socket connection. A client can connect to the given port 
   %% and sends the name of the required service. It gets back the hostname and the
   %% port the service is running.

import
   Connection
   Open(socket)
   OS(unlink)

   XApplication             at 'x-ozlib://mathweb/share/XApplication.ozf'
   XURL                     at 'x-ozlib://mathweb/share/XURL.ozf'
   File                     at 'x-ozlib://mathweb/share/File.ozf'
   ServiceRegistry          at 'x-ozlib://mathweb/sys/ServiceRegistry.ozf'
   Env(home:Home host:Host) at 'x-ozlib://mathweb/share/Env.ozf'
   
define
   AppName = 'portd'
   App = {New XApplication.manager init}
   
   %% args
   Spec = record(port(single char:&p type:int default:20000)
		 ticketfile(single char:[&t &d] type:string
			    default:Home#"/public_html/"#"port-daemon@"#Host))
   Args
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {App.showError AppName#': '#S}
      {App.exit 1}
   end

   VS2A = VirtualString.toAtom

   ServicePort = Args.port
   PortTicketFile = Args.ticketfile
   
   {App.showInfo "This is `"#AppName#"' on `"#Host#"'."} 
 
   %%
   %% main class
   %%
      
   class PortPoint
         
      feat
	 TicketFile
	 Services
            
      meth init(ThisTicketFile)
	 self.TicketFile = ThisTicketFile
	 self.Services  = {New ServiceRegistry.'class' init(App)}
      end
         
      meth offer(ServiceName Service ?ServiceID
		 mode:      Mode       <= procedure
		 host:      Host       <= "<unknown>"
		 signature: Signature  <= unit
		 done:      Done       <= _)
	 try
	    Entry
	 in
	    if Mode == atom then
	       Available = proc{$ Sock}
			      {Sock write(vs: "data('"#Service#"')")}
			   end
	    in
	       Entry = entry(Available
			     data:      Service
			     name:      ServiceName
			     mode:      Mode
			     host:      Host
			     signature: Signature)
	    else
	       Entry = entry(Service
			     name:      ServiceName
			     mode:      Mode
			     host:      Host
			     signature: Signature)
	    end		       
	    
	    {App.printInfo 'Service \''#ServiceName#
	     '\' is offered by \''#Host#'\'...'}
	    {self.Services register(ServiceName Entry ?ServiceID)}
	    {App.showInfo ' done.'}
	    Done = unit
	 catch Err then
	    {App.showInfo ' error:'}
	    {App.show Err}
	    Done = Err
	 end
      end
      
      meth retire(ServiceID
		  signature: Signature <= unit
		  done:      Done      <= _)
	 try
	    Entry             = {self.Services getID(ServiceID $)}
	    ServiceName       = Entry.name
	    OriginalSignature = Entry.signature
	 in
	    {App.printInfo 'Service \''#ServiceName#'\' is retired...'}
	    if Signature == OriginalSignature
	    then
	       {self.Services unregister(ServiceName ServiceID)}
	       {App.showInfo ' done.'}
	       Done = unit
	    else
	       {App.showInfo ' Signature mismatch!'}
	       Done = signature(mismatch)
	    end
	 catch Err then
	    {App.showInfo ' error:'}
	    {App.show Err}
	    Done = Err
	 end
      end

      meth request(ServiceName ?Procedure_Sock_H_P
		   timeOut: TimeOut    <= 0
		   done:    Done       <= _)
	 try
	    ServiceIDs = {self.Services getIDsByName(ServiceName $)}
	 in
	    {App.printInfo 'Service \''#ServiceName#'\' is requested...'}
	    case ServiceIDs
	    of nil then
	       {App.showInfo ' not found.'}
	       Procedure_Sock_H_P = false
	       Done = false
	    [] FirstID|_ then
	       {App.showInfo ' found.'}
	       Entry = {self.Services getID(FirstID $)}
	    in
	       case Entry.mode
	       of procedure then
		  P  = Entry.1
		  Ar = {Procedure.arity P} 
	       in
		  case Ar
		  of 3 then
		     Procedure_Sock_H_P = P
		  [] 1 then 
		     proc {Procedure_Sock_H_P Sock _ _} {P Sock} end
		  else
		     raise mode(procedureArity expected:[1 3] got:Ar) end
		  end
	       [] atom then
		  if {HasFeature Entry data} then
		     Procedure_Sock_H_P = Entry.data
		  else 
		     Procedure_Sock_H_P = false
		     raise mode(atom 'no data') end
		  end
	       else
		  raise mode(error(Entry)) end
	       end
	       Done = unit
	    end
	 catch Err then
	    {App.showInfo ' error:'}
	    {App.show Err}
	    Done = Err
	 end
      end

      meth entries
	 Entries = {self.Services entries($)}
      in
	 {App.showInfo 'Known Services: '}
	 {ForAll Entries
	  proc{$ E}
	     {App.show E}
	  end}
      end
      
      meth ping($)
	 'PortPoint is alive'
      end
         
   end % of class PortPoint


   %% exit if portticket exists
   try
      OldTicket
   in
      {Delay 1000}
      {App.showInfo ''}
      {App.showInfo 'Looking for an existing portticket:'} 
      {App.printInfo 'Loading `'#PortTicketFile#'\'...'}
      OldTicket = {VS2A {XURL.loadVS PortTicketFile}}
      {App.showInfo ' done.'}
      {App.printInfo 'Connecting to \''#OldTicket#'\'...'}
      _ = {Connection.take OldTicket}
      {App.showInfo ' done.'}
      {App.showInfo 'A portdaemon is already running. Exiting...'}
      {App.exit 0}
   catch _ then
      {Delay 1000}
      {App.showInfo ' failed.'}
   end
   %% no portticket found

   %% exit if port already active
   try
      TestClient = {New Open.socket init}
   in
      {App.printInfo "Connecting to port "#ServicePort#" at "#Host#"..."}
      {TestClient connect(port:ServicePort host:Host)}
      {App.showInfo " done. "}
      {TestClient shutDown}
      {TestClient close} 
      {App.showInfo "There is already a server using this port. Exiting..."}
      {App.exit 0}
   catch _ then
      {Delay 1000}
      {App.showInfo " failed."}
   end
   %% port not active

   %% install new port server
   {App.showInfo "Creating new server socket."}
   S = {New Open.socket init}
   {App.printInfo "Binding socket to port Port "#ServicePort#"..."}  
   local
      proc {Loop N}
	 try
	    {S bind(takePort:ServicePort)}
	 catch _ then
	    {Delay 1000}
	    {App.printInfo " "#N+1}
	    {Loop N+1}
	 end
      end
   in
      {Loop 0}
   end
   {App.showInfo " done."}
   {S listen}
   {App.showInfo "Listening..."}
   %% port server installed

   {App.printInfo 'Creating new server...'}
   Server    = {App newServer(PortPoint init(PortTicketFile) $
			      detach: false
			      info:   'portpoint.server')} 
   {App.showInfo ' done.'}
   {App.showInfo 'Creating new portticket...'}
   Gate      = {New Connection.gate init(Server)}
   Ticket    = {Gate getTicket($)}
   {App.showInfo 'New ticket is \''#Ticket#'\'.'}
   %% Write ticket to directory accessable via URL
   {App.printInfo 'Saving ticket in file \''#PortTicketFile#'\'...'}
   {File.saveVS Ticket PortTicketFile}
   {App.showInfo ' done.'}
      
   fun {ReadLoop Sock}
      %% read only ONE character every time !!!
      case {Sock read(list:$ size:1)} 
      of nil then nil %raise closeConnection end
      [] C|nil then
	 case C
	 of &\n then nil
	 [] &\r then nil
	 else C|{ReadLoop Sock}
	 end
      end
   end
   
   proc {DoConnection No Sock H P}
      MsgA in
      {App.showInfo No#"Waiting for message..."}
      MsgA = {StringToAtom {ReadLoop Sock}}
      {App.showInfo No#"Received `"#MsgA#"'."}
      case MsgA
      of 'shutdown!' then
	 {Sock write(vs:"Shutting down...\n")}
	 {App.showInfo No#"Exiting with ExitStatus = 0 ..."}
	 {App.exit 0}
      [] 'info' then
	 {Sock write(vs:"There is no information available yet.\n")}
      else
	 case {Server request(MsgA $)} of false then
	    {Sock write(vs:"Service `"#MsgA#"' unknown or unavailable.\n")}
	 elseof Procedure_Sock_H_P then
	    {Procedure_Sock_H_P Sock H P}
	 end
      end
   end

   class AcceptClass
      from
	 Open.socket
      attr
	 no:""
      meth setNo(No)
	 no<-No
      end
      meth write(vs:VS)
	 {App.showInfo @no#"Sending `"#VS#"'"}
	 Open.socket, write(vs:VS) 
      end
   end
   
   proc {AcceptLoop}
      Sock P H ThrID No StartSignal
      proc {DoThread}
	 {Wait StartSignal}
	 {DoConnection No Sock H P}
      end
      proc {DoError Err}
	 {App.showError No#"Exception catched:"}
	 {App.showE ThrID#Err}
      end
      proc {DoExit ExitStatus}
	 {App.showInfo No#"Closing..."}
	 {Sock shutDown}
	 {Sock close}
	 {App.showInfo No#"Closing... done."}
      end
   in
      {S accept(acceptClass:AcceptClass accepted:?Sock port:?P host:?H)}
      {App newThread(DoThread ?ThrID info:'connection('#H#','#P#')'
		     onError:DoError onExit:DoExit)}   
      {Wait ThrID}
      No = "["#ThrID#"]: "
      {Sock setNo(No)}
      {App.showInfo No#"Connection accepted from port "#P#" at "#H#"."}  
      StartSignal = unit
      {AcceptLoop}
   end
in
   try
      {AcceptLoop}
   catch Err then
      {App.printInfo 'Removing portticket...'}
      {OS.unlink PortTicketFile}
      {App.showInfo ' done.'}
      {App.printInfo 'Shutting down main socket...'}
      try {S shutDown}
      catch system(os(os "shutdown" 134...)...) then skip end
      {App.showInfo ' done.'}
      {App.printInfo 'Closing main socket...'}
      {S close}
      {App.showInfo ' done.'}
      case Err of system(kernel(terminate...)...)
      then
	 {App.showInfo "*** TERM signal catched."}
	 {App.showInfo "Terminating all subthreads..."}
	 {App tryTerminateAll}
	 {App.showInfo "Waiting for all threads to unregister..."}
	 Empty = {App empty($)}
      in
	 {WaitOr Empty {Alarm 2000}}
	 if {IsDet Empty}
	 then {App.showInfo "All connections closed. Exiting..."}
	 else {App.showInfo "Resistent connections: "#{App keys($)}}
	 end
	 {App.exit 0}
      else
	 {App.showInfo "Exception catched:"}
	 {App.show Err}
	 raise Err end
      end
   end
end







